# $Id: //depot/prod/test/nacldev/lib/NACL/C/StorageAggregate.pm#92 $
#
# Copyright (c) 2001-2017 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary StorageAggregate Component Module
## @author lincoln@netapp.com, dl-nacl-dev@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::StorageAggregate

=head1 DESCRIPTION

C<NACL::C::StorageAggregate> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access
to an ONTAP StorageAggregate.  A related class is
L<NACL::CS::StorageAggregate|lib-NACL-CS-StorageAggregate-pm>, which represents
the past state of an ONTAP StorageAggregate.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 aggregate

The name of the aggregate, used to identify it uniquely.  (StorageAggregate
names are assumed to be unique within a node or within a cluster)

=cut

package NACL::C::StorageAggregate;

use strict;
use warnings;

use Params::Validate qw(validate_with validate :types);
use NACL::CS::StorageAggregate;
use NACL::CS::StorageAggregateVerifyStatus;
use NACL::ComponentUtils qw (_optional_scalars Dumper _dump_one);
use base 'NACL::C::Component::ONTAP';
use NACL::C::Job;
use NACL::C::UnitNormalization;
use NACL::APISet::Exceptions::CommandFailedException qw(:try);
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NATE::ParamSet qw(param_global);
use NACL::C::_Mixins::AggregateVolume qw(_translate_options_values);
use NACL::C::_Mixins::StorageAggregate qw(:all);
use NACL::Exceptions::InvalidChoice ();
use NACL::GeneralUtils qw();

# Component exceptions
use NACL::Exceptions::NoElementsFound;
use NACL::C::Exceptions::StorageAggregate::AlreadyModified                 ();
use NACL::C::Exceptions::StorageAggregate::CurrentlyBusy                   ();
use NACL::C::Exceptions::StorageAggregate::AggregateHasMountedVolumes      ();
use NACL::C::Exceptions::StorageAggregate::AlreadyExists                   ();
use NACL::C::Exceptions::StorageAggregate::NotEnoughMatchingDisksAvailable ();
use
    NACL::C::Exceptions::StorageAggregate::InsufficientDisksToStartNewRaidGroup (
    );
use NACL::C::Exceptions::StorageAggregate::NoMatchingDisksAvailable     ();
use NACL::C::Exceptions::StorageAggregate::FreeSpaceReallocationEnabled ();
use NACL::C::Exceptions::StorageAggregate::SnaplockEnterpriseNotEnabled;
use NACL::C::Exceptions::StorageAggregate::SnaplockNotEnabled;
use NACL::C::Exceptions::StorageAggregate::StoragePoolBusy;
use NACL::C::Exceptions::StorageAggregate::ComplianceClockNotSet;
use NACL::C::Exceptions::StorageAggregate::ContainsVolumes;
use NACL::C::Exceptions::StorageAggregate::DoesNotExist;

=head1 Linked components

(See
http://wikid.netapp.com/w/QA/projects/Libraries_Initiative/Project/Common_Infrastructure/Users_Guide/Component_Layer_Users_Guide#Linked_components
for a more detailed explanation on linked components)

This component is linked to the following components:

=begin html

<table border="1">
<tr> <th>Prefix</th> <th>Linked Component</th> </tr>
<tr> <td>wafliron</td> <td>StorageAggregateWafliron</td> </tr>
<tr> <td>sixtyfourbitupgrade</td> <td>StorageAggregate64bitUpgrade</td> </tr>
<tr> <td>member</td> <td>StorageAggregateMember</td> </tr>
<tr> <td>plex</td> <td>StorageAggregatePlex</td> </tr>
<tr> <td>snapshot</td> <td>AggregateSnapshot</td> </tr>
</table>

=end html

Examples:

    # Invoking the "stop" method of the StorageAggregateWafliron component
    $aggr_obj->wafliron_stop(%opts);

    # Invoking the "start" method of the StorageAggregate64bitUpgrade component
    $aggr_obj->sixtyfourbitupgrade_start(%opts);

    # Invoking the "reserve_get" method of AggregateSnapshot component
    $aggr_obj->snapshot_reserve_get(%opts);

=cut

sub can {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my @opts       = $pkg_or_obj->_can_validation(@_);

    my $code;
    my $method    = $opts[0];
    my $error_ptr = $opts[1];
    if ($code = $pkg_or_obj->NACL::C::Component::can(@opts)) {
        $Log->exit() if $may_exit;
        return $code;
    } else {
        $code = $pkg_or_obj->_find_wrapper_generator(@opts);
        if ($code) {
            $Log->exit() if $may_exit;
            return $code;
        } else {

            # See if it can be found in the linked components
            $code = $pkg_or_obj->_handle_component_aggregation(
                method    => $method,
                error_ptr => $error_ptr,
                table     => [
                    {   component => 'NACL::C::StorageAggregateWafliron',
                        prefix    => 'wafliron'
                    },
                    {   component => 'NACL::C::StorageAggregate64bitUpgrade',
                        prefix    => 'sixtyfourbitupgrade'
                    },
                    {   component => 'NACL::C::StorageAggregateMember',
                        prefix    => 'member'
                    },
                    {   component => 'NACL::C::StorageAggregatePlex',
                        prefix    => 'plex'
                    },
                    {   component => 'NACL::C::AggregateSnapshot',
                        prefix    => 'snapshot'
                    },
                ],
            );
            if ($code) {
                $Log->exit() if $may_exit;
                return $code;
            } elsif ($code = $pkg_or_obj->SUPER::can(@opts)) {
                $Log->exit() if $may_exit;
                return $code;
            } else {
                $Log->exit() if $may_exit;
                return undef;
            }
        }
    }
}

=head1 METHODS

=head2 new

  my $aggregate = NACL::C::StorageAggregate->new(command_interface=>$ci,
                                         aggregate =>$aggr_name);

Return a new StorageAggregate component object with the given attributes.

=cut

use Class::MethodMaker [scalar => 'aggregate',];

sub isa {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, $kind) = @_;
    my $isa = $pkg_or_obj->_build_isa(
        kind  => $kind,
        alias => 'NACL::C::Aggregate'
    );
    $Log->exit() if $may_exit;
    return $isa;
}

=head2 find

    my $Aggregate  = NACL::C::StorageAggregate->find(command_interface => $ci, ...);
    my @Aggregates = NACL::C::StorageAggregate->find(command_interface => $ci, ...);

(Class method)
A constructor of sorts, which discovers which elements are present (by
interacting with the given command interface).  In an array context it
creates a component object for each present element and returns them.
In a scalar context it creates a component object for just one
element.

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::find|lib-NACL-C-Component-pm/find> for a more
detailed description along with explanation of all the options accepted.

=head2 CANNED FILTERS

See
http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Canned_filters
for a description of canned filters.

=head3 online

Returns aggregates which are online.

=head3 root

Returns the root aggregate.

=head3 not_root

Returns all the non-root aggregates.

=head3 online_not_root

Returns all the online, non-root aggregates.

=head3 online_not_root_64_bit

Returns all the online, non-root, 64 bit aggregates.

=head3 online_not_root_32_bit

Returns all the online, non-root, 32 bit aggregates.

=head2 64_bit

Returns all 64-bit aggregates.

=head2 32_bit

Returns all 32-bit aggregates.

=head2 FIND WRAPPERS

Each of the canned filters have a wrapper method (implemented through
AUTOLOAD) of the form:

    find_<canned-filter-name>

For example, there is a C<find_root> method for the C<root>
canned filter.

    NACL::C::StorageAggregate->find_root(...);

These methods accept all of the arguments normally accepted by find()
but automatically apply the canned filter contained in its name.

In addition to the wrapper methods created for the canned filters, the
following find wrappers are also available:

=head3 find_online_non_root_arranged_by_most_availsize

    # In list context returns all online, non-root aggregates with
    # them arranged from most to least available space
    my @aggrs = NACL::C::StorageAggregate->find_online_non_root_arranged_by_most_availsize(
        command_interface => $node,
        %other_opts
    );

    # In scalar context returns the aggregate which is online, non-root
    # and has the most available space
    my $aggr = NACL::C::StorageAggregate->find_online_non_root_arranged_by_most_availsize(
        command_interface => $node,
        %other_opts
    );

Returns aggregates which are online, non-root and are arranged from most
to least available space.

This accepts all the options accepted by find().

=cut

sub find_online_non_root_arranged_by_most_availsize {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my @objs = $pkg->find(
        @args,
        canned_filters  => [qw(online_not_root)],
        sort_descending => ['availsize']
    );
    $Log->exit() if $may_exit;
    return wantarray ? @objs : $objs[0];
}

=head3 find_online_non_root_64_bit_arranged_by_most_availsize

    # In list context returns all online, non-root aggregates with
    # them arranged from most to least available space
    my @aggrs = NACL::C::StorageAggregate->find_online_non_root_64_bit_arranged_by_most_availsize(
        command_interface => $node,
        %other_opts
    );

    # In scalar context returns the aggregate which is online, non-root, 64 bit
    # and has the most available space
    my $aggr = NACL::C::StorageAggregate->find_online_non_root_64_bit_arranged_by_most_availsize(
        command_interface => $node,
        %other_opts
    );

Returns aggregates which are online, non-root, 64 bit and are arranged from most
to least available space.

This accepts all the options accepted by find().

=cut

sub find_online_non_root_64_bit_arranged_by_most_availsize {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my @objs = $pkg->find(
        @args,
        canned_filters  => [qw(online_not_root_64_bit)],
        sort_descending => ['availsize']
    );
    $Log->exit() if $may_exit;
    return wantarray ? @objs : $objs[0];
}

=head3 find_online_non_root_32_bit_arranged_by_most_availsize

    # In list context returns all online, non-root aggregates with
    # them arranged from most to least available space
    my @aggrs = NACL::C::StorageAggregate->find_online_non_root_32_bit_arranged_by_most_availsize(
        command_interface => $node,
        %other_opts
    );

    # In scalar context returns the aggregate which is online, non-root, 32 bit
    # and has the most available space
    my $aggr = NACL::C::StorageAggregate->find_online_non_root_32_bit_arranged_by_most_availsize(
        command_interface => $node,
        %other_opts
    );

Returns aggregates which are online, non-root, 32 bit and are arranged from most
to least available space.

This accepts all the options accepted by find().

=cut

sub find_online_non_root_32_bit_arranged_by_most_availsize {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my @objs = $pkg->find(
        @args,
        canned_filters  => [qw(online_not_root_32_bit)],
        sort_descending => ['availsize']
    );
    $Log->exit() if $may_exit;
    return wantarray ? @objs : $objs[0];
}

=head2 find_aggrs_not_containing_vsroot_vols

    # In list context, returns an array of objects
    my @aggrs_c = NACL::C::StorageAggregate->find_aggrs_not_containing_vsroot_vols(
        command_interface => $node);

    # In scalar context, returns an array of objects
    my $aggr_c = NACL::C::StorageAggregate->find_aggrs_not_containing_vsroot_vols(
        command_interface => $node);

    # The type of object returned depends on whether invoked with a C or STask:
    # if invoked with a C package name, then returned Component object. If
    # invoked with the STask, then STask objects returned.
    my @aggrs_stask = NACL::STask::Aggregate->find_aggrs_not_containing_vsroot_vols(
        command_interface => $node);

    # This accepts the options typically accepted by find(): filter,
    # allow_empty etc
    # This example returns only online aggregates that do not contain a vsroot
    # volume
    my @aggrs_stask = NACL::STask::Aggregate->find_aggrs_not_containing_vsroot_vols(
        command_interface => $node,
        filter            => { state => 'online' }
    );

    # If no matching aggregates are found, then a
    # NACL::Exceptions::NoElementsFound exception is thrown
    try {
        my @aggrs_stask = NACL::STask::Aggregate->find_aggrs_not_containing_vsroot_vols(
            command_interface => $node,
            filter            => { availsize => '>1TB' }
        );
    } catch NACL::Exceptions::NoElementsFound with {
        # No matching aggregates
    };

    # allow_empty is also supported: if set to 1, then no matching aggregates
    # results in nothing being returned (no exception thrown)
    my @aggrs_stask = NACL::STask::Aggregate->find_aggrs_not_containing_vsroot_vols(
        command_interface => $node,
        filter            => { availsize => '>1TB' },
        allow_empty       => 1
    );

Returns objects for aggregates that do not contain the rootvolume for any
vserver. (For 7Mode, simply returns all aggregates that match the filter)

This supports all of the facilities that find() supports, such as:

=over

=item *

Returns an array in list context and the first object in scalar context

=item *

Accepts all of the options accepted by find() method (such as C<filter>,
C<allow_empty> etc)

=item *

Throws a C<NACL::Exceptions::NoElements> found if not matching aggregates
are found. This can be suppressed by using C<allow_empty>.

=back

=cut

sub find_aggrs_not_containing_vsroot_vols {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            command_interface => {
                type => OBJECT,
                isa  => 'NACL::C::CommandInterface::ONTAP'
            }
        },
        allow_extra => 1
    );

    # This hash is keyed by the names of the aggregates which DO contain
    # a vserver rootvolume
    my %aggrs_with_vsroot;

    if ($opts{command_interface}->is_cmode()) {
        my %common_opts_with_ci;
        $pkg->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_opts_with_ci
        );

        my @vol_cs = NACL::CS::Volume->fetch(
            %common_opts_with_ci,
            canned_filters   => [qw(vsroot_vols)],
            requested_fields => ['aggregate'],
            allow_empty      => 1
        );

        foreach my $single_vol (@vol_cs) {
            $aggrs_with_vsroot{$single_vol->aggregate()} = 1;
        }
    }

    my @aggrs = $pkg->find(%opts);

    my @objs;
    foreach my $aggr (@aggrs) {
        unless (exists $aggrs_with_vsroot{$aggr->aggregate()}) {
            push @objs, $aggr;
        }
    }

    if (!$opts{allow_empty} && !@objs) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::NoElementsFound->throw('Could not find any '
                . 'aggregates that do not contain the rootvolume of any '
                . "vserver and satisfy the filter criteria:\n\t"
                . _dump_one($opts{filter})
                . "\n");
    }

    $Log->exit() if $may_exit;
    return wantarray ? @objs : $objs[0];
}

=head2 state

    my $Aggr_state = $aggr_obj->state();

(Instance method)
Fetch the current state of the element associated
with this component, and return a Component State object corresponding
to this component (calling this on NACL::C::StorageAggregate would return
a NACL::CS::StorageAggregate object)

This method is generally not implemented in individual components. The base
class (L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm> provides an implementation.
See <L<NACL::C::Component::state|lib-NACL-C-Component-pm/state> for a more
detailed description along with explanation of all the options accepted.

=cut

=head2 create

  my $aggregate  = NACL::C::StorageAggregate->create(command_interface=>$ci,
                                              aggregate=>$aggr_name,...);

(Class method) Request creation of a new aggregate element, and  return a new
aggregate component object that refers to it.  Does not wait for aggregate
creation to complete.

Uses a 7Mode CLI/ZAPI, CMode CLI/ZAPI APIset.

Invokes "aggr-create" API for CMode ZAPI,7Mode ZAPI.

Invokes "aggr create" command for 7Mode CLI.

=over

=item Parameters

The following command-line parameters can be specified to affect the behaviour of the call at run-time.

=over

=item NACL_SET_AGGREGATE_BLOCK_TYPE

Global flags that affect the behavior of create command
This flag will make the library set the block-type value to the value set for this flag.
This will override user passed values.

=item NACL_SNAPLOCK_TYPE

This will make the aggregate be created of the specified snaplock-type.
Uses a 7Mode CLI/ZAPI.

=back

=back

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required)See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=>$aggr_name >>

(Required) Name of aggregate to create.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: aggregate

7M CLI : aggregate name (Programmatic name: "aggregate")

7M ZAPI: aggregate

=item C<< disklist => [$disk1, $disk2, $disk3, ... ] >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: disks

7M CLI : -d <value> (Programmatic name: "disklist")

7M ZAPI: disks

=item C<< "mirror-disks" => [$disk1, $disk2, $disk3, ... ] >>

Applicable for CMode CLI, 7Mode CLI/ZAPI.

Maps to:

7M CLI : -d <value> (Programmatic name: "mirror-disks")

7M ZAPI: mirror-disks

=item C<< 'node' => $node_name >>

Node on which to create the aggregate.

Providing this value as an array-reference is deprecated since striped
aggregates no longer exist.

Applicable for CMode CLI/ZAPI.

=item C<< nodes => [$node1_name, $node2_name, $node3_name, ... ] >>

B<DEPRECATED.>

Use C<node> instead.

=item C<< "force-spare-pool" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -f (Programmatic name: "force-spare-pool")

7M ZAPI: force-spare-pool

=item C<< language => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -l <value> (Programmatic name: "language-code")

7M ZAPI: language-code

=item C<< diskrpm => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: rpm

7M CLI : -R <value> (Programmatic name: "rpm")

7M ZAPI: rpm

=item C<< verbose => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -v (Programmatic name: "verbose")

7M ZAPI: verbose

=item C<< "block-type" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: block-type

7M CLI : -B <value> (Programmatic name: "block-type")

7M ZAPI: block-type

=item C<< "is-mirrored" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -m (Programmatic name: "is-mirrored")

7M ZAPI: is-mirrored

=item C<< "is-display-command" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -n (Programmatic name: "is-display-command")

7M ZAPI: is-display-command

=item C<< maxraidsize => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: raid-size

7M CLI : -r <value> (Programmatic name: "raid-size")

=item C<< disktype => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: disk-type

7M CLI : -T <value> (Programmatic name: "disk-type")

7M ZAPI: disk-type

=item C<< raidtype => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: raid-type

7M CLI : -t <value> (Programmatic name: "raid-type")

7M ZAPI: raid-type

=item C<< diskcount => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: disk-count

7M CLI : disk count (Programmatic name: "disk-count")

7M ZAPI: disk-count

=item C<< disksize => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: disk-size

7M CLI : disk size (Programmatic name: "disk-size")

7M ZAPI: disk-size

=item C<< raidgroup => $string >>

Applicable for CMode CLI.

=item C<< voltype => $string >>

This field was removed from ONTAP in FS.0. Any scripts that still pass in
this option for FS should be updated to remove this option. Support for this
option will be removed from the library soon.

Applicable for CMode CLI.

=item C<< "disk-size-with-unit" => $string >>

Applicable for 7Mode ZAPI.

Maps to:

7M ZAPI: disk-size-with-unit

=item C<< "allow-mixed" => $string >>

Applicable for CMode CLI.

=item C<< "ignore-inconsistent" => $string >>

Applicable for CMode CLI.

=item C<< "snaplock-type" => $string >>

Snaplock Type

Applicable for 7Mode CLI/ZAPI and CMode CLI/ZAPI, non-iter.

Maps to:

7M CLI : -L <value> (Programmatic name: "snaplock-type")

7M ZAPI: snaplock-type

=item C<< "raid-block-checksum" => $string >>

Applicable for CMode CLI.

=item C<< "raid-zoned-checksum" => $string >>

Applicable for CMode CLI.

=item C<< "snapshot-autodelete" => $string >>

Applicable for CMode CLI.

=item C<< "thorough-scrub" => $string >>

Applicable for CMode CLI.

=item C<< "volume-style" => $string >>

This field was removed from ONTAP in FS.0. Any scripts that still pass in
this option for FS should be updated to remove this option. Support for this
option will be removed from the library soon.

Applicable for CMode CLI.

=item C<< "is-snaplock" => $string >>

Is Snaplock possible value(s) are, true,false

Applicable for 7Mode ZAPI and CMode CLI/ZAPI, non-iter.

Maps to:

7M ZAPI: is-snaplock

=item C<< "raid-size" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -r <value> (Programmatic name: "raid-size")

7M ZAPI: raid-size

=item C<< "striping" => $string >>

Applicable to 7Mode ZAPI.

Maps to:

7M ZAPI: striping

=item C<< "type" => $string >>

Applicable to 7Mode ZAPI.

Maps to:

7M ZAPI: type

=item C<< "ha-policy" => $string >>

Applicable to 7Mode ZAPI.

Maps to:

7M ZAPI: ha-policy

=item C<< "chksumstyle" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: checksum-style

7M CLI : -c (Programmatic name: "checksum-style")

=item C<< "job_component"=> \$scalar >>

Applicable for CMode CLI.

=item C<< 'disksize-in-blks' => $string >>

Applicable for CMode CLI.

=item C<< 'unmount-volumes' => $string >>

Applicable for CMode CLI.

=item C<< 'group-selection-mode' => $string >>

Applicable for CMode CLI.

=item C<< 'disksize-with-unit' => $string >>

Applicable for CMode CLI.

=item C<< 'simulate' => $string >>

Applicable for CMode CLI.

=item C<< 'mirror-disklist' => [ $mirror-disk1, $mirror-disk2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI.

=item C<< 'mirror' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

7M CLI : -m (Programmatic name: "is-mirrored")

CM ZAPI : is-mirrored

=item C<< 'ignore-pool' => $string >>

Applicable for CMode CLI.

=item C<< 'allow-mixed-rpms' => $string >>

Applicable for CMode CLI/ZAPI.

=item C<< 'allow-same-carrier' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to :

CM ZAPI : allow-same-carrier

=item C<< 'allow-mixed-rpm' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to :

CM ZAPI : allow-mixed-rpm

=item C<< 'ignore-pool-checks' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to :

CM ZAPI : ignore-pool-checks

=item C<< 'pre-check' => $string >>

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=item C<< 'skip-32bit-warning' => $string >>

Skip Warning When Creating 32-bit Format Aggregate
Applicable for CMode CLI

=item C<< force-small-aggregate => $boolean >>

Force the Creation of a Small Aggregate
Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: force-small-aggregate

=item C<< 'mirror-partitionlist' => [ $mirror-partition1, $mirror-partition2, ... ] >>

partitions for Mirrored Plex
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI

=item C<< 'partitionlist' => [ $partition1, $partition2, ... ] >>

Partitions for First Plex
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI

=item C<< 'excluded-from-balancer' => $string >>

Is the Capacity Balancer Excluded possible value(s) are, true,false
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"storage aggregate create" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::StorageAggregate::UnmirroredAggregatesNotSupported>

This exception is raised when a mirrored aggregate is creating without passing the -mirror option.

=item C<NACL::C::Exceptions::StorageAggregate::ComplianceClockNotSet>

This exception is thrown when an attempt to create a Snaplock aggregate is
being made with the Snaplock ComplianceClock not set.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "mirror-partitionlist" => {type => ARRAYREF,  optional => 1},
            "partitionlist"        => {type => ARRAYREF,  optional => 1},
            
            # "nodes" applicable till striped aggregates were removed (SN);
            # node used to be a list till then, is now a scalar
            'node'                 => {type => SCALAR|ARRAYREF,  optional => 1},
            nodes                  => {type => ARRAYREF,  optional => 1},

            "mirror-disklist"      => {type => ARRAYREF,  optional => 1},
            disklist               => {type => ARRAYREF,  optional => 1},
            "mirror-disks"         => {type => ARRAYREF,  optional => 1},            
            job_component          => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( force-spare-pool language diskrpm verbose
                    block-type is-mirrored is-display-command
                    maxraidsize  disktype raidtype diskcount
                    disksize raidgroup voltype ha-policy striping
                    allow-mixed ignore-inconsistent snaplock-type
                    raid-block-checksum raid-zoned-checksum type
                    snapshot-autodelete thorough-scrub volume-style
                    disk-size-with-unit is-snaplock raid-size
                    chksumstyle
                    disksize-in-blks unmount-volumes group-selection-mode disksize-with-unit
                    simulate
                    mirror ignore-pool allow-mixed-rpms trad-volume

                    allow-same-carrier

                    allow-mixed-rpm ignore-pool-checks

                    pre-check

                    skip-32bit-warning
                    force-small-aggregate

                    excluded-from-balancer
                    )
            ),
        },
        cli_cmd => 'storage aggregate create',
    );

    if (defined($opts{disksize})) {
        my $size = delete $opts{disksize};
        $size = NACL::C::UnitNormalization::normalize($size, 'i');
        $opts{disksize} = $size;
    }

    my $ci = $opts{command_interface};
    
    if (defined $opts{nodes}) {
        $Log->warn("The option 'nodes' was provided in the call to "
            . "$pkg->create. However, since striped aggregates were "
            . 'removed in SN, it is no longer possible to create an '
            . "aggregate on multiple nodes. Switch to using 'node' "
            . "instead of 'nodes'"
        );
    }
    
    if (defined $opts{node} && ref $opts{node} eq 'ARRAY') {
        $Log->warn("The option 'node' was provided as an array-reference "
            . "in the call $pkg->create. However, since striped aggregates "
            . 'were removed in SN, it is no longer possible to create an '
            . "aggregate on multiple nodes. Provide the value of 'node' "
            . 'as a scalar instead of as an array-reference'
        );
    }

    # Striped aggregates were removed in SN. The fields related to it
    # (i.e. volume-style and voltype) were removed in FS. Log a warning
    # if either of these options are passed for a call to an FS build.
    foreach my $removed_field (qw(voltype volume-style)) {
        if (exists $opts{$removed_field}) {
            $Log->warn("The option '$removed_field' was removed in FS, "
                    . "but it was passed in the call to $pkg->create. "
                    . 'Please remove this option from the call because '
                    . 'support for this option will be removed from the '
                    . 'library soon.');
        }
    }

    my $global_block_type = param_global->get('NACL_SET_AGGREGATE_BLOCK_TYPE');
    if($global_block_type) {
        $Log->comment("Setting block-type to '$global_block_type' because "
            . "parameter 'NACL_SET_AGGREGATE_BLOCK_TYPE' was provided");
        $opts{'block-type'} = $global_block_type;
    }

    if ( (defined $opts{'block-type'}) && ( $opts{command_interface}->version_manager()->is_promoted() ) ) {
	delete $opts{'block-type'};
	# as per burt#1075425, in FS and later block type is now deprecated
	# since all aggrs are now 64 bit
	$Log->warn("'block-type' option is not supported on promoted builds.")
    }

    my $global_snaplock_type = param_global->get('NACL_SNAPLOCK_TYPE');
    if($global_snaplock_type) {
        $Log->comment("Setting snaplock-type to '$global_snaplock_type' because "
            . "parameter 'NACL_SNAPLOCK_TYPE' was provided");
        $opts{'snaplock-type'} ||= $global_snaplock_type;
    }

    my $ret;
    try {
        $ret = $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_create_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_create_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    check     => '_create_zapi_check',
                },
                {   method    => '_create_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_create_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
            ],
            frontend => 'NACL::C::StorageAggregate::create',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        foreach my $type (qw(AlreadyExists NotEnoughMatchingDisksAvailable
            UnmirroredAggregatesNotSupported ComplianceClockNotSet)) {
            my $exception_name = "NACL::C::Exceptions::StorageAggregate::$type";
            $exception_name->detect_convert_and_throw(exception => $exception);
        }

        # The FeatureNotEnabled exceptions need the command_interface as well
        foreach my $type (qw(SnaplockEnterpriseNotEnabled SnaplockNotEnabled)) {
            my $exception_name = "NACL::C::Exceptions::StorageAggregate::$type";
            $exception_name->detect_convert_and_throw(
                exception         => $exception,
                command_interface => $opts{command_interface},
            );
        }

        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub create

sub _create_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    if (defined($opts{disklist})) {
        $opts{disklist} = join(" ", @{$opts{disklist}});
    }
    if (defined($opts{"mirror-disks"})) {
        $opts{"mirror-disks"} = join(" ", @{$opts{"mirror-disks"}});
    }
    my $aggr_args = $pkg->_hash_copy(
        source => \%opts,
        copy   => [
            qw(aggregate force-spare-pool snaplock-type block-type raid-size trad-volume
                is-mirrored is-display-command verbose mirror-disks)
        ],
        map => {
            "language"    => "language-code",
            "disktype"    => "disk-type",
            "raidtype"    => "raid-type",
            "diskcount"   => "disk-count",
            "disksize"    => "disk-size",
            "disklist"    => "disks",
            "diskrpm"     => "rpm",
            "chksumstyle" => "checksum-style",
            "mirror"      => "is-mirrored",
            "maxraidsize" => "raid-size",
        }
    );

    $pkg->_translate_block_type(
        '32b_value' => '32',
        '64b_value' => '64',
        api_opts    => $aggr_args
    );

    $apiset->aggr_create(%$aggr_args);
    $Log->exit() if $may_exit;
    return $pkg->new(
        aggregate         => $aggr_args->{aggregate},
        command_interface => $command_interface
    );
} ## end sub _create_7mode_cli

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    my $command_interface = $opts{command_interface};
    $pkg->_base_backend_cmode_cli(%opts, api => 'storage_aggregate_create',);
    $Log->exit() if $may_exit;
    return $pkg->new(
        aggregate         => $opts{aggregate},
        command_interface => $command_interface
    );
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = $pkg->_validate_backend_opts(@args);
    my $volstyle = $opts{'volume-style'} || $opts{voltype};

    $pkg->_translate_block_type(
        '32b_value' => '32_bit',
        '64b_value' => '64_bit',
        api_opts    => \%opts,
    );

    if (!defined $opts{nodes} && defined $opts{node}) {
        my $node = $opts{node};
        if (ref $node eq 'ARRAY') {
            $opts{nodes} = $node;
        } else {
            # ZAPI requires it to be provided as an array
            $opts{nodes} = [$node];
        }
    }
    
    # Provided as a scalar
    if (defined $opts{nodes} && !ref $opts{nodes}) {
        $opts{nodes} = [$opts{nodes}];
    }

    # The mapping for disklist should be [make_zapi_array('disks'), 'name'].
    # However, because of an existing bug in _zapi_hash_copy, it constructs
    # the ZAPI XML with only the first value in the array. Hence, we are
    # manually constructing the mapping here.
    if (defined $opts{disklist}) {
        # The structure formed will be like:
        # disks => [ {name => $disk[0]}, {name => $disk[1]}, ... ]
        my @disk_info = map { {name => $_} } @{$opts{disklist}};
        $opts{additional_args} = {disks => \@disk_info};
    }

    $pkg->_base_backend_cmode_zapi(
        %opts,
        api  => 'aggr-create',
        copy => [qw(aggregate block-type nodes force-small-aggregate
            allow-mixed-rpm ignore-pool-checks allow-same-carrier striping
            snaplock-type)
        ],
        map  => {
            'diskcount'   => 'disk-count',
            'chksumstyle' => 'checksum-style',
            'disktype'    => 'disk-type',
            'maxraidsize' => 'raid-size',
            'raidtype'    => 'raid-type',
            'disksize'    => 'disk-size',
            'diskrpm'     => 'rpm',
            'mirror'      => 'is-mirrored',
        },
        inapplicable_for_help_xml => [qw(node)],
    );

    my $obj = $pkg->new(
        aggregate         => $opts{aggregate},
        command_interface => $opts{command_interface},
    );
    $Log->exit() if $may_exit;
    return $obj;
} ## end sub _create_cmode_zapi

# Cannot use ZAPI if:
# 1. "partition-list" specified (is test-privilege, so CLI-only)
# 2. raidtype => 'raid0' (again test-privilege)
sub _create_zapi_check {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_is_partitionlist_specified(%opts);
    if (defined ($opts{raidtype}) && $opts{raidtype} =~ /raid0/i) {
        my $msg = "Creating an aggregate with raidtype as 'raid0' is "
            . 'possible only through the CLI, so we cannot use the ZAPI '
            . 'implementation.';
        $Log->comment($msg);
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw($msg);
    }
    $Log->exit() if $may_exit;
}

sub _create_7mode_zapi {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $aggr_args = $pkg->_hash_copy(
        source => \%opts,
        copy   => [
            qw(aggregate force-spare-pool snaplock-type block-type
                is-mirrored is-display-command  verbose mirror-disks ha-policy
                disk-size-with-unit is-snaplock raid-size striping type)
        ],
        map => {
            "language"  => "language-code",
            "disktype"  => "disk-type",
            "raidtype"  => "raid-type",
            "diskcount" => "disk-count",
            "disksize"  => "disk-size",
            "disklist"  => "disks",
            "diskrpm"   => "rpm",
        },
    );

    $pkg->_translate_block_type(
        '32b_value' => '32_bit',
        '64b_value' => '64_bit',
        api_opts    => $aggr_args
    );

    $apiset->aggr_create(%$aggr_args);
    $Log->exit() if $may_exit;
    return $pkg->new(
        aggregate         => $aggr_args->{aggregate},
        command_interface => $command_interface
    );
} ## end sub _create_7mode_zapi

=head2 delete

  NACL::C::StorageAggregate->delete(command_interface=>$ci,
                             aggregate=>$aggr_name,...);

                            or

  $aggregate_object->delete(...);

(Class or instance method) This method deletes an aggregate element.

Uses a 7Mode/ZAPI, CMode/ZAPI APIset.

Invokes "aggr-destroy" API for CMode ZAPI, 7Mode ZAPI.

Invokes "aggr destroy" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: aggregate

7M CLI : aggregate name (Programmatic name: "aggregate")

7M ZAPI: aggregate

=item C<< force => $boolean >>

(Optional) A boolean when true will do a forceful delete of the aggregate
component.

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -f (Programmatic name: "force")

7M ZAPI: force

=item C<< "job_component"=> \$scalar >>

Applicable for CMode CLI.

=item C<< 'ignore-vldb' => $string >>

Applicable for CMode CLI.

=item C<< 'plex' => $string >>

Applicable for CMode CLI.

=item C<< 'dblade-only' => $string >>

Applicable for CMode CLI.

=item C<< 'hidden-plex' => $string >>

Applicable for CMode CLI.

=item C<< 'preserve-config-data' => $string >>

=item C<< 'ignore-disks-check' => $string >>

Do Not Check for Disks Associated with the Aggregate
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"storage aggregate delete" command in ONTAP.

Applicable for CMode CLI.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( ignore-vldb

                    plex dblade-only hidden-plex

                    preserve-config-data

                    ignore-disks-check
                    )
            ),
            force         => {type => BOOLEAN,   optional => 1},
            job_component => {type => SCALARREF, optional => 1},
        },
        cli_cmd => 'storage aggregate delete',
    );
    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_delete_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::StorageAggregate::delete',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        foreach my $type (qw(ContainsVolumes DoesNotExist)) {
            my $exception_name = "NACL::C::Exceptions::StorageAggregate::$type";
            $exception_name->detect_convert_and_throw(exception => $exception);
        }
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(
         %opts,
        api     => 'storage_aggregate_delete',
    );

    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate force)],
    );
    %opts = %$aggr_args;
    $opts{force} = 1;
    $Log->debug("Opts to API call 'aggr_destroy' are: \n" . Dumper(\%opts));
    $apiset->aggr_destroy(%opts);
    $Log->exit() if $may_exit;
} ## end sub _delete_7mode_cli

sub _delete_7mode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate force)],
    );
    %opts = %$aggr_args;
    $opts{force} = 1;

    $apiset->aggr_destroy(%opts);
    $Log->exit() if $may_exit;
} ## end sub _delete_7mode_zapi

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
    );
    %opts = %$aggr_args;

    $apiset->aggr_destroy(%opts);
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

=head2 modify

  NACL::C::StorageAggregate->modify(command_interface=>$ci,
                             aggregate=>$aggr_name,...);

                            or

  $aggregate_object->modify(...);

(Class or instance method) Modify an aggregate element.

Uses a  7Mode CLI/ZAPI, CMode CLI/ZAPI APIset.

Invokes "aggr-online" API for CMode ZAPI if $state option passed in is online.

Invokes "aggr-offline" API for CMode ZAPI if $state option passed in is offline.

Invokes "aggr-restrict" API for CMode ZAPI if $state option passed in is
restricted.

Invokes "aggr-set-option" API for CMode ZAPI.

Invokes "command" command for 7Mode CLI,7Mode ZAPI.

For CMode-CLI, the capabilities of the "storage aggregate modify" command are
available.  For CMode-ZAPI and 7Mode , "modify" only knows so far how to
modify  the "state" attribute of the  aggregate element, setting it to be
"online", "offline", or  "restricted".

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: aggregate

7M CLI : aggregate name (Programmatic name: "aggregate")

7M ZAPI: aggregate

=item C<< state => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: state

=item C<< raidtype => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: raidtype

=item C<< maxraidsize => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: raidsize

=item C<< "ignore-inconsistent" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ignore_inconsistent

=item C<< "raid-block-checksum" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: raid_cv

=item C<< "raid-zoned-checksum" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: raid_cv

=item C<< "ha-policy" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ha_policy

=item C<< "snapshot-autodelete" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: snapshot_autodelete

=item C<< "thorough-scrub" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: thorough_scrub

=item C<< "cifs-delay-seconds" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -t <value> (Programmatic name: "cifs-delay-seconds")

7M ZAPI: cifs-delay-seconds

=item C<< "unmount-volumes" => $boolean >>

Applicable for 7Mode ZAPI.

Maps to:

7M ZAPI: unmount-volumes

=item C<< "keep-same-uuid" => $string >>

Applicable for 7Mode ZAPI.

Maps to:

7M ZAPI: keep-same-uuid

=item C<< 'hybrid-enabled' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: hybrid_enabled

=item C<< 'raid-lost-write' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: raid_lost_write

=item C<< 'percent-snapshot-space' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: percent_snapshot_space

=item C<< 'free-space-realloc' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: free_space_realloc

=item C<< 'auto-snapshots' => $string >>

Applicable for CMode CLI.

=item C<< 'resync-snap-time' => $string >>

Applicable for CMode CLI.

=item C<< 'resyncsnaptime' => $string >>

Applicable for CMode CLI.

=item C<< 'snapshot-reserve' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: snapshot_reserve

=item C<< 'max-write-alloc-blocks' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: max_write_alloc_blocks

=item C<< 'force-hybrid-enabled' => $string >>

Applicable for CMode CLI.

=item C<< 'disktype' => $string >>

Applicable for CMode CLI.

=item C<< 'nosnap'  => $string >>

Applicable for 7Mode/Nodescope

=item C<< 'lost_write_protect'  => $string >>

Applicable for CMode CLI.

=item C<< 'snapmirrored'  => $string >>

Applicable for CMode CLI.

=item C<< 'fs_size_fixed'  => $string >>

Applicable for CMode CLI.

=item C<< 'no_delete_log'  => $string >>

Applicable for CMode CLI.

=item C<< 'metadata_checking'  => $string >>

Applicable for CMode CLI.

=item C<< 'root'  => "" >>

Applicable for CMode CLI.

=item C<< 'pre-check' => $string >>

Applicable for CMode CLI

=item C<< 'cache-raid-group-size' => $string >>

Hybrid Aggregate SSD Tier Maximum RAID Group Size

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI:cache_raid_group_size

=item C<< 'is-flash-pool-caching-enabled' => $string >>

Is Flash Pool Caching possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'excluded-from-balancer' => $string >>

Is the Capacity Balancer Excluded possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'is-balance-eligible' => $string >>

Is Eligible for the Balancer possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'space-full-threshold-percent' => $string >>

Aggregate Full Threshold Percent
Applicable for CMode CLI

=item C<< 'space-nearly-full-threshold-percent' => $string >>

Aggregate Nearly Full Threshold Percent
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"storage aggregate modify" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::StorageAggregate::AlreadyModified>

This type of exception is thrown when an attempt is made to modify the
attributes of an aggregate into a state it's already in.

=item C<NACL::C::Exceptions::StorageAggregate::CurrentlyBusy>

This type of exception is thrown when an attempt is made to modify the
attributes of an aggregate when the aggregate is "dynamically growing state" or busy.

=item C<NACL::C::Exceptions::StorageAggregate::AggregateHasMountedVolumes>

This type of exception is thrown when an attempt is made to modify an aggregate to an offline state.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'unmount-volumes' => {type => SCALAR,  optional => 1},
            'keep-same-uuid'  => {type => BOOLEAN, optional => 1},
            _optional_scalars(
                qw(state raidtype maxraidsize ignore-inconsistent
                    raid-block-checksum raid-zoned-checksum ha-policy
                    snapshot-autodelete thorough-scrub cifs-delay-seconds
                    hybrid-enabled raid-lost-write percent-snapshot-space free-space-realloc
                    auto-snapshots

                    resync-snap-time resyncsnaptime

                    snapshot-reserve

                    max-write-alloc-blocks

                    force-hybrid-enabled

                    disktype
                    nosnap
                    lost_write_protect
                    snapmirrored
                    fs_size_fixed
                    no_delete_log
                    metadata_checking
                    root

                    pre-check

                    cache-raid-group-size

                    is-flash-pool-caching-enabled

                    excluded-from-balancer

                    is-balance-eligible

                    space-full-threshold-percent space-nearly-full-threshold-percent
                    )
            ),
        },
        cli_cmd => 'storage aggregate modify',
    );

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_modify_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_modify_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode|Nodescope',
                },
                {   method    => '_modify_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
                {   method    => '_modify_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::StorageAggregate::modify',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        my $text      = $exception->text();
        NACL::C::Exceptions::StorageAggregate::AggregateHasMountedVolumes->detect_convert_and_throw(exception => $exception);
        foreach my $type (qw(ContainsVolumes DoesNotExist)) {
            my $exception_name = "NACL::C::Exceptions::StorageAggregate::$type";
            $exception_name->detect_convert_and_throw(exception => $exception);
        }
        if ($text =~ /is already/) {
            $Log->exit() if $may_exit;
            NACL::C::Exceptions::StorageAggregate::AlreadyModified
                ->convert_and_throw(exception => $exception);
        } elsif ($text =~ /(aggregate is currently busy|dynamic create|is in use by other operations)/) {
            $Log->exit() if $may_exit;
            NACL::C::Exceptions::StorageAggregate::CurrentlyBusy
                ->convert_and_throw(exception => $exception);
        }else {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };

    $Log->exit() if $may_exit;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    my $unmount_volume = delete $opts{'unmount-volumes'};
    my $rex = qr/.*?unmount all of the volumes hosted by the aggregate.*/i;
    my $prompt_answers = [$rex => 'yes',];
    my $apiset = $opts{apiset};

    my $set_confirmation = 0;
    if (defined $unmount_volume && $unmount_volume eq 'true') {
        $apiset->set(confirmation => 'on');
        $set_confirmation = 1;
    }

    try {
        $pkg_or_obj->_base_backend_cmode_cli(
            %opts,
            api => 'storage_aggregate_modify',
            additional_args => {
                'connectrec-match_table' => $prompt_answers,
            },
        );
    } finally {
        if ($set_confirmation) {
            $apiset->set(confirmation => 'off');
        }
    };

    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $state             = delete $opts{state};
    my $command;

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
    );

    # Copy and translate the option names
    my $aggropts_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => _aggr_options_copy(),
        map    => _aggr_options_map(),
    );

    _translate_options_values(
        hashref   => $aggropts_args,
        direction => 'CLUSTERTOSEVEN',
        values_to_translate =>
            _option_values_to_translate(mode_of_options => 'SEVEN')
    );

    # Call 'aggr options' to set each of the options provided
    while (my ($key, $value) = each %$aggropts_args) {
        $apiset->aggr_options(
            aggregate      => $opts{aggregate},
            'option-name'  => $key,
            'option-value' => $value
        );
    } ## end while ( my ( $key, $value...

    if (defined $state) {
        try {
            if ($state eq 'online') {
                $command = 'aggr_online';
                my %args = (%$aggr_args, force => 1);

                $apiset->$command(%args);
            } elsif ($state eq 'offline' || $state eq 'restricted') {
                my %args = (%$aggr_args, force => 1);
                $pkg_or_obj->_hash_copy(
                    source => \%opts,
                    target => \%args,
                    copy   => [qw(cifs-delay-seconds)]
                );
                if (defined $opts{'unmount-volumes'}
                    && $opts{'unmount-volumes'} ne 'false')
                {
                    $args{'aggr-with-vols'} = 1;
                }
                if ($state eq 'offline') {
                    $command = 'aggr_offline';
                } else {
                    $command = 'aggr_restrict';
                }

                $apiset->$command(%args);
            } else {
                $Log->exit() if $may_exit;
                NATE::Exceptions::Argument->throw(
                          "Modify can change aggregate state to either"
                        . " online or offline or restricted. Other "
                        . "modifications to Type,Ignore inconsistent,"
                        . "Raid block checksum,snapshot autodelete, "
                        . "Thorough scrub,HA-Policy are not supported"
                        . "for 7mode.");
            } ## end else [ if ( $state eq 'online')
        } ## end try
        catch NACL::APISet::Exceptions::CommandFailedException with {
            _modify_7m_exception_handler(
                exception         => $_[0],
                command_interface => $command_interface,
                command           => $command
            );
        };
    } ## end if ( defined $state )
    $Log->exit() if $may_exit;
} ## end sub _modify_7mode_cli

sub _modify_7mode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $state             = delete $opts{state};

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
    );
    my $command;
    try {
        if ($state eq 'online') {
            $command = 'aggr_online';
            my %args = %$aggr_args;
            $pkg_or_obj->_hash_copy(
                source => \%opts,
                target => \%args,
                copy   => ['keep-same-uuid']
            );

            $apiset->$command(%args);
        } elsif ($state eq 'offline' || $state eq 'restricted') {
            my %args = %$aggr_args;
            $pkg_or_obj->_hash_copy(
                source => \%opts,
                target => \%args,
                copy   => [qw(cifs-delay-seconds unmount-volumes)]
            );
            if ($state eq 'offline') {
                $command = 'aggr_offline';
            } else {
                $command = 'aggr_restrict';
            }

            $apiset->$command(%args);
        } else {
            NATE::Exceptions::Argument->throw(
                      "Modify can change aggregate state to either"
                    . " online or offline or restricted. Other "
                    . "modifications to Type,Ignore inconsistent,"
                    . "Raid block checksum,snapshot autodelete, "
                    . "Thorough scrub,HA-Policy are not supported"
                    . "for 7mode.");
        } ## end else [ if ( $state eq 'online')
    } ## end try
    catch NACL::APISet::Exceptions::CommandFailedException with {
        _modify_7m_exception_handler(
            exception         => $_[0],
            command_interface => $command_interface,
            command           => $command
        );
    };
    $Log->exit() if $may_exit;
} ## end sub _modify_7mode_zapi

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $state             = delete $opts{state};
    my $aggregate         = delete $opts{aggregate};

    my %options = (aggregate => $aggregate);
    if ($state) {

        # Modify the state first
        if ($state eq 'online') {
            $apiset->aggr_online(%options);
        } elsif ($state eq 'offline') {
            $apiset->aggr_offline(%options, 'unmount-volumes' => 'true');
        } elsif ($state eq 'restricted') {
            $apiset->aggr_restrict(%options, 'unmount-volumes' => 'true');
        } else {
            NATE::Exceptions::Argument->throw(
                      "'state' can be any of 'online', "
                    . "'offline' or 'restricted' but it was provided as '$state'"
            );
        } ## end else [ if ( $state eq 'online')
    } ## end if ($state)
    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate raidtype)],
        map    => {
            "maxraidsize"            => "raidsize",
            "hybrid-enabled"         => "hybrid_enabled",
            "ignore-inconsistent"    => "ignore_inconsistent",
            "raid-zoned-checksum"    => "raid_zoned",
            "raid-block-checksum"    => "raid_cv",
            "snapshot-autodelete"    => "snapshot_autodelete",
            "thorough-scrub"         => "thorough_scrub",
            "ha-policy"              => "ha_policy",
            "raid-lost-write"        => "raid_lost_write",
            "free-space-realloc"     => "free_space_realloc",
            "snapshot-reserve"       => "snapshot_reserve",
            "max-write-alloc-blocks" => "max_write_alloc_blocks",
            "percent-snapshot-space" => "percent_snapshot_space",
        },
    );
    %opts = %$aggr_args;

    while (my ($name, $value) = each %$aggr_args) {
        $apiset->aggr_set_option(
            aggregate      => $aggregate,
            'option-name'  => $name,
            'option-value' => $value
        );

    } ## end while ( my ( $name, $value...
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

=head2 add_members

  NACL::C::StorageAggregate->add_members(command_interface=>$ci,
                             aggregate=>$aggr_name,...);

                            or

  $aggregate_object->add_members(...);

(Class or instance method) This method adds aggregate memebers to an
aggregate element.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

Applicable for CMode CLI

=item C<< disklist => [$disk1, $disk2, $disk3, ... ] >>

Applicable for CMode CLI

=item C<< nodes => [$node1_name, $node2_name, $node3_name, ... ] >>

Applicable for CMode CLI

=item C<< "add-members-volume" => $string >>

Applicable for CMode CLI

=item C<< foreground => $string >>

Applicable for CMode CLI

=item C<< "method-timeout"=> $scalar >>

(Optional)The new timeout value.

=item C<< '*' >>

This method takes all other options supported by
"storage aggregate add-members" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub add_members {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            nodes                => {type => ARRAYREF},
            "add-member-volumes" => {type => SCALAR},
            disklist             => {type => ARRAYREF, optional => 1},
            _optional_scalars(qw(foreground )),
        },
        api => 'storage_aggregate_add_members',
    );
    if (defined($opts{disklist})) {
        $opts{disklist} = join(",", @{$opts{disklist}});
    }
    if (defined($opts{nodes})) {
        $opts{nodes} = join(",", @{$opts{nodes}});
    }
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_add_members_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::StorageAggregate::add_members',
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub add_members

sub _add_members_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
     $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api     => 'storage_aggregate_add_members',
    );

    $Log->exit() if $may_exit;
} ## end sub _add_members_cmode_cli

=head2 add_disks

  NACL::C::StorageAggregate->add_disks(command_interface=>$ci,
                             aggregate=>$aggr_name,...);

                            or

  $aggregate_object->add_disks(...);

(Class or instance method)This method adds disks to an aggregate.

Uses a 7Mode CLI/ZAPI, CMode CLI/ZAPI APISet.

Invokes "aggr-add" API for CMode ZAPI, 7Mode ZAPI.

Invokes "aggr add" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: aggregate

7M CLI : aggregate name (Programmatic name: "aggregate")

7M ZAPI: aggregate

=item C<< disklist => [$disk1, $disk2, $disk3, ... ] >>

This is an arrayref of disk names to be added to the aggregate.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: disks

7M CLI : -d <value> (Programmatic name: "disklist")

7M ZAPI: disks

=item C<< "mirror-disks" => [$disk1, $disk2, $disk3, ... ] >>

=item C<< force => $boolean >>

Disks in a plex are not normally permitted to span spare pools. This behavior is
overridden with this option when it is set.
Permissible values: true/false or 1/0.

Applicable for CMode CLI, 7Mode CLI/ZAPI.

Maps to:

7Mode CLI: "-f" of "aggr add"

7Mode ZAPI: "force-spare-pool" of "aggr-add"

=item C<< diskcount => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: disk-count

7M CLI : disk count (Programmatic name: "disk-count")

7M ZAPI: disk-count

=item C<< raidgroup => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: raid-group

7M CLI : -g (Programmatic name: "raid-group")

7M ZAPI: raid-group

=item C<< "allow-mixed" => $string >>

Applicable for CMode CLI.

=item C<< "is-display-command" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -n (Programmatic name: "is-display-command")

7M ZAPI: is-display-command

=item C<< "start-tag" => $string >>

Applicable for 7Mode ZAPI.

Maps to:

7M ZAPI: start-tag

=item C<< "upgrade-64bit-mode" => $string >>

This option is deprecated. The option "64bit-upgade" should be used instead.
Support for this option will be removed in a future sprint.

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -64bit-upgrade (Programmatic name: "64bit-upgrade")

7M ZAPI: upgrade-64bit-mode

=item C<< "force-spare-pool" => $string >>

This option is deprecated. The option "force" should be used instead. Support
for "force-spare-pool" will be removed in a future sprint.

Applicable for 7Mode ZAPI.

Maps to:

7M ZAPI: force-spare-pool

=item C<< "disk-size-with-unit" => $string >>

Applicable for 7Mode ZAPI.

Maps to:

7M ZAPI: disk-size-with-unit

=item C<< "disk-size" => $string >>

Size of disks in bytes.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: disk-size

7M CLI : disk size (Programmatic name: "disk-size")

7M ZAPI: disk-size

=item C<< "group-selection-mode" => $string >>

Applicable for CMode CLI, 7Mode ZAPI.

Maps to:

7M ZAPI: group-selection-mode

=item C<< "max-records" => $string >>

=item C<< "64bit-upgrade" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: upgrade-64bit-mode

7M CLI : -64bit-upgrade <value> (Programmatic name: "64bit-upgrade")

7M ZAPI: upgrade-64bit-mode

The option values are mapped

Mode for upgrade to 64 bit

'normal' maps to:
    CMode ZAPI: "grow_none" of "aggr_add"

'grow-reserved' maps to:
    CMode ZAPI: "grow_reserved" of "aggr_add"

'grow-all' maps to:
    CMode ZAPI: "grow_all" of "aggr_add"

=item C<< 'bad-disklist' => $string >>

Applicable for CMode CLI.

=item C<< '64bit-upgrade-zapi' => $string >>

Applicable for CMode CLI.

=item C<< 'disksize-blk' => $string >>

Applicable for CMode CLI.

=item C<< 'disksize-in-unit' => $string >>

Applicable for CMode CLI.

=item C<< 'disktype' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: disk-type

7M CLI : -T <value> (Programmatic name: "disk-type")

=item C<< 'chksumstyle' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: checksum-style

7M CLI : -c (Programmatic name: "checksum-style")

=item C<< 'selected-disklist' => $string >>

Applicable for CMode CLI.

=item C<< 'cookie' => $string >>

Applicable for CMode CLI.

=item C<< 'simulate' => $string >>

Applicable for CMode CLI.

=item C<< 'mirror-disklist' => [ $mirror-disk1, $mirror-disk2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI.

=item C<< 'ignore-pool' => $string >>

Applicable for CMode CLI.

=item C<< 'allow-mixed-rpms' => $string >>

Applicable for CMode CLI.

=item C<< 'group-select-mode' => $string >>

Applicable for CMode CLI.

=item C<< 'allow-same-carrier' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: allow-same-carrier

=item C<< 'allow-mixed-rpm' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: allow-mixed-rpm

=item C<< 'ignore-pool-checks' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ignore-pool-checks

=item C<< 'raidtype' => $string >>

Applicable for CMode CLI.

=item C<< 'pre-check-results-add' => $string >>

Applicable for CMode CLI

=item C<< 'pre-check' => $string >>

Applicable for CMode CLI

=item C<< 'force-cache-size' => $string >>

*Force Addition of SSD Disks to Hybrid Enabled Aggregate

Applicable for CMode CLI

=item C<< 'cache-raid-group-size' => $string >>

RAID Group Size for Cache Tier
Applicable for CMode CLI

=item C<< 'mirror-partitionlist' => [ $mirror-partition1, $mirror-partition2, ... ] >>

Partitions for Mirrored Plex
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI

=item C<< 'partitionlist' => [ $partition1, $partition2, ... ] >>

Partitions
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI

=item C<< 'allocation-units' => $string >>

Allocation Units ranges from 1 to 4
Applicable for CMode CLI

=item C<< 'storage-pool' => $string >>

Shared Storage Pool
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"storage aggregate add-disks" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::StorageAggregate::InsufficientDisksToStartNewRaidGroup>

This exception is thrown when we attempt to add disks to an aggregate and it fails
because the count of disks is not sufficient to create a new raidgroup.

=item C<NACL::C::Exceptions::StorageAggregate::NoMatchingDisksAvailable>

This exception is thrown when there are no matching disks available

=item C<NACL::C::Exceptions::StorageAggregate::StoragePoolBusy>

This exception is thrown when provisioning disks from the Storage pool
to an aggregate fails with 'Storage Pool is busy with another operation...'

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub add_disks {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "mirror-partitionlist" => {type => ARRAYREF, optional => 1},
            "partitionlist"        => {type => ARRAYREF, optional => 1},
            "mirror-disklist"      => {type => ARRAYREF, optional => 1},
            disklist               => {type => ARRAYREF, optional => 1},
            "mirror-disks"         => {type => ARRAYREF, optional => 1},
            force                  => {type => BOOLEAN,  optional => 1},
            _optional_scalars(
                qw(diskcount raidgroup allow-mixed is-display-command
                    start-tag upgrade-64bit-mode diskcount raidgroup
                    is-display-command force-spare-pool disk-size-with-unit
                    disk-size group-selection-mode max-records 64bit-upgrade
                    bad-disklist 64bit-upgrade-zapi disksize-blk disksize-in-unit disktype
                    chksumstyle selected-disklist cookie simulate

                    ignore-pool allow-mixed-rpms

                    group-select-mode

                    allow-same-carrier

                    allow-mixed-rpm ignore-pool-checks

                    raidtype

                    pre-check-results-add pre-check

                    force-cache-size

                    cache-raid-group-size

                    allocation-units storage-pool
                    )
            )
        },
        cli_cmd => 'storage aggregate add-disks',
    );

    if (exists $opts{'upgrade-64bit-mode'}) {
        $pkg_or_obj->_print_deprecated_option_used_warning(
            deprecated_option => 'upgrade-64bit-mode',
            new_option        => '64bit-upgrade'
        );
    } ## end if ( exists $opts{'upgrade-64bit-mode'...

    if (exists $opts{'force-spare-pool'}) {
        $pkg_or_obj->_print_deprecated_option_used_warning(
            deprecated_option => 'force-spare-pool',
            new_option        => 'force'
        );
    } ## end if ( exists $opts{'force-spare-pool'...

    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_add_disks_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_add_disks_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_add_disks_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
                {   method    => '_add_disks_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    check     => '_is_partitionlist_specified',
                },
            ],
            frontend => 'NACL::C::StorageAggregate::add_disks',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = shift;
        NACL::C::Exceptions::StorageAggregate::InsufficientDisksToStartNewRaidGroup
            ->detect_convert_and_throw('exception' => $exception,);
        NACL::C::Exceptions::StorageAggregate::NoMatchingDisksAvailable
            ->detect_convert_and_throw('exception' => $exception,);
        NACL::C::Exceptions::StorageAggregate::StoragePoolBusy
            ->detect_convert_and_throw('exception' => $exception,);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub add_disks

sub _add_disks_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api     => 'storage_aggregate_add_disks',
    );

    $Log->exit() if $may_exit;
} ## end sub _add_disks_cmode_cli

sub _add_disks_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    if (defined($opts{disklist})) {
        $opts{disklist} = join ' ', @{$opts{disklist}};
    }

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate is-display-command force 64bit-upgrade)],
        map    => {
            'disktype'           => 'disk-type',
            'diskcount'          => 'disk-count',
            'raidtype'           => 'raid-type',
            'raidgroup'          => 'raid-group',
            'disklist'           => 'disks',
            'upgrade-64bit-mode' => '64bit-upgrade',
            'chksumstyle'        => 'checksum-style',
        }
    );
    %opts = %$aggr_args;

    $pkg_or_obj->_handle_boolean_options(
        api_options     => \%opts,
        boolean_options => [qw(force)],
        is_7m_cli       => 1
    );

    # "aggr add" accepts size in GB
    if (exists $opts{'disk-size'}) {
        $opts{'disk-size'} = $opts{'disk-size'} / (1024 * 1024);
    }
    $apiset->aggr_add(%opts);
    $Log->exit() if $may_exit;
} ## end sub _add_disks_7mode_cli

sub _add_disks_7mode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset    = delete $opts{apiset};
    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [
            qw(aggregate is-display-command force-spare-pool
                disk-size-with-unit disk-size group-selection-mode
                max-records mirror-disks start-tag upgrade-64bit-mode)
        ],
        map => {
            'diskcount'     => 'disk-count',
            'raidgroup'     => 'raid-group',
            'disklist'      => 'disks',
            '64bit-upgrade' => 'upgrade-64bit-mode',
            'force'         => 'force-spare-pool',
        }
    );
    %opts = %$aggr_args;

    $pkg_or_obj->_handle_boolean_options(
        api_options     => \%opts,
        boolean_options => [qw(force-spare-pool)],
        is_7m_cli       => 0
    );

    # "aggr-add" accepts size in 4KB blocks
    if (exists $opts{'disk-size'}) {
        $opts{'disk-size'} = $opts{'disk-size'} / (4 * 1024);
    }

    $apiset->aggr_add(%opts);
    $Log->exit() if $may_exit;
} ## end sub _add_disks_7mode_zapi

sub _add_disks_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    
    if (defined $opts{'upgrade-64bit-mode'}) {
        $opts{'upgrade-64bit-mode'} = _map_modes($opts{'upgrade-64bit-mode'});
    }
    
    # The mapping for disklist should be [make_zapi_array('disks'), 'name'].
    # However, because of an existing bug in _hash_copy, the mapping do not
    # work and fails with the following error:
    # The 'value' parameter ("ARRAY(0xe97f260)") to NACL::ComponentUtils::__ANON__
    # was an 'arrayref', which is not one of the #allowed types: scalar.
    # Hence, we are manually constructing the mapping here.

    if (defined $opts{disklist}) {
        # The structure formed will be like:
        # disks => [ {name => $disk[0]}, {name => $disk[1]}, ... ]
        my @disk_info = map { {name => $_} } @{$opts{disklist}};
        $opts{additional_args} = {disks => \@disk_info};
    }


    $pkg_or_obj->_base_backend_cmode_zapi(
        %opts,
        api  => 'aggr-add',
        copy => [qw(aggregate allocation-units allow-mixed-rpm
            allow-same-carrier cache-raid-group-size disksize-in-unit
            force-cache-size group-selection-mode ignore-pool-checks
            pre-check storage-pool)],
        map  => {
            raidtype          => 'cache-raid-type',
            chksumstyle       => 'checksum-style',
            diskcount         => 'disk-count',
            'disksize-blk'    => 'disk-size',
            disktype          => 'disk-type',
            'mirror-disklist' => 'mirror-disks',
            raidgroup         => 'raid-group',
            '64bit-upgrade'   => 'upgrade-64bit-mode',
        },
    );

    $Log->exit() if $may_exit;
} ## end sub _add_disks_cmode_zapi

#The Param partitionlist is supported only in CLI
#Hence there is no support for this in ZAPI, 
#if it is passed it should not go through ZAPI

sub _is_partitionlist_specified {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    if ( defined $opts{partitionlist} ) {
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw(
            "Param partitionlist is not supported in ZAPI" );
    }
    $Log->exit() if $may_exit;
}

=head2 remove_members

  NACL::C::StorageAggregate->remove_members(command_interface=>$ci,
                             aggregate=>$aggr_name,...);

                            or

  $aggregate_object->remove_members(...);

(Class or instance method) This method removes aggregate members from
an aggregate element.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

Applicable for CMode CLI

=item C<< members => [$member1, $member2, $member3, ...] >>

Applicable for CMode CLI

=item C<< foreground => $string >>

Applicable for CMode CLI

=item C<< "method-timeout"=> $scalar >>

(Optional)The new timeout value.

=item C<< '*' >>

This method takes all other options supported by
"storage aggregate remove-members" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub remove_members {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            members => {type => ARRAYREF | SCALAR},
            _optional_scalars(qw(foreground)),
        },
        api => 'storage_aggregate_remove_members',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_remove_members_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::StorageAggregate::remove_members',
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub remove_members

sub _remove_members_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api     => 'storage_aggregate_remove_members',
    );
    $Log->exit() if $may_exit;
} ## end sub _remove_members_cmode_cli

=head2 rename

  NACL::C::StorageAggregate->rename(command_interface=>$ci,
                             aggregate=>$aggr_name,newname => $new_name );

                            or

  $aggregate_object->rename(newname => $new_name);

(Class or instance method) This method renames an aggregate element.

Uses a 7Mode CLI/ZAPI, CMode/ZAPI APIset.

Invokes "aggr-rename" API for CMode ZAPI.

Invokes "aggr rename" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: aggregate

7M CLI : aggregate name (Programmatic name: "aggregate")

=item C<< newname => $string >>

(Required) The new name of the aggregate element.

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: new-aggregate-name

7M CLI : new aggregate name (Programmatic name: "new-aggregate-name")

=item C<< "job_component"=> \$scalar >>

=item C<< '*' >>

This method takes all other options supported by
"storage aggregate rename" command in ONTAP.

Applicable for CMode CLI.

Applicable for CMode CLI

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub rename {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            newname       => {type => SCALAR},
            job_component => {type => SCALARREF, optional => 1},
        },
        cli_cmd => 'storage aggregate rename',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_rename_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_rename_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_rename_zapi',
                interface => 'ZAPI',
            },
        ],
        frontend => 'NACL::C::StorageAggregate::rename',
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub rename

sub _rename_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api     => 'storage_aggregate_rename',
    );

    $pkg_or_obj->aggregate($opts{newname}) if ref($pkg_or_obj);
    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_cli

sub _rename_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset    = delete $opts{apiset};
    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "aggregate" => "aggregate",
            "newname"   => "new-aggregate-name",
        }
    );
    %opts = %$aggr_args;

    $apiset->aggr_rename(%opts);
    $pkg_or_obj->aggregate($opts{"new-aggregate-name"}) if ref($pkg_or_obj);
    $Log->exit() if $may_exit;
} ## end sub _rename_7mode_cli

sub _rename_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset    = delete $opts{apiset};
    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "aggregate" => "aggregate",
            "newname"   => "new-aggregate-name",
        }
    );
    %opts = %$aggr_args;

    $apiset->aggr_rename(%opts);
    $pkg_or_obj->aggregate($opts{"new-aggregate-name"}) if ref($pkg_or_obj);
    $Log->exit() if $may_exit;
} ## end sub _rename_zapi

=head2 scrub

  NACL::C::StorageAggregate->scrub(command_interface=>$ci,
                             aggregate=>$aggr_name,... );

                            or

  $aggregate_object->scrub(...);

(Class or instance method) This method performs various scrub functions
on an aggregate element.

Supports 7Mode CLI/ZAPI, CMode CLI/ZAPI.

Invokes "aggr-scrub-start" API for CMode ZAPI or 7Mode ZAPI if action is start

Invokes "aggr-scrub-stop" API for CMode ZAPI or 7Mode ZAPI if action is stop

Invokes "aggr-scrub-pause" API for CMode ZAPI or 7Mode ZAPI if action is pause

Invokes "aggr-scrub-resume" API for CMode ZAPI or 7Mode ZAPI if action is resume

Invokes "aggr-scrub-suspend" API for CMode ZAPI or 7Mode ZAPI if action is suspend

Invokes "aggr scrub start" command for 7Mode CLI
        "aggr scrub stop" command for 7Mode CLI
        "aggr scrub pause" command for 7Mode CLI
        "aggr scrub resume" command for 7Mode CLI

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

Applicable for CMode CLI/ZAPI or 7Mode CLI/ZAPI

Maps to:

CM ZAPI: aggregate

7M CLI : aggregate name (preogrammatic name: aggregate)

7M ZAPI: aggregate

=item C<< action => $string >>

(Required) Takes the following values : start , stop , resume and suspend.

Applicable for CMode CLI

=item C<< raidgroup => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: raidgroup

=item C<< 'plex' => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: plex

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"storage aggregate scrub" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub scrub {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            action => {type => SCALAR},
            _optional_scalars(
                qw(raidgroup
                    plex
                    )
            ),
        },
        cli_cmd => 'storage aggregate scrub',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_scrub_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_scrub_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_scrub_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
            {   method    => '_scrub_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            }
        ],
        frontend => 'NACL::C::StorageAggregate::scrub',
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub scrub

sub _scrub_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api     => 'storage_aggregate_scrub',
    );

    $Log->exit() if $may_exit;
} ## end sub _scrub_cmode_cli

sub _scrub_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset  = delete $opts{apiset};
    my $command = delete $opts{action};

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
    );
    %opts    = %$aggr_args;
    $command = "aggr_scrub_" . $command;
    $Log->debug("Opts to API call $command are: \n" . Dumper(\%opts));
    $apiset->$command(%opts);
    $Log->exit() if $may_exit;
} ## end sub _scrub_7mode_cli

sub _scrub_7mode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset  = delete $opts{apiset};
    my $command = delete $opts{action};

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
    );
    %opts    = %$aggr_args;
    $command = "aggr_scrub_" . $command;
    $apiset->$command(%opts);
    $Log->exit() if $may_exit;
} ## end sub _scrub_7mode_zapi

sub _scrub_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};
    my $command;
    if ($opts{action}) {
        $command = $opts{action};
    } else {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument->throw(
            "Mandatory argument 'action' was not provided");
    }

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw( aggregate plex raidgroup )],
    );
    $command = "aggr_scrub_" . $command;
    $apiset->$command(%$aggr_args);
    $Log->exit() if $may_exit;
} ## end sub _scrub_cmode_zapi

=head2 verify_start

  NACL::C::StorageAggregate->verify_start(command_interface=>$ci,
                             aggregate=>$aggr_name,...);

                            or

  $aggregate_object->verify_start(...);

(Class or instance method) This method starts verifying  an aggregate element.

Uses a 7Mode CLI/ZAPI , CMode ZAPI APISet.

Invokes "aggr-verify-start" API for CMode ZAPI.

Invokes "aggr verify start" command for 7Mode CLI.

Invokes "aggr-verify-start" API for 7Mode ZAPI.

Note: To get mode agnostic 7Mode and CMode support, we recommend you to use
L<NACL::C::StorageAggregate->verify|lib-NACL-C-StorageAggregate-pm/verify>

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

Applicable for 7Mode CLI/ZAPI or CMode ZAPI.

Maps to:

7M ZAPI: aggregate

7M CLI: aggregate name (programmatic name: aggregate)

CM ZAPI: aggregate

=item C<< "plex-number"  => $string >>

Applicable for 7Mode CLI/ZAPI or CMode ZAPI.

Maps to:

7M ZAPI: fix-plex

7M CLI: -f <value> (programmatic name: plex-number)

CM ZAPI: fix-plex

=item C<< "log-only"  => $string >>

Applicable for 7Mode CLI/ZAPI or CMode ZAPI.

Maps to:

7M ZAPI: log-only

CM ZAPI: log-only

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub verify_start {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {_optional_scalars(qw(plex-number log-only))},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_verify_start_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_verify_start_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
            {   method    => '_verify_start_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub verify_start

sub _verify_start_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset    = delete $opts{apiset};
    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate plex-number)],
    );
    %opts = %$aggr_args;
    $Log->debug(
        "Opts to API call 'aggr_verify_start' are: \n" . Dumper(\%opts));
    $apiset->aggr_verify_start(%opts);
    $Log->exit() if $may_exit;
} ## end sub _verify_start_7mode_cli

sub _verify_start_7mode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset    = delete $opts{apiset};
    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate log-only)],
        map    => {"plex-number" => "fix-plex",}
    );
    %opts = %$aggr_args;
    $apiset->aggr_verify_start(%opts);
    $Log->exit() if $may_exit;
} ## end sub _verify_start_7mode_zapi

sub _verify_start_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);

    my $apiset = $opts{apiset};

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate log-only)],
        map    => {"plex-number" => "fix-plex",}
    );
    $apiset->aggr_verify_start(%$aggr_args);
    $Log->exit() if $may_exit;
} ## end sub _verify_start_cmode_zapi

=head2 verify_stop

  NACL::C::StorageAggregate->verify_stop(command_interface=>$ci,
                             aggregate=>$aggr_name);

                            or

  $aggregate_object->verify_stop();

(Class or instance method) This method stops verifying  an aggregate element.

Uses a 7Mode CLI/ZAPI , CMode ZAPI APISet.

Invokes "aggr-verify-stop" API for CMode ZAPI.

Invokes "aggr verify stop" command for 7Mode CLI.

Invokes "aggr-verify-stop" API for 7Mode ZAPI.

Note: To get mode agnostic 7Mode and CMode support, we recommend you to use
L<NACL::C::StorageAggregate->verify|lib-NACL-C-StorageAggregate-pm/verify>

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

Applicable for 7Mode CLI/ZAPI or CMode ZAPI.

Maps to:

7M ZAPI: aggregate

7M CLI: aggregate name (programmatic name: aggregate)

CM ZAPI: aggregate

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub verify_stop {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_verify_stop_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_verify_stop_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
            {   method    => '_verify_stop_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub verify_stop

sub _verify_stop_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
    );
    %opts = %$aggr_args;
    $Log->debug(
        "Opts to API call 'aggr_verify_stop' are: \n" . Dumper(\%opts));
    $apiset->aggr_verify_stop(%opts);
    $Log->exit() if $may_exit;
} ## end sub _verify_stop_7mode_cli

sub _verify_stop_7mode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
    );
    %opts = %$aggr_args;
    $apiset->aggr_verify_stop(%opts);
    $Log->exit() if $may_exit;
} ## end sub _verify_stop_7mode_zapi

sub _verify_stop_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
    );
    $apiset->aggr_verify_stop(%$aggr_args);
    $Log->exit() if $may_exit;
} ## end sub _verify_stop_cmode_zapi

=head2 verify_suspend

  NACL::C::StorageAggregate->verify_suspend(command_interface=>$ci,
                             aggregate=>$aggr_name);

                            or

  $aggregate_object->verify_suspend();

(Class or instance method) This method suspends verifying  an aggregate element.

Uses a 7Mode CLI/ZAPI , CMode ZAPI APISet.

Invokes "aggr-verify-suspend" API for CMode ZAPI.

Invokes "aggr verify suspend" command for 7Mode CLI.

Invokes "aggr-verify-suspend" API for 7Mode ZAPI.

Note: To get mode agnostic 7Mode and CMode support, we recommend you to use
L<NACL::C::StorageAggregate->verify|lib-NACL-C-StorageAggregate-pm/verify>

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

Applicable for 7Mode CLI/ZAPI or CMode ZAPI.

Maps to:

7M ZAPI: aggregate

7M CLI: aggregate name (programmatic name: aggregate)

CM ZAPI: aggregate

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub verify_suspend {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_verify_suspend_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_verify_suspend_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
            {   method    => '_verify_suspend_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub verify_suspend

sub _verify_suspend_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;

    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
    );
    %opts = %$aggr_args;
    $Log->debug(
        "Opts to API call 'aggr_verify_suspend' are: \n" . Dumper(\%opts));
    $apiset->aggr_verify_suspend(%opts);
    $Log->exit() if $may_exit;
} ## end sub _verify_suspend_7mode_cli

sub _verify_suspend_7mode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;

    my %opts = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
    );
    %opts = %$aggr_args;
    $apiset->aggr_verify_suspend(%opts);
    $Log->exit() if $may_exit;
} ## end sub _verify_suspend_7mode_zapi

sub _verify_suspend_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
    );
    $apiset->aggr_verify_suspend(%$aggr_args);
    $Log->exit() if $may_exit;
} ## end sub _verify_suspend_cmode_zapi

=head2 verify_resume

  NACL::C::StorageAggregate->verify_resume(command_interface=>$ci,
                             aggregate=>$aggr_name);

                            or

  $aggregate_object->verify_resume();

(Class or instance method) This method resumes verifying  an aggregate element.

Uses a 7Mode CLI/ZAPI , CMode ZAPI APISet.

Invokes "aggr-verify-resume" API for CMode ZAPI.

Invokes "aggr verify resume" command for 7Mode CLI.

Invokes "aggr-verify-resume" API for 7Mode ZAPI.

Note: To get mode agnostic 7Mode and CMode support, we recommend you to use
L<NACL::C::StorageAggregate->verify|lib-NACL-C-StorageAggregate-pm/verify>

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

Applicable for 7Mode CLI/ZAPI or CMode ZAPI.

Maps to:

7M ZAPI: aggregate

7M CLI: aggregate name (programmatic name: aggregate)

CM ZAPI: aggregate

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub verify_resume {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_verify_resume_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_verify_resume_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
            {   method    => '_verify_resume_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub verify_resume

sub _verify_resume_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
    );
    %opts = %$aggr_args;
    $Log->debug(
        "Opts to API call 'aggr_verify_resume' are: \n" . Dumper(\%opts));
    $apiset->aggr_verify_resume(%opts);
    $Log->exit() if $may_exit;
} ## end sub _verify_resume_7mode_cli

sub _verify_resume_7mode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
    );
    %opts = %$aggr_args;
    $apiset->aggr_verify_resume(%opts);
    $Log->exit() if $may_exit;
} ## end sub _verify_resume_7mode_zapi

sub _verify_resume_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts   = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
    );
    $apiset->aggr_verify_resume(%$aggr_args);
    $Log->exit() if $may_exit;
} ## end sub _verify_resume_cmode_zapi

=head2 offline

  NACL::C::StorageAggregate->offline(command_interface=>$ci,aggregate =>$aggr_name,
                           ...);

                            or

  $aggregate_object->offline(...);

(Class or instance method) This method offlines a StorageAggregate element.

Uses a 7Mode CLI, CMode CLI/ZAPI APIset.

Invokes "aggr-offline" API for CMode ZAPI.

Invokes "aggr offline" command for 7Mode CLI.

Same operation can be performed using modify method also.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=> $string >>

(Required) Name of aggregate.

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: aggregate

7M CLI : aggregate name (Programmatic name: "aggregate")

=item C<< "unmount-volumes"=>$boolean >>

Unmount Volumes - A boolean value (true or false)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: unmount-volumes

=item C<< 'plex' => $string >>

Applicable for CMode CLI.

=item C<< 'hidden-plex' => $string >>

=item C<< 'force-offline' => $string >>

Force offline
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"storage aggregate offline" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub offline {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            aggregate     => {type => SCALAR},
            _optional_scalars qw (unmount-volumes
                plex hidden-plex

                force-offline
                ),
        },
        cli_cmd => 'storage aggregate offline',
    );
    $Log->exit() if $may_exit;
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_offline_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_offline_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_offline_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::StorageAggregate::offline',
    );
    return $ret;
} ## end sub offline

sub _offline_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api     => 'storage_aggregate_offline',
    );

    $Log->exit() if $may_exit;
} ## end sub _offline_cmode_cli

sub _offline_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    $pkg_or_obj->_base_backend_cmode_zapi(
        %opts,
        api     => 'aggr-offline',
    );

    $Log->exit() if $may_exit;
} ## end sub _offline_cmode_zapi

sub _offline_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [qw(aggregate)];

    my %mode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%mode_args,
        copy   => $allowed_7mode_options,
    );

    $apiset->aggr_offline(%mode_args);

    $Log->exit() if $may_exit;
} ## end sub _offline_7mode_cli

=head2 online

  NACL::C::StorageAggregate->online(command_interface=>$ci,aggregate =>$aggr_name,
                           ...);

                            or

  $aggregate_object->online(...);

(Class or instance method) This method onlines a volume element.

Uses a 7Mode CLI, CMode CLI/ZAPI APIset.

Invokes "aggr-online" API for CMode ZAPI.

Invokes "aggr online" command for 7Mode CLI.

Same function can be performed using modify method also.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=> $string >>

(Required) Name of aggregate.

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: aggregate

7M CLI : aggregate name (Programmatic name: "aggregate")

=item C<< keep-same-uuid => $boolean >>

(Optional) A boolean (true or false).

Applicable for CMode CLI

=item C<< 'plex' => $string >>

Applicable for CMode CLI.

=item C<< 'hidden-plex' => $string >>

Applicable for CMode CLI.

=item C<< 'force' => $string >>

Applicable for CMode CLI

=item C<< 'force-online' => $string >>

=item C<< '*' >>

This method takes all other options supported by
"storage aggregate online" command in ONTAP.

Applicable for CMode CLI.


=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub online {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( plex hidden-plex force force-online
                    )
            ),
            aggregate        => {type => SCALAR},
            'keep-same-uuid' => {type => BOOLEAN, optional => 1},
        },
        cli_cmd => 'storage aggregate online',
    );
    $Log->exit() if $may_exit;
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_online_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_online_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_online_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::StorageAggregate::online',
    );
    return $ret;
} ## end sub online

sub _online_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api     => 'storage_aggregate_online',
    );

    $Log->exit() if $may_exit;
} ## end sub _online_cmode_cli

sub _online_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset    = delete $opts{apiset};
    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
    );
    %opts = %$aggr_args;
    $apiset->aggr_online(%opts);
    $Log->exit() if $may_exit;
} ## end sub _online_cmode_zapi

sub _online_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [qw(aggregate)];

    my %mode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%mode_args,
        copy   => $allowed_7mode_options,
    );

    $apiset->aggr_online(%mode_args);

    $Log->exit() if $may_exit;
} ## end sub _online_7mode_cli

=head2 restrict

  NACL::C::StorageAggregate->restrict(command_interface=>$ci,
                             aggregate=>$aggr_name);

                            or

  $aggregate_object->restrict();

(Class or instance method) This method restricts a volume element.

Uses a 7Mode CLI, CMode CLI/ZAPI APIset.

Invokes "aggr-restrict" API for CMode ZAPI.

Invokes "aggr restrict" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=> $string >>

(Required) Name of aggregate.

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: aggregate

7M CLI : aggregate name (Programmatic name: "aggregate")

=item C<< 'unmount-volumes' => $string >>

=item C<< '*' >>

This method takes all other options supported by
"storage aggregate restrict" command in ONTAP.

Applicable for CMode CLI/ZAPI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub restrict {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw(aggregate
                unmount-volumes
                ),
        },
        cli_cmd => 'storage aggregate restrict',
    );
    $Log->exit() if $may_exit;
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_restrict_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_restrict_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_restrict_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::StorageAggregate::restrict',
    );
    return $ret;
} ## end sub restrict

sub _restrict_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api     => 'storage_aggregate_restrict',
    );

    $Log->exit() if $may_exit;
} ## end sub _restrict_cmode_cli

sub _restrict_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $aggr_args         = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate unmount-volumes)],
    );
    %opts = %$aggr_args;
    $apiset->aggr_restrict(%opts);
    $Log->exit() if $may_exit;
} ## end sub _restrict_cmode_zapi

sub _restrict_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [qw(aggregate)];

    my %mode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%mode_args,
        copy   => $allowed_7mode_options,
    );

    $apiset->aggr_restrict(%mode_args);

    $Log->exit() if $may_exit;
} ## end sub _restrict_7mode_cli

=head2 split

  NACL::C::StorageAggregate->split(command_interface=>$ci,
                             aggregate=>$aggr_name,
                             new-aggregate-name=>$new_aggr_name);

                            or

  $aggregate_object->split(new-aggregate-name=>$new_aggr_name);

(Class or instance method) This method split an mirrored aggregate to a new aggregate element.

Uses a 7Mode CLI APIset.

Invokes "aggr split" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=> $string >>

(Required) Name of aggregate.

Applicable for 7Mode CLI.

=item C<< new-aggregate-name=> $string >>

(Required) Name of the new aggregate.

Applicable for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub split {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;

    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@_,
        additional_spec =>
            {_optional_scalars qw(plex-name new-aggregate-name),},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_split_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}

sub _split_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [qw(aggregate new-aggregate-name)];

    my %mode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%mode_args,
        copy   => $allowed_7mode_options,
    );

    my $plex;
    if (defined($opts{'plex-name'})) {
        $plex = $opts{'plex-name'};

    }

    $mode_args{'aggregate'} = $opts{aggregate} . '/' . $plex;
    $apiset->aggr_split(%mode_args);
    $Log->exit() if $may_exit;
    return $pkg_or_obj->new(
        command_interface => $command_interface,
        aggregate         => $mode_args{'new-aggregate-name'}
    );
} ## end sub _split_7mode_cli

=head2 add_disks_64bitupgrade_check

  my @objs = NACL::C::StorageAggregate->add_disks_64bitupgrade_check(command_interface=>$ci,
                             aggregate=>$aggr_name,...);

                            or

  my $objs = $aggregate_object->add_disks_64bitupgrade_check(...);

(Class or instance method)Displays a summary of the space impact which would
result from upgrading the aggregate to 64-bit.
It returns an arrayref, which contains objects of type StorageAggregateAddDisks64bitupgradeCheck.pm
See L<NACL::C::StorageAggregateAddDisks64bitupgradeCheck|lib-NACL-CS-StorageAggregateAddDisks64bitupgradeCheck-pm>

Uses a CMode CLI/ZAPI, 7Mode CLI/ZAPI APISet.

Invokes "aggr-add" API for CMode ZAPI,7Mode ZAPI.

Invokes "aggr add" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: aggregate

7M CLI : aggregate name (Programmatic name: "aggregate")

7M ZAPI: aggregate

=item C<< disklist => [$disk1, $disk2, $disk3, ... ] >>

This is an arrayref of disk names to be added to the aggregate.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: disks

7M CLI : -d <value> (Programmatic name: "disklist")

7M ZAPI: disks

=item C<< diskcount => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: disk-count

7M CLI : disk count (Programmatic name: "disk-count")

7M ZAPI: disk-count

=item C<< raidgroup => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: raid-group

7M CLI : -g <value> (Programmatic name: "raid-group")

7M ZAPI: raid-group

=item C<< allow-mixed => $string >>

Applicable for CMode CLI.

=item C<< disk-size => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

7M CLI : disk size (Programmatic name: "disk-size")

7M ZAPI: disk-size

=item C<< disktype => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: disktype

=item C<< chksumstyle => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: checksum-style

=item C<< force => $boolean >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI: -f (programmatic name: "force")

7M ZAPI: force-spare-pool

=item C<< "64bit-upgrade" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: upgrade-64bit-mode

7M CLI : -64bit-upgrade <value> (Programmatic name: "64bit-upgrade")

7M ZAPI: upgrade-64bit-mode

The option values are mapped

'normal' maps to:
    CMode ZAPI: "grow_none" of "aggr_add"

'grow-reserved' maps to:
    CMode ZAPI: "grow_reserved" of "aggr_add"

'grow-all' maps to:
    CMode ZAPI: "grow_all" of "aggr_add"

Applicable only for 7Mode CLI
Disks in a plex are not normally permitted to span spare pools. This behavior
is overridden with this option when it is set.
Permissible values: true/false or 1/0.

Maps to:
    7Mode CLI: "-f" of "aggr add"

=item C<< "is-display-command" => $string >>

Applicable only for 7Mode CLI

<< '*' >>

This method takes all other options supported by
"storage aggregate add-disks" command in ONTAP.

Applicable for CMode CLI

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub add_disks_64bitupgrade_check {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            disklist => {type => ARRAYREF, optional => 1},
            force    => {type => BOOLEAN,  optional => 1},
            _optional_scalars(
                qw(diskcount raidgroup allow-mixed raidgroup disk-size
                    disktype is-display-command)
            )
        },
        cli_cmd => 'storage aggregate add-disks',
    );
    my @objs = $pkg_or_obj->call_on_apiset(
        %opts,
        '64bit-upgrade' => 'check',
        choices         => [
            {   method    => '_add_disks_64bitupgrade_check_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_add_disks_64bitupgrade_check_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_add_disks_64bitupgrade_check_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
            {   method    => '_add_disks_64bitupgrade_check_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::StorageAggregate::add_disks_64bitupgrade_check',
    );
    $Log->exit() if $may_exit;
    return wantarray ? @objs : $objs[0];
} ## end sub add_disks_64bitupgrade_check

sub _add_disks_64bitupgrade_check_cmode_cli {
    $Log->enter() if $may_enter;
    require NACL::CS::StorageAggregateAddDisks64bitupgradeCheck;
    my ($pkg_or_obj, %opts) = @_;
    my @state_objs;
    my $response;

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api     => 'storage_aggregate_add_disks',
        apiset_response_obj => \$response,
    );

    my $output   = $response->get_parsed_output();
    foreach my $row (@$output) {
        my $obj = NACL::CS::StorageAggregateAddDisks64bitupgradeCheck->new();
        $obj->_set_fields(row => $row);
        push(@state_objs, $obj);
    }
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _add_disks_64bitupgrade_check_cmode_cli

sub _add_disks_64bitupgrade_check_7mode_cli {
    $Log->enter() if $may_enter;
    my @state_objs;
    require NACL::CS::StorageAggregateAddDisks64bitupgradeCheck;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    if (defined($opts{disklist})) {
        $opts{disklist} = join ' ', @{$opts{disklist}};
    }

    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate is-display-command force 64bit-upgrade)],
        map    => {
            'diskcount' => 'disk-count',
            'raidgroup' => 'raid-group',
            'disklist'  => 'disks',
        }
    );
    %opts = %$aggr_args;

    $pkg_or_obj->_handle_boolean_options(
        api_options     => \%opts,
        boolean_options => [qw(force)],
        is_7m_cli       => 1
    );

    # "aggr add" accepts size in GB
    if (exists $opts{'disk-size'}) {
        $opts{'disk-size'} = $opts{'disk-size'} / (1024 * 1024);
    }

    my $response = $apiset->aggr_add(%opts);
    my $output   = $response->get_parsed_output();
    foreach my $row (@$output) {
        my $obj = NACL::CS::StorageAggregateAddDisks64bitupgradeCheck->new();
        $obj->_set_fields(row => $row);
        push(@state_objs, $obj);
    }
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _add_disks_64bitupgrade_check_7mode_cli

sub _add_disks_64bitupgrade_check_cmode_zapi {
    $Log->enter() if $may_enter;
    require NACL::CS::StorageAggregateAddDisks64bitupgradeCheck;
    my @state_objs;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $aggr_args         = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
        map    => {
            "diskcount"     => "disk-count",
            "disktype"      => "disk-type",
            "chksumstyle"   => "checksum-style",
            "disklist"      => "disks",
            "raidgroup"     => "raid-group",
            "64bit-upgrade" => "upgrade-64bit-mode",
        }
    );
    %opts = %$aggr_args;

    $opts{'upgrade-64bit-mode'} = _map_modes($opts{'upgrade-64bit-mode'})
        if (defined $opts{'upgrade-64bit-mode'});

    my $response = $apiset->aggr_add(%opts);
    my $output   = $response->get_parsed_output();
    my $integer  = $output->[0]->{'upgrade-64bit-cookie'};
    my @objs     = NACL::CS::Aggregate->fetch(
        command_interface => $command_interface,
        filter            => {aggregate => $opts{aggregate}}
    );

    foreach my $obj (@objs) {
        if ($obj->upgrade_64bit_cookie() eq $integer) {
            my $state_fields = $pkg_or_obj->_hash_copy(
                source => \%$obj,
                map    => {"upgrade_64bit_added_space" => "total"}
            );
            my $new
                = NACL::CS::StorageAggregateAddDisks64bitupgradeCheck->new();
            $new->_set_fields(row => $state_fields);
            push(@state_objs, $new);
        } ## end if ( $obj->upgrade_64bit_cookie...
    } ## end foreach my $obj (@objs)
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _add_disks_64bitupgrade_check_cmode_zapi

sub _add_disks_64bitupgrade_check_7mode_zapi {
    $Log->enter() if $may_enter;
    my @state_objs;
    require NACL::CS::StorageAggregateAddDisks64bitupgradeCheck;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $aggr_args         = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [
            qw(aggregate is-display-command force-spare-pool
                disk-size-with-unit disk-size group-selection-mode
                max-records mirror-disks start-tag)
        ],
        map => {
            'diskcount'     => 'disk-count',
            'raidgroup'     => 'raid-group',
            'disklist'      => 'disks',
            '64bit-upgrade' => 'upgrade-64bit-mode',
            'force'         => 'force-spare-pool',
        }
    );
    %opts = %$aggr_args;

    $pkg_or_obj->_handle_boolean_options(
        api_options     => \%opts,
        boolean_options => [qw(force-spare-pool)],
        is_7m_cli       => 0
    );

    # "aggr-add" accepts size in 4KB blocks
    if (exists $opts{'disk-size'}) {
        $opts{'disk-size'} = $opts{'disk-size'} / (4 * 1024);
    }
    my $response = $apiset->aggr_add(%opts);
    my $output   = $response->get_parsed_output();
    my $integer  = $output->[0]->{'upgrade-64bit-cookie'};
    my @objs     = NACL::CS::Aggregate->fetch(
        command_interface => $command_interface,
        filter            => {aggregate => $opts{aggregate}}
    );
    foreach my $obj (@objs) {
        if ($obj->upgrade_64bit_cookie() eq $integer) {
            my $state_fields = $pkg_or_obj->_hash_copy(
                source => \%$obj,
                map    => {"upgrade_64bit_added_space" => "total"}
            );
            my $new
                = NACL::CS::StorageAggregateAddDisks64bitupgradeCheck->new();
            $new->_set_fields(row => $state_fields);
            push(@state_objs, $new);
        } ## end if ( $obj->upgrade_64bit_cookie...
    } ## end foreach my $obj (@objs)
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _add_disks_64bitupgrade_check_7mode_zapi

=head2 get_volumes

  my @volume_objects = $aggr_object->get_volumes();

(Instance method only) This method returns a STask/Component object
corresponding to all the children volumes present in the aggregate
modeled by this object.
If the aggregate object is a Component object, then object of type
NACL::C::Volume will be returned. If the aggregate object is a STask
object, then object of type NACL::STask::Volume will be returned.

=cut

sub get_volumes {
    my $self = shift;
    $Log->enter() if $may_enter;

    $self->_verify_invocation(style => 'instance_only');

    # Find whom the caller is - STask/C??.
    ref($self) =~ /NACL::(.+)::/;
    my $t_or_c = $1;

    # Get volume objects
    my @vols = $self->get_children_objects(type => "NACL::${t_or_c}::Volume");
    $Log->exit() if $may_exit;
    return @vols;
} ## end sub get_volumes

=head2 get_nodes

  my @node_objects = $aggr_object->get_nodes();

(Instance method only) This method returns a STask/Component object
corresponding to the nodes to which the aggregate belong to.
Normally an aggregate belongs to only one node except if it is a
striped aggregate (CMode only feature)
If the aggregate object is a Component object, then object of type
NACL::C::Node will be returned. If the aggregate object is a STask
object, then object of type NACL::STask::Node will be returned.

=cut

sub get_nodes {
    my $self = shift;
    $Log->enter() if $may_enter;

    $self->_verify_invocation(style => 'instance_only');

    # Find whom the caller is - STask/C??.
    ref($self) =~ /NACL::(.+)::/;
    my $t_or_c = $1;

    # Get node objects
    my @node_objs = ();

    #my @nodes = $self->get_one_state_attribute('nodes');
    # TODO:
    # CM ZAPI backend fails to construct 'desired-attributes' on specifying
    # requested_fields. For now, the mapping in CS file works only for
    # output side. Hence we are not limiting the fields in the output.
    my @nodes = $self->state()->nodes();
    $Log->debug("The obtained node names are:\n" . Dumper(\@nodes));
    foreach my $node (@nodes) {
        my $pkg_name = "NACL::${t_or_c}::Node";
        push(
            @node_objs,
            $pkg_name->new(
                name              => $node,
                command_interface => $self->command_interface()
            )
        );
        $Log->debug("created the $t_or_c object for node '$node'");
    }
    $Log->exit() if $may_exit;
    return @node_objs;
}

=head2 is_hybrid

    $ret = $aggregate->is_hybrid();

    or

    my $raid_status = NACL::C::StorageAggregate->is_hybrid('command_interface' => $ci, 'aggregate' => $aggr_name);

(Instance or Class method) This method checks if the aggregate is hybrid or not.
Will return 1 if its hybrid and 0 otherwise.

=cut

sub is_hybrid {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts = $pkg_or_obj->_common_validate_with(params => \@_,);

    my %filter_opts;
    $pkg_or_obj->_hash_move(
        source => \%opts,
        target => \%filter_opts,
        move   => [qw(aggregate)]
    );

    # Check if aggregate is hybrid
    my $state_obj = NACL::CS::StorageAggregate->fetch(
        %opts,
        filter           => \%filter_opts,
        requested_fields => [qw(hybrid)]
    );

    if ($state_obj->hybrid() eq 'true') {
        $Log->exit() if $may_exit;
        return 1;
    } else {
        $Log->exit() if $may_exit;
        return 0;
    }
}

=head2 show_scrub_status

    my $method_obj  = NACL::C::StorageAggregate->show_scrub_status(command_interface => $ci,
                          [ filter => { %filter } ,]
                          [ requested_fields => [ @req_fields ]);

                    or

    my $method_objs = NACL::C::StorageAggregate->show_scrub_status(command_interface => $ci,
                          [ filter => { %filter } ,]
                          [ requested_fields => [ @req_fields ]);

(Class method)
This method returns objects of the type L<NACL::CS::StorageAggregateShowScrubStatus|lib-NACL-CS-StorageAggregateShowScrubStatus-pm>.

This method accepts all the parameters accepted by L<fetch|lib-NACL-CS-ComponentState-pm/fetch>
(requested_fields, filter, allow_empty) and has all of the other properties of
fetch methods (callable in scalar or list context, throws a NoElementsFound
exception if no matching elements are found)

=cut

sub show_scrub_status {
    $Log->enter() if $may_enter;
    my $pkg = shift;
    require NACL::CS::StorageAggregateShowScrubStatus;
    my @state_objs = NACL::CS::StorageAggregateShowScrubStatus->fetch(@_);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub show_scrub_status

=begin COMMENTED_OUT

    # This is an example of using this method for 7M CLI
    sub _create_7mode_cli {
        ...
        $pkg->_translate_block_type)
            '32b_value' => '32',
            '64b_value' => '64',
            api_opts    => \%opts
        );
        $apiset->aggr_create(%opts);
        ...
    }

(For developers only)
Map the value of "block-type" from the various combinations understood
by CMode CLI (32-bit,  32-bi, 32-b, 32, 32_bit, 64*) to what the other
modes/interfaces understand. This example is for 7M CLI, in which the
block-type value can be "32" or "64".

=over

=item Options

=over

=item C<< '32b_value' => $32b_value >>

(Mandatory)
The value that denotes 32-bit for that mode/interface

=item C<< '64b_value' => $64b_value >>

(Mandatory)
The value that denotes 64-bit for that mode/interface

=item C<< api_opts => \%opts >>

(Mandatory)
A reference to the has of options which will get passed to the API call.

=item C<< block_type_field_name => $field_name >>

(Optional, defaults to "block-type")
The option name used by that mode/interface for the "block-type" parameter.

=back

=back

=end COMMENTED_OUT

=cut

sub _translate_block_type {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = validate @_,
        {
        '32b_value'           => {type => SCALAR},
        '64b_value'           => {type => SCALAR},
        block_type_field_name => {type => SCALAR, default => 'block-type'},
        api_opts              => {type => HASHREF}
        };

    my $block_type_field_name = $opts{block_type_field_name};
    my $api_opts              = $opts{api_opts};

    if (defined $api_opts->{$block_type_field_name}) {
        my $b32_value = $opts{'32b_value'};
        my $b64_value = $opts{'64b_value'};
        if ($api_opts->{$block_type_field_name} =~ /32/) {
            $api_opts->{$block_type_field_name} = $b32_value;
        } elsif ($api_opts->{$block_type_field_name} =~ /64/) {
            $api_opts->{$block_type_field_name} = $b64_value;
        } else {

            # If any other value is passed then don't translate it.
            # Let it pass down to the filer and fail there.
        }
    }

    $Log->exit() if $may_exit;
}

sub _map_modes {

    # Extract parameters
    my ($mode) = @_;
    my %mappings = (
        'normal'        => 'grow_none',
        'grow-reserved' => 'grow_reserved',
        'grow-all'      => 'grow_all'
    );
    if (exists $mappings{$mode}) {
        $mode = $mappings{$mode};
    }
    return $mode;
} ## end sub _map_modes

# The Params::Validate spec describing the primary keys of the
# StorageAggregate Component

=head2 verify

    NACL::C::StorageAggregate->verify(
                                command_interface => $ci,
                                'aggregate' => $aggregate,
                                ...);

                                 or

    $StorageAggregate_obj->verify(...);

(Class or instance method)
This method is used to verify an aggregate.

Uses a CMode CLI/ZAPI, 7Mode CLI/ZAPI APISet.

Invokes "command" API for CMode ZAPI.

Invokes "command" command for 7Mode CLI/ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'aggregate' => $aggregate >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: aggregate

7M CLI : aggregate name (Programmatic name: "aggregate")

7M ZAPI: aggregate

=item C<< 'plex-to-fix' => $string >>

=item C<< 'action' => $string >>
The possible values are 'start', stop', 'suspend', 'resume'. To specify
the action as 'status', use
See L<NACL::C::StorageAggregate->verify_status|lib-NACL-C-StorageAggregate-pm/verify_status>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: fix-plex

7M ZAPI: fix-plex

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

<< '*' >>

This method takes all other options supported by
"storage aggregate verify" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub verify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( plex-to-fix
                    )
            ),
            action => {type => SCALAR},
        },
        cli_cmd => 'storage aggregate verify',
    );
    if ($opts{action} eq 'status') {
        $Log->exit() if $may_exit;
        NATE::Exceptions::Argument(
            "Use NACL::C::StorageAggregate->verify_status() to get the status"
                . " of verify operation");
    }
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_verify_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_verify_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_verify_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
            {   method    => '_verify_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
    );

    $Log->exit() if $may_exit;
}

sub _verify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api     => 'storage_aggregate_verify',
    );

    $Log->exit() if $may_exit;
}

sub _verify_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $ci        = delete $opts{command_interface};
    my $apiset    = delete $opts{apiset};
    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
        map    => {'plex-to-fix' => 'plex-number'}
    );
    my $action  = $opts{action};
    my $command = 'aggr_verify_' . $action;
    %opts = %$aggr_args;

    if (   (defined $opts{'plex-number'})
        && ($opts{'plex-number'} =~ /plex(\d+)/i))
    {

        # 7M CLI and ZAPI takes the plex information as just number whereas
        # CM CLI takes it in the form 'plex<plex_number>'
        $opts{'plex-number'} = $1;
    }

    $apiset->$command(%opts);
    $Log->exit() if $may_exit;
} ## end sub _verify_7mode_cli

sub _verify_7mode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $ci        = delete $opts{command_interface};
    my $apiset    = delete $opts{apiset};
    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate)],
        map    => {'plex-to-fix' => 'fix-plex'}
    );
    my $action  = $opts{action};
    my $command = 'aggr_verify_' . $action;
    %opts = %$aggr_args;

    if (   (defined $opts{'fix-plex'})
        && ($opts{'fix-plex'} =~ /plex(\d+)/i))
    {

        # 7M CLI and ZAPI takes the plex information as just number whereas
        # CM CLI takes it in the form 'plex<plex_number>'
        $opts{'fix-plex'} = $1;
    }

    $apiset->$command(%opts);
    $Log->exit() if $may_exit;
} ## end sub _verify_7mode_zapi

sub _verify_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $ci        = delete $opts{command_interface};
    my $apiset    = delete $opts{apiset};
    my $aggr_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(aggregate log-only)],
        map    => {'plex-to-fix' => 'fix-plex'}
    );
    my $action  = $opts{action};
    my $command = 'aggr_verify_' . $action;
    %opts = %$aggr_args;

    if (   (defined $opts{'fix-plex'})
        && ($opts{'fix-plex'} =~ /plex(\d+)/i))
    {

        # CM ZAPI takes the plex information as just number
        $opts{'fix-plex'} = $1;
    } ## end if ( ( defined $opts{'fix-plex'...

    $apiset->$command(%opts);
    $Log->exit() if $may_exit;
} ## end sub _verify_cmode_zapi

=head2 verify_status

    my $verify_status_obj = NACL::C::StorageAggregate->verify_status(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @verify_status_objs = NACL::C::StorageAggregate->verify_status(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $verify_status_obj = $StorageAggregate_obj->verify_status();

(Class or Instance method)
This method is used to get the status of the verification process started/
stopped/suspended/resumed on an aggregate using
NACL::C::StorageAggregate->verify().

This method returns objects of the type
L<NACL::CS::StorageAggregateVerifyStatus|lib-NACL-CS-StorageAggregateVerifyStatus-pm>.

This method accepts all the parameters accepted by L<fetch|lib-NACL-CS-ComponentState-pm/fetch>
(requested_fields, filter, allow_empty) and has all of the other properties of
fetch methods (callable in scalar or list context, throws a NoElementsFound
exception if no matching elements are found). For backwards compatibility
reasons, this method does not throw a NoElementsFound exception by default
(i.e. it defaults C<allow_empty> to 1). For a NoElementsFound exception
to be thrown, pass in C<< allow_empty => 0 >>.

Uses a CMode CLI/ZAPI, 7Mode CLI/ZAPI APISet.

=cut

sub verify_status {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@args);

    # Provide support for Aggregate->verify_status(aggregate => $aggr)
    if (defined $opts{aggregate}) {
        $opts{filter}{aggregate} = delete $opts{aggregate};
    }

    my @status_obj = NACL::CS::StorageAggregateVerifyStatus->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @status_obj : $status_obj[0];
} ## end sub verify_status

=head2 mirror

    NACL::C::StorageAggregate->mirror(
                                command_interface => $ci,
                                'aggregate' => $aggregate,
                                ...);

                                 or

    $StorageAggregate_obj->mirror(...);

(Class or instance method)
This method is used to mirror an existing aggregate.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "aggr-mirror" API for CMode ZAPI.

Invokes "aggr mirror" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'aggregate' => $aggregate >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: aggregate

7M CLI : aggregate name (Programmatic name: "aggregate")

=item C<< 'ignore-pool' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: ignore-spare-pool

=item C<< 'allow-mixed-rpms' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: allow-mixed-rpms

=item C<< 'mirror-disklist' => [ $mirror-disk1, $mirror-disk2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: mirror-disks

7M CLI: -d <value> (programmatic name: "disks")

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'pre-check' => $string >>

Applicable for CMode CLI

=item C<< 'pre-check-results-mirror' => $string >>

Applicable for CMode CLI

=item C<< 'mirror-partitionlist' => [ $mirror-partition1, $mirror-partition2, ... ] >>

Partitions for Mirrored Plex
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)
Applicable for CMode CLI

=item C<< 'allow-same-carrier' => $string >>

Allow Same RAID Group Within Carrier
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"storage aggregate mirror" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub mirror {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "mirror-partitionlist" => {type => ARRAYREF, optional => 1},
            'mirror-disklist'      => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw( ignore-pool allow-mixed-rpms

                    ignore-pool-checks allow-mixed-rpm

                    pre-check pre-check-results-mirror victim-aggr

                    allow-same-carrier
                    )
            ),
        },
        cli_cmd => 'storage aggregate mirror',
    );
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_mirror_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_mirror_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
                {   method    => '_mirror_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
            ],
            frontend => 'NACL::C::StorageAggregate::mirror',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::StorageAggregate::FreeSpaceReallocationEnabled
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
} ## end sub mirror

sub _mirror_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api     => 'storage_aggregate_mirror',
    );

    $Log->exit() if $may_exit;
} ## end sub _mirror_cmode_cli

sub _mirror_cmode_zapi {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $allowed_cmode_options = [
        qw( aggregate allow-mixed-rpms force ignore-pool-checks
            )
    ];
    my %cmode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_cmode_options,
        map    => {
            'mirror-disklist' => 'mirror-disks',
            'ignore-pool'     => 'ignore-spare-pool'
        },
        target => \%cmode_args,
    );

    $apiset->aggr_mirror(%cmode_args);
    $Log->exit() if $may_exit;
} ## end sub _mirror_cmode_zapi

sub _mirror_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};

    my $allowed_7mode_options = [qw(aggregate force victim-aggr)];

    my %mode_args;
    if (defined($opts{'mirror-disklist'})) {
        $opts{'mirror-disklist'} = join ' ', @{$opts{'mirror-disklist'}};
    }

    $opts{'force'} = 1;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%mode_args,
        map    => {"mirror-disklist" => "disks",},
        copy   => $allowed_7mode_options,
    );
    my $aggr = $mode_args{aggregate};

    $apiset->aggr_mirror(%mode_args);
    $Log->exit() if $may_exit;
} ## end sub _mirror_7mode_cli

=head2 show_resync_status

    my $show_resync_status_obj = NACL::C::StorageAggregate->show_resync_status(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @show_resync_status_objs = NACL::C::StorageAggregate->show_resync_status(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $show_resync_status_obj = $StorageAggregate_obj->show_resync_status();

(Class or Instance method)
This method returns objects of the type L<NACL::CS::StorageAggregateShowResyncStatus|lib-NACL-CS-StorageAggregateShowResyncStatus-pm>.

This method accepts all the parameters accepted by L<fetch|lib-NACL-CS-ComponentState-pm/fetch>
(requested_fields, filter, allow_empty) and has all of the other properties of
fetch methods (callable in scalar or list context, throws a NoElementsFound
exception if no matching elements are found)

=cut

sub show_resync_status {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::StorageAggregateShowResyncStatus;
    my @state_objs = NACL::CS::StorageAggregateShowResyncStatus->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 undestroy

  my $aggr_obj = NACL::C::StorageAggregate->undestroy(command_interface=>$ci,
                             aggregate=>$aggr_name,...);

                            or

  my $aggr_obj = $aggregate_object->undestroy(...);

(Class or instance method) This method undestroys an aggregate element and
returns the aggregate object.

Uses 7Mode CLI APISet.

Invokes "aggr undestroy" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< aggregate=>$aggr_name >>

(Required for class method, Not Applicable for instance method)
Name of aggregate.

Applicable for 7Mode CLI.

=item C<< raidtree_id => $raidtree_id >>

(Optional) A Raidtree Id of the aggregate to be undestroyed.

Applicable for 7Mode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub undestroy {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;

    $pkg_or_obj->_frontend_log_debug_opts(@_);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {_optional_scalars(qw( raidtree_id )),},
    );

    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_undestroy_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
}

sub _undestroy_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg  = shift;
    my %opts = $pkg->_validate_backend_opts(@_);

    my $command_interface = delete $opts{command_interface};
    my $raidtree_id       = delete $opts{raidtree_id}
        if (defined $opts{raidtree_id});

    my $apiset = delete $opts{apiset};
    if (defined $raidtree_id) {
        $opts{'connectrec-match_table'}
            = [qr/\[(\d+)\].*($raidtree_id)/ => sub { $1; }];
    } else {
        $opts{'connectrec-match_table'} = [qr/Selection/ => 2];
    }

    $apiset->aggr_undestroy(%opts);
    my $obj = $pkg->new(
        aggregate         => $opts{aggregate},
        command_interface => $command_interface
    );
    $Log->exit() if $may_exit;
    return $obj;
}

# Return a boolean denoting whether the command interface is running an RR1
# build. Currently required since calling modify to online/offline/restrict
# results in an exception only for RR.1, so the determination of whether
# to throw an exception is made based on the build.
sub _check_if_post_rr1 {
    $Log->enter() if $may_enter;
    my %opts = validate @_, {command_interface => _ci_validate_spec()};
    my $command_interface = $opts{command_interface};
    my $version_mgr       = $command_interface->get_version_manager();
    my $release = $version_mgr->get_version_attribute(attribute => 'release');
    my $codename = $version_mgr->map_release_to_codename(release => $release);
    if ($codename =~ /Rollingrock_0/i || $codename =~ /Boilermaker/i) {
        $Log->exit() if $may_exit;
        return 0;
    } else {
        $Log->exit() if $may_exit;
        return 1;
    }
}

# Common exception handler for 7Mode CLI/ZAPI implementations for modify

sub _modify_7m_exception_handler {
    $Log->enter() if $may_enter;
    my %opts = validate(
        @_,
        {   command_interface => _ci_validate_spec(),
            exception         => {isa => 'NATE::BaseException'},
            command           => {type => SCALAR}
        }
    );
    my $exception = $opts{exception};
    my $command   = $opts{command};
    $Log->debug("Invocation of $command failed:\n" . $exception->text());
    if ($exception->text() !~ /is already/) {
        $Log->exit() if $may_exit;
        $exception->throw();

    } elsif (
        _check_if_post_rr1(command_interface => $opts{command_interface}))
    {
        $Log->trace(
            'Running RR.1 or a later release, hence propagating the exception'
        );
        $Log->exit() if $may_exit;
        $exception->throw();

    }
    $Log->exit() if $may_exit;
}

sub _ci_validate_spec {
    return {isa => 'NACL::C::CommandInterface::ONTAP', type => OBJECT};
}

=head2 show_space

    my $show_space_obj = NACL::C::StorageAggregate->show_space(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @show_space_objs = NACL::C::StorageAggregate->show_space(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $show_space_obj = $StorageAggregate_obj->show_space();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-StorageAggregateShowSpace-pm/ATTRIBUTES>
of NACL::CS::StorageAggregateShowSpace as input.
This method returns objects of the type L<NACL::CS::StorageAggregateShowSpace|lib-NACL-CS-StorageAggregateShowSpace-pm>.

=cut

sub show_space {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::StorageAggregateShowSpace;
    my @state_objs = NACL::CS::StorageAggregateShowSpace->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 check_spare_low

    NACL::C::StorageAggregate->check_spare_low(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to check for spare low condition.

Uses a CMode CLI/ZAPI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'node-name' => $string >>

Node  Name possible value(s) are, <nodename>,local

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI:

<aggr-check-spare-low>
<node-name>$value</node-name>
</aggr-check-spare-low>

=item C<< 'cmd_execution_time' => \$scalar >>

(Optional) Scalar reference , it will be updated with the command execution
time returned by the apiset response object

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< "extended_query" => $string  >>

=item C<< '*' >>

This method takes all other options supported by
"storage aggregate check_spare_low" command in ONTAP.

Applicable for CMode CLI.

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub check_spare_low {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'cmd_execution_time' => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( node-name
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'storage aggregate check_spare_low',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_check_spare_low_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_check_spare_low_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::StorageAggregate::check_spare_low',
    );

    $Log->exit() if $may_exit;
} ## end sub check_spare_low

sub _check_spare_low_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_base_backend_cmode_cli(
        %opts,
        api     => 'storage_aggregate_check_spare_low',
    );
    $Log->exit() if $may_exit;
} ## end sub _check_spare_low_cmode_cli

sub _check_spare_low_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    my %opts   = $pkg->_validate_backend_opts(@opts);
    my $apiset = $opts{apiset};

    my %cmode_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => [qw (node-name)],
        target => \%cmode_args,
    );

    $apiset->aggr_check_spare_low(%cmode_args);

    $Log->exit() if $may_exit;
} ## end sub _check_spare_low_cmode_zapi

=head2 show_df

    my $show_df_obj = NACL::C::StorageAggregate->show_df(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

    my @show_df_objs = NACL::C::StorageAggregate->show_df(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

   my $show_df_obj = $StorageAggregate_obj->show_df();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-StorageAggregateShowDf-pm/ATTRIBUTES>
of NACL::CS::StorageAggregateShowDf as input.
This method returns objects of the type L<NACL::CS::StorageAggregateShowDf|lib-NACL-CS-StorageAggregateShowDf-pm>.

=cut

sub show_df {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::StorageAggregateShowDf;
    my @state_objs = NACL::CS::StorageAggregateShowDf->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 show_raidgroup

    my $show_raidgroup_obj = NACL::C::StorageAggregate->show_raidgroup(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

    my @show_raidgroup_objs = NACL::C::StorageAggregate->show_raidgroup(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

   my $show_raidgroup_obj = $StorageAggregate_obj->show_raidgroup();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-StorageAggregateShowRaidgroup-pm/ATTRIBUTES>
of NACL::CS::StorageAggregateShowRaidgroup as input.
This method returns objects of the type L<NACL::CS::StorageAggregateShowRaidgroup|lib-NACL-CS-StorageAggregateShowRaidgroup-pm>.

=cut

sub show_raidgroup {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::StorageAggregateShowRaidgroup;
    my @state_objs = NACL::CS::StorageAggregateShowRaidgroup->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 show_view

    my $show_view_obj = NACL::C::StorageAggregate->show_view(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

    my @show_view_objs = NACL::C::StorageAggregate->show_view(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

   my $show_view_obj = $StorageAggregate_obj->show_view();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-StorageAggregateShowView-pm/ATTRIBUTES>
of NACL::CS::StorageAggregateShowView as input.
This method returns objects of the type L<NACL::CS::StorageAggregateShowView|lib-NACL-CS-StorageAggregateShowView-pm>.

=cut

sub show_view {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::StorageAggregateShowView;
    my @state_objs = NACL::CS::StorageAggregateShowView->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 plex_show

    my $plex_show_obj = NACL::C::StorageAggregate->plex_show(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

    my @plex_show_objs = NACL::C::StorageAggregate->plex_show(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or

   my $plex_show_obj = $StorageAggregate_obj->plex_show(); 

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of 
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty) 
and has all of the other properties of fetch methods (callable in scalar or list 
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-StorageAggregatePlexShow-pm/ATTRIBUTES>
of NACL::CS::StorageAggregatePlexShow as input. 
This method returns objects of the type L<NACL::CS::StorageAggregatePlexShow|lib-NACL-CS-StorageAggregatePlexShow-pm>.

=cut

sub plex_show {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::StorageAggregatePlexShow;
    my @state_objs = NACL::CS::StorageAggregatePlexShow->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 show_status

    my $show_status_obj = NACL::C::StorageAggregate->show_status(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);
						  
                    or
					
    my @show_status_objs = NACL::C::StorageAggregate->show_status(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or
  
   my $show_status_obj = $StorageAggregate_obj->show_status(); 

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of 
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty) 
and has all of the other properties of fetch methods (callable in scalar or list 
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-StorageAggregateShowStatus-pm/ATTRIBUTES>
of NACL::CS::StorageAggregateShowStatus as input. 
This method returns objects of the type L<NACL::CS::StorageAggregateShowStatus|lib-NACL-CS-StorageAggregateShowStatus-pm>.

=cut

sub show_status {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::StorageAggregateShowStatus;
    my @state_objs = NACL::CS::StorageAggregateShowStatus->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 show_spare_disks

    my $show_spare_disks_obj = NACL::C::StorageAggregate->show_spare_disks(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);
						  
                    or
					
    my @show_spare_disks_objs = NACL::C::StorageAggregate->show_spare_disks(command_interface => $ci,
                          [ filter => { %filter } ,]
			  [ requested_fields => [ @req_fields ]);

                    or
  
   my $show_spare_disks_obj = $StorageAggregate_obj->show_spare_disks(); 
						  
(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of 
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty) 
and has all of the other properties of fetch methods (callable in scalar or list 
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-StorageAggregateShowSpareDisks-pm/ATTRIBUTES>
of NACL::CS::StorageAggregateShowSpareDisks as input. 
This method returns objects of the type L<NACL::CS::StorageAggregateShowSpareDisks|lib-NACL-CS-StorageAggregateShowSpareDisks-pm>.

=cut

sub show_spare_disks {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::StorageAggregateShowSpareDisks;
    my @state_objs = NACL::CS::StorageAggregateShowSpareDisks->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 remove_stale_instance

    NACL::C::StorageAggregate->remove_stale_instance(
                                command_interface => $ci,
	                            'aggregate' => $aggregate,
                                ...);
	                             
								 or

    $StorageAggregate_obj->remove_stale_instance(...);

(Class or instance method)
This method is used to remove a stale aggregate record.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'aggregate' => $aggregate >>

(Required for class method, Not Applicable for instance method)
Aggregate
Applicable for CMode CLI

=item C<< 'nodename' => $string >>

Node Name
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "storage aggregate remove stale instance" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub remove_stale_instance {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( nodename
                    )
            ),
        },
        api => 'storage_aggregate_remove_stale_instance',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_remove_stale_instance_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::StorageAggregate::remove_stale_instance',
    );

    $Log->exit() if $may_exit;
}

sub _remove_stale_instance_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'storage_aggregate_remove_stale_instance',);
    $Log->exit() if $may_exit;
}

=head2 remove_stale_record

    NACL::C::StorageAggregate->remove_stale_record(
                                command_interface => $ci,
	                            'aggregate' => $aggregate,
                                ...);
	                             
								 or

    $StorageAggregate_obj->remove_stale_record(...);

(Class or instance method)
This method is used to remove a stale aggregate record.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'aggregate' => $aggregate >>

(Required for class method, Not Applicable for instance method)
Aggregate
Applicable for CMode CLI

=item C<< 'nodename' => $string >>

Node Name
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "storage aggregate remove stale record" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub remove_stale_record {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( nodename
                    )
            ),
        },
        cli_cmd => 'storage aggregate remove-stale-record',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_remove_stale_record_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::StorageAggregate::remove_stale_record',
    );

    $Log->exit() if $may_exit;
}

sub _remove_stale_record_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'storage_aggregate_remove_stale_record',);
    $Log->exit() if $may_exit;
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (aggregate => {type => SCALAR},);
}

=head2 name_attribute

Returns 'aggregate'.

=cut

sub name_attribute {
    return 'aggregate';
}

=head2 get_local_node_name

    my $node_name = $aggr_obj->get_local_node_name();

(Applicable only for CMode)
Determines the node on which the aggregate is present and returns the node
name. This is necessary for Nodescope operations, since they need to
be performed in the context of the node where the aggregate is present.

=cut

sub get_local_node_name {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts = $pkg_or_obj->_common_validate_with(params => \@_);
    if ($opts{command_interface}->mode() ne 'CMode') {
        $Log->exit() if $may_exit;
        return $opts{command_interface}->hostrec->hostname;
    }

    # Move the primary keys into the filter, all other options are
    # sent through directly to fetch (these should just be the common options)
    my %filter;
    $pkg_or_obj->_hash_move(
        source => \%opts,
        target => \%filter,
        move   => [$pkg_or_obj->get_primary_keys()]
    );

    my $cs_object = NACL::CS::StorageAggregate->fetch(
        %opts,
        filter           => \%filter,
        requested_fields => [qw(nodes)]

    );

    my @nodes = $cs_object->nodes();
    $Log->exit() if $may_exit;
    return $nodes[0];

} ## end sub get_local_node_name

=head2 get_local_node_obj

    my $node_obj = $aggr_obj->get_local_node_obj();

Similar to C<get_local_node_name>, but returns a C<NACL::STask::Node>
object for the node hosting the aggregate.

=cut

# Implementation comes from the base-class

=head2 file_create

    # Instance call
    my $aggregate_file = $aggregate_obj->file_create(
        filename => $filename,
        %common_option
    );

(instance method)

Returns the AggregateFile which is created on the aggregate.

This method creates a file on the aggregate and returns the
L<NACL::C::AggregateFile|lib-NACL-C-AggregateFile-pm/> object.

=cut

sub file_create {
    $Log->enter() if $may_enter;
    my ( $self, @args ) = @_;
    my %opts = $self->_common_validate_with(
        params            => \@args,
        additional_spec => { filename => { type => SCALAR, optional => 1 } },

        # Other options which is valid for NACL::C::AggregateFile->create
        allow_extra => 1
    );

    my $filename = delete $opts{filename}
      || NACL::GeneralUtils::random_name_generator( prefix => 'file' );
    my $aggregate = delete $opts{aggregate};
    $opts{path} = "/aggr/$aggregate/$filename";

    my $c_or_t = $self->is_c_or_t();
    my $file_pkg = "NACL::${c_or_t}::AggregateFile";

    eval "require $file_pkg";

    # If "filecount" is provided, then multiple file objects are returned
    my @files = $file_pkg->create(%opts);

    $Log->exit() if $may_exit;
    return wantarray ? @files : $files[0];
}

=head2 wafl_buftree_check

(Instance Method) Executes the "wafl buftree check" command on the Node

    my $response = $Aggregate->wafl_buftree_check(
        inodespec        => $location,
        %other_options,
    );

=over

=item Options

=over

=item C<< inodespec => $location >>

(Required) is one of:

        <inum>
        priv:<inum>
        vvol:<vvolname>/label
        vvol:<vvolname>/container
        vvol:<vvolname>/uuid
        vvol:<vvolname>/uuid/<pathname>
        path:<pathname>
        meta:<pathname>

=item C<< snapid => $snapID >>

(Optional, default AFS)
snapshot ID

=item C<< level-check => $level >>

(Optional, default 0)
lowest level to check

=item C<< first-last-fbn => <first>[,<last>] >>

(Optional)
first and last FBNs to check

=item C<< no-check => 0|1 >>

(Optional)
don't check wi_blk_cnt and wi_indirect_blk_cnt

=item C<< no-cross-check => 0|1 >>

(Optional)
don't cross-check vvol indirect formats against container map parent format

=item C<< no-panic => 0|1 >>

(Optional)
don't panic on when an error is found

=item C<< verbose => 0|1 >>

(Optional)
display extra statistics

=item C<< snapdiff => 0|1 >>

(Optional)
test snapdiff interfaces

=back

=back

=cut

__PACKAGE__->_nodescope_builder(
    command => 'wafl_buftree_check',
);

=head2 emap_insert emap_delete emap_lookup

    # Instance call
    my $eid_info = $aggregate_obj->emap_lookup(
        eid => $eid
        %common_option
    );

    $aggregate_obj->emap_insert(%opts);

    $aggregate_obj->emap_delete(%opts);

(instance method)

Returns the eid and pvbn info of the given eid.

=cut

__PACKAGE__->_nodescope_builder(command => 'mw_emap_insert',  method => 'emap_insert');
__PACKAGE__->_nodescope_builder(command => 'mw_emap_delete',  method => 'emap_delete');

__PACKAGE__->_cs_call_method_builder(method => 'emap_lookup', extra_filters => ['eid', 'random', 'region', 'count']);


=head2 filersio_ereg

    $aggregate_obj->filersio_ereg(%opts);

(instance method)

=cut

__PACKAGE__->_nodescope_builder(command => 'filersio_ereg');

=head2 stage_test, stage_stat stage_zerostat

    # Instance call
    my $stat = $aggregate_obj->mw_stage_stat(
        'aggregate' => $aggr
    );

    $aggregate_obj->mw_stage_test(%opts);

    $aggregate_obj->mw_stage_zerostat(%opts);


(instance method)

Returns the stage stats.

=cut

__PACKAGE__->_nodescope_builder(command => 'mw_stage_test',);
__PACKAGE__->_nodescope_builder(command => 'mw_stage_zerostat',);
__PACKAGE__->_cs_call_method_builder( method => 'mw_stage_stat',);
=head
=cut
1;
