#
# Copyright (c) 2013 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Volume Component Module
## @author dl-nacl-dev@netapp.com, pkj@netapp.com
## @status shared
## @pod here

=head1 NAME

NACL::C::Volume

=head1 DESCRIPTION

C<NACL::C::Volume> is a derived class of
L<NACL::C::Component::ONTAP|lib-NACL-C-Component-ONTAP-pm>.  It represents access
to an ONTAP Volume.  A related class is
L<NACL::CS::Volume|lib-NACL-CS-Volume-pm>, which represents
the past state of an ONTAP volume.

=head1 ATTRIBUTES

=head2 command_interface

See L<NACL::C::Component::command_interface|lib-NACL-C-Component-pm/command_interface>

=head2 volume

The name of the volume, used to identify it uniquely.  (Volume
names are assumed to be unique within a node or within a cluster in 7mode
and in a vserver in Cmode)

=head2 vserver

(Required for CMode, Not Applicable for 7Mode) The name of the vserver
in which the volume is present.

=cut

package NACL::C::Volume;

use strict;
use warnings;
use Params::Validate qw(:all);
use NACL::CS::Volume;
use NACL::CS::VolumeExplore;
use NACL::ComponentUtils qw(_optional_scalars Dumper);
use NATE::BaseException qw(:try);
use NATE::Exceptions::Argument ();
use base 'NACL::C::Component::ONTAP';
use NACL::C::Job;
use NACL::STask::VolumeMove;
use NACL::STask::VolumeClone;
use NACL::STask::Mkfile;
use NACL::C::Vserver;
use NATE::Log qw(log_global);
my $Log = log_global();
my $may_enter = $Log->may_enter();
my $may_exit  = $Log->may_exit();
use NACL::Exceptions::InvalidFilterField;
use NACL::APISet::Exceptions::CommandFailedException ();
use NACL::C::UnitNormalization;
use NACL::CS::VserverServicesUnixUser;
use NACL::CS::VserverServicesUnixGroup;
use Data::Dumper ();
use NACL::Exceptions::InvalidChoice;
use NATE::ParamSet qw(param_global);

# Imports from mix-ins
use NACL::C::_Mixins::AggregateVolume qw(_translate_options_values);
use NACL::C::_Mixins::Volume qw(:all);
use NACL::C::_Mixins::VolumeFileDirectory qw(:find_state);
use NACL::C::_Mixins::Common qw(:iv_methods);

use NACL::STask::VolumeFile;
use NACL::C::VolumeDirectory;
use NACL::CS::VolumeListFilesInDirectory;
use NACL::STask::VolumeFlexcache;

# Component exceptions
use NACL::C::Exceptions::NotEnoughSpaceOnAggregate               ();
use NACL::C::Exceptions::Volume::AlreadyExists                   ();
use NACL::C::Exceptions::Volume::DoesNotExist                    ();
use NACL::C::Exceptions::Volume::UnmountRoot                     ();
use NACL::C::Exceptions::Volume::AlreadyModified                 ();
use NACL::C::Exceptions::Volume::UnsupportedUnmount              ();
use NACL::C::Exceptions::Volume::StateUnknown                    ();
use NACL::C::Exceptions::Volume::FileOperationsInProgress        ();
use NACL::C::Exceptions::Volume::CurrentlyBusy                   ();
use NACL::C::Exceptions::Volume::ConstituentsInInconsistentState ();
use NACL::C::Exceptions::Volume::DeleteSnapMirrorRelationship    ();
use NACL::C::Exceptions::Volume::AlreadyMounted                  ();
use NACL::C::Exceptions::Volume::VolumeOffline                   ();
use NACL::C::Exceptions::Volume::VserverLocked                   ();
use NACL::C::Exceptions::Volume::NotExported                     ();
use NACL::C::Exceptions::Volume::HasClones                       ();
use NACL::C::Exceptions::Volume::CannotDeleteRootVolume          ();
use NACL::C::Exceptions::Volume::AlreadyRestricted               ();
use NACL::C::Exceptions::Volume::MustBeUnmounted                 ();
use NACL::C::Exceptions::Volume::AggregateRequiredForCreate      ();
use NACL::C::Exceptions::Volume::InvalidDataConstituentName      ();
use NACL::C::Exceptions::Volume::FailedToDeleteJunction          ();
use NACL::C::Exceptions::Volume::VserverNotInAggrList            ();
use NACL::C::Exceptions::Volume::HasFlexcacheOrigin              ();
use NACL::C::Exceptions::Volume::RebalancerRunning               ();
use NACL::C::Exceptions::Volume::OperationNotPermittedOnDrVolume ();
use NACL::C::Exceptions::Volume::UnsupportedDELOperation         ();
use NACL::C::Exceptions::Volume::ApplicationVolumeRemove         ();

=head1 Linked components

(See
http://wikid.netapp.com/w/QA/projects/Libraries_Initiative/Project/Common_Infrastructure/Users_Guide/Component_Layer_Users_Guide#Linked_components
for a more detailed explanation on linked components)

This component is linked to the following components:

=begin html

<table border="1">
<tr> <th>Prefix</th> <th>Linked Component</th> </tr>
<tr> <td>efficiency</td> <td>VolumeEfficiency</td> </tr>
<tr> <td>snapshot</td> <td>VolumeSnapshot</td> </tr>
<tr> <td>move</td> <td>VolumeMove</td> </tr>
<tr> <td>qtree</td> <td>VolumeQtree</td> </tr>
<tr> <td>lun</td> <td>Lun</td> </tr>
<tr> <td>snaplock</td> <td>VolumeSnaplock</td> </tr>
</table>

=end html

Examples:

    # Invoking the "start" method of the VolumeEfficiency component
    $vol_obj->efficiency_start(%opts);

    # Invoking the "create" method of the VolumeSnapshot component
    my $snapshot_obj = $vol_obj->snapshot_create(snapshot => $snap_name, %opts);

=head2 Linkage to file and directory components

Functionality of the L<NACL::C::VolumeFile|lib-NACL-C-VolumeFile-pm>
component can be accessed by using the "file_" prefix and provided
the filename through the C<filename> argument. (Note that this
should be the filename, not the full path)

Examples:

    my $file_obj = $vol_obj->file_create(filename => 'file', %other_opts);

    or,

    my $file_obj = $vol_obj->file_create(%other_opts);

    # filename is optional for file_create method. A random name is generated
    # with the suffix 'filename'.

    $vol_obj->file_delete(filename => 'file');

Functionality of the L<NACL::C::VolumeDirectory|lib-NACL-C-VolumeDirectory-pm>
component can be accessed by using the "directory_" prefix and provided
the filename through the C<dirname> argument. (Note that this
should be the directory name, not the full path)

Examples:

    my $dir_obj = $vol_obj->directory_create(dirname => 'dir', %other_opts);

    $vol_obj->directory_delete(dirname => 'dir');

=head2 Linkage to VolumeFlexcache component

Functionality of the L<NACL::C::VolumeFlexcache|lib-NACL-C-VolumeFlexcache-pm>
component can be accessed by using the C<flexcache_> prefix. The C<volume>
attribute of the volume object is used as the value of the C<origin-volume>
argument for the Flexcache methods. Examples:

    my $cache_obj = $vol_obj->flexcache_create();

    my @caches = $vol_obj->flexcache_find();

=cut

no warnings qw(redefine);
sub can {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my @opts = $pkg_or_obj->_can_validation(@args);

    my $code;
    my ($method, $error_ptr) = @opts;
    if ($code = $pkg_or_obj->NACL::C::Component::can(@opts)) {
        $Log->exit() if $may_exit;
        return $code;
    }

    # Link to implementations in NACL::C::VolumeFile and VolumeDirectory
    my $file_directory_handle = sub {
        my %args = validate_with(
            params => \@_,
            spec   => {
                linked_component => {type => SCALAR},
                linked_method    => {type => SCALAR},
                name_argument    => {type => SCALAR},
            }
        );
        my $linked_component = $args{linked_component};
        my $linked_method    = $args{linked_method};
        my $name_argument    = $args{name_argument};

        if ($linked_component->can($linked_method, $error_ptr)) {
            $code = sub {
                $Log->enter() if $may_enter;
                my ($self, %args) = @_;
                if ($method eq 'file_create' && ! exists $args{filename}) {
                    $args{filename} = $pkg_or_obj->random_name_generator(suffix => 'filename');
                }
                my %opts = $self->_common_validate_with(
                    params          => [%args],
                    additional_spec => {$name_argument => {type => SCALAR}},

                    # Allow other options, which are those accepted by the
                    # NACL::C::Volume(File|Directory) method
                    allow_extra => 1
                );

                # The form of the file path is /vol/$vol_name/$file_name
                $opts{path} = "/vol/$opts{volume}/$opts{$name_argument}";
                delete $opts{volume};
                delete $opts{$name_argument};
                if ( $linked_method eq "new" ) {
                    my %pk_with_ci;
                    $linked_component->_copy_primary_keys(
                        source => \%opts,
                        target => \%pk_with_ci
                    );
                    $pk_with_ci{command_interface} = $opts{command_interface};
                    %opts = %pk_with_ci;
                }

                my @ret = $linked_component->$linked_method(%opts);
                $Log->exit() if $may_exit;
                return wantarray ? @ret : $ret[0];
            };
        } else {
            $error_ptr .= "Could not find the method '$linked_method' in "
                . "$linked_component\n";
        }
    };

    if ($method =~ /file_(.*)/) {
        my $method = $1;
        my $linked_component;
        if ($pkg_or_obj->isa('NACL::STask::STask')) {
            $linked_component = 'NACL::STask::VolumeFile';
        } else {
            $linked_component = 'NACL::C::VolumeFile';
        }
        $file_directory_handle->(
            linked_component => $linked_component,
            linked_method    => $method,
            name_argument    => 'filename'
        );
    } elsif ($method =~ /directory_(.*)/) {
        my $method = $1;
        $file_directory_handle->(
            linked_component => 'NACL::C::VolumeDirectory',
            linked_method    => $method,
            name_argument    => 'dirname'
        );
    }

    if ($code) {
        $Log->exit() if $may_exit;
        return $code;
    }

    if ($method =~ /flexcache_/) {
        if ($method eq 'flexcache_find') {
            $code = sub {
                $Log->enter() if $may_enter;
                my ($self, @args) = @_;
                my %opts = $self->_common_validate_with(
                    params      => \@args,
                    allow_extra => 1
                );
                my %filter;
                $self->_hash_move(
                    source => \%opts,
                    target => \%filter,
                    move   => ['vserver'],
                    map    => {volume => 'origin-volume'}
                );
                my @caches = NACL::C::VolumeFlexcache->find(%opts,
                    filter => \%filter);
                $Log->exit() if $may_exit;
                return wantarray ? @caches : $caches[0];
            };
        } else {
            $method =~ /flexcache_(.*)/;
            my $specific_method = $1;
            $code = sub {
                $Log->enter() if $may_enter;
                my ($self, @args) = @_;
                my %opts = $self->_common_validate_with(
                    params      => \@args,
                    allow_extra => 1
                );
                $opts{'origin-volume'} = delete $opts{volume};
                my $c_or_t    = $self->is_c_or_t();
                my $cache_pkg = "NACL::${c_or_t}::VolumeFlexcache";
                my @ret       = $cache_pkg->$specific_method(%opts);
                $Log->exit() if $may_exit;
                return wantarray ? @ret : $ret[0];
            };
        }
    }

    if ($code) {
        $Log->exit() if $may_exit;
        return $code;
    }

    $code = $pkg_or_obj->_handle_component_aggregation(
        method    => $method,
        error_ptr => $error_ptr,
        table     => [
            {   component => 'NACL::C::VolumeEfficiency',
                prefix    => 'efficiency'
            },
            {   component => 'NACL::C::VolumeSnapshot',
                prefix    => 'snapshot'
            },
            {   component => 'NACL::C::VolumeMove',
                prefix    => 'move'
            },
            {   component => 'NACL::C::VolumeQtree',
                prefix    => 'qtree'
            },
            {   component => 'NACL::C::Lun',
                prefix    => 'lun'
            },
            {   component => 'NACL::C::VolumeSnaplock',
                prefix    => 'snaplock',
            },
        ],
    );

    if ($code) {
        $Log->exit() if $may_exit;
        return $code;
    } elsif ($code = $pkg_or_obj->SUPER::can(@opts)) {
        $Log->exit() if $may_exit;
        return $code;
    } else {
        $Log->exit() if $may_exit;
        return undef;
    }
} ## end sub can

=head2 clone_create

    my $clone_c_obj     = $vol_c_obj->clone_create(flexclone => $clone_name);

    my $clone_stask_obj = $vol_stask_obj->clone_create();

Creates a clone of the volume on which the method was invoked. The
C<command_interface>, C<vserver> and C<parent-volume> arguments do not
need to be passed since these are extracted from the volume object.

If invoked with a component object, then the arguments that can be passed
to this method are those accepted by
L<the component|lib-NACL-C-VolumeClone-pm/create> and the object returned is
of type C<NACL::C::VolumeClone>.

If invoked with a task object, then the arguments that can be passed
to this method are those accepted by
L<the task|lib-NACL-STask-VolumeClone-pm/create> and the object returned is
of type C<NACL::STask::VolumeClone>.

=cut

sub clone_create {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params      => \@args,
        allow_extra => 1
    );
    $opts{'parent-volume'} = delete $opts{volume};
    my $clone_pkg
        = $pkg_or_obj->_get_package_at_same_layer(file => 'VolumeClone');
    my $clone_obj = $clone_pkg->create(%opts);
    $Log->exit() if $may_exit;
    return $clone_obj;
}

=head2 clone_find

    # Scalar context, called with component object
    my $clone_c_obj = $vol_c_obj->clone_find();

    # List context, called with component object
    my @clone_c_objs = $vol_c_obj->clone_find();

    # Scalar context, called with task object
    my $clone_stask_obj = $vol_stask_obj->clone_find();

    # List context context, called with task object
    my @clone_stask_objs = $vol_stask_obj->clone_find();

    # Accepts all the options of find(). (Example here shows using
    # allow_empty)
    $vol_obj->clone_find(allow_empty => 1);

    # Also callable on a VolumeClone object, provides the same features
    # (scalar/list context; returns C/STask object based on type of object
    # with which method was called; accepts all options of find)
    my @clones = $clone_obj->clone_find();

Finds and returns objects for any flexclones of this volume.

=cut

sub clone_find {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {filter => {type => HASHREF, default => {}}},
        allow_extra     => 1
    );
    my %filter;
    $pkg_or_obj->_move_primary_keys(
        source => \%opts,
        target => \%filter
    );
    $filter{'parent-volume'} = delete $filter{volume};
    $opts{filter} = {%{$opts{filter}}, %filter};
    my $clone_pkg
        = $pkg_or_obj->_get_package_at_same_layer(file => 'VolumeClone');
    my @clones = $clone_pkg->find(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @clones : $clones[0];
}

=head2 get_recovery_queue_obj

    $vol->delete();
    my $recovery_queue_obj = $vol_obj->get_recovery_queue_obj();
    # ... Perform operations on $recovery_queue_obj ...

With the introduction of the volume undelete feature in FS.0 (see
https://wikid.netapp.com/w/WAFL/Volume_Undelete/TDS), it is possible
to partially delete a volume. Partially deleted volumes get moved to a
recovery queue. To get the
L<NACL::C::VolumeRecoveryQueue|lib-NACL-C-VolumeRecoveryQueue-pm> object
related to the volume that was partially deleted, this method can be used.

If the entry is not found in the recovery queue, then a
C<NACL::Exceptions::NoElementsFound> exception is thrown.

=cut

sub get_recovery_queue_obj {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(params => \@args);
    require NACL::C::VolumeRecoveryQueue;
    my %primary_keys;
    $pkg_or_obj->_move_primary_keys(
        source => \%opts,
        target => \%primary_keys
    );

    # The DEL volume will be of name "<origname>_<some_integer>"
    $primary_keys{volume} .= '_*';
    my $recovery_queue = NACL::C::VolumeRecoveryQueue->find(%opts,
        filter => \%primary_keys,);
    $Log->exit() if $may_exit;
    return $recovery_queue;
}
use warnings qw(redefine);

=head1 METHODS

=head2 new

  my $volume = NACL::C::Volume->new(command_interface=>$ci,volume => $vol_name,
                                  vserver => $vserver_name);

Return a new Volume component object with the given attributes.

=cut

use Class::MethodMaker [
    scalar => 'volume',
    scalar => 'vserver',

    # Not a primary key, but used to determine which ZAPI to invoke.
    # Hence, is a private attribute and is undocumented.
    scalar => '_is_infinivol',

    # FlexGroups require different settings
    scalar => '_is_flexgroup',

    # Old attribute, will be removed in an upcoming change. All the code
    # using this attribute will also be updated in that change
    scalar => '_is_repository',

    scalar => [
        {   '-read_cb' => sub {
                my $self     = $_[0];
                my $aggr_obj = $self->get_aggregate();
                return $aggr_obj->aggregate();
            },
            '-store_cb' => sub {
                my ($self, $value) = @_;
                $Log->warn('Trying to store \'aggregate\' information in '
                        . 'volume object persistently may cause this information '
                        . 'to go stale at any point. Eg: volume move operation '
                        . "\nPlease update your call not to store this "
                        . 'information in the object. When you invoke '
                        . '$obj->aggregate, Component will automatically '
                        . 'fetch the current value by querying the filer.');
                return $value;
            },
        },
        'aggregate'
    ],
];

# Backwards compatibility for the "is_repository" method
no strict 'refs';
*is_repository = *_is_repository;
use strict 'refs';

__PACKAGE__->_load_event_class();

sub init {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    $self->SUPER::init(@args);
    $self->_vserver_context_init(@args);
    $Log->exit() if $may_exit;
} ## end sub init

# For CMode, fill in the _is_infinivol attribute as well. Do so by
# requesting for the "volume-style" field from the Volume CS and if its
# value is infinitevol, set the attribute to 1.
sub find {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $pkg = $pkg_or_obj->get_package_name();

    # If invoked with an object, extract the primary keys and send that
    # as the filter. Also extract the command interface.
    if (ref $pkg_or_obj) {
        my %filter = $pkg_or_obj->get_primary_keys_hash();
        @args = (
            @args,
            filter            => \%filter,
            command_interface => $pkg_or_obj->command_interface()
        );
    }

    my %opts = validate @args, $pkg->_find_validate_spec();
    my @primary_keys
        = (
        $pkg->get_primary_keys(command_interface => $opts{command_interface})
        );

    my @extra_req_fields;
    if ($opts{command_interface}->is_cmode()) {
        push @extra_req_fields, 'volume-style';
    }

    my @state_objs = NACL::CS::Volume->fetch(%opts,
        requested_fields => [@primary_keys, @extra_req_fields]);
    my @components;

    foreach my $state_obj (@state_objs) {
        my %new_hash;
        foreach my $primary_key (@primary_keys) {
            my $value = $state_obj->$primary_key();
            $new_hash{$primary_key} = $value if (defined $value);
        }

        my $is_infinivol = $state_obj->is_infinivol();
        $new_hash{_is_infinivol} = $is_infinivol if (defined $is_infinivol);

        push @components,
            $pkg->new(%new_hash,
            command_interface => $opts{command_interface},);
    }

    $Log->exit() if $may_exit;
    return wantarray ? @components : $components[0];
}

# Make "is_system_vol" a valid option.
sub _find_validate_spec {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    my %spec = %{$pkg->SUPER::_find_validate_spec(@opts)};
    $spec{is_system_vol} = {type => SCALAR, optional => 1};
    $Log->exit() if $may_exit;
    return {%spec};
}

=head2 create

  my $volume  = NACL::C::Volume->create(command_interface=>$ci,volume =>$vol_name,
                                      vserver => $vserver_name);

(Class method) Request creation of a new volume element, and  return a new
volume component object that refers to it.  Does not wait for volume
creation to complete.

Uses a 7Mode CLI/ZAPI, CMode CLI/ZAPI APISet.

Invokes volume-create-async/volume-create API's for CMode ZAPI, 7Mode ZAPI.

Invokes "vol create" command for 7Mode ZAPI.

=over

=item NACL_VGE_ENCRYPT

Global flags that affect the behavior of create command
All the volumes created would be encrypted if this parameter is provided

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: volume-name

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: volume

=item C<< vserver=>$string >>

(Required for CMode, Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< disklist => [$disk1, $disk2, $disk3, ... ] >>

Applicable for CMode 7Mode CLI/ZAPI.

Maps to:

7M CLI : -d <value> (Programmatic name: "disks")

7M ZAPI: disks

=item C<< "mirror-disks" => [$disk1, $disk2, $disk3, ... ] >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: mirror-disks

7M CLI : mirror disks (Programmatic name: "mirror-disks")

=item C<< "force" => $boolean >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -f (Programmatic name: "force")

7M ZAPI: force

=item C<< size => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: size

7M CLI : volume size (Programmatic name: "size")

7M ZAPI: size

=item C<< state => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume-state

=item C<< type => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume-type

=item C<< policy => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: export-policy

=item C<< user => $string >>

If the user specified 'user' as string(user name, instead of user-id)
then CMode ZAPI executes NACL::CS::VserverServicesUnixUser->fetch()
method to get the user-id.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: user-id


=item C<< group => $string >>

If the user specified 'group' as string(group name, instead of group id)
then CMode ZAPI executes NACL::CS::VserverServicesUnixGroup->fetch()
method to get the group-id.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: group-id

=item C<< "security-style" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume-security-style

=item C<< "unix-permissions" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: unix-permissions

=item C<< "junction-path" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: junction-path

=item C<< "junction-active" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-junction-active

=item C<< vsroot => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-vserver-root

=item C<< comment => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume-comment

=item C<< "maxdir-size" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: max-dir-size

=item C<< "space-guarantee" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: space-reserve

7M CLI : space reserved (Programmatic name: "space-reserve")

=item C<< "NACL_SPACE_GUARANTEE" => ntest parameter >>

It is a global parameter which sets the space-guarantee in volume create when it is passed in the parameter       NACL_SPACE_GUARANTEE

=item C<< "snapshot-policy" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: snapshot-policy

=item C<< "stripe-count" => $string >>

Applicable for CMode CLI

=item C<< "stripe-width" => $string >>

Applicable for CMode CLI

=item C<< concurrency => $string >>

Applicable for CMode CLI

=item C<< optimize => $string >>

Applicable for CMode CLI

=item C<< foreground => $string >>

Applicable for CMode CLI

=item C<< "antivirus-on-access-policy" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: antivirus-on-access-policy

=item C<< "stripe-format" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: antivirus-on-access-policy

=item C<< "index-dir-enabled" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: index-dir-enabled

=item C<< "max-write-alloc-blocks" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: max-write-alloc-blocks

=item C<< aggregate => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: containing-aggr-name

7M CLI : aggregate name (Programmatic name: "containing-aggr-name")

7M ZAPI: containing-aggr-name

=item C<< language => $string >>

Applicable for 7Mode CLI/ZAPI. Also Applicable for CMode CLI/ZAPI for builds
that support the Vserver Multiple Languages feature. (i.e. from SN.0 onwards)

Maps to:

CM ZAPI: language-code

7M CLI : -l <value> (Programmatic name: "language-code-for-create")

7M ZAPI: language-code

=item C<< "is-mirrored" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -m (Programmatic name: "is-mirrored")

7M ZAPI: is-mirrored

=item C<< "space-reserve" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: space-reserve

7M CLI : -s <value>(Programmatic name: "space-reserve")

7M ZAPI: space-reserve

=item C<< "remote-location" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -S (Programmatic name: "remote-location")

7M ZAPI: remote-location

=item C<< "snaplock-type" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -S <value> (Programmatic name: "remote-location")

7M ZAPI: remote-location

=item C<< "raid-size" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -r <value> (Programmatic name: "raid-size")

7M ZAPI: raid-size

=item C<< "is-display-command" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -n (Programmatic name: "is-display-command")

7M ZAPI: is-display-command

=item C<< diskcount => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : number of disks (Programmatic name: "disk-count")

7M ZAPI: disk-count

=item C<< "disk-size" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : disk size (Programmatic name: "disk-size")

7M ZAPI: disk-size

=item C<< raidtype => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -t <value> (Programmatic name: "volume-raid-type")

7M ZAPI: volume-raid-type

=item C<<"disk-size" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : disk size (Programmatic name: "disk-size")

7M ZAPI: disk-size

=item C<< "is-snaplock" => $string >>

Applicable for 7Mode ZAPI.

=item C<< "disk-size-with-unit" => $string >>

Applicable for 7Mode ZAPI.

=item C<< "flexcache-origin-volume" => $string >>

Flexcache Origin Volume Name

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: flexcache-origin-volume-name

=item C<< "job_component"=> \$scalar >>

Applicable for CMode CLI

=item C<< 'autosize' => $string >>

Applicable for CMode CLI

=item C<< 'max-autosize' => $string >>

Applicable for CMode CLI

=item C<< 'nvfail' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-nvfail-enabled

=item C<< 'flexcache-fill-policy' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: flexcache-fill-policy

=item C<< 'flexcache-cache-policy' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: flexcache-cache-policy

=item C<< 'autosize-increment' => $string >>

Not supported for DOT:dev and above builds

=item C<< 'directory-bitwidth' => $string >>

Applicable for CMode CLI

=item C<< 'flexcache-min-reserve' => $string >>

Applicable for CMode CLI

=item C<< 'constituent-role' => $string >>

Applicable for CMode CLI

=item C<< 'vm-align-suffix' => $string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -u <value> (Programmatic name: "vm-align-suffix")

=item C<< 'vm-align-sector' => $string >>

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : -e <value> (Programmatic name: "vm-align-sector")

=item C<< 'percent-snapshot-space' => $string >>

Applicable for CMode CLI

=item C<< 'is-repository' => $string >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< 'autosize-mode' => $string >>

Applicable for CMode CLI

=item C<< 'min-autosize' => $string >>

Applicable for CMode CLI

=item C<< 'autosize-shrink-threshold-percent' => $string >>

Applicable for CMode CLI

=item C<< 'autosize-grow-threshold-percent' => $string >>

Applicable for CMode CLI

=item C<< 'ols-aggr-list' => [ $ols-aggr-1, $ols-aggr-2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI

=item C<< 'object-write-sync-period' => $string >>

Applicable for CMode CLI

=item C<< 'is-reserve-constituent' => $string >>

Applicable for CMode CLI

=item C<< 'enable-object-store' => $string >>

Applicable for CMode CLI

=item C<< 'ols-constituent-count' => $string >>

Applicable for CMode CLI

=item C<< 'ols-constituent-size' => $string >>

Applicable for CMode CLI

=item C<< 'qos-policy-group' => $string >>

Applicable for CMode CLI

=item C<< 'is-managed-by-service' => $string >>

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: is-managed-by-service

=item C<< 'storage-service' => $string >>

Applicable for CMode CLI

=item C<< 'enable-snapdiff' => $string >>

Applicable for CMode CLI

=item C<< 'external-cache' => $string >>

External Cache Policy Name

Applicable for CMode CLI

=item C<< 'data-aggr-list' => [ $data-aggr-1, $data-aggr-2, ... ] >>

*List of Aggregates for Data Constituents
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI

=item C<< 'ns-mirror-aggr-list' => [ $ns-mirror-aggr-1, $ns-mirror-aggr-2, ... ] >>

*List of Aggregates for Namespace Mirrors
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI

=item C<< 'max-namespace-constituent-size' => $string >>

*Maximum Size of Namespace Constituent, possible value(s) are
<integer>[KB,MB,GB,TB,PB]

Applicable for CMode CLI

=item C<< 'namespace-aggregate' => $string >>

*Namespace Aggregate

Applicable for CMode CLI

=item C<< 'use-new-algorithm' => $string >>

*Use New Algorithm, possible value(s) are true,false

Applicable for CMode CLI

=item C<< 'percent-reserve-constituent' => $string >>

*Percent Space For Reserve Data Constituent

Applicable for CMode CLI

=item C<< 'max-data-constituent-size' => $string >>

*Maximum Size of each Data Constituent, possible value(s) are
<integer>[KB,MB,GB,TB,PB]

Applicable for CMode CLI

=item C<< 'autosize-increment-percent' => $string >>

Autosize Increment Percent (for flexvols only)

Applicable for CMode CLI

=item C<< 'unreachable-attr-action' => $string >>


Action When Attributes Are Not Reachable possible value(s) are, return-generated,wait
Applicable for CMode CLI

=item C<< 'caching-policy' => $string >>

Caching Policy Name
Applicable for CMode CLI/ZAPI

=item C<< 'junctionFileHandle' => $string >>

Junction File Handle
Applicable for CMode CLI

=item C<< 'snapshot-policy-uuid' => $string >>

Snapshot policy uuid
Applicable for CMode CLI

=item C<< "efficiency-policy" => $string >>

Applicable for CMode CLI/ZAPI.

It is a global parameter which takes preference  when it is passed in the parameter NACL_EFFICIENCY_POLICY

CM ZAPI: efficiency-policy

=item C<< "application-io-size" => $string >>

Applicable for CMode CLI.

It is a global parameter which sets the application-io-size in volume create when it is passed in the parameter      NACL_APPLICATION_IO_SIZE

=item C<< "default-extent-size" => $string >>

Applicable for CMode CLI/ZAPI.

It is a global parameter which sets the extent-size in volume create when it is passed in the parameter NACL_SET_DEFAULT_EXTENT_MVOL_SIZE
Values for this parameter could be: 4K, 8K, 16K, 32K

CM ZAPI:default-extent-size

=item C<< 'qos-policy-group-uuid' => $string >>

QoS Policy Group uuid
Applicable for CMode CLI

=item C<< 'config-replication-flexcache-origin-volume' => $string >>

FlexCache Origin Volume Name for config-replication
Applicable for CMode CLI

=item C<< 'excluded-from-balancer' => $string >>

Is Capacity Balancer Disabled possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'is-vol-on-hybrid-aggr' => $string >>

Is Volume on a hybrid aggregate possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'is-balance-eligible' => $string >>

Is Eligible for the Balancer possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'auto-select-aggregates' => $string>>

It automatically select aggregates for FlexGroup Constituents.
If this is defined then aggr-list, aggregate or aggr-list-multiplier cannot be used with it.
Its possible value(s) are, true,false.

=item C<< 'aggr-list' => $arrayref >>

List of Aggregates for FlexGroup constituents

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI/ZAPI

=item C<< 'aggr-list-multiplier' => $string >>

Aggregate List Repeat Count

Applicable for CMode CLI/ZAPI

=item C<< 'encrypt' => $string >>

Is volume granular encryption enabled. Possible values(s) are true,false
All the volumes created would be encrypted if the global parameter NACL_VGE_ENCRYPT
is provided

Applicable for CMode CLI


=item C<< '*' >>

This method takes all other options supported by
"volume create" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::Volume::AlreadyExists>

This type of exception gets thrown if an attempt is made to create a
volume that already exists.

=item C<NACL::C::Exceptions::Volume::AggregateRequiredForCreate>

This type of exception gets thrown if an attempt is made to create a
volume without providing the "aggregate".

=item C<NACL::C::Exceptions::Volume::InvalidDataConstituentName>

This type of exception gets thrown if an attempt is made to create a
constituent volume with an invalid name.

=item C<NACL::C::Exceptions::Volume::VserverNotInAggrList>

This type of exception gets thrown if an attempt is made to create a volume
on an aggregate which is not added to the vserver aggr-list

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub create {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);

    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "data-aggr-list"      => {type => ARRAYREF, optional => 1},
            "ns-mirror-aggr-list" => {type => ARRAYREF, optional => 1},
            "ols-aggr-list" => {type => ARRAYREF | SCALAR, optional => 1},
            disklist        => {type => ARRAYREF,          optional => 1},
            "mirror-disks"  => {type => ARRAYREF,          optional => 1},
            force           => {type => BOOLEAN,           optional => 1},
            "is-snaplock"   => {type => BOOLEAN,           optional => 1},
            job_component   => {type => SCALARREF,         optional => 1},
            'aggr-list'     => {type => ARRAYREF, optional => 1},

            # 1) "volume" made optional because constituent volumes of an
            # infinivol can be created without specifying a volume name (ONTAP
            # chooses one if not provided)
            # 2) is_infinivol used to specify if the volume is an infinite
            # volume
            _optional_scalars qw(volume is_infinivol
                size state type policy user group security-style
                unix-permissions junction-path junction-active vsroot comment
                maxdir-size space-guarantee snapshot-policy stripe-count stripe-width
                concurrency optimize foreground antivirus-on-access-policy stripe-format
                index-dir-enabled max-write-alloc-blocks aggregate language
                is-mirrored space-reserve remote-location snaplock-type raid-size
                is-display-command diskcount disk-size raidtype disk-size
                is-snaplock disk-size-with-unit flexcache-origin-volume autosize
                max-autosize nvfail flexcache-fill-policy flexcache-cache-policy
                autosize-increment directory-bitwidth flexcache-min-reserve
                constituent-role vm-align-suffix vm-align-sector
                percent-snapshot-space is-repository
                autosize-mode min-autosize autosize-shrink-threshold-percent
                autosize-grow-threshold-percent
                object-write-sync-period is-reserve-constituent enable-object-store
                ols-constituent-count ols-constituent-size qos-policy-group
                is-managed-by-service storage-service enable-snapdiff
                external-cache
                max-namespace-constituent-size namespace-aggregate use-new-algorithm
                percent-reserve-constituent max-data-constituent-size
                autosize-increment-percent unreachable-attr-action
                caching-policy junctionFileHandle snapshot-policy-uuid

                qos-policy-group-uuid

                config-replication-flexcache-origin-volume

                excluded-from-balancer

                is-vol-on-hybrid-aggr

                is-balance-eligible

                aggr-list-multiplier

                encrypt

                default-extent-size

                ),
        },
        cli_cmd => 'volume create',
    );

    my $param_global = param_global();
    my $efficiency_policy = $param_global->get('NACL_EFFICIENCY_POLICY');
    if ($efficiency_policy) {
               $Log->comment("Setting 'efficiency-policy' to '$efficiency_policy' "
                    . "because parameter 'NACL_EFFICIENCY_POLICY' was provided");
            $opts{'efficiency-policy'} = $efficiency_policy;

    }
    my $default_extent_size = $param_global->get('NACL_SET_DEFAULT_EXTENT_MVOL_SIZE');
    if($default_extent_size) {
        $Log->comment("Setting 'default-extent-size' to '$default_extent_size' "
                . "because parameter 'NACL_SET_DEFAULT_EXTENT_MVOL_SIZE' was provided");
        $opts{'default-extent-size'} = $default_extent_size;
    }
    my $space_guarantee = $param_global->get('NACL_SPACE_GUARANTEE');
    if ($space_guarantee) {
        $Log->comment("Setting 'space-guarantee' to '$space_guarantee' "
                . "because parameter 'NACL_SPACE_GUARANTEE' was provided");
        $opts{'space-guarantee'} = $space_guarantee;
    }

    my $application_io_size = $param_global->get('NACL_APPLICATION_IO_SIZE');
    if ($application_io_size) {
        $Log->comment("Setting 'application-io-size' to '$application_io_size' "
                . "because parameter 'NACL_APPLICATION_IO_SIZE' was provided");
        $opts{'application-io-size'} = $application_io_size;
    }
    my (@existing_vols, %common_opts);
    if (  !defined $opts{volume}
        && defined $opts{'constituent-role'}
        && $opts{command_interface}->is_cmode())
    {

        # It is possible to create constituent volumes without specifying
        # the name. If "volume" is not supplied we check if the user wants
        # to create a constituent volume (requires "constituent-role" to be
        # provided and is possible only for CMode). If so, we check the list
        # of volumes currently existing in the vserver. After the call to
        # create, we will query for the list again so that we can determine
        # which volume got created.
        # Note that if "volume" is not provided and we're not creating a
        # constituent volume, then we can just let it go down to the API
        # call which would end up failing.
        $pkg->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_opts
        );
        @existing_vols = $pkg->find(
            %common_opts,
            filter => {
                vserver          => $opts{vserver},
                'is-constituent' => 'true'
            },
            allow_empty => 1
        );
    }

    my $global_encrypted_volume = param_global->get('NACL_VGE_ENCRYPT');
    if ( $global_encrypted_volume ) {
        $Log->comment("Only encrypted volumes would be created because"
            . " the parameter 'NACL_VGE_ENCRYPT' was provided");

        $opts{'encrypt'} = 'true';
    }

    my $is_infinivol_ref;
    $opts{is_infinivol_ref} ||= \$is_infinivol_ref;

    try {
        $pkg->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_create_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_create_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
	                check     => '_check_zapi_option',
                },
                {   method    => '_create_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_create_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },

            ],
            frontend => 'NACL::C::Volume::create',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];

        foreach my $type (
            qw(AlreadyExists AggregateRequiredForCreate
            InvalidDataConstituentName VserverNotInAggrList
            )
            )
        {
            my $full_type = "NACL::C::Exceptions::Volume::$type";
            $full_type->detect_convert_and_throw(exception => $exception);
        }

        NACL::C::Exceptions::NotEnoughSpaceOnAggregate->detect_convert_and_throw(
            exception => $exception);

        $Log->exit() if $may_exit;
        $exception->throw();
    };

    unless (defined $opts{volume}) {

        # If we've reached here, then the API must have passed, hence it
        # must have been a creation on a constituent volume. Let's re-query
        # the list of volumes in the vserver and compare against the original
        # list so that we can determine which one got created.
        my @new_vols = $pkg->find(
            %common_opts,
            filter => {
                vserver          => $opts{vserver},
                'is-constituent' => 'true',
            },
        );
        my %new_vols_hash = map { $_->volume() => 1 } @new_vols;
        foreach my $existing_vol (@existing_vols) {
            delete $new_vols_hash{$existing_vol->volume()};
        }

        # The one existing key in the hash is the name of the newly created
        # volume
        my @new_vol_arr = keys %new_vols_hash;
        $opts{volume} = $new_vol_arr[0];
    }

    my %new_hash = (volume => $opts{volume});
    if ($opts{command_interface}->is_cmode()) {
        $new_hash{vserver} = $opts{vserver};
    }

    my $self
        = $pkg->new(%new_hash, command_interface => $opts{command_interface});

# "aggr-list" or  "auto-select-aggregates" can be supplied for FlexGroups
    if ($opts{'aggr-list'} || $opts{'auto-select-aggregates'}) {
        $self->_is_flexgroup(1);
        $self->_is_infinivol(0);
    } else {
        $self->_is_flexgroup(0);
        $self->_update_infinivol_val_in_self(%opts);
    }

    $Log->exit() if $may_exit;
    return $self;

} ## end sub create

sub _create_7mode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts   = $pkg->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my %vol_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => [
            qw(volume force is-mirrored space-reserve size
                remote-location snaplock-type raid-size mirror-disks
                is-display-command disk-size vm-align-sector vm-align-suffix)
        ],
        map => {
            "disklist"        => "disks",
            "aggregate"       => "containing-aggr-name",
            "raidtype"        => "volume-raid-type",
            "diskcount"       => "disk-count",
            'space-guarantee' => 'space-reserve',
            "language"        => "language-code-for-create",
        },
        target => \%vol_args
    );

    foreach my $list_field (qw(disks mirror-disks)) {
        if (defined $vol_args{$list_field}) {
            $vol_args{$list_field} = join ' ', @{$vol_args{$list_field}};
        }
    }

    $pkg->_translate_size_suffixes(
        options_to_translate => [qw(size)],
        api_opts             => \%vol_args,
    );

    $apiset->vol_create(%vol_args);
    $Log->exit() if $may_exit;
} ## end sub _create_7mode_cli

sub _create_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my $vserver;
    $pkg->_base_backend_cmode_cli(
        @args,
        api                 => 'volume_create',
        vserver_operated_on => \$vserver
    );
    $Log->exit() if $may_exit;
} ## end sub _create_cmode_cli

sub _create_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts              = $pkg->_validate_backend_opts(@args);
    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $copy = [
        qw(antivirus-on-access-policy constituent-role
            flexcache-cache-policy flexcache-fill-policy
            junction-path max-data-constituent-size
            max-namespace-constituent-size max-write-alloc-blocks
            size snapshot-policy storage-service unix-permissions
            vm-align-sector vm-align-suffix caching-policy)
    ];
    my $map = {
        'type'                    => 'volume-type',
        'state'                   => 'volume-state',
        'security-style'          => 'volume-security-style',
        'comment'                 => 'volume-comment',
        'user'                    => 'user-id',
        'maxdir-size'             => 'max-dir-size',
        'vsroot'                  => 'is-vserver-root',
        'nvfail'                  => 'is-nvfail-enabled',
        'junction-active'         => 'is-junction-active',
        'group'                   => 'group-id',
        'flexcache-origin-volume' => 'flexcache-origin-volume-name',
        'policy'                  => 'export-policy',
        'aggregate'               => 'containing-aggr-name',
        'percent-snapshot-space'  => 'percentage-snapshot-reserve',
        'qos-policy-group'        => 'qos-policy-group-name',
        'language'                => 'language-code',
    };

    my $autosize;
    my $autosize_args = $pkg->_hash_move(
        source => \%opts,
        move   => [qw(autosize autosize-mode max-autosize min-autosize
            autosize-grow-threshold-percent autosize-shrink-threshold-percent
            autosize-increment autosize-increment-percent)],
    );
    if (keys %$autosize_args) {
        $pkg->_hash_copy(
            source => \%opts,
            target => $autosize_args,
            copy   => [qw(command_interface volume vserver)]
        );
        $autosize = 1;
    }

    # This code gets the user-id, and group-id,
    # If the user specified these options as strings(names, instead of id's)
    if (defined $opts{'user'} && $opts{'user'} !~ /\d+/) {
        my $cs_obj = NACL::CS::VserverServicesUnixUser->fetch(
            command_interface => $command_interface,
            filter            => {'user' => $opts{'user'}},
            requested_fields  => ['id'],
            apiset_should     => {interface => 'ZAPI'},
        );
        $opts{'user'} = $cs_obj->id();
    }

    if (defined $opts{'group'} && $opts{'group'} !~ /\d+/) {
        my $cs_obj = NACL::CS::VserverServicesUnixGroup->fetch(
            command_interface => $command_interface,
            filter            => {'name' => $opts{'group'}},
            requested_fields  => ['id'],
            apiset_should     => {interface => 'ZAPI'},
        );
        $opts{'group'} = $cs_obj->id();
    }

    my $response;

    # The -async ZAPI will need to be used if we're attempting to create an:
    # infinivol or a FlexGroup (presence of 'aggr-list' indicates creation of
    # an FG).
    my ($use_async, $is_infinivol);
    if (exists $opts{'aggr-list'}) {
        $use_async = 1;
        $is_infinivol = 0;
    } else {
        $is_infinivol = $pkg->_determine_if_infinivol_being_created(%opts);
        $use_async = $is_infinivol;
    }

    ${$opts{is_infinivol_ref}} = $is_infinivol;
    my $convert_to_bytes_list = [qw(max-data-constituent-size
        max-namespace-constituent-size maxdir-size)];

    my $api;
    if ($use_async) {
        $api = 'volume-create-async';
        $map->{volume} = 'volume-name';
        push @$copy, ('space-guarantee', 'is-managed-by-service', 'aggr-list');
        # async ZAPI requires size to be provided in bytes; does not support
        # size suffixes (burt 929101)
        push @$convert_to_bytes_list, 'size';
    } else {
        $api = 'volume-create';
        push @$copy, 'volume';
        $map->{'space-guarantee'} = 'space-reserve';
    }

    my $vserver;
    $pkg->_base_backend_cmode_zapi(
        %opts,
        api                     => $api,
        vserver_operated_on     => \$vserver,
        copy                    => $copy,
        map                     => $map,
        zapi_field_translations => {
            convert_to_bytes  => $convert_to_bytes_list,
            remove_percentage => [qw(percent-snapshot-space)],
            on_off_boolean    => [qw(nvfail)],
        },
        frontend_method     => 'create',
        inapplicable_for_help_xml => [qw(foreground)],
    );

    if ($autosize) {
        # burt 1081604 - if we are going to perform a modify after the create,
        # we should ensure that the volume is created prior to running the modify
        # operation. We can only do this if a job component object was provided
        # and if the instance being used has the ability to wait
        if ($opts{job_component} and $pkg->can("wait_for_creation")) {
            $pkg->wait_for_creation(job_component     => ${$opts{job_component}},
                                    volume            => $opts{volume},
                                    vserver           => $vserver,
                                    polling_interval  => 5, # poll every 5 seconds
                                    apiset_should     => {interface => 'ZAPI'},
                                    command_interface => $command_interface);
        }
        $pkg->modify(%$autosize_args, apiset_should => {interface => 'ZAPI'})
    }

    $Log->exit() if $may_exit;
} ## end sub _create_cmode_zapi

sub _create_7mode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts   = $pkg->_validate_backend_opts(@args);
    my $apiset = $opts{apiset};

    my $vol_args = [
        qw(volume force is-mirrored space-reserve size
            remote-location snaplock-type raid-size mirror-disks
            is-display-command disk-size disk-size-with-unit is-snaplock)
    ];

    my %vol_args;
    $pkg->_hash_copy(
        source => \%opts,
        copy   => $vol_args,
        map    => {
            "language"  => "language-code",
            "disklist"  => "disks",
            "aggregate" => "containing-aggr-name",
            "raidtype"  => "volume-raid-type",
            "diskcount" => "disk-count",
        },
        target => \%vol_args
    );

    $pkg->_translate_size_suffixes(
        options_to_translate => [qw(size)],
        api_opts             => \%vol_args,
    );

    $apiset->volume_create(%vol_args);

    $Log->exit() if $may_exit;
} ## end sub _create_7mode_zapi

sub _check_zapi_option {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    foreach my $field (qw(index-dir-enabled)) {
       if (exists $opts{$field}) {
          my $msg = 'Could not use ZAPI implementation because field '
             . "'$field' is not accepted by the ZAPI command ";
            $Log->comment($msg);
            $Log->exit() if $may_exit;
            NACL::Exceptions::InvalidChoice->throw($msg);
        }
    }
    $Log->exit() if $may_exit;
}

=head2 delete

  NACL::C::Volume->delete(command_interface=>$ci,volume =>$aggr_name,
                           vserver => $vserver_name);

                            or

  $volume_object->delete(...);

(Class or instance method)This method deletes a volume element.

Uses a 7Mode CLI/ZAPI, CMode CLI/ZAPI APISet.

Invokes "volume-destroy,volume-destroy-async" API for CMode ZAPI,7Mode ZAPI.

Invokes "vol destroy" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: name

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: name

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< foreground =>$string >>

Applicable for CMode CLI

=item C<< "disable-offline-check " =>$string >>

Applicable for CMode CLI

=item C<< force => $boolean >>

(Optional) A boolean when true will do a forceful deleting of the volume .

Applicable for CMode CLI, 7Mode CLI/ZAPI.

Maps to:

7M CLI : -f (Programmatic name: "force")

7M ZAPI: force

=item C<< destroy_corrupt_volume => $boolean >>

(Optional defaults to 0)

If 7-Mode CLI command is used to delete the inconsistent volume,filer throws the
prompt 'This volume is marked corrupt and could be useful for debugging
 purposes, are you sure you want to proceed? ' and asks for user confirmation.

Set destroy_corrupt_volume to 1 to delete inconsistent volume.

This option is not applicable and ignored for 7-Mode ZAPI and C-Mode CLI/ZAPI.

=item C<< "job_component"=> \$scalar >>

=item C<< 'is-repository' => 0 | 1 >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< '*' >>

This method takes all other options supported by
"volume delete" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::Volume::DoesNotExist>

This type of exception is thrown when an attempt is made to delete
a non-existent volume.

=item C<NACL::C::Exceptions::Volume::DeleteSnapMirrorRelationship>

This type of exception gets thrown if an attempt is made to delete a
volume that has a snapmirror relationship.

=item C<NACL::C::Exceptions::Volume::HasFlexcacheOrigin>

This type of exception gets thrown if an attempt is made to delete a
volume that has a flex cache

=item C<NACL::C::Exceptions::Volume::RebalancerRunning>

This type of exception gets thrown if an attempt is made to delete
the volumes when capacity balancer is running on infinite volumes.

=item C<NACL::C::Exceptions::Volume::UnsupportedDELOperation>

This type of exception is thrown when an attempt is made to delete the
volume of type "DEL".

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub delete {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            volume                 => {type => SCALAR},
            force                  => {type => BOOLEAN, optional => 1},
            job_component          => {type => SCALARREF, optional => 1},
            destroy_corrupt_volume => {type => BOOLEAN, default => 0},
            _optional_scalars
                qw( foreground disable-offline-check is-repository
                ),
        },
        cli_cmd => 'volume delete',
    );
    my $ret;

    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_delete_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_delete_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_delete_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
                {   method    => '_delete_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    check     => '_check_cmode_zapi_delete'
                },

            ],
            frontend => 'NACL::C::Volume::delete',
        );

    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        $Log->enter() if $may_enter;
        my $exception_object = $_[0];
        foreach my $type (
            qw(DoesNotExist DeleteSnapMirrorRelationship
            HasClones CannotDeleteRootVolume HasFlexcacheOrigin
            RebalancerRunning UnsupportedDELOperation ApplicationVolumeRemove)
            )
        {
            my $full_type = "NACL::C::Exceptions::Volume::$type";
            $full_type->detect_convert_and_throw(
                exception => $exception_object);
        }
        $Log->exit() if $may_exit;
        $exception_object->throw();
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub delete

sub _delete_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli (
        %opts,
        api => 'volume_delete',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_cli

sub _delete_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface   = delete $opts{command_interface};
    my $apiset              = delete $opts{apiset};
    my $destroy_corrupt_vol = delete $opts{destroy_corrupt_volume};

    my $prompt_answer_yes = ['.*volume is marked corrupt .*' => 'y'];
    my $prompt_answer_no  = ['.*volume is marked corrupt .*' => 'n'];

    my $vol_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume force)],
    );
    %opts = %$vol_args;
    $opts{force} = 1;

    if ($destroy_corrupt_vol) {
        $opts{'connectrec-match_table'} = $prompt_answer_yes;
    } else {
        $opts{'connectrec-match_table'} = $prompt_answer_no;
    }

    $apiset->vol_destroy(%opts);
} ## end sub _delete_7mode_cli

sub _delete_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface   = delete $opts{command_interface};
    my $apiset              = delete $opts{apiset};
    my $destroy_corrupt_vol = delete $opts{destroy_corrupt_volume};

    my $vol_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            volume => "name",
            force  => "force",
        }
    );
    %opts = %$vol_args;
    $opts{force} = 1;

    $apiset->volume_destroy(%opts);
} ## end sub _delete_7mode_zapi

sub _delete_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    delete $opts{force} if ( defined $opts{force} );

    $pkg_or_obj->_backend_cmode_zapi_infinivol_builder(
        %opts,
        api => 'volume-destroy',
        map => {volume => 'name'}
    );

    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

=head2 unmount

  NACL::C::Volume->unmount(command_interface=>$ci,volume =>$aggr_name,
                           vserver => $vserver_name);

                            or

  $volume_object->unmount(...);

(Class or instance method) This method unmounts a volume element.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "volume-unmount" API for CMode ZAPI.

Invokes "exportfs" command for 7Mode CLI.

For agnostic purposes, this function is a no-op for 7Mode

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: volume-name

7M CLI : volume name (Programmatic name: "volume")

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< force => $boolean >>

(Optional) A boolean when true will do a forceful unmounting of the volume .

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: force

7M CLI : -f (Programmatic name: "force")

=item C<< "extended_query" => $string >>

(Optional)

Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume unmount" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::Volume::FailedToDeleteJunction>

This type of exception gets thrown if an attempt is made to unmount a
vserver's volume when root volume is offline.

=item C<NACL::C::Exceptions::Volume::UnmountRoot>

This type of exception gets thrown if an attempt is made to unmount a
vserver's root volume.

=item C<NACL::C::Exceptions::Volume::DoesNotExist>

This type of exception gets thrown if an attempt is made to unmount a
non-existing volume.

=item C<NACL::C::Exceptions::Volume::UnsupportedUnmount>

This type of exception gets thrown if an attempt is made to unmount a
volume of type DP or DC.

=item C<NACL::C::Exceptions::Volume::NotExported>

This type of exception gets thrown if an attempt is made to unmount a
volume which is not exported.

=item C<NACL::C::Exceptions::Volume::OperationNotPermittedOnDrVolume>

This type of exception gets thrown if an operation is not permitted on
Vserver that is onfigured as the destination for Vserver DR.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub unmount {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {force => {type => BOOLEAN, optional => 1},},
        cli_cmd         => 'volume unmount',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_unmount_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_unmount_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
                {   method    => '_unmount_7mode_cli',
                    set       => '7Mode',
                    interface => 'CLI',
                },
            ],
            frontend => 'NACL::C::Volume::unmount',
        );
    } ## end try
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        foreach my $type (
            qw(UnmountRoot FailedToDeleteJunction DoesNotExist UnsupportedUnmount NotExported OperationNotPermittedOnDrVolume)
            )
        {
            my $full_type = "NACL::C::Exceptions::Volume::$type";
            $full_type->detect_convert_and_throw(exception => $exception);
        }

        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
} ## end sub unmount

sub _noop { }    # null action for 7Mode use of CMode-only actions.

sub _unmount_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
        api    => 'volume_unmount',
        copy   => [qw(force)],
        map    => {volume => "volume-name",}
    );
    $Log->exit() if $may_exit;
} ## end sub _unmount_cmode_zapi

sub _unmount_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api     => 'volume_unmount',
    );
    $Log->exit() if $may_exit;
} ## end sub _unmount_cmode_cli

sub _unmount_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %cmd_args;
    $cmd_args{path} = "/vol/" . $opts{volume};
    if (exists($opts{force}) and $opts{force} =~ /true/) {
        $cmd_args{zap} = "true";
    } else {
        $cmd_args{unexport} = "true";
    }

    my $response = $apiset->exportfs(%cmd_args);
    $Log->exit() if $may_exit;
} ## end sub _unmount_7mode_cli

=head2 mount

  NACL::C::Volume->mount(command_interface=>$ci,volume =>$aggr_name,
                           vserver => $vserver_name);

                            or

  $volume_object->mount(...);

(Class or instance method)

For CMode, this method mounts a volume element on a particular vserver's
junction point. (not to be confused with I<client>-side mounting, such as
over NFS).

In 7Mode, exports the volume using "exportfs"

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "volume-mount" API for CMode ZAPI.

Invokes "exportfs,rdfile" commands for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume-name

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< junction-path=>$string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: junction-path

=item C<< policy-override=>$string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: export-policy-override

=item C<< active=>$string >>

Applicable for CMode CLI

=item C<< activate-junction=>$string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: activate-junction

=item C<< '*' >>

This method takes all other options supported by
"volume mount" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item L<NACL::C::Exceptions::Volume::AlreadyMounted|lib-NACL-C-Exceptions-Volume-AlreadyMounted-pm>

This type of exception is thrown when an attempt is made to mount a volume
that has already been mounted at a different junction-path.

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub mount {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw ( junction-path active policy-override
                activate-junction ),
        },
        cli_cmd => 'volume mount',
    );

    my $vserver;
    $opts{vserver_operated_on} = \$vserver;

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_mount_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_mount_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_mount_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::Volume::mount'
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Volume::AlreadyMounted
            ->detect_convert_and_throw(exception => $exception);

        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
} ## end sub mount

sub _mount_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args,
        api     => 'volume_mount',
    );
    $Log->exit() if $may_exit;
} ## end sub _mount_cmode_cli

sub _mount_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_base_backend_cmode_zapi(%opts,
        api    => 'volume_mount',
        copy   => [qw(junction-path activate-junction)],
        map    => {
            "volume"          => "volume-name",
            "policy-override" => "export-policy-override",
        }
    );
    $Log->exit() if $may_exit;
} ## end sub _mount_cmode_zapi

sub _mount_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my %cmd_args;

    my $response = $apiset->rdfile(file => "/etc/exports");
    my $output = $response->get_parsed_output();

    if (grep (/\/vol\/$opts{volume}/, split(/\n/, $output->[0]->{contents})))
    {
        $cmd_args{path} = "/vol/$opts{volume}";
        $apiset->exportfs(%cmd_args);
    } else {
        $cmd_args{persistent} = "/vol/" . $opts{volume};
        $apiset->exportfs(%cmd_args);
    } ## end else [ if ( grep ( /\/vol\/$opts{volume}/...))]

    $Log->exit() if $may_exit;
} ## end sub _mount_7mode_cli

=head2 modify

  NACL::C::Volume->modify(command_interface=>$ci,volume =>$aggr_name,
                           vserver => $vserver_name,...);

                            or

  $volume_object->modify(...);

(Class or instance method) Modify a volume element.

For CMode, the capabilities of the "volume modify" command are available.
For 7Mode, "modify" only knows so far how to modify
the "state" attribute of the  volume element, setting it to be "online",
"offline", or  "restricted".

Uses a 7Mode CLI/ZAPI, CMode CLI/ZAPI APISet.

Invokes "volume-set-total-files" API for CMode ZAPI.

Invokes "volume-online,volume-offline,volume-restrict" API's for 7Mode ZAPI.

Invokes "vol online,vol offline,vol restrict,vol autosize"
commands for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: name

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: name

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: owning-vserver-name

=item C<< size => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: size

7M CLI : volume size (Programmatic name: "size")

7M ZAPI: size

=item C<< state => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: state

7M ZAPI: state

=item C<< policy => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: policy

=item C<< user => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: user-id

=item C<< group => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: group-id

=item C<< "security-style" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: style

=item C<< "unix-permissions" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: permissions

=item C<< "junction-active" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-junction-active

=item C<< comment => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: comment

=item C<< autosize => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: is-enabled

7M ZAPI: is-enabled

=item C<< "max-autosize" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: maximum-size

7M CLI : -m <value> (Programmatic name: "maximum-size")

7M ZAPI: maximum-size

=item C<< "autosize-increment" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: increment-size

7M CLI : -i <value> (Programmatic name: "increment-size")

7M ZAPI: increment-size

Not supported for DOT:dev and above builds

=item C<< files => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: files-total

=item C<< "maxdir-size" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: max-dir-size

=item C<< "space-guarantee" => $string >>

Applicable for CMode CLI

=item C<< "min-readahead" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: minimal-read-ahead

=item C<< "NACL_SPACE_GUARANTEE" => ntest parameter >>

It is a global parameter which sets the space-guarantee in volume create when it is passed in the parameter       NACL_SPACE_GUARANTEE

=item C<< "atime-update" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-atime-update-enabled

=item C<< "snapdir-access" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: snapdir-access-enabled

=item C<< "percent-snapshot-space" => $string >>

Applicable for CMode CLI

=item C<< "snapshot-policy" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: snapshot-policy

=item C<< "stripe-width" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: stripe-width

=item C<< concurrency => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: concurrency

=item C<< optimize => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: optimize

=item C<< "index-dir-enabled" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: index-dir-enabled

=item C<< "max-write-alloc-blocks" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: max-write-alloc-blocks

=item C<< "antivirus-on-access-policy" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume-antivirus-attributes

=item C<< "cifs-delay" => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -t <value> (Programmatic name: "cifs-delay")

7M ZAPI: cifs-delay

=item C<< language => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: language

=item C<< "stripe-format" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: stripe-format

=item C<< "upgrade-64bit-available-space" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: available-space

=item C<< "upgrade-64bit-capacity" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: capacity

=item C<< "upgrade-64bit-grow-space" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: grow-space

=item C<< "upgrade-64bit-percent-used-space" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: used-space

=item C<< "root-dir-gen" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: root-dir-gen

=item C<< "block-type" => $string  >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: block-type

=item C<< "files-private-used" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: files-private-used

=item C<< "files" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: files-total

=item C<< "files-used" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: files-used

item C<< inodefile-private-capacity => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: inodefile-private-capacity

=item C<< inodefile-public-capacity => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: inodefile-public-capacity

=item C<< "job_component"=> \$scalar >>

Applicable for CMode CLI

=item C<< 'read-realloc' => $string >>

Valid values are "on/off/space-optimized".

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM CLI : "read-realloc"

CM ZAPI: "read-realloc"

<volume-attributes>
<volume-performance-attributes>
<read-realloc>space-optimized</read-realloc>
</volume-performance-attributes>
</volume-attributes>

Invokes "volume-modify-iter" API for CMode ZAPI.

7M CLI: "read_realloc"

Invokes "vol options $vol_name read_realloc $val" command for 7Mode CLI.

7M ZAPI: "read_realloc"

<volume-set-option>
<option-name>read_realloc</option-name>
<option-value>space_optimized</option-value>
<volume>$vol_name</volume>
</volume-set-option>

Invokes "volume-set-option" API for 7Mode ZAPI.

=item C<< 'sched-snap-name' => $string >>

Naming Scheme for Automatic Snapshots. Valid values are "create_time" and "ordinal".

Applicable for CMode CLI, 7Mode CLI/ZAPI.

Maps to:

7Mode CLI: "schedsnapname" of "vol options"

7Mode ZAPI: "schedsnapname" option-name of "volume-set-options"

=item C<< 'extent-enabled' => $string >>

Extent Option. Valid values are "on", "off" and "space-optimized".

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CMode ZAPI:

<volume-performance-attributes>
<extent-enabled>
</extent-enabled>
</volume-performance-attributes>

7Mode CLI: "extent" option of "vol options"

7Mode ZAPI: "extent" option-name of "volume-set-options"

Note that in 7Mode, "space_optimized" is a valid value, not
"space-optimized". So, if a call is made like:
NACL::C::Volume->modify(%opts, 'extent-enabled' => 'space-optimized');
the value automatically gets translated to 'space_optimized' in 7Mode.

=item C<< 'snapshot-clone-dependency' => $string >>

Snapshot Cloning Dependency. Valid values are "on" and "off".

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CMode ZAPI:

<volume-snapshot-attributes>
<snapshot-clone-dependency>
</snapshot-clone-dependency>
</volume-snapshot-attributes>
of "volume-modify-iter"

7Mode CLI: "snapshot_clone_dependency" option of "vol options"

7Mode ZAPI: "snapshot_clone_dependency" option-name of "volume-set-option"

=item C<< 'auto-snapshots-enabled' => $string >>

Should Automatic Snapshots be enabled? Valid values are "true" and "false".

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CMode ZAPI:

<volume-snapshot-attributes>
<auto-snapshots-enabled>
</auto-snapshots-enabled>
</volume-snapshot-attributes>
of "volume-modify-iter"

7Mode CLI: "nosnap" option of "vol options"

7Mode ZAPI: "nosnap" option-name of "volume-set-option"

Note that since "nosnap" means the opposite of "auto-snapshots-enabled"
("nosnap" means "should automatic snapshots be disabled?")
in 7Mode the value are flipped, i.e.
'auto-snapshots-enabled' => 'true' internally gets translated to
'nosnap' => 'false'.

=item C<< 'create-ucode' => $true_or_false >>

Use Unicode Directory Format. Valid values are "true" and "false".

Applicable for CMode ZAPI and 7Mode CLI.

Maps to:

CMode ZAPI:

<volume-language-attributes>
<create-ucode> $value </create-ucode>
</volume-language-attributes>
of "volume-modify-iter"

7Mode CLI: "create_ucode" option of "vol options"

=item C<< 'instance-uuid' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: instance-uuid

=item C<< 'provenance-uuid' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: provenance-uuid

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'flexcache-min-reserve' => $string >>

=item C<< 'is-repository' => 0 | 1 >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< 'autosize-mode' => $string >>

Applicable for CMode CLI, 7Mode CLI/ZAPI.

Maps to:

7M CLI : autosize mode (Programmatic name: "autosize-mode")

7M ZAPI: mode

=item C<< 'force-nvfail-on-dr' => $string >>

Applicable for CMode CLI

=item C<< 'min-autosize' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: minimum-size

7M CLI : -minimum-size <value> (Programmatic name: "minimum-size")

7M ZAPI: minimum-size

=item C<< 'autosize-grow-threshold-percent' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: grow-threshold-percent

=item C<< 'autosize-shrink-threshold-percent' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: shrink-threshold-percent

=item C<< 'ols-aggr-list' => [ $ols-aggr-1, $ols-aggr-2, ... ] >>

(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI

=item C<< 'object-write-sync-period' => $string >>

Applicable for CMode CLI

=item C<< 'is-reserve-constituent' => $string >>

Applicable for CMode CLI

=item C<< 'enable-object-store' => $string >>

Applicable for CMode CLI

=item C<< 'ols-constituent-count' => $string >>

Applicable for CMode CLI

=item C<< 'ols-constituent-size' => $string >>

Applicable for CMode CLI

=item C<< 'minimum-size' => $string >>

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : -minimum-size <value> (Programmatic name: "minimum-size")

7M ZAPI: minimum-size

=item C<< 'grow-threshold-percent' => $string >>

Applicable for CMode CLI

=item C<< 'shrink-threshold-percent' => $string >>

Applicable for CMode CLI

=item C<< 'shrink' => $string >>

Applicable for CMode CLI

=item C<< 'on' => $string >>

Applicable for CMode CLI

=item C<< 'off'=> $string >>

Applicable for CMode CLI

=item C<< 'grow-shrink'=> $string >>

Applicable for CMode CLI

=item C<< 'reset'=> $string >>

Applicable for CMode CLI

=item C<< 'qos-policy-group' => $string >>

Applicable for CMode CLI

=item C<< 'enable-snapdiff' => $string >>

Applicable for CMode CLI/ZAPI

=item C<< 'is-managed-by-service' => $string >>

Applicable for CMode CLI/ZAPI

=item C<< 'storage-service' => $string >>

Applicable for CMode CLI/ZAPI

=item C<< 'fc-delegs-enabled' => $string >>

*FlexCache Delegations, possible value(s) are true,false

Applicable for CMode CLI

=item C<< 'autosize-reset' => $string >>

Autosize Reset

Applicable for CMode CLI

=item C<< 'external-cache' => $string >>

External Cache Policy Name

Applicable for CMode CLI

=item C<< 'ns-mirror-aggr-list' => [ $ns-mirror-aggr-1, $ns-mirror-aggr-2, ... ] >>

*List of Aggregates for Namespace Mirrors
(Arrayref, even if only a single value is to be sent it should specified
as an arrayref containing that single value)

Applicable for CMode CLI

=item C<< 'max-namespace-constituent-size' => $string >>

*Maximum Size of Namespace Constituent, possible value(s) are
<integer>[KB,MB,GB,TB,PB]

Applicable for CMode CLI/ZAPI

=item C<< 'use-new-algorithm' => $string >>

*Use New Algorithm, possible value(s) are true,false

Applicable for CMode CLI

=item C<< 'percent-reserve-constituent' => $string >>

*Percent Space For Reserve Data Constituent

Applicable for CMode CLI

=item C<< 'max-data-constituent-size' => $string >>

*Maximum Size of each Data Constituent, possible value(s) are
<integer>[KB,MB,GB,TB,PB]

Applicable for CMode CLI/ZAPI

=item C<< 'space-nearly-full-threshold-percent' => $string >>

Volume Nearly Full Threshold Percent

Applicable for CMode CLI

=item C<< 'space-full-threshold-percent' => $string >>

Volume Full Threshold Percent

Applicable for CMode CLI

=item C<< 'autosize-increment-percent' => $string >>

Autosize Increment Percent (for flexvols only)

Applicable for CMode CLI

=item C<< 'unreachable-attr-action' => $string >>


Action When Attributes Are Not Reachable possible value(s) are, return-generated,wait
Applicable for CMode CLI

=item C<< 'caching-policy' => $string >>

Caching Policy Name
Applicable for CMode CLI/ZAPI

=item C<< 'snapshot-policy-uuid' => $string >>

Snapshot policy uuid
Applicable for CMode CLI

=item C<< 'junctionFileHandle' => $string >>

Junction File Handle
Applicable for CMode CLI

=item C<< 'qos-policy-group-uuid' => $string >>

QoS Policy Group uuid
Applicable for CMode CLI

=item C<< 'config-replication-flexcache-origin-volume' => $string >>

FlexCache Origin Volume Name for config-replication
Applicable for CMode CLI

=item C<< 'filesys-size-fixed' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CMode ZAPI: is-filesys-size-fixed

=item C<< 'excluded-from-balancer' => $string >>

Is Capacity Balancer Disabled possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'is-vol-on-hybrid-aggr' => $string >>

Is Volume on a hybrid aggregate possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'flexgroup-index' => $string >>

FlexGroup Index of the constituent
Applicable for CMode CLI

=item C<< 'is-balance-eligible' => $string >>

Is Eligible for the Balancer possible value(s) are, true,false
Applicable for CMode CLI

=item C<< 'namespace-aggregate' => $string >>

Namespace Aggregate
Applicable for CMode CLI

=item C<< 'snaplock-default-period' => $string >>

Default retention period for files
Applicable for CMode and 7Mode CLI

=item C<< 'snaplock-minimum-period' => $string >>

Minimum retention period for files
Applicable for CMode and 7Mode CLI

=item C<< 'snaplock-maximum-period' => $string >>

Maximum retention period for files
Applicable for CMode and 7Mode CLI

=item C<< 'snaplock-autocommit-period' => $string >>

Set default retention value for files
Applicable for CMode and 7Mode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume modify" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::Volume::DoesNotExist>

This type of exception is thrown when an attempt is made to modify the
attributes of a non-existent volume.

=item C<NACL::C::Exceptions::Volume::AlreadyModified>

This type of exception is thrown when an attempt is made to modify the
attributes of a volume into a state it's already in. For example, attempting
to offline an already offline volume, or force-online an already online
volume will both result in this sort of exception.

=item C<NACL::C::Exceptions::Volume::StateUnknown>

This type of exception is thrown when an attempt is made to modify the
attributes of a volume whose state is unknown ("-")

=item C<NACL::C::Exceptions::Volume::CurrentlyBusy>

This type of exception is thrown when an attempt is made to modify the
attributes of a volume when the volume is "dynamically growing state" or busy.

=item C<NACL::C::Exceptions::Volume::VolumeOffline>

This type of exception is thrown when an attempt is made to modify the
attributes of a volume which is in offline state.

=item C<NACL::C::Exceptions::Volume::VserverLocked>

This type of exception is thrown when an attempt is made to modify the
attributes of a volume when its vserver is locked (either config-locked
or baseline is being generated. See burt 637356)

=item C<NACL::C::Exceptions::Volume::UnsupportedDELOperation>

This type of exception is thrown when an attempt is made to offline the
volume of type "DEL".

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub modify {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            "ns-mirror-aggr-list" => {type => ARRAYREF, optional => 1},
            "ols-aggr-list" => {type => ARRAYREF | SCALAR, optional => 1},
            job_component   => {type => SCALARREF,         optional => 1},
            _optional_scalars(
                qw (size state policy user group security-style unix-permissions
                    junction-active comment autosize max-autosize autosize-increment
                    files maxdir-size space-guarantee min-readahead atime-update
                    snapdir-access percent-snapshot-space snapshot-policy
                    stripe-width concurrency optimize antivirus-on-access-policy
                    flexcache-cache-policy flexcache-fill-policy nvfail
                    filesys-size-fixed fractional-reserve index-dir-enabled
                    max-write-alloc-blocks convert-ucode space-mgmt-try-first
                    in-nvfailed-state snap-autodelete-enabled
                    snap-autodelete-commitment snap-autodelete-defer-delete
                    snap-autodelete-delete-order snap-autodelete-prefix
                    snap-autodelete-target-free-space snap-autodelete-trigger
                    snap-autodelete-destroy-list cifs-delay auto-snapshots-enabled
                    read-realloc extent-enabled
                    sched-snap-name snapshot-clone-dependency instance-uuid provenance-uuid
                    foreground
                    flexcache-min-reserve is-repository
                    autosize-mode force-nvfail-on-dr min-autosize
                    autosize-grow-threshold-percent autosize-shrink-threshold-percent i2p-enabled
                    object-write-sync-period is-reserve-constituent enable-object-store
                    ols-constituent-count ols-constituent-size
                    qos-policy-group
                    enable-snapdiff is-managed-by-service storage-service
                    fc-delegs-enabled autosize-reset
                    external-cache
                    max-namespace-constituent-size use-new-algorithm
                    percent-reserve-constituent max-data-constituent-size
                    space-nearly-full-threshold-percent space-full-threshold-percent
                    autosize-increment-percent create-ucode

                    unreachable-attr-action
                    caching-policy snapshot-policy-uuid junctionFileHandle

                    qos-policy-group-uuid

                    config-replication-flexcache-origin-volume

                    excluded-from-balancer

                    is-vol-on-hybrid-aggr

                    flexgroup-index

                    is-balance-eligible namespace-aggregate
                    snaplock-default-period snaplock-minimum-period
                    snaplock-maximum-period snaplock-autocommit-period
                    )
            ),
        },
        cli_cmd => 'volume modify',
    );
    my $param_global = param_global();
    my $space_guarantee = $param_global->get('NACL_SPACE_GUARANTEE');
    if ($space_guarantee) {
        $Log->comment("Setting 'space-guarantee' to '$space_guarantee' "
                 . "because parameter 'NACL_SPACE_GUARANTEE' was provided");
        $opts{'space-guarantee'} = $space_guarantee;
    }

    try {
        $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_modify_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_modify_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_modify_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
                {   method    => '_modify_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    check     => '_infinivol_vol_vs_check',
                },
            ],
            frontend => 'NACL::C::Volume::modify',
        );
    } ## end try
    catch NACL::C::Exceptions::Component with {
        my $exception = $_[0];
        $Log->exit() if $may_exit;
        $exception->throw();
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        if (defined $opts{state} && $opts{state} eq "offline") {
            if ($exception->text() =~ /volume\s+(.*?\s)*is offline/i) {
                NACL::C::Exceptions::Volume::AlreadyModified
                    ->convert_and_throw(exception => $exception);
            }
        }
        foreach my $type (
            qw(DoesNotExist AlreadyModified UnsupportedDELOperation StateUnknown VolumeOffline
            VserverLocked MustBeUnmounted CurrentlyBusy ConstituentsInInconsistentState )
            )
        {
            my $full_type = "NACL::C::Exceptions::Volume::$type";
            $full_type->detect_convert_and_throw(exception => $exception);
        }

        my $fileops = 'NACL::C::Exceptions::Volume::FileOperationsInProgress';
        if ($fileops->detect(exception => $exception)) {

            # This error implies that "force-offline" is supported for
            # "volume modify -state <state>"
            $opts{command_interface}->version_manager()
                ->features_set('force_offline_for_volume' => 1);

            $Log->exit() if $may_exit;
            $fileops->convert_and_throw(exception => $exception);
        }

        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
} ## end sub modify

sub _modify_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    my ($enabled_confirmations);
  MODIFY: {
        use warnings qw(exiting);
        try {
            $pkg_or_obj->_base_backend_cmode_cli(
                %opts,
                api => 'volume_modify',
                vserver_operated_on => \$vserver,
            );
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            my $e = shift;
            if ($e->text() =~
                /not\spermitted\swhen\scommand\sconfirmations\sare\sdisabled/i)
            {
                $opts{apiset}->set(confirmations => 'on');
                $enabled_confirmations = 1;
                $opts{additional_args}->{'connectrec-match_table'} =
                  ['want to proceed?' => sub { 'y' }];
                no warnings qw(exiting);
                redo MODIFY;
            } else {
                $e->throw();
            }
        } finally {
            $opts{apiset}->set(confirmations => 'off') if ($enabled_confirmations);
        };
    }
    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_cli

sub _modify_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $volume            = $opts{volume};

    # This code must be run before running any of the other APIs
    # This is because, in CMode, if we attempt modifying the state along
    # with other fields, it attempts to change the state first before
    if (exists $opts{state}) {
        my $state = delete $opts{state};
        if ($state eq 'online') {
            $apiset->vol_online(
                volume => $volume,
                force  => $opts{force}
            );
        } else {
            my %args_offline_restrict;
            $pkg_or_obj->_hash_copy(
                source => \%opts,
                copy   => [qw(volume cifs-delay)],
                target => \%args_offline_restrict
            );
            if ($state eq 'offline') {
                $apiset->vol_offline(%args_offline_restrict);
            } elsif ($state eq 'restricted') {
                $apiset->vol_restrict(%args_offline_restrict);
            } else {
                NATE::Exceptions::Argument->throw("'state' can be any of "
                        . "'online', 'offline' or 'restricted' "
                        . "but the value supplied was '$state'");
            } ## end else [ if ( $state eq 'offline')]
        } ## end else [ if ( $state eq 'online')]
    } ## end if ( exists $opts{state...})

    if (exists $opts{'files'}) {
        my %args_maxfiles;
        $pkg_or_obj->_hash_copy(
            source  => \%opts,
            map     => {volume => "vol-name",files  => "max"},
            target  => \%args_maxfiles
        );
        $apiset->maxfiles(%args_maxfiles, force  => 1 );
    }

    # Modifying the snapshot policy here. CMode allows us to set a snapshot
    # policy for each volume. 7Mode doesn't allow this, but it does allow
    # us to control whether a volume should have automatic snapshot enabled
    # or not. A CMode snapshot-policy of 'none' would in 7Mode mean
    # "switch off auto-snapshots", and any other snapshot-policy would mean
    # "switch on snapshots".

    if (exists $opts{'read-realloc'}) {
        $opts{'read-realloc'} =~ s/-/_/;
    }

    if (exists $opts{'snapshot-policy'}) {
        if ($opts{'snapshot-policy'} eq 'none') {
            $apiset->vol_options(
                volume         => $opts{volume},
                'option-name'  => 'nosnap',
                'option-value' => 'on'
            );
        } else {
            $apiset->vol_options(
                volume         => $opts{volume},
                'option-name'  => 'nosnap',
                'option-value' => 'off'
            );
        } ## end else [ if ( $opts{'snapshot-policy'...})]
        delete $opts{'snapshot-policy'};
    } ## end if ( exists $opts{'snapshot-policy'...})

    # Translate size suffixes to the 7Mode equivalents
    $pkg_or_obj->_translate_size_suffixes(
        api_opts => \%opts,
        options_to_translate =>
            [qw(size max-autosize autosize-increment min-autosize)],
    );

    if (exists $opts{size}) {
        $apiset->vol_size(size => $opts{size}, volume => $volume);
    }

    # Copy and translate the option names
    my $volopts_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => _vol_options_copy(),
        map    => _vol_options_map(),
    );

    my @options_to_flip_then_translate
        = @{_options_to_flip(mode_of_options => 'SEVEN')};

    _flip_option_values(
        hashref        => $volopts_args,
        mode_of_values => 'CLUSTER',
        values_to_flip => \@options_to_flip_then_translate
    );

    _translate_options_values(
        hashref             => $volopts_args,
        direction           => 'CLUSTERTOSEVEN',
        values_to_translate => \@options_to_flip_then_translate
    );

    # Call 'vol options' to set each of the options provided
    while (my ($key, $value) = each %$volopts_args) {
        $apiset->vol_options(
            volume         => $volume,
            'option-name'  => $key,
            'option-value' => $value
        );
    } ## end while ( my ( $key, $value...))

    # Update the hashref to have on/off/reset to be set to 1 depending on the
    # value of the key specified through "variable_name".
    # $hashref->{$variable_name} = 'true' becomes $hashref->{on} = 1
    # $hashref->{$variable_name} = 'false' becomes $hashref->{off} = 1
    # $hashref->{$variable_name} = 'reset' becomes $hashref->{reset} = 1
    my $populate_hash_true_false = sub {
        my %opts = validate(
            @_,
            {   hashref       => {type => HASHREF},
                variable_name => {type => SCALAR},
            }
        );
        my $hash          = $opts{hashref};
        my $variable_name = $opts{variable_name};
        my $variable      = delete $hash->{$variable_name};

        if ($variable eq 'true') {
            $hash->{on} = 1;
        } elsif ($variable eq 'false') {
            $hash->{off} = 1;
        } else {
            NATE::Exceptions::Argument->throw(
                      "Value of '$variable_name' must be 'true' or 'false' "
                    . "but value provided was '$variable'");
        }
    };

    my %autosize_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%autosize_args,
        copy   => [qw(autosize shrink autosize-mode)],
        map    => {
            'max-autosize'                      => 'maximum-size',
            'autosize-increment'                => 'increment-size',
            'min-autosize'                      => 'minimum-size',
            'autosize-grow-threshold-percent'   => 'grow-threshold-percent',
            'autosize-shrink-threshold-percent' => 'shrink-threshold-percent',
        },
    );

    if (exists $autosize_args{autosize} && !$autosize_args{'autosize-mode'}) {
        $populate_hash_true_false->(
            hashref       => \%autosize_args,
            variable_name => 'autosize'
        );
    } ## end if ( exists $autosize_args...)

    if (keys %autosize_args) {
        $apiset->vol_autosize(volume => $volume, %autosize_args);
    }

    if (exists $opts{'snap-autodelete-enabled'}) {

        # Use _translate_options_values to convert true/false to on/off
        _translate_options_values(
            hashref             => \%opts,
            direction           => 'CLUSTERTOSEVEN',
            values_to_translate => [qw(snap-autodelete-enabled)],
        );
        $apiset->snap_autodelete(
            'vol-name' => $volume,
            action     => delete $opts{'snap-autodelete-enabled'}
        );
    } ## end if ( exists $opts{'snap-autodelete-enabled'...})

    my %snap_autodel_map;
    foreach my $snap_autodel_opt (@{_snap_autodelete_fields()}) {
        my $sevenmode_equivalent = $snap_autodel_opt;
        $sevenmode_equivalent =~ s/^snap\-autodelete\-//;
        $sevenmode_equivalent =~ s/-/_/g;
        $snap_autodel_map{$snap_autodel_opt} = $sevenmode_equivalent;
    } ## end foreach my $snap_autodel_opt...
    my %snap_autodelete_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%snap_autodelete_args,
        map    => \%snap_autodel_map,
    );

    # Call 'snap autodelete' to set each of the options provided
    # while ( my ( $key, $value ) = each %snap_autodelete_args ) {
    # $apiset->snap_autodelete(
    # volume         => $volume,
    # 'option-name'  => $key,
    # 'option-value' => $value
    # );
    # }
} ## end sub _modify_7mode_cli

sub _modify_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $volume            = $opts{volume};

    # This code must be run before running any of the other APIs
    # This is because, in CMode, if we attempt modifying the state along
    # with other fields, it attempts to change the state first before
    if (exists $opts{state}) {
        my $state = delete $opts{state};
        if ($state eq 'online') {
            $apiset->volume_online(name => $volume);
        } else {
            my %args_offline_restrict;
            $pkg_or_obj->_hash_copy(
                source => \%opts,
                map    => {'volume' => 'name'},
                copy   => [qw(cifs-delay)],
                target => \%args_offline_restrict
            );
            if ($state eq 'offline') {
                $apiset->volume_offline(%args_offline_restrict);
            } elsif ($state eq 'restricted') {
                $apiset->volume_restrict(%args_offline_restrict);
            } else {
                NATE::Exceptions::Arguments->throw(
                          "'state' can be any of "
                        . "'online', 'offline' or 'restricted' "
                        . "but the value supplied was '$state'");
            } ## end else [ if ( $state eq 'offline')]
        } ## end else [ if ( $state eq 'online')]
    } ## end if ( exists $opts{state...})

    if (exists $opts{'read-realloc'}) {
        $opts{'read-realloc'} =~ s/-/_/;
    }

    # Translate size suffixes to the 7Mode equivalents
    $pkg_or_obj->_translate_size_suffixes(
        api_opts => \%opts,
        options_to_translate =>
            [qw(size max-autosize autosize-increment min-autosize)],
    );

    if (exists $opts{size}) {
        $apiset->volume_size(volume => $volume, 'new-size' => $opts{size});
    }

    # Copy and translate the option names
    my $volopts_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => _vol_options_copy(),
        map    => _vol_options_map(),
    );

    _translate_options_values(
        hashref   => $volopts_args,
        direction => 'CLUSTERTOSEVEN',
        values_to_translate =>
            _option_values_to_translate(mode_of_options => 'SEVEN')
    );

    my @options_to_flip_then_translate
        = @{_options_to_flip(mode_of_options => 'SEVEN')};

    _flip_option_values(
        hashref        => $volopts_args,
        mode_of_values => 'CLUSTER',
        values_to_flip => \@options_to_flip_then_translate
    );

    _translate_options_values(
        hashref             => $volopts_args,
        direction           => 'CLUSTERTOSEVEN',
        values_to_translate => \@options_to_flip_then_translate
    );

    # Call 'volume-set-option' to set each of the options provided
    while (my ($key, $value) = each %$volopts_args) {
        $apiset->volume_set_option(
            volume         => $volume,
            'option-name'  => $key,
            'option-value' => $value
        );
    } ## end while ( my ( $key, $value...))

    my %autosize_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%autosize_args,
        map    => {
            'max-autosize'                      => 'maximum-size',
            'autosize-increment'                => 'increment-size',
            'autosize'                          => 'is-enabled',
            'min-autosize'                      => 'minimum-size',
            'autosize-grow-threshold-percent'   => 'grow-threshold-percent',
            'autosize-shrink-threshold-percent' => 'shrink-threshold-percent',
            'autosize-mode'                     => 'mode'
        },
    );

    if (keys %autosize_args) {
        $apiset->volume_autosize_set(volume => $volume, %autosize_args);
    }

    if (exists $opts{files}) {
        $apiset->volume_set_total_files(
            volume                  => $volume,
            'requested-total-files' => $opts{files}
        );
    } ## end if ( exists $opts{files...})

    my %snap_autodel_map;
    foreach my $snap_autodel_opt (@{_snap_autodelete_fields()}) {
        my $sevenmode_equivalent = $snap_autodel_opt;
        $sevenmode_equivalent =~ s/^snap\-autodelete\-//;
        $sevenmode_equivalent =~ s/-/_/g;
        $snap_autodel_map{$snap_autodel_opt} = $sevenmode_equivalent;
    } ## end foreach my $snap_autodel_opt...
    my %snap_autodelete_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%snap_autodelete_args,
        map    => \%snap_autodel_map,
    );

    # Call 'snap autodelete' to set each of the options provided
    while (my ($key, $value) = each %snap_autodelete_args) {
        $apiset->snapshot_autodelete_set_option(
            volume         => $volume,
            'option-name'  => $key,
            'option-value' => $value
        );
    } ## end while ( my ( $key, $value...))
} ## end sub _modify_7mode_zapi

sub _modify_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);

    my $apiset = $opts{apiset};
    my $volume = $opts{volume} || '';

    my $value;

    # Guaranteed to be a single value.
    # XXX Currently not handling the case of "vserver" not being provided:
    # which would be possible if extended_query is provided
    if (exists $opts{vserver}) {

        # We should invoke the async form of the ZAPI for repository vservers
        $value = $pkg_or_obj->_determine_is_infinivol(%opts);
    }

    my $api;
    if ($value) {
        $api = 'volume-modify-iter-async';
    } else {
        $api = 'volume-modify-iter';
    }

    my $common_map = $pkg_or_obj->_common_zapi_map(
        command_interface => $opts{command_interface});

    # The "volume modify" CLI allows providing values like "+1M", "-5M" to
    # grow/shrink the volume, but the ZAPI only accepts size as an integer.
    # So, if size is provided and begins with +/-, then we determine the
    # current size of the volume, and add/subtract the provided value
    # to determine the final size of the volume.
    my $size = $opts{size};
    if (defined $size && $size =~ /^([+-])(.*)/) {
        my ($sign, $size_without_sign) = ($1, $2);
        my $vol_cs = NACL::CS::Volume->fetch(
            command_interface => $opts{command_interface},
            filter            => {
                vserver => $opts{vserver},
                volume  => $opts{volume},
            },
            requested_fields => ['size'],
            allow_empty      => 1,
            apiset_should    => {interface => 'ZAPI'},
        );
        if ($vol_cs) {
            my $curr_size = $vol_cs->size();
            my $size_provided_in_bytes = $pkg_or_obj->convert_to_bytes($size_without_sign);
            my $final_size = eval "$curr_size $sign $size_provided_in_bytes";
            $opts{size} = $final_size;
        }
    }
    try {
        $pkg_or_obj->_iter_helper(
            %opts,
            api                     => $api,
            show_cmd                => 'volume show',
            map                     => $common_map,
            zapi_field_translations => {
                convert_to_bytes => [
                    qw(upgrade-64bit-grow-space
                        upgrade-64bit-percent-used-space autosize-increment
                        min-autosize max-autosize maxdir-size compression-space-saved
                        dedupe-space-saved sis-space-saved overwrite-reserve
                        overwrite-reserve-required overwrite-reserve-used
                        overwrite-reserve-used-actual size available
                        size-available-for-snapshots total used
                        snapshot-reserve-size)
                ],
                on_off_boolean => [qw(nvfail)],
                remove_percentage => [qw(autosize-grow-threshold-percent
                    autosize-shrink-threshold-percent autosize-increment-percent)],
            },
            does_not_exist_error => {
                sub => sub {
                    my ($vol_name) = validate_pos(@_, {type => SCALAR});
                    return "Volume '$vol_name' not found. Reason: entry "
                        . "doesn't exist";
                },
                args => [$volume]
            },
            does_not_exist_exception =>
                'NACL::C::Exceptions::Volume::DoesNotExist'
        );
    } catch NACL::C::Exceptions::Volume::DoesNotExist with {
        my $exception = $_[0];
        # Attempt to determine DoesNotExist was thrown because it was a DEL
        # volume (DEL volumes cannot be seen by ZAPI).

        my $vol_cs = NACL::CS::Volume->fetch(
            command_interface => $opts{command_interface},
            filter            => {
                vserver => $opts{vserver},
                volume  => $opts{volume},
            },
            requested_fields => ['type'],
            apiset_should    => {interface => 'CLI'},
            allow_empty      => 1,
        );
        if (defined $vol_cs && $vol_cs->type() =~ /DEL/i) {
            $Log->exit() if $may_exit;
            NACL::C::Exceptions::Volume::UnsupportedDELOperation->convert_and_throw(
                exception => $exception);
        } else {
            $Log->exit() if $may_exit;
            $exception->throw();
        }
    };

    $Log->exit() if $may_exit;
} ## end sub _modify_cmode_zapi

=head2 start_check

  NACL::C::Volume->start_check(command_interface=>$ci,volume =>$aggr_name,
                           vserver => $vserver_name,...);

                            or

  $volume_object->start_check(...);

(Class or instance method) This method starts checking a volume element.

Uses a CMode CLI, 7Mode CLI/ZAPI APIset.

Invokes "volume-verify-start" API for 7Mode ZAPI.

Invokes "vol verify start" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI, 7Mode CLI/ZAPI.

Maps to:

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: volume

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI

=item C<< foreground => $string >>

Applicable for CMode CLI

=item C<< repair => $string >>

Applicable for CMode CLI

=item C<< resume => $string >>

Applicable for CMode CLI

=item C<< inode => $string >>

Applicable for CMode CLI

=item C<< "plex-number" => $string >>

Applicable for CMode CLI, 7Mode CLI/ZAPI.

Maps to:

7M CLI : -f <value> (Programmatic name: "plex-number")

7M ZAPI: fix-plex

=item C<< "log-only" => $string >>

Applicable for 7Mode ZAPI.

=item C<< '*' >>

This method takes all other options supported by
"volume start-check" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub start_check {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            volume => {type => SCALAR},
            _optional_scalars qw( repair resume foreground inode plex-number
                log-only)
        },
        api => 'volume_start_check',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_start_check_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_start_check_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_start_check_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::Volume::start_check',
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub start_check

sub _start_check_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume log-only)],
        map    => {"plex-number" => "fix-plex",}
    );
    %opts = %$vol_args;

    $apiset->volume_verify_start(%opts);

} ## end sub _start_check_7mode_zapi

sub _start_check_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],
    );
    %opts = %$vol_args;

    $apiset->vol_verify_start(%opts);

} ## end sub _start_check_7mode_cli

sub _start_check_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'volume_start_check',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _start_check_cmode_cli

=head2 stop_check

  NACL::C::Volume->stop_check(command_interface=>$ci,volume =>$aggr_name,
                              vserver => $vserver_name);

                            or

  $volume_object->stop_check(...);

(Class or instance method) This method stops checking a volume element.

Uses a CMode CLI APISet, 7Mode CLI/ZAPI APISet.

Invokes "volume-verify-stop" API for 7Mode ZAPI.

Invokes "vol verify stop" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI, 7Mode CLI/ZAPI.

Maps to:

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: volume

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume stop-check" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub stop_check {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        api    => 'volume_stop_check',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_stop_check_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_stop_check_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_stop_check_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::Volume::stop_check',
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub stop_check

sub _stop_check_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],
    );
    %opts = %$vol_args;

    $apiset->vol_verify_stop(%opts);

} ## end sub _stop_check_7mode_cli

sub _stop_check_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],
    );
    %opts = %$vol_args;

    $apiset->volume_verify_stop(%opts);

} ## end sub _stop_check_7mode_zapi

sub _stop_check_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli (
        %opts,
        api => 'volume_stop_check',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _stop_check_cmode_cli

=head2 suspend_check

  NACL::C::Volume->suspend_check(command_interface=>$ci,volume=>$vol_name);

                            or

  $volume_object->suspend_check();

(Class or instance method) This method suspends checking a volume element.

Uses a 7Mode CLI/ZAPI APIset.

Invokes "volume-verify-suspend" API for 7Mode ZAPI.

Invokes "vol verify suspend" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: volume

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub suspend_check {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {volume => {type => SCALAR},},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_suspend_check_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_suspend_check_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub suspend_check

sub _suspend_check_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],
    );
    %opts = %$vol_args;

    $apiset->vol_verify_suspend(%opts);

} ## end sub _suspend_check_7mode_cli

sub _suspend_check_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],
    );
    %opts = %$vol_args;

    $apiset->volume_verify_suspend(%opts);

} ## end sub _suspend_check_7mode_zapi

=head2 resume_check

  NACL::C::Volume->resume_check(command_interface=>$ci,volume=>$vol_name);

                            or

  $volume_object->resume_check();

(Class or instance method) This method resumes checking a volume element.

Uses a 7Mode CLI/ZAPI APIset.

Invokes "volume-verify-resume" API for 7Mode ZAPI.

Invokes "vol verify resume" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for 7Mode CLI/ZAPI.

Maps to:

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: volume

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub resume_check {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {volume => {type => SCALAR},},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_resume_check_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_resume_check_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub resume_check

sub _resume_check_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],
    );
    %opts = %$vol_args;

    $apiset->vol_verify_resume(%opts);

} ## end sub _resume_check_7mode_cli

sub _resume_check_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],
    );
    %opts = %$vol_args;

    $apiset->volume_verify_resume(%opts);

} ## end sub _resume_check_7mode_zapi

=head2 offline

  NACL::C::Volume->offline(command_interface=>$ci,volume =>$aggr_name,
                           vserver => $vserver_name,...);

                            or

  $volume_object->offline(...);

(Class or instance method) This method offlines a volume element.

Uses a CMode CLI/ZAPI, 7Mode CLI/ZAPI APIset.

Invokes "volume-offline-async,volume-offline" API for CMode ZAPI.

Invokes "volume-offline" API for 7Mode ZAPI.

Invokes "volume offline" command for 7Mode CLI.

Same operation can be performed using modify method also.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: <option name>

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: name

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI

=item C<< "cifs-delay" => $string >>

Applicable for CMode CLI

=item C<< "job_component"=> \$scalar >>

Applicable for CMode CLI

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'force' => $string >>

Applicable for CMode CLI

=item C<< 'is-repository' => 0 | 1 >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< is_staging_volume => 0 | 1 >>

(Optional)
Needs to be specified as 1 for staging volumes.
This is necessary because offlining staging volumes is possible
only through CLI and it requires confirmations to be enabled.

=item C<< '*' >>

This method takes all other options supported by
"volume offline" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<NACL::C::Exceptions::Volume::CurrentlyBusy>

This type of exception is thrown when an attempt is made to offline the
volume when the volume is "dynamically growing state" or busy.

=item C<NACL::C::Exceptions::Volume::UnsupportedDELOperation>

This type of exception is thrown when an attempt is made to offline the
volume of type "DEL".

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub offline {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            'aggr-list'     => {type => ARRAYREF, optional => 1},
            _optional_scalars(
                qw(is_staging_volume cifs-delay foreground force is-repository )
            ),
        },
        api => 'volume_offline',
    );

    my $ret;

    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_offline_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_offline_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_offline_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
                {   method    => '_offline_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                    check     => '_check_if_staging_volume',
                },
            ],
            frontend => 'NACL::C::Volume::offline',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        foreach my $type (
            qw(CurrentlyBusy UnsupportedDELOperation)
            )
        {
            my $full_type = "NACL::C::Exceptions::Volume::$type";
            $full_type->detect_convert_and_throw(
                exception => $exception);
        }
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub offline

sub _offline_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    my $is_staging_volume = $opts{is_staging_volume};
    my $apiset            = $opts{apiset};
    my $prompts_answers   = ['.*Do you want to proceed? {y|n}:' => 'y'];
    my $priv;

    try {

        if ($is_staging_volume) {
            $apiset->set(confirmations => 'on');
        }
        $pkg_or_obj->_base_backend_cmode_cli(
            %opts,
            api             => 'volume_offline',
            additional_args => {'connectrec-match_table' => $prompts_answers},
        );
    }
    finally {
        if ($is_staging_volume) {
            $apiset->set(confirmations => 'off');
        }
    };
    $Log->exit() if $may_exit;
} ## end sub _offline_cmode_cli

sub _offline_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset   = delete $opts{apiset};
    my $vol_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume cifs-delay)],
    );
    %opts = %$vol_args;

    $apiset->vol_offline(%opts);

} ## end sub _offline_7mode_cli

sub _offline_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset   = delete $opts{apiset};
    my $vol_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(cifs-delay)],
        map    => {"volume" => "name",}
    );
    %opts = %$vol_args;

    $apiset->volume_offline(%opts);

} ## end sub _offline_7mode_zapi

sub _offline_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $api = 'volume-offline';
   # The -async ZAPI will need to be used if we're attempting to create an:
   # FlexGroup (presence of 'aggr-list' indicates creation of an FG).
    if (exists $opts{'aggr-list'}) {
        $api = 'volume-offline-async';
    }
    $pkg_or_obj->_backend_cmode_zapi_infinivol_builder(
        %opts,
        api => $api,
        map => {volume => 'name'}
    );
    $Log->exit() if $may_exit;
}

=head2 online

  NACL::C::Volume->online(command_interface=>$ci,volume =>$aggr_name,
                           vserver => $vserver_name,...);

                            or

  $volume_object->online(...);

(Class or instance method) This method onlines a volume element.

Uses a CMode CLI/ZAPI, 7Mode CLI/ZAPI APIset.

Invokes "volume-online, volume-online-async" API for CMode ZAPI.

Invokes "volume-online" API for 7Mode ZAPI.

Invokes "vol online" command for 7Mode CLI.

Same function can be performed using modify method also.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: name

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: name

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< force => $boolean >>

(Optional) A boolean when true forcefully onlines the volume element.

Applicable for 7Mode CLI.

Maps to:

7M CLI: -f (Programmatic name: "force")

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'is-repository' => 0 | 1 >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< is_staging_volume => 0 | 1 >>

(Optional)
Needs to be specified as 1 for staging volumes.
This is necessary because offlining staging volumes is possible
only through CLI and it requires confirmations to be enabled.

=item C<< '*' >>

This method takes all other options supported by
"volume online" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub online {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( is_staging_volume foreground is-repository
                    )
            ),
            volume => {type => SCALAR},
            force  => {type => BOOLEAN, optional => 1},
            'aggr-list'     => {type => ARRAYREF, optional => 1},

        },
        api => 'volume_online',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_online_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_online_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_online_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
            {   method    => '_online_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
                check     => '_check_if_staging_volume',
            },
        ],
        frontend => 'NACL::C::Volume::online',
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub online

sub _online_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    my $is_staging_volume = $opts{is_staging_volume};
    my $apiset            = $opts{apiset};
    my $prompts_answers   = ['.*Do you want to proceed? {y|n}:' => 'y'];

    try {

        if ($is_staging_volume) {
            $apiset->set(confirmations => 'on');
        }
        $pkg_or_obj->_base_backend_cmode_cli(
            %opts,
            api             => 'volume_online',
            additional_args => {'connectrec-match_table' => $prompts_answers},
        );
    }
    finally {
        if ($is_staging_volume) {
            $apiset->set(confirmations => 'off');
        }
    };
    $Log->exit() if $may_exit;
}

sub _online_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset   = delete $opts{apiset};
    my $vol_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume force)],
    );
    %opts = %$vol_args;
    $apiset->vol_online(%opts);
} ## end sub _online_7mode_cli

sub _online_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    delete $opts{command_interface};
    my $apiset   = delete $opts{apiset};
    my $vol_args = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {"volume" => "name",}
    );
    %opts = %$vol_args;
    $apiset->volume_online(%opts);

} ## end sub _online_7mode_zapi

sub _online_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $api = 'volume-online';
    # The -async ZAPI will need to be used if we're attempting to create an:
    # FlexGroup (presence of 'aggr-list' indicates creation of an FG).
    if (exists $opts{'aggr-list'}) {
        $api = 'volume-online-async';
    }

    $pkg_or_obj->_backend_cmode_zapi_infinivol_builder(
        %opts,
        api => $api,
        map => {volume => 'name'}
    );
    $Log->exit() if $may_exit;
}

=head2 rename

  NACL::C::Volume->rename(command_interface=>$ci,volume =>$aggr_name,
                           vserver => $vserver_name,...);

                            or

  $volume_object->rename(...);

(Class or instance method) This method renames a volume element.

Uses a CMode CLI/ZAPI, 7Mode CLI/ZAPI APIset.

Invokes "volume-rename,volume-rename-async" API for CMode ZAPI.

Invokes "volume-rename" API for CMode ZAPI.

Invokes "vol rename" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: volume

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: volume

=item C<< vserver=>$string >>

(Required for CMode , Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< newname => $new_name >>

(Required) A scalar to which the volume element is to be renamed.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: new-volume-name

7M CLI : new volume name (Programmatic name: "new-volume-name")

7M ZAPI: new-volume-name

=item C<< "job_component"=> \$scalar >>

Applicable for CMode CLI

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'is-repository' => 0 | 1 >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< '*' >>

This method takes all other options supported by
"volume rename" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub rename {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( foreground is-repository
                    )
            ),
            newname       => {type => SCALAR},
            job_component => {type => SCALARREF, optional => 1},
        },
        cli_cmd => 'volume rename',
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_rename_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_rename_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_rename_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
            {   method    => '_rename_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::rename',
    );

    $pkg_or_obj->volume($opts{newname}) if (ref $pkg_or_obj);
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub rename

sub _rename_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'volume_rename',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _rename_cmode_cli

sub _rename_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume"  => "volume",
            "newname" => "new-volume-name",
        }
    );
    %opts = %$vol_args;

    $apiset->vol_rename(%opts);
    $pkg_or_obj->volume($opts{"new-volume-name"}) if ref($pkg_or_obj);

} ## end sub _rename_7mode_cli

sub _rename_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume"  => "volume",
            "newname" => "new-volume-name",
        }
    );
    %opts = %$vol_args;

    $apiset->volume_rename(%opts);

} ## end sub _rename_7mode_zapi

sub _rename_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_backend_cmode_zapi_infinivol_builder(
        @args,
        api => 'volume-rename',
        map => {
            volume  => 'volume',
            newname => 'new-volume-name',
        }
    );
    $Log->exit() if $may_exit;
}

=head2 autosize

    NACL::C::Volume->autosize(
                                command_interface => $ci,
                                volume => $volume,
                                vserver => $vserver,
                                increment-size => $value,
                            ...,
                            );


    # Instance method
    $vol->autosize('increment-size' => $value);

(Class or instance method)

This method is used to set the value of autosize related inputs.
B<NOTE:> Using this method to get the value of autosize-related fields
is deprecated. Either use $vol_obj->state() or NACL::CS::Volume->fetch()
to get these details.

Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "volume-autosize-set" API for CMode ZAPI.

Invokes "vol autosize" command for 7Mode CLI.

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< "volume" => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: volume

7M CLI : volume name (Programmatic name: "volume")

=item C<< "vserver" => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< "increment-size" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: increment-size

7M CLI : -i <value> (Programmatic name: "increment-size")

=item C<< "maximum-size" => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: maximum-size

7M CLI : -m <value> (Programmatic name: "maximum-size")

=item C<< "is-enabled" => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-enabled

=item C<< 'shrink-threshold-percent' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: shrink-threshold-percent

7M CLI : -shrink-threshold-percent <value>
(Programmatic name: "shrink-threshold-percent")

=item C<< 'mode' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: mode

7M CLI : -mode <value> (Programmatic name: "mode")

=item C<< 'grow-threshold-percent' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: grow-threshold-percent

7M CLI : -grow-threshold-percent <value>
(Programmatic name: "grow-threshold-percent")

=item C<< 'minimum-size' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: minimum-size

7M CLI : -minimum-size <value> (Programmatic name: "minimum-size")

=item C<< 'reset' => $string >>

Autosize Reset

Applicable for CMode CLI, 7Mode CLI

Maps to:

7M CLI: reset (Programmatic name: reset)

=item C<< 'increment-percent' => $string >>

Increment Percentage

Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume autosize" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub autosize {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( increment-size maximum-size is-enabled
                    shrink-threshold-percent mode grow-threshold-percent minimum-size
                    reset
                    increment-percent
                    )
            ),
        },
        api => 'volume_autosize',
    );

    my $vserver;
    $opts{vserver_operated_on} ||= \$vserver;
    my @ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_autosize_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_autosize_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
                check     => '_check_zapi_attrs'
            },
            {   method    => '_autosize_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
        ],
        frontend => 'NACL::C::Volume::autosize',
    );

    $Log->exit() if $may_exit;
    return wantarray ? @ret : $ret[0];
} ## end sub autosize

sub _autosize_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);

    # If any field other than volume/vserver is provided, then we're setting
    # a value.
    my $apiset = $opts{apiset};
    my @allowed_options = $apiset->get_allowed_options_for_command(
        command => 'volume_autosize');
    my %allowed_options_hash = map {$_ => 1} @allowed_options;
    delete $allowed_options_hash{volume};
    delete $allowed_options_hash{vserver};
    my $type = 'get';
    foreach my $set_field (keys %allowed_options_hash) {
        if (exists $opts{$set_field}) {
            $type = 'set';
            last;
        }
    }

    my $response;
    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api                 => 'volume_autosize',
        apiset_response_obj => \$response,
    );

    my @state_objs;
    if ($type eq 'get') {
        $pkg_or_obj->_autosize_warning();
        my $row = $response->get_parsed_output()->[0];
        require NACL::CS::VolumeAutosize;
        my $cs = NACL::CS::VolumeAutosize->new();
        $cs->_set_fields(row => $row);
        push @state_objs, $cs;
    }

    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _autosize_cmode_cli

sub _autosize_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);

    # If any field other than volume/vserver is provided, then we're setting
    # a value.
    my @fields_for_set = (
        qw(increment-size maximum-size is-enabled
            shrink-threshold-percent mode grow-threshold-percent minimum-size)
    );
    my @copy = (@fields_for_set, 'volume');
    my $api_type = 'get';
    foreach my $field (@fields_for_set) {
        if (exists $opts{$field}) {
            $api_type = 'set';
            last;
        }
    }
    my $api = "volume-autosize-$api_type";

    my $response;
    $pkg_or_obj->_base_backend_cmode_zapi(
        %opts,
        api                     => $api,
        apiset_response_obj     => \$response,
        copy                    => \@copy,
        zapi_field_translations => {
            on_off_boolean    => [qw(is-enabled)],
            remove_percentage => [qw(shrink-threshold-percent grow-threshold-percent)],
        },
    );

    my @state_objs;
    if ($api_type eq 'get') {
        $pkg_or_obj->_autosize_warning();
        my $row = $response->get_parsed_output()->[0];

        # CLI value has %, so add it on
        foreach my $percent_field (qw(shrink grow)) {
            my $full_percent_field = "$percent_field-threshold-percent";
            my $val = $row->{$full_percent_field};
            if (defined $val) {
                my $cli_val = "$val%";
                $row->{$full_percent_field} = $cli_val;
            }
        }

        # ZAPI returns the value in terms of KB. CLI returns the value with
        # the most appropriate unit, like 4k, 5m etc. We will perform this
        # conversion here.
        foreach my $size_field (qw(minimum maximum increment)) {
            my $full_size_field = "$size_field-size";
            my $val = $row->{$full_size_field};
            if (defined $val) {
                my $orders_magnitude = -1;

                my $final_val;
                while ($val > 0) {
                    $final_val = $val;
                    $val = int ($val / 1024);
                    $orders_magnitude++;
                }

                my @magnitude_list = ('k', 'm', 'g', 't', 'p');
                my $new_val = $final_val . $magnitude_list[$orders_magnitude];
                $row->{$full_size_field} = $new_val;
            }
        }

        # CLI returns state=ON/OFF, which we can determine by looking
        # at the is-enabled value.
        if ($row->{'is-enabled'} eq 'true') {
            $row->{state} = 'ON';
        } else {
            $row->{state} = 'OFF';
        }

        my $vserver = $opts{vserver} || $opts{command_interface}->vserver();
        $row->{volume} = "$vserver:$opts{volume}";

        require NACL::CS::VolumeAutosize;
        my $cs = NACL::CS::VolumeAutosize->new();
        $cs->_set_fields(row => $row);
        push @state_objs, $cs;
    }

    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _autosize_cmode_zapi

sub _autosize_warning {
    my ($pkg_or_obj) = @_;
    my $warn_message = 'Using ' . $pkg_or_obj->get_package_name()
        . '->autosize() to get the autosize details is deprecated. '
        . 'Use $vol_obj->state() or NACL::CS::Volume->fetch() instead '
        . 'to get the autosize details';
    NACL::C::Component->_print_deprecation_warning($warn_message);
}

sub _autosize_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    my $command_interface = $opts{command_interface};
    my $apiset            = $opts{apiset};

    my $allowed_7mode_options = [
        qw( volume increment-size maximum-size
            minimum-size shrink-threshold-percent grow-threshold-percent
            )
    ];

    $pkg_or_obj->_translate_size_suffixes(
        options_to_translate =>
            [qw(maximum-size increment-size minimum-size)],
        api_opts => \%opts,
    );

    my %mode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_7mode_options,
        map    => {'mode' => 'autosize-mode'},
        target => \%mode_args,
    );

    if ($opts{'is-enabled'} && !$opts{'mode'}) {
        $mode_args{'on'}    = 1 if ($opts{'is-enabled'} eq "on");
        $mode_args{'off'}   = 1 if ($opts{'is-enabled'} eq "off");
        $mode_args{'reset'} = 1 if ($opts{'is-enabled'} eq "reset");
    }

    if ($opts{reset} && $opts{reset} eq "true") {
        $mode_args{'reset'} = 1;
    }
    my $response   = $apiset->vol_autosize(%mode_args);
    my $output     = $response->get_parsed_output();
    my @state_objs = $pkg_or_obj->_vol_autosize(%opts, output => $output);
    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _autosize_7mode_cli

=head2 scrub_start

  NACL::C::Volume->scrub_start(command_interface=>$ci,
                             volume=>$vol_name);

                            or

  $volume_object->scrub_start();

(Class or instance method) This method starts scrubbing a volume element.

Uses a 7Mode CLI/ZAPI APIset.

Invokes "volume-scrub-start" API for 7Mode ZAPI.

Invokes "vol scrub start" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : volume name <value> (Programmatic name: "volume")

7M ZAPI: name

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub scrub_start {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_scrub_start_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_scrub_start_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub scrub_start

sub _scrub_start_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],

    );
    %opts = %$vol_args;

    $apiset->vol_scrub_start(%opts);

} ## end sub _scrub_start_7mode_cli

sub _scrub_start_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {"volume" => "name",}
    );
    %opts = %$vol_args;

    $apiset->volume_scrub_start(%opts);

} ## end sub _scrub_start_7mode_zapi

=head2 scrub_stop

  NACL::C::Volume->scrub_stop(command_interface=>$ci,
                             volume=>$vol_name);

                            or

  $volume_object->scrub_stop();

(Class or instance method) This method stops scrubbing a volume element.

Uses a 7Mode CLI/ZAPI APIset.

Invokes "volume-scrub-stop" API for 7Mode ZAPI.

Invokes "vol scrub stop" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : volume name <value> (Programmatic name: "volume")

7M ZAPI: name

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub scrub_stop {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {volume => {type => SCALAR},},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_scrub_stop_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_scrub_stop_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub scrub_stop

sub _scrub_stop_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],

    );
    %opts = %$vol_args;

    $apiset->vol_scrub_stop(%opts);

} ## end sub _scrub_stop_7mode_cli

sub _scrub_stop_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {"volume" => "name",}
    );
    %opts = %$vol_args;

    $apiset->volume_scrub_stop(%opts);

} ## end sub _scrub_stop_7mode_zapi

=head2 scrub_resume

  NACL::C::Volume->scrub_resume(command_interface=>$ci,
                             volume=>$vol_name);

                            or

  $volume_object->scrub_resume();

(Class or instance method) This method resumes scrubbing a volume element.

Uses a 7Mode CLI/ZAPI APIset.

Invokes "volume-scrub-resume" API for 7Mode ZAPI.

Invokes "vol scrub resume" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : volume name <value> (Programmatic name: "volume")

7M ZAPI: name

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub scrub_resume {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {volume => {type => SCALAR},},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_scrub_resume_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_scrub_resume_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub scrub_resume

sub _scrub_resume_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],

    );
    %opts = %$vol_args;

    $apiset->vol_scrub_resume(%opts);

} ## end sub _scrub_resume_7mode_cli

sub _scrub_resume_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {"volume" => "name",}
    );
    %opts = %$vol_args;

    $apiset->volume_scrub_resume(%opts);

} ## end sub _scrub_resume_7mode_zapi

=head2 scrub_suspend

  NACL::C::Volume->scrub_suspend(command_interface=>$ci,
                             volume=>$vol_name);

                            or

  $volume_object->scrub_suspend();

(Class or instance method) This method suspends scrubbing a volume element.

Uses a 7Mode CLI/ZAPI APIset.

Invokes "volume-scrub-suspend" API for 7Mode ZAPI.

Invokes "vol scrub suspend" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI, 7Mode CLI.

Maps to:

7M CLI : volume name <value> (Programmatic name: "volume")

7M ZAPI: name

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub scrub_suspend {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {volume => {type => SCALAR},},
    );
    my $ret = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_scrub_suspend_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_scrub_suspend_7mode_zapi',
                interface => 'ZAPI',
                set       => '7Mode',
            },
        ],
    );
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub scrub_suspend

sub _scrub_suspend_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume)],

    );
    %opts = %$vol_args;

    $apiset->vol_scrub_suspend(%opts);

} ## end sub _scrub_suspend_7mode_cli

sub _scrub_suspend_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {"volume" => "name",}
    );
    %opts = %$vol_args;

    $apiset->volume_scrub_suspend(%opts);

} ## end sub _scrub_suspend_7mode_zapi

=head2 restrict

  NACL::C::Volume->restrict(command_interface=>$ci,
                             volume=>$vol_name);

                            or

  $volume_object->restrict();

(Class or instance method) This method restricts a volume element.

Uses a 7Mode CLI/ZAPI, CMode CLI/ZAPI APISet.

Invokes "volume-restrict, volume-restrict-async" API for CMode ZAPI.

Invokes "volume-restrict" API for 7Mode ZAPI.

Invokes "vol restrict" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI/ZAPI, 7Mode CLI/ZAPI.

Maps to:

CM ZAPI: name

7M CLI : volume name (Programmatic name: "volume")

7M ZAPI: name

=item C<< vserver=>$string >>

(Required for CMode, Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'is-repository' => 0 | 1 >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< '*' >>

This method takes all other options supported by
"volume restrict" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::Volume::AlreadyRestricted>

This type of exception gets thrown if an attempt is made to restrict a
volume which is already restricted.

=back

=back

=cut

sub restrict {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars qw(cifs-delay
                foreground is-repository
                ),
        },
        cli_cmd => 'volume restrict',
    );
    my $ret;
    try {
        $ret = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_restrict_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode',
                },
                {   method    => '_restrict_7mode_zapi',
                    interface => 'ZAPI',
                    set       => '7Mode',
                },
                {   method    => '_restrict_cmode_cli',
                    interface => 'CLI',
                    set       => 'CMode',
                },
                {   method    => '_restrict_cmode_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
            frontend => 'NACL::C::Volume::restrict',
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Volume::AlreadyRestricted
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub restrict

sub _restrict_7mode_cli {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => [qw(volume cifs-delay)],

    );
    %opts = %$vol_args;

    $apiset->vol_restrict(%opts);

} ## end sub _restrict_7mode_cli

sub _restrict_7mode_zapi {
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );
    my $command_interface = delete $opts{command_interface};
    my $apiset            = delete $opts{apiset};
    my $vol_args          = $pkg_or_obj->_hash_copy(
        source => \%opts,
        map    => {
            "volume"     => "name",
            "cifs-delay" => "cifs-delay",
        }
    );
    %opts = %$vol_args;

    $apiset->volume_restrict(%opts);

} ## end sub _restrict_7mode_zapi

sub _restrict_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_backend_cmode_zapi_infinivol_builder(
        @args,
        api => 'volume-restrict',
        map => {volume => 'name'}
    );
    $Log->exit() if $may_exit;
} ## end sub _delete_cmode_zapi

sub _restrict_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'volume_restrict',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _restrict_cmode_cli

=head2 move

  NACL::C::Volume->move(command_interface=>$ci,
                             volume=>$vol_name ...);

                            or

  $volume_object->move();

(Class or instance method) This method moves a volume element to another
aggregate.

Uses a CMode CLI APIset.

=over

=item Options

=over

=item C<< command_interface=>$command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< volume=> $string >>

(Required) Name of volume.

Applicable for CMode CLI

=item C<< vserver=>$string >>

(Required for CMode, Not Applicable for 7Mode)The name of the vserver
in which the volume is present.

Applicable for CMode CLI

=item C<< "destination-aggregate" => $string >>

Applicable for CMode CLI

=item C<< "destination-volume" => $string >>

Applicable for CMode CLI

=item C<< "destination-stripe-format" => $string >>

Applicable for CMode CLI

=item C<< "cutover-window" => $string >>

Applicable for CMode CLI

=item C<< "cutover-attempts" => $string >>

Applicable for CMode CLI

=item C<< "foreground" => $string >>

Applicable for CMode CLI

=item C<< "cutover-action" => $string >>

Applicable for CMode CLI

=item C<< "perform-validation-only" => $string >>

Applicable for CMode CLI

=item C<< "extended_query" => $string  >>

(Optional)
L<< NACL::C::VolumeMove->start|lib-NACL-C-VolumeMove-pm/start >>
If invoked with an STask package or object,
then see the implementation and documentation of provided functionality is present in
L<< NACL::STask::VolumeMove->start|lib-NACL-STask-VolumeMove-pm/start >>

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub move {
    $Log->enter() if $may_enter;
    $Log->warn(
        "volume move should not be used ,volume move start should be used");
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params      => \@args,
        allow_extra => 1
    );
    my $is_c_or_t = $pkg_or_obj->is_c_or_t();
    my $move_pkg  = "NACL::${is_c_or_t}::VolumeMove";

    my $ret = $move_pkg->start(%opts);

    $Log->exit() if $may_exit;
    return $ret;
}

=head2 make_vsroot

    NACL::C::Volume->make_vsroot(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->make_vsroot(...);

(Class or instance method)
This method is used to designate a non-root volume as a root volume of the Vserver.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'aggregate' => $string >>

Aggregate Name
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume make-vsroot" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub make_vsroot {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        cli_cmd    => 'volume make-vsroot',
        ,
        additional_spec => {
            _optional_scalars(
                qw( aggregate
                    )
            )
        },
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_make_vsroot_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::make_vsroot',
    );

    $Log->exit() if $may_exit;
} ## end sub make_vsroot

sub _make_vsroot_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli (
        %opts,
        api => 'volume_make_vsroot',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _make_vsroot_cmode_cli

=head2 add_members

    NACL::C::Volume->add_members(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->add_members(...);

(Class or instance method)
This method is used to add members to an existing striped volume.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< "job_component"=> \$scalar >>

Applicable for CMode CLI

=item C<< 'count' => $string >>

Applicable for CMode CLI

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"volume add-members" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub add_members {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( count foreground
                    )
            ),
        },
        api => 'volume_add_members',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_add_members_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::add_members',
    );

    $Log->exit() if $may_exit;
} ## end sub add_members

sub _add_members_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli (
        %opts,
        api => 'volume_add_members',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _add_members_cmode_cli

=head2 add_other_volumes

    NACL::C::Volume->add_other_volumes(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to add 7-mode volumes to the cluster database.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'node' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"volume add-other-volumes" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub add_other_volumes {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( node
                    )
            ),
        },
        ignore_primary_keys => 1,
        api                 => 'volume_add_other_volumes',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_add_other_volumes_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::add_other_volumes',
    );

    $Log->exit() if $may_exit;
} ## end sub add_other_volumes

sub _add_other_volumes_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'volume_add_other_volumes');
    $Log->exit() if $may_exit;
} ## end sub _add_other_volumes_cmode_cli

=head2 transition

    NACL::C::Volume->transition(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->transition(...);

(Class or instance method)
This method is used to transition a volume from 7-Mode to Cluster-Mode.

Uses a CMode CLI/ZAPI APISet.

Invokes 'volume-transition' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume-name

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'affinity-node' => $string >>

Applicable for CMode CLI/ZAPI.

=item C<< 'check-only' => $string >>

Applicable for CMode CLI

=item C<< 'junction-path' => $string >>

Applicable for CMode CLI/ZAPI.

=item C<< 'source-node' => $string >>

Applicable for CMode CLI/ZAPI.

=item C<< 'nondisruptive' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: non-disruptive

=item C<< 'suppress-warnings' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: override-warnings

=item C<< 'operation-type' => $string >>

Applicable for CMode CLI/ZAPI.

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'allow-7mode' => $string >>

Applicable for CMode CLI

=item C<< 'destination-vserver' => $string >>

*Destination vserver for a 7-mode transition

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: destination-vserver-name

=item C<< '*' >>

This method takes all other options supported by
"volume transition" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub transition {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( affinity-node check-only junction-path source-node nondisruptive
                    suppress-warnings operation-type foreground

                    allow-7mode

                    destination-vserver
                    )
            ),
        },
        cli_cmd => 'volume transition',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_transition_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_transition_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::transition',
    );

    $Log->exit() if $may_exit;
} ## end sub transition

sub _transition_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli (
        %opts,
        api => 'volume_transition',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _transition_cmode_cli

sub _transition_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg, %opts) = @_;
    $pkg->_base_backend_cmode_zapi(
        %opts,
        api => 'volume-transition',
        copy =>
            [qw( affinity-node operation-type source-node junction-path )],
        map => {
            'destination-vserver' => 'destination-vserver-name',
            'nondisruptive'       => 'non-disruptive',
            'suppress-warnings'   => 'override-warnings',
            'volume'              => 'volume-name',
        },
        frontend_method => 'transition'
    );
    $Log->exit() if $may_exit;
} ## end sub _transition_cmode_zapi

=head2 size
     $vol_obj->get_one_state_attribute('size');
or
    NACL::CS::Volume->fetch(
        command_interface => $command_interface,
        filter => {
            volume  => $vol,
            vserver => $vs,
        },
        requested_fields => [qw(size)]
    );
or
   NACL::C::Volume->size(
                         command_interface => $ci,
                         'volume' => $volume,
                         'vserver' => $vserver,
                         ...);

(Class or instance method)
This method is used to set/Display the size of flexible volume..
Uses a CMode CLI/ZAPI, 7Mode CLI APISet.

Invokes "volume-size-async, volume-size" API for CMode ZAPI.

Invokes "vol size" command for 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: volume-name

7M CLI : volume name (Programmatic name: "volume-name")

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< 'vol-size' => $string >>

Applicable for CMode CLI

=item C<< 'is-replica' => $string >>

Applicable for CMode CLI

=item C<< 'is-rdonly' => $string >>

Applicable for CMode CLI

=item C<< 'new-size' => $string >>

Applicable for CMode CLI/ZAPI, 7Mode CLI.

Maps to:

CM ZAPI: new-size

7M CLI : volume size (Programmatic name: "size")

=item C<< 'is-fixed' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'is-repository' => 0 | 1 >>

B<DEPRECATED>. Use C<is_infinivol> instead.

=item C<< is_infinivol => 0 | 1 >>

(Optional, Applicable for CMode ZAPI)

This boolean option describes whether the volume is an infinite volume or not.
For CMode ZAPI, the "async" form of the ZAPI must be used for the infinite
volume. If this option is not specified, the library will attempt to
determine if an infinite volume is being operated on (this could involve
querying the filer).

=item C<< is_infinivol_ref => \$scalar >>

(Optional, SCALARREF)

If the method was invoked through CMode ZAPI, the variable gets updated
describing whether it operated on an infinivol (used the "async" ZAPI) or not
(used the normal ZAPI).

=item C<< '*' >>

This method takes all other options supported by
"volume size" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub size {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( vol-size is-replica is-rdonly new-size is-fixed is-repository
                    )
            ),
        },
        cli_cmd => 'volume size',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_size_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_size_7mode_cli',
                interface => 'CLI',
                set       => '7Mode',
            },
            {   method    => '_size_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::size',
    );

    $Log->exit() if $may_exit;
} ## end sub size

sub _size_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'volume_size',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _size_cmode_cli

sub _size_7mode_cli {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = validate_with(
        params      => \@_,
        spec        => $pkg_or_obj->_backend_validate_spec(),
        allow_extra => 1
    );

    delete $opts{command_interface};
    my $apiset = delete $opts{apiset};

    my $allowed_mode_options = [qw(volume)];
    my %mode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_mode_options,
        target => \%mode_args,
        map    => {'new-size' => 'size',},
    );

    $apiset->vol_size(%mode_args);
    $Log->exit() if $may_exit;
} ## end sub _size_7mode_cli

sub _size_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_backend_cmode_zapi_infinivol_builder(
        @args,
        api  => 'volume-size',
        map  => {volume => 'volume'},
        copy => [qw(new-size)],
    );
    $Log->exit() if $may_exit;
}

=head2 remove_other_volume

    NACL::C::Volume->remove_other_volume(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->remove_other_volume(...);

(Class or instance method)
This method is used to remove 7-mode volume from the cluster database.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"volume remove-other-volume" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub remove_other_volume {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params  => \@args,
        cli_cmd => 'volume remove-other-volume',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_remove_other_volume_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::remove_other_volume',
    );

    $Log->exit() if $may_exit;
} ## end sub remove_other_volume

sub _remove_other_volume_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@args,
        api     => 'volume_remove_other_volume',
    );
    $Log->exit() if $may_exit;
} ## end sub _remove_other_volume_cmode_cli

=head2 remove_junction

    NACL::C::Volume->remove_junction(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                ...);

(Class method)
This method is used to remove a dead junction by junction path.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required)

Applicable for CMode CLI

=item C<< 'junction-path' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"volume remove-junction" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub remove_junction {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( vserver junction-path
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'volume remove-junction',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_remove_junction_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::remove_junction',
    );

    $Log->exit() if $may_exit;
} ## end sub remove_junction

sub _remove_junction_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args,
        api     => 'volume_remove_junction',
    );
    $Log->exit() if $may_exit;
} ## end sub _remove_junction_cmode_cli

=head2 destroy

    NACL::C::Volume->destroy(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->destroy(...);

(Class or instance method)
This method is used to destroy an existing volume.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'force' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'disable-offline-check' => $string >>

=item C<< is_staging_volume => 0 | 1 >>

(Optional)
Needs to be specified as 1 for staging volumes.
This is necessary because offlining staging volumes is possible
only through CLI and it requires confirmations to be enabled.

=item C<< '*' >>

This method takes all other options supported by
"volume destroy" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub destroy {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( force is_staging_volume

                    disable-offline-check
                    )
            ),
        },
        cli_cmd => 'volume destroy',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_destroy_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::destroy',
    );

    $Log->exit() if $may_exit;
} ## end sub destroy

sub _destroy_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;

    my $is_staging_volume = $opts{is_staging_volume};
    my $apiset            = $opts{apiset};
    my $prompts_answers   = [
        '.*Do you want to proceed? {y|n}:'         => 'y',
        'Are you sure you want to delete volume.*' => 'y',
    ];

    try {

        if ($is_staging_volume) {
            $apiset->set(confirmations => 'on');
        }
        $pkg_or_obj->_base_backend_cmode_cli(
            %opts,
            api             => 'volume_destroy',
            additional_args => {'connectrec-match_table' => $prompts_answers},
        );
    }
    finally {
        if ($is_staging_volume) {
            $apiset->set(confirmations => 'off');
        }
    };
    $Log->exit() if $may_exit;
}    #sub _destroy_cmode_cli

=head2 transition_recover

    NACL::C::Volume->transition_recover(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->transition_recover(...);

(Class or instance method)
This method is used to initiates recovery for a volume which is in an incompletely transitioned state.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI

=item C<< 'affinity-node' => $string >>

Applicable for CMode CLI

=item C<< 'action' => $string >>

Applicable for CMode CLI

=item C<< 'foreground' => $string >>

Applicable for CMode CLI

=item C<< 'source-node' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"volume transition-recover" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub transition_recover {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( affinity-node action foreground source-node
                    )
            ),
        },
        api => 'volume_transition_recover',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_transition_recover_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::transition_recover',
    );

    $Log->exit() if $may_exit;
} ## end sub transition_recover

sub _transition_recover_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $vserver = $opts{vserver};

    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'volume_transition_recover',
        vserver_operated_on => \$vserver,
    );

    $Log->exit() if $may_exit;
} ## end sub _transition_recover_cmode_cli

=head2 explore
    package call:
    my $vol_explore_obj = NACL::C::Volume->explore(
                                command_interface => $ci,
                                format => 'volume',
                                ...);
                or

    my @vol_explore_objs = NACL::C::Volume->explore(
                                command_interface => $ci,
                                format => 'volume',
                                ...);
    Instance call:
    my $explore_cs_obj = $vol_obj->explore();

    The returned object contains all the details specific to that volume.
    Returns the obj of type NACL::CS::VolumeExplore
    and is equivalent to the following call :
    NACL::C::Volume->explore(command_interface => $ci, format => 'volume', scope => $vol_name);

(Class or Instance method)
This method is used to view meta-data within volumes.

Uses a CMode CLI APISet.

In an array context it creates a Component State object of the type
L<NACL::CS::VolumeExplore|lib-NACL-CS-VolumeExplore-pm>. for each
present element and returns them.
In scalar context it creates a Component State object of the type
L<NACL::CS::VolumeExplore|lib-NACL-CS-VolumeExplore-pm>. for each
present present but returns only the first element.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'verbose' => $string >>

Applicable for CMode CLI

=item C<< 'dump' => $string >>

Applicable for CMode CLI

=item C<< 'find' => $string >>

Applicable for CMode CLI

=item C<< 'noinomap' => $string >>

Applicable for CMode CLI

=item C<< 'export-file' => $string >>

Applicable for CMode CLI

=item C<< 'print' => $string >>

Applicable for CMode CLI

=item C<< 'rawtimes' => $string >>

Applicable for CMode CLI

=item C<< 'recursive' => $string >>

Applicable for CMode CLI

=item C<< 'fill' => $string >>

Applicable for CMode CLI

=item C<< 'format' => $string >>

Applicable for CMode CLI

=item C<< 'import-file' => $string >>

Applicable for CMode CLI

=item C<< 'set' => $string >>

Applicable for CMode CLI

=item C<< 'scope' => $string >>

Applicable for CMode CLI

=item C<< 'list' => $string >>

Applicable for CMode CLI

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'ondisk' => $string >>

Omit in-core-only data when possible
Applicable for CMode CLI

=item C<< 'limit' => $string >>

Limit the number of lines of data printed
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by
"volume explore" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub explore {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( verbose dump find noinomap export-file print rawtimes recursive fill format
                    import-file set scope list

                    ondisk limit volume vserver
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'volume explore',
    );

    if(ref  $pkg_or_obj){
      $opts{'format'} = 'volume';
      $opts{'scope'}  = $pkg_or_obj->vserver().":".$pkg_or_obj->volume();
    }

    my @objs = $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_explore_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::explore',
    );

    $Log->exit() if $may_exit;
    return wantarray ? @objs : $objs[0];
} ## end sub explore

sub _explore_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my $response;
    $pkg_or_obj->_base_backend_cmode_cli(
        %opts,
        api => 'volume_explore',
        apiset_response_obj => \$response
    );
    my $output = $response->get_parsed_output();

    my @state_objs;
    foreach my $row (@$output) {
        my $obj = NACL::CS::VolumeExplore->new(
            command_interface => $opts{command_interface});
        $obj->_set_fields(row => $row);
        push(@state_objs, $obj);
    }

    $Log->exit() if $may_exit;
    return @state_objs;
} ## end sub _explore_cmode_cli

=head2 transition_protect

    NACL::C::Volume->transition_protect(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->transition_protect(...);

(Class or instance method)
This method is used to fence access to volume during Transition..

Uses a CMode CLI/ZAPI APISet.

Invokes 'volume-transition-protect' API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: volume

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: vserver

=item C<< 'is-enabled' => $string >>

Applicable for CMode CLI/ZAPI.

Maps to:

CM ZAPI: is-enabled

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"volume transition-protect" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub transition_protect {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( is-enabled
                    )
            ),
        },
        cli_cmd => 'volume transition-protect',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_transition_protect_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_transition_protect_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::transition_protect',
    );

    $Log->exit() if $may_exit;
} ## end sub transition_protect

sub _transition_protect_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args,
        api     => 'volume_transition_protect',
    );
    $Log->exit() if $may_exit;
} ## end sub _transition_protect_cmode_cli

sub _transition_protect_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_validate_backend_opts(@args);

    my $apiset = $opts{apiset};

    my $allowed_cmode_options = [qw( volume vserver is-enabled )];

    my %cmode_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        copy   => $allowed_cmode_options,
        target => \%cmode_args,
    );

    my $convert_opts = {'on' => 'true', 'off' => 'false'};

    if (defined $convert_opts->{$cmode_args{'is-enabled'}}) {
        $cmode_args{'is-enabled'}
            = $convert_opts->{$cmode_args{'is-enabled'}};
    }

    $apiset->volume_transition_protect(%cmode_args);
    $Log->exit() if $may_exit;
} ## end sub _transition_protect_cmode_zapi

=head2 set_msid

    NACL::C::Volume->set_msid(
                                command_interface => $ci,
                                'volume' => $volume,
                                'vserver' => $vserver,
                                ...);

                                 or

    $Volume_obj->set_msid(...);

(Class or instance method)
This method is used to set MSID of a volume.

Uses a CMode CLI/ZAPI APISet.

Invokes "volume-set-msid" API for CMode ZAPI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'volume' => $volume >>

Applicable for CMode CLI/ZAPI

Maps to :

CM ZAPI : name

(Required for class method, Not Applicable for instance method)

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)

Applicable for CMode CLI/ZAPI

Maps to:

CM ZAPI: ZAPI will be invoked in the context of the vserver specified to this
option.

=item C<< 'new-msid' => $string >>

Applicable for CMode CLI/ZAPI

Maps to :

CM ZAPI : new-msid

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< 'extended_query' => $string >>

(Optional)

Applicable for CMode CLI.

=item C<< '*' >>

This method takes all other options supported by
"volume set-msid" command in ONTAP.

Applicable for CMode CLI.

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub set_msid {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( new-msid
                    )
            ),
        },
        cli_cmd => 'volume set-msid',
    );
    my $vserver;
    $opts{vserver_operated_on} = \$vserver;
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_set_msid_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
            {   method    => '_set_msid_cmode_zapi',
                interface => 'ZAPI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::set_msid',
    );

    $Log->exit() if $may_exit;
} ## end sub set_msid

sub _set_msid_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_base_backend_cmode_cli(@args,
        api     => 'volume_set_msid',
    );
    $Log->exit() if $may_exit;
} ## end sub _set_msid_cmode_cli

sub _set_msid_cmode_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my $allowed_cmode_options = [qw( new-msid )];
    $pkg_or_obj->_base_backend_cmode_zapi(@args,
        api    => 'volume_set_msid',
        copy   => $allowed_cmode_options,
        map    => {'volume' => 'name',},
    );
    $Log->exit() if $may_exit;
} ## end sub _set_msid_cmode_zapi

=head2 show_space

    my $show_space_obj = NACL::C::Volume->show_space(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @show_space_objs = NACL::C::Volume->show_space(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $show_space_obj = $Volume_obj->show_space();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeShowSpace-pm/ATTRIBUTES>
of NACL::CS::VolumeShowSpace as input.
This method returns objects of the type L<NACL::CS::VolumeShowSpace|lib-NACL-CS-VolumeShowSpace-pm>.

=cut

sub show_space {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::VolumeShowSpace;
    my @state_objs = NACL::CS::VolumeShowSpace->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub show_space

=head2 show_footprint

    my $show_footprint_obj = NACL::C::Volume->show_footprint(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @show_footprint_objs = NACL::C::Volume->show_footprint(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $show_footprint_obj = $Volume_obj->show_footprint();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeShowFootprint-pm/ATTRIBUTES>
of NACL::CS::VolumeShowFootprint as input.
This method returns objects of the type L<NACL::CS::VolumeShowFootprint|lib-NACL-CS-VolumeShowFootprint-pm>.

=cut

sub show_footprint {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::VolumeShowFootprint;
    my @state_objs = NACL::CS::VolumeShowFootprint->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub show_footprint

=head2 show_space_old

    my $show_space_old_obj = NACL::C::Volume->show_space_old(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @show_space_old_objs = NACL::C::Volume->show_space_old(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $show_space_old_obj = $Volume_obj->show_space_old();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeShowSpaceOld-pm/ATTRIBUTES>
of NACL::CS::VolumeShowSpaceOld as input.
This method returns objects of the type L<NACL::CS::VolumeShowSpaceOld|lib-NACL-CS-VolumeShowSpaceOld-pm>.

=cut

sub show_space_old {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::VolumeShowSpaceOld;
    my @state_objs = NACL::CS::VolumeShowSpaceOld->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub show_space_old

=head2 show_footprint_old

    my $show_footprint_old_obj = NACL::C::Volume->show_footprint_old(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @show_footprint_old_objs = NACL::C::Volume->show_footprint_old(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $show_footprint_old_obj = $Volume_obj->show_footprint_old();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeShowFootprintOld-pm/ATTRIBUTES>
of NACL::CS::VolumeShowFootprintOld as input.
This method returns objects of the type L<NACL::CS::VolumeShowFootprintOld|lib-NACL-CS-VolumeShowFootprintOld-pm>.

=cut

sub show_footprint_old {
    $Log->enter() if $may_enter;
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_handle_obj_cs_like_calls(@_);
    require NACL::CS::VolumeShowFootprintOld;
    my @state_objs = NACL::CS::VolumeShowFootprintOld->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
} ## end sub show_footprint_old

=head2 cluster_max_volume

    NACL::C::Volume->cluster_max_volume(
                                command_interface => $ci,
                                ...);

(Class method)
This method is used to get or set the maximum number of volumes allowed in the cluster.

Uses a CMode CLI APISet.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'cluster-max-volume' => $string >>

Maximum number of volumes allowed in the cluster possible value(s) are, integer

Applicable for CMode CLI

=item C<< 'cmd_execution_time' => \$scalar >>

(Optional) Scalar reference , it will be updated with the command execution
time returned by the apiset response object

=item C<< 'method-timeout' => $time_in_seconds >>

(Optional)
The default NATE timeout is 60 seconds. If the command might take more
time to complete, this option can be used to specify a larger timeout value.
The value should be provided in seconds.

=item C<< '*' >>

This method takes all other options supported by
"volume cluster-max-volume" command in ONTAP.

Applicable for CMode CLI.

=item C<< "extended_query" => $string  >>

(Optional)

Applicable for CMode CLI

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub cluster_max_volume {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'cmd_execution_time' => {type => SCALARREF, optional => 1},
            _optional_scalars(
                qw( cluster-max-volume
                    )
            ),
        },
        ignore_primary_keys => 1,
        cli_cmd             => 'volume cluster-max-volume',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_cluster_max_volume_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::cluster_max_volume',
    );

    $Log->exit() if $may_exit;
} ## end sub cluster_max_volume

sub _cluster_max_volume_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;

    $pkg->_base_backend_cmode_cli(
        @args,
        api => 'volume_cluster_max_volume',
    );

    $Log->exit() if $may_exit;
} ## end sub _cluster_max_volume_cmode_cli

=head2 language

    my $language_obj = NACL::C::Volume->language(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @language_objs = NACL::C::Volume->language(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $language_obj = $Volume_obj->language();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeLanguage-pm/ATTRIBUTES>
of NACL::CS::VolumeLanguage as input.
This method returns objects of the type L<NACL::CS::VolumeLanguage|lib-NACL-CS-VolumeLanguage-pm>.

=cut

sub language {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VolumeLanguage;
    my @state_objs = NACL::CS::VolumeLanguage->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 show_df

    my $show_df_obj = NACL::C::Volume->show_df(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @show_df_objs = NACL::C::Volume->show_df(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $show_df_obj = $Volume_obj->show_df();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeShowDf-pm/ATTRIBUTES>
of NACL::CS::VolumeShowDf as input.
This method returns objects of the type L<NACL::CS::VolumeShowDf|lib-NACL-CS-VolumeShowDf-pm>.

=cut

sub show_df {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VolumeShowDf;
    my @state_objs = NACL::CS::VolumeShowDf->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 status

    my $status_obj = NACL::C::Volume->status(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

    my @status_objs = NACL::C::Volume->status(command_interface => $ci,
                          [ filter => { %filter } ,]
              [ requested_fields => [ @req_fields ]);

                    or

   my $status_obj = $Volume_obj->status();

(Class or Instance method)

This method is a CS-like fetch method which accepts all the parameters of
L<fetch|lib-NACL-CS-ComponentState-pm/fetch> (requested_fields, filter, allow_empty)
and has all of the other properties of fetch methods (callable in scalar or list
context, throws a NoElementsFound exception if no matching elements are found).
The requested fields and filter takes L<ATTRIBUTES|lib-NACL-CS-VolumeStatus-pm/ATTRIBUTES>
of NACL::CS::VolumeStatus as input.
This method returns objects of the type L<NACL::CS::VolumeStatus|lib-NACL-CS-VolumeStatus-pm>.

=cut

sub status {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_handle_obj_cs_like_calls(@args);
    require NACL::CS::VolumeStatus;
    my @state_objs = NACL::CS::VolumeStatus->fetch(%opts);
    $Log->exit() if $may_exit;
    return wantarray ? @state_objs : $state_objs[0];
}

=head2 obtain_cr_ops_lock

    NACL::C::Volume->obtain_cr_ops_lock(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                ...);

(Class method)
This method is used to obtain the CR Job conflict framework operations lock.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required)
Vserver
Applicable for CMode CLI

=item C<< 'name' => $string >>

Name
Applicable for CMode CLI

=item C<< 'tag' => $string >>

Operations Tag
Applicable for CMode CLI

=item C<< 'job-uuid' => $string >>

Job UUID
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "volume obtain cr ops lock" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub obtain_cr_ops_lock {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( vserver name tag job-uuid
                    )
            ),
        },
        ignore_primary_keys => 1,
        api                 => 'volume_obtain_cr_ops_lock',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_obtain_cr_ops_lock_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::obtain_cr_ops_lock',
    );

    $Log->exit() if $may_exit;
}

sub _obtain_cr_ops_lock_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'volume_obtain_cr_ops_lock',);
    $Log->exit() if $may_exit;
}

=head2 release_cr_ops_lock

    NACL::C::Volume->release_cr_ops_lock(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                ...);

(Class method)
This method is used to release the CR Job conflict framework operations lock.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required) See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required)
Vserver
Applicable for CMode CLI

=item C<< 'tag' => $string >>

Operations Tag
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "volume release cr ops lock" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub release_cr_ops_lock {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    $pkg->_frontend_log_debug_opts(@args);
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( vserver tag
                    )
            ),
        },
        ignore_primary_keys => 1,
        api                 => 'volume_release_cr_ops_lock',
    );
    $pkg->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_release_cr_ops_lock_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::release_cr_ops_lock',
    );

    $Log->exit() if $may_exit;
}

sub _release_cr_ops_lock_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg, @opts) = @_;
    $pkg->_base_backend_cmode_cli(@opts, api => 'volume_release_cr_ops_lock',
    );
    $Log->exit() if $may_exit;
}

=head2 sdiftest

    NACL::C::Volume->sdiftest(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                'volume' => $volume,
                                ...);

                                 or

    $Volume_obj->sdiftest(...);

(Class or instance method)
This method is used to test the Snapdiff RPC interface (check log for results).
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver
Applicable for CMode CLI

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)
Volume name
Applicable for CMode CLI

=item C<< 'snapshot' => $string >>

Snapshot name
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "volume sdiftest" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub sdiftest {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( snapshot
                    )
            ),
        },
        api => 'volume_sdiftest',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_sdiftest_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::sdiftest',
    );

    $Log->exit() if $may_exit;
}

sub _sdiftest_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts, api => 'volume_sdiftest',);
    $Log->exit() if $may_exit;
}

=head2 check_And_Get_RepositoryId

    NACL::C::Volume->check_And_Get_RepositoryId(
                                command_interface => $ci,
                                'vserver' => $vserver,
                                'volume' => $volume,
                                ...);

                                 or

    $Volume_obj->check_And_Get_RepositoryId(...);

(Class or instance method)
This method is used to get a Repository ID of Infinite Volume.
Uses CMode CLI .

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Vserver
Applicable for CMode CLI

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)
Volume name
Applicable for CMode CLI

=item C<< '*' >>

This method takes all other options supported by "volume check And Get RepositoryId" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub check_And_Get_RepositoryId {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params => \@args,
        api    => 'volume_check_And_Get_RepositoryId',
    );
    $pkg_or_obj->call_on_apiset(
        %opts,
        choices => [
            {   method    => '_check_And_Get_RepositoryId_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
            },
        ],
        frontend => 'NACL::C::Volume::check_And_Get_RepositoryId',
    );

    $Log->exit() if $may_exit;
}

sub _check_And_Get_RepositoryId_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts,
        api => 'volume_check_And_Get_RepositoryId',);
    $Log->exit() if $may_exit;
}


__PACKAGE__->_load_event_class();

sub _primary_keys_validate_spec {
    return (
        volume  => {type => SCALAR},
        vserver => {type => SCALAR | UNDEF, optional => 1},
    );
} ## end sub _primary_keys_validate_spec

=head2 get_aggregate

  my $aggr_object = $volume_object->get_aggregate();

(Instance method only) This method returns a STask/Component object
corresponding to the parent aggregate for the volume modeled by this object.
If the volume object is a Component object, then object of type
NACL::C::StorageAggregate will be returned. If the volume object is a STask
object, then object of type NACL::STask::Aggregate will be returned.

=cut

sub get_aggregate {
    my $self = shift;
    $Log->enter() if $may_enter;

    $self->_verify_invocation(style => 'instance_only');

    # Find whom the caller is - STask/C??.
    my $c_or_t = $self->is_c_or_t();

    # Get aggregate object
    my $aggr_obj = $self->build_associated_object(
        type => "NACL::${c_or_t}::Aggregate");
    $Log->exit() if $may_exit;
    return $aggr_obj;
} ## end sub get_aggregate

=head2 get_vserver

  my $vserver_object = $volume_object->get_vserver();

(Instance method only) This method returns a Component object
corresponding to the vserver to which the volume modeled by this object belongs
to. It returns an object of type NACL::C::Vserver irrespective of whether
this volume object is a STask or Component object. When a STask for Vserver
is added, this method will return a Vserver STask object for the callers
from STask layer.

=cut

sub get_vserver {
    my $self = shift;
    $Log->enter() if $may_enter;

    $self->_verify_invocation(style => 'instance_only');

    if ($self->command_interface()->mode() eq 'CMode') {

        my $c_or_t = $self->is_c_or_t();
        my $type   = "NACL::${c_or_t}::Vserver";

        # Get vserver object
        my $vserver_obj = $self->emerge_contained_object(type => $type);
        $Log->exit() if $may_exit;
        return $vserver_obj;
    } else {

        # Do nothing for 7Mode
        $Log->exit() if $may_exit;
        return;
    } ## end else [ if ( $self->command_interface...)]
} ## end sub get_vserver

=head2 is_hybrid

    $ret = $volume->is_hybrid();

(Instance method only) This method checks if the volume lives on a
hybrid aggregate or not.  Will return 1 if its hybrid and 0 otherwise.

=cut

sub is_hybrid {
    $Log->enter() if $may_enter;
    my $self = shift;

    # Get aggregate containing the volume
    my $aggr = $self->get_aggregate();

    # Check if container aggregate is hybrid
    my $ret = $aggr->is_hybrid();
    $Log->exit() if $may_exit;
    return $ret;
} ## end sub is_hybrid

# Implementation for the following find/state methods are from
# NACL::C::_Mixins::VolumeFileDirectory

=head3 find_files

    # List context
    my @file_objs = $vol_obj->find_files();

    # Scalar context
    my $file_obj = $vol_obj->find_files();

    # Accepts the same options as normal "find" calls
    my $file_obj = $vol_obj->find_files(allow_empty => 1);

    my $file_obj = $vol_obj->find_files(filter => { 'file-size' => '>1MB' });

    my $file_obj = $vol_obj->find_files(recursive => 1);

Finds files present in the top-level of the volume and returns objects of type
L<NACL::C::VolumeFile|lib-NACL-C-VolumeFile-pm>. Accepts arguments accepted
by "find". If passed with recursive => 1,(default 0) finds all the files
present in the volume.

=head3 find_directories

    # List context
    my @dir_objs = $vol_obj->find_directories();

    # Scalar context
    my $dir_obj = $vol_obj->find_directories();

    # Accepts the same options as normal "find" calls
    my $dir_obj = $vol_obj->find_directories(allow_empty => 1);

    my $dir_obj = $vol_obj->find_directories(filter => { 'owner-id' => $id });

Finds directories present in the volume and returns objects of type
L<NACL::C::VolumeDirectory|lib-NACL-C-VolumeDirectory-pm>. Accepts arguments accepted
by "find".

=head3 find_files_and_directories

    # List context
    my @objs = $vol_obj->find_files_and_directories();

    # Scalar context
    my $obj = $vol_obj->find_files_and_directories();

    # Accepts the same options as normal "find" calls
    my $obj = $vol_obj->find_files_and_directories(allow_empty => 1);

    my $obj = $vol_obj->find_files_and_directories(filter => { 'acl-type' => $acl });

Finds files and directories present in the volume and returns objects of type
L<NACL::C::VolumeFile|lib-NACL-C-VolumeFile-pm> for files found and objects
of type
L<NACL::C::VolumeDirectory|lib-NACL-C-VolumeDirectory-pm> for directories found.
Accepts arguments accepted by "find".

=head3 state_files

    # List context
    my @files_cs = $vol_obj->state_files();

    # Scalar context
    my $file_cs = $vol_obj->state_files();

    # Accepts arguments accepted by "fetch"
    my @files_cs = $vol_obj->state_files(
        requested_fields => [ qw(acl-type) ],
        filter           => { 'file-size' => '>1MB' },
        allow_empty      => 1
    );

Fetches the state of contained files and returns objects of type
L<NACL::CS::VolumeListFilesInDirectory|lib-NACL-CS-VolumeListFilesInDirectory-pm>.
Accepts the arguments accepted by the "fetch" method.

=head3 state_directories

    # List context
    my @dirs_cs = $vol_obj->state_directories();

    # Scalar context
    my $dir_cs = $vol_obj->state_directories();

    # Accepts arguments accepted by "fetch"
    my @dirs_cs = $vol_obj->state_directories(
        requested_fields => [ qw(acl-type) ],
        filter           => { 'inode-number' => num },
        allow_empty      => 1
    );

Fetches the state of contained directories and returns objects of type
L<NACL::CS::VolumeListFilesInDirectory|lib-NACL-CS-VolumeListFilesInDirectory-pm>.
Accepts the arguments accepted by the "fetch" method.

=head3 state_files_and_directories

    # List context
    my @cs = $vol_obj->state_files_and_directories();

    # Scalar context
    my $cs = $vol_obj->state_files_and_directories();

    # Accepts arguments accepted by "fetch"
    my @cs = $vol_obj->state_files_and_directories(
        requested_fields => [ qw(acl-type) ],
        filter           => { 'inode-number' => num },
        allow_empty      => 1
    );

Fetches the state of contained files and directories and returns objects of type
L<NACL::CS::VolumeListFilesInDirectory|lib-NACL-CS-VolumeListFilesInDirectory-pm>.
Accepts the arguments accepted by the "fetch" method.

=cut

sub _file_dir_find_state_wrapper {
    $Log->enter() if $may_enter;
    my ($self, @args) = @_;
    my %opts = $self->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'file-type' => {type => SCALAR,  optional => 1},
            filter      => {type => HASHREF, default  => {}},
        },

        # Allow other options for find
        allow_extra => 1
    );
    my $volume = delete $opts{volume};
    my %filter = %{$opts{filter}};
    $filter{'directory-path'} = "/vol/$volume";
    $self->_hash_move(
        source => \%opts,
        target => \%filter,
        move   => [qw(vserver file-type)]
    );
    $opts{filter} = \%filter;
    my @cs = NACL::CS::VolumeListFilesInDirectory->fetch(%opts);
    $Log->exit() if $may_exit;
    return @cs;
}

=head2 name_attribute

Returns 'volume'.

=cut

sub name_attribute {
    return 'volume';
}

=head2 get_local_node_name

    # Class method call

    my $node_name = NACL::C::Volume->get_local_node_name(
        command_interface => $ci,
        volume            => $vol_name,
        vserver           => $vs_name   # Necessary only for CMode
    );

    # Instance call
    my $node_name = $vol_obj->get_local_node_name();

(Applicable only for CMode, for 7Mode it simply returns the name of the
node modeled by the command_interface)

Determines the node on which the volume is present and returns the node
name. This is necessary for Nodescope operations, since they need to
be performed in the context of the node where the volume is present.

B<NOTE:> For FlexGroups, this returns the node which hosts the root
constituent of the FlexGroup.

=over

=item Options

=over

=item C<< is_flexgroup => 0|1 >>

(Optional) A value of 1 indicates that this is a FlexGroup volume.

=back

=back

=cut

sub get_local_node_name {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {is_flexgroup => {type => SCALAR, optional => 1}},
    );
    if ($opts{command_interface}->is_7mode()) {
        $Log->exit() if $may_exit;
        return $opts{command_interface}->hostrec->hostname;
    } elsif (!defined $opts{vserver}) {
        $Log->warn('For a CMode system, the vserver value '
                . 'is required to uniquely identify the volume but was not '
                . 'passed. Please update the call to get_local_node_name to pass '
                . 'the vserver value. Note that the value returned for this '
                . 'method is not guaranteed to be correct since the vserver '
                . 'value was not provided');
    }

    # Move the primary keys into the filter, all other options are
    # sent through directly to fetch (these should just be the common options)
    my %filter;
    $pkg_or_obj->_move_primary_keys(
        source => \%opts,
        target => \%filter,
    );

    my $is_fg = delete $opts{is_flexgroup} // $pkg_or_obj->_is_flexgroup();
    my $node_name;

    # For FlexGroups, the "node" value is "-", so instead we'll return
    # the node on which the root constituent is present
    my $get_fg_node_name = sub {
        my $cs_object = NACL::CS::Volume->fetch(
            %opts,
            filter           => {
                vserver            => $filter{vserver},

                # The FG's constituents are named <vol>__0001, <vol>__0002
                volume             => $filter{volume} . '*',
                'is-constituent'   => 'true',
                'constituent-role' => 'root'
            },
            requested_fields => [qw(node)],
        );
        return $cs_object->node();
    };
    if ($is_fg) {
        $node_name = $get_fg_node_name->();
    } else {
        my $cs_object = NACL::CS::Volume->fetch(
            %opts,
            filter           => \%filter,
            requested_fields => [qw(node)],
        );
        $node_name = $cs_object->node();
        if ($node_name eq '-') {
            # Might be a FlexGroup
            $node_name = $get_fg_node_name->();
        }
    }

    if (!defined $node_name || $node_name eq '-') {
        $Log->exit() if $may_exit;
        NATE::BaseException->throw('Could not determine node on which volume '
            . "'$filter{volume}' (vserver='$filter{vserver}') is present");
    }

    $Log->exit() if $may_exit;
    return $node_name;
} ## end sub get_local_node_name

=head2 get_local_node_obj

    my $node_obj = $vol_obj->get_local_node_obj();

Similar to C<get_local_node_name>, but returns a C<NACL::STask::Node> object
for the node hosting the volume.

=cut

# Implementation comes from the base-class.

=head2 mkfile

B<DEPRECATED>. Please use the method C<file_create> instead.

    $vol_obj->mkfile(filename => $name, %other_opts);

(Instance method only)
This method is used to make files on a volume. It uses the volume name
and the file-name provided to construct the file path (of the form
/vol/$vol_name/$file_name).

This method is a wrapper around
L<< NACL::C::Mkfile->mkfile|lib-NACL-C-Mkfile-pm/mkfile >>. Therefore,
this method accepts all of the options accepted by
the C<NACL::C::Mkfile> method.

=over

=item Options

=over

=item C<< filename => $filename >>

(Required)
This is the name of the file to be created. Note that this is <not> the
path to the file. The path to the file gets constructed internally based
on the volume name and file name.

=item Other options

All of the options accepted by
L<< NACL::C::Mkfile->mkfile|lib-NACL-C-Mkfile-pm/mkfile >> are accepted
by this method as well.

=back

=back

=cut

sub mkfile {
    $Log->enter() if $may_enter;
    my $self = shift;
    my $warn_message = 'This method is deprecated. Please use the method '
            . "'file_create' instead";
    NACL::C::Component->_print_deprecation_warning($warn_message);
    $self->_verify_invocation(style => 'instance_only');
    my %opts = $self->_common_validate_with(
        params          => \@_,
        additional_spec => {filename => {type => SCALAR}},

        # Allow other options, which are those accepted by mkfile
        allow_extra => 1
    );

    # The form of the file path is /vol/$vol_name/$file_name
    # The "filename" argument accepted by the mkfile APISet is more
    # correctly the file-path. Here we're constructing the file path
    # given the file name.
    $opts{filename} = "/vol/$opts{volume}/$opts{filename}";
    delete $opts{volume};

    # Figure out whether to invoke NACL::C::Mkfile or NACL::STask::Mkfile
    # based on whether this was called by a C or STask Volume object.
    my $is_c_or_t  = $self->is_c_or_t();
    my $mkfile_pkg = "NACL::${is_c_or_t}::Mkfile";
    my $ret        = $mkfile_pkg->mkfile(%opts);

    $Log->exit() if $may_exit;
    return $ret;
} ## end sub mkfile

=head2 construct_dblade_path

    # Class method call
    my $dblade_path = NACL::C::Volume->construct_dblade_path(
        command_interface => $ci,
        volume            => $vol_name,
        vserver           => $vs_name
    );

                  or

    # Instance call
    my $dblade_path = $vol_obj->construct_dblade_path();

(Class or instance method)
Returns the dblade path of a volume while accounting for the "name-ordinal"
value of the volume. For 7Mode, it simply returns "/vol/$vol_name" (there's
no concept of a dblade or a "name-ordinal" for 7Mode)

While most scripts assume the path to be "/vol/$vol_name", this assumption
is not always correct. It is possible to create multiple volumes of the same
name on the same node. In this case, the "name-ordinal" value has to be
accounted for. The first volume created will have a "name-ordinal" value of
"base", while successive volumes of that name will be assigned consecutive
integer "name-ordinal" values. For example, let's say we create 3 volumes on a
node, all named "myvol". In this case, the respective "name-ordinal" and
dblade paths are:

First creation:
    name-ordinal: base
    dblade path:  /vol/myvol

Second creation:
    name-ordinal: 1
    dblade path:  /vol/myvol(1)

Third creation:
    name-ordinal: 2
    dblade path:  /vol/myvol(2)

=over

=item Options

=over

=item Class method call

The primary keys (i.e volume (and vserver if CMode)) along with the
command_interface are mandatory. The "common options" (such as
method-timeout, apiset_must, apiset_should) are also accepted.

=item Instance method call

The "common options" (such as method-timeout, apiset_must, apiset_should) are
accepted.

=back

=back

=cut

# This and "construct_dblade_file_path" are unit tested in
# VolumeFileDirectory.thpl since they're closely tied to files.
sub construct_dblade_path {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(params => \@args);
    my $path = "/vol/$opts{volume}";

    # No concept of "name-ordinal" in 7Mode, so do this only for CMode
    if ($opts{command_interface}->is_cmode()) {
        my %primary_keys;
        $pkg_or_obj->_move_primary_keys(
            source => \%opts,
            target => \%primary_keys
        );

        my $vol_cs = NACL::CS::Volume->fetch(
            %opts,
            filter           => \%primary_keys,
            requested_fields => ['name-ordinal']
        );
        my $name_ordinal = $vol_cs->name_ordinal();
        # name_ordinal() is '-' for a FlexGroup which does not have a real
        # name-ordinal. So, the path should not contain (-) for a FlexGroup
        # or something which does not have a real name-ordianal.
        # See burt995288 for more info.
        if (($name_ordinal ne 'base') && ($name_ordinal ne '-')) {

            # If it's not "base", then it's a number. The dblade path will
            # contain this name-ordinal within brackets
            $path .= "($name_ordinal)";
        }
    }

    $Log->exit() if $may_exit;
    return $path;
}

=head2 construct_dblade_file_path

    # Class method call
    my $dblade_file_path = NACL::C::Volume->construct_dblade_file_path(
        command_interface => $ci,
        volume            => $vol_name,
        vserver           => $vs_name,
        filename          => $filename
    );

                  or

    # Instance call
    my $dblade_file_path = $vol_obj->construct_dblade_file_path(filename => $filename);

(Class or instance method)
Returns the dblade file path for a given filename. This internally uses
L<construct_dblade_path> to account for the "name-ordinal".

In addition to the options accepted by L<construct_dblade_path>, this method
accepts C<filename> as argument.

=cut

sub construct_dblade_file_path {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {filename => {type => SCALAR}},

        # All other options get passed through to "construct_dblade_path"
        allow_extra => 1
    );
    my $filename = delete $opts{filename};

    my $dblade_path = $pkg_or_obj->construct_dblade_path(%opts);
    $dblade_path .= "/$filename";

    $Log->exit() if $may_exit;
    return $dblade_path;
}

sub _determine_if_infinivol_being_created {
    $Log->enter() if $may_enter;
    my ($pkg, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            command_interface => {
                type => OBJECT,
                isa  => 'NACL::C::CommandInterface::ONTAP'
            },
            vserver => {type => SCALAR | OBJECT, optional => 1},
            _optional_scalars(qw(nacltask_is_infinivol is_infinivol)),
            %{$pkg->_common_validate_spec()},
        },
        allow_extra => 1
    );
    my $vserver           = $opts{vserver};
    my $command_interface = $opts{command_interface};
    my $is_infinivol = $opts{nacltask_is_infinivol} || $opts{is_infinivol};

    my %common_opts_with_ci;
    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts_with_ci
    );

    unless (defined $is_infinivol) {
        if (   exists $opts{aggregate}
            || exists $opts{'vsroot'}
            || exists $opts{'constituent-role'})
        {

            # These options are applicable only for flex vols, so set to use
            # the normal ZAPI. If not provided, do not assume that the async
            # ZAPI should be used. It could be negative testing where not
            # providing these options is expected to result in the normal ZAPI
            # failing.
            $is_infinivol = 0;
        } else {

            # If an attempt is made to create a volume on a repository vserver
            # without aggregate and constituent-role, then the assumption
            # is that an attempt is being made to create/re-create the
            # infinivol
            if ($command_interface->isa('NACL::C::Vserver')) {
                $is_infinivol = $command_interface->_is_repository();
            } else {
                my %common_opts_with_ci;
                $pkg->_copy_common_component_params_with_ci(
                    source => \%opts,
                    target => \%common_opts_with_ci
                );

                my $vs_cs = NACL::CS::Vserver->fetch(
                    %common_opts_with_ci,
                    filter           => {vserver => $vserver},
                    requested_fields => ['is-repository'],

                    # Allow for negative testing: if vserver does not exist,
                    # we'll just execute the normal ZAPI (which will fail)
                    allow_empty => 1
                );
                if (defined $vs_cs && $vs_cs->is_repository() eq 'true') {
                    $is_infinivol = 1;
                } else {
                    $is_infinivol = 0;
                }
            }
        }
    }

    $Log->exit($is_infinivol);
    return $is_infinivol;
}

sub _backend_cmode_zapi_infinivol_builder {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts = validate_with(
        params => \@args,
        spec   => {
            api              => {type => SCALAR},
            map              => {type => HASHREF, default => {}},
            is_infinivol_ref => {type => SCALARREF, optional => 1},
        },
        allow_extra => 1
    );
    my $is_infinivol_ref = delete $opts{is_infinivol_ref};

    my $is_infinivol = $pkg_or_obj->_determine_is_infinivol(%opts,
        volume => $opts{volume});

    if ($is_infinivol) {
        $opts{api} .= '-async';
        $opts{map}{volume} = 'volume-name';
    } else {
        $opts{map}{volume} ||= 'volume';
    }

    $$is_infinivol_ref = $is_infinivol if ($is_infinivol_ref);

    my $vserver;
    $opts{vserver_operated_on} ||= \$vserver;
    $pkg_or_obj->_base_backend_cmode_zapi(
        %opts,
        inapplicable_for_help_xml => ['foreground']
    );

    $Log->exit() if $may_exit;
}

# These fields cannot be added to _common_validate_spec since they end up
# causing validation failures on the CS side, hence this extra spec for
# options that are common to all methods on this Component on the C side.
sub _additional_common_validate_spec {
    my ($pkg_or_obj, @opts) = @_;
    my $spec = $pkg_or_obj->SUPER::_additional_common_validate_spec(@opts);
    my $iv_spec = $pkg_or_obj->_add_infinivol_params_to_additional_common_validate_spec(@opts);
    $spec = {%$iv_spec, %$spec};
    return $spec;
}

sub _check_zapi_attrs {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    foreach my $field (qw(increment-percent reset)) {
        if (exists $opts{$field}) {
            my $msg = 'Could not use ZAPI implementation because field '
                . "'$field' is available only through the volume autosize "
                . 'CLI command';
            $Log->comment($msg);
            $Log->exit() if $may_exit;
            NACL::Exceptions::InvalidChoice->throw($msg);
        }
    }
    $Log->exit() if $may_exit;
}

sub _check_cmode_zapi_delete {
    $Log->enter() if $may_enter;
    my @array;
    my ($pkg_or_obj, %opts) = @_;
    foreach my $key (qw(force foreground disable-offline-check)) {
        if (defined $opts{$key} && $opts{$key} =~ /true/i) {
            push @array, $key;
        }
    }
    if (@array) {
        my $str = join('; ', @array);
        my $msg = 'The following options were provided but are not accepted '
            . "by the ZAPI:\n$str\nFor this reason, the ZAPI back-end "
            . 'could not be used for delete';
        $Log->comment($msg);
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw($msg);
    }

    $pkg_or_obj->_infinivol_vol_vs_check(%opts);

    $Log->exit() if $may_exit;
} ## end sub _check_cmode_zapi_delete

sub _check_if_staging_volume {

    $Log->enter() if $may_enter;
    my ($pkg_or_obj, %opts) = @_;
    my @array;
    foreach my $key (qw(force is_staging_volume )) {
    if ($opts{$key}) {
        push @array, $key;
    }
   }
    if (@array) {
        my $str = join('; ', @array);
        my $msg = 'The following options were provided but are not accepted '
            . "by the ZAPI:\n$str\nFor this reason, the ZAPI back-end "
            . 'could not be used for offline';
        $Log->comment($msg);
        $Log->exit() if $may_exit;
        NACL::Exceptions::InvalidChoice->throw($msg);
    }
    $Log->exit() if $may_exit;
}

=head2 clear_error

    NACL::C::Volume->clear_error(
        command_interface  => $ci,
        volume             => $volume_name,
        vserver            => $vserver_name, # Required for CMode
        inaccessible       => $boolean,      # default '1'
        %other_option
    );

            or

    $vol_obj->clear_error(
        %other_option
    );

(Class or Instance Method)
clears iron marked inaccessible block ranges from AFS of the volume.

Supports 7Mode/Nodescope CLI.
Invokes "vol_clear_error" for 7Mode/Nodescope CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for Instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< volume => $volume_name >>

(Required for class method, Not Applicable for Instance method)
Volume name.

=item C<< vserver => $vserver_name >>

(Required for class method in CMode, Not Applicable for Instance method or for 7Mode)
The vserver on which the file is present.

=item C<< inaccessible => 0|1 >>

(Optional)
inaccessible option clears iron marked inaccessible block ranges.

=item C<< nodescope-node-name => $node_name >>

(Optional)
When run in nodescope, the component method defaults to running on the "local"
node. This argument can be used to run the method on node other than "local".

Applicable for 7Mode CLI/Nodescope .

=item C<< determine_appropriate_nodescope_node => 0|1 >>

(Optional, defaults to 0)
When set to 1, the method first determines which node the volume is present
and makes the call in the context of that node. Note that when
C<nodescope-node-name> overrides this argument, meaning that if both
C<nodescope-node-name> is provided and C<determine_appropriate_nodescope_node> is
set to 1 then the node name provided in C<nodescope-node-name> is used.

This argument defaults to 0.

Applicable for 7Mode CLI/Nodescope.

=item C<< handle_name_ordinal => 0 | 1 >>

(Optional, defaults to 1)
(This is applicable only for CMode)

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=item C<NACL::C::Exceptions::Volume::DoesNotExist>

This type of exception gets thrown if an attempt is made to clear error
on a volume which does not exists.

=back

=back

=cut


sub clear_error {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    $pkg_or_obj->_frontend_log_debug_opts(@args);

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            inaccessible  => {type => BOOLEAN, default => 1},
            handle_name_ordinal => {type => SCALAR, default => 1},
            $pkg_or_obj->_nodescope_options_validate_spec(),
        }
    );

    $pkg_or_obj->_get_node_and_handle_ordinal(\%opts);

    try {
        my $objs = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_clear_error_7mode_cli',
                    interface => 'CLI',
                    set       => '7Mode|Nodescope',
                },
            ],
        );
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Volume::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
}

sub _clear_error_7mode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts              = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset            = $opts{apiset};

    my @allowed_options =
        $apiset->get_allowed_options_for_command(command => 'vol_clear_error');
    push @allowed_options, 'nodescope-node-name';

    my %api_args;
    $pkg_or_obj->_hash_copy(
        source => \%opts,
        target => \%api_args,
        copy   => \@allowed_options,
    );

    $apiset->vol_clear_error(%api_args);

    $Log->exit() if $may_exit;
}

=head2 force_cp

    NACL::C::Volume->force_cp(
        command_interface  => $ci,
        volume             => $volume_name,
        vserver            => $vserver_name,
        %other_option
    );

            or

    $vol_obj->force_cp(
        %other_option
    );

(Class or Instance Method). Supports CMode ZAPI.
It invokes the wafl_sync on volume and returns the wafl_sync handle
for it.


=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for Instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< volume => $volume_name >>

(Required for class method, Not Applicable for Instance method)
Volume name.

=item C<< vserver => $vserver_name >>

(Required for class method in CMode, Not Applicable for Instance method or for 7Mode)
The vserver on which the file is present.

=back

=back

=cut

sub force_cp {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
    );

    my $objs;
    try {
         $objs = $pkg_or_obj->call_on_apiset(
            %opts,
            choices => [
                {   method    => '_force_cp_zapi',
                    interface => 'ZAPI',
                    set       => 'CMode',
                },
            ],
        );
    } catch NACL::APISet::Exceptions::ResponseException with {
        my $exception = $_[0];
        NACL::C::Exceptions::Volume::DoesNotExist
            ->detect_convert_and_throw(exception => $exception);
        $Log->exit() if $may_exit;
        $exception->throw();
    };

    $Log->exit() if $may_exit;
    return $objs;
}

sub _force_cp_zapi {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    my %opts    = $pkg_or_obj->_validate_backend_opts(@args);
    my $apiset  = $opts{apiset};
    my $vserver = $opts{vserver};
    my $volume  = $opts{volume};

    my $parsed = $apiset->wafl_sync(
        volumes => [$volume],
        'vserver-override-name' => $vserver,
    )->get_parsed_output();

    my $handle = $parsed->[0]->{handles}->[0]->{'wafl-sync-handle'}->[0];
    $Log->exit() if $may_exit;
    return $handle;
}

=head2 wafl_buftree_check

(Instance Method) Executes the "wafl buftree check" command on the Node

    my $response = $Volume->wafl_buftree_check(
        inodespec        => $location,
        %other_options,
    );

=over

=item Options

=over

=item C<< inodespec => $location >>

(Required) is one of:

        <inum>
        priv:<inum>
        vvol:<vvolname>/label
        vvol:<vvolname>/container
        vvol:<vvolname>/uuid
        vvol:<vvolname>/uuid/<pathname>
        path:<pathname>
        meta:<pathname>

=item C<< snapid => $snapID >>

(Optional, default AFS)
snapshot ID

=item C<< level-check => $level >>

(Optional, default 0)
lowest level to check

=item C<< first-last-fbn => <first>[,<last>] >>

(Optional)
first and last FBNs to check

=item C<< no-check => 0|1 >>

(Optional)
don't check wi_blk_cnt and wi_indirect_blk_cnt

=item C<< no-cross-check => 0|1 >>

(Optional)
don't cross-check vvol indirect formats against container map parent format

=item C<< no-panic => 0|1 >>

(Optional)
don't panic on when an error is found

=item C<< verbose => 0|1 >>

(Optional)
display extra statistics

=item C<< snapdiff => 0|1 >>

(Optional)
test snapdiff interfaces

=back

=back

=cut

__PACKAGE__->_nodescope_builder(
    command => 'wafl_buftree_check',
);

=head1 HELPER METHODS

=head2 _get_node_and_handle_ordinal

Helper Method to get node name and to handle the name-ordinal.

(See the concept of "name-ordinal" first)

$pkg_or_obj->_get_node_and_handle_ordinal(\%opts);

Adds the 'nodescope-node-name' key in opts hash and
update the 'volume' key in opts hash with name-ordinal.

=cut

sub _get_node_and_handle_ordinal {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, $opts) = @_;


    my %common_option_with_ci;

    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => $opts,
        target => \%common_option_with_ci
    );

    my @requested_fields;
    if (!$opts->{'nodescope-node-name'}
        && $opts->{determine_appropriate_nodescope_node}) {
        push @requested_fields, 'node';
    }

    if ($opts->{handle_name_ordinal} && $opts->{'volume'} !~ /\(\d+\)/
        && $opts->{'volume'} !~ /^\//) {
        push @requested_fields, 'name-ordinal';
        $opts->{'volume'} = "/vol/$opts->{volume}";
    }

    # name-ordianal id only for C-Mode and vserver option is mandotary
    if ( @requested_fields && $opts->{command_interface}->is_cmode()
        && $opts->{vserver} ) {

        my $vol_cs = NACL::CS::Volume->fetch(
            %common_option_with_ci,
            filter           => {
                volume => $opts->{'volume'}, vserver => $opts->{vserver}
            },
            requested_fields => \@requested_fields
        );

        # Update volume key in %opts hash with name-ordianal
        if ($vol_cs->name_ordinal_isset() && $vol_cs->name_ordinal() ne 'base'
            && $opts->{handle_name_ordinal}) {
            $opts->{'volume'} .= "(".$vol_cs->name_ordinal().")";
        }

        # update the nodescope-node-name in %opts, only when user passes
        # determine_appropriate_nodescope_node => 1
        if ( $opts->{determine_appropriate_nodescope_node} ) {
           $opts->{'nodescope-node-name'} = $vol_cs->node();
        }
    }

    $Log->exit() if $may_exit;
}

=head2 rehost

    NACL::C::Volume->rehost(
                                command_interface => $ci,
                                    'vserver' => $vserver,
                                    'volume' => $volume,
                                ...);

                                                                 or

    $Volume_obj->rehost(...);

(Class or instance method)
This method is used to rehost a volume from one Vserver into another Vserver.
Uses CMode CLI/ZAPI .
Invokes  volume-rehost  for CMode Zapi non-iter

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should => $ruleset >>

(Optional) See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>


=item C<< 'vserver' => $vserver >>

(Required for class method, Not Applicable for instance method)
Source Vserver name
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'volume' => $volume >>

(Required for class method, Not Applicable for instance method)
Target volume name
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'force-unmap-luns' => $string >>

Unmap LUNs in volume possible value(s) are, true,false
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'destination-vserver' => $string >>

Destination Vserver name
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< 'auto-remap-luns' => $string >>

Automatic Remap of LUNs possible value(s) are, true,false
Applicable for CMode CLI/ZAPI, non-iter.

=item C<< '*' >>

This method takes all other options supported by "volume rehost" command in ONTAP

=item C<< other_options >>

This method takes other common options described in L<Common method options|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#Other_options_valid_on_component_methods>

=back

=back

=over

=item Exceptions

=over

=item C<APISet exceptions>

In general, Component methods propagate the exceptions thrown from
underlying API layer and the details of those exceptions are described in the
L<Exceptions section|http://thebrewery-web.corp.netapp.com/Brewery/brewery/QA/TestAutomation/NACL/UsersGuide/ComponentLayerUsersGuide/index.html#How_do_I_know_which_exceptions_are_thrown_by_a_component_method>
of user guide.

=back

=back

=cut

sub rehost {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @args) = @_;
    $pkg_or_obj->_frontend_log_debug_opts(@args);
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _optional_scalars(
                qw( force-unmap-luns destination-vserver auto-remap-luns
                    )
            ),
        },
        cli_cmd => 'volume rehost',
    );
    $pkg_or_obj->call_on_apiset_with_exceptions(
        %opts,
        choices => [
            {   method    => '_rehost_cmode_cli',
                interface => 'CLI',
                set       => 'CMode',
                zapi_type => 'none',
            },
            {   method    => '_rehost_cmode_zapi_non_iter',
                interface => 'ZAPI',
                set       => 'CMode',
                zapi_type => 'non-iter',
            },
        ],
        frontend => 'NACL::C::Volume::rehost',
    );

    $Log->exit() if $may_exit;
}

sub _rehost_cmode_cli {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    $pkg_or_obj->_base_backend_cmode_cli(@opts, api => 'volume_rehost',);
    $Log->exit() if $may_exit;
}

sub _rehost_cmode_zapi_non_iter {
    $Log->enter() if $may_enter;
    my ($pkg_or_obj, @opts) = @_;
    my $copy
        = [
        qw(force-unmap-luns destination-vserver volume auto-remap-luns vserver)
        ];
    $pkg_or_obj->_base_backend_cmode_zapi(
        @opts,
        api  => 'volume-rehost',
        copy => $copy,
    );
    $Log->exit() if $may_exit;
}


1;
