# Copyright (c) 2012-2016 NetApp, Inc., All Rights Reserved
#   Any use, modification, or distribution is prohibited
#   without prior written consent from NetApp, Inc.
#

## @summary Snapmirror Task Module
## @author dl-nacl-dev@netapp.com
## @status shared
## @pod here

package NACL::STask::Snapmirror;

use strict;
use warnings;
use POSIX qw/ceil/;
use NACL::ComponentUtils qw(Dumper);
use Readonly;
use Scalar::Util qw(blessed);

# This block is a hack to make "perl -wc" work despite the circular dependency
# on this module:
BEGIN {
    use File::Spec;
    $INC{'NACL/STask/Snapmirror.pm'} ||= File::Spec->rel2abs(__FILE__);
}

use base qw(NACL::C::Snapmirror NACL::STask::STask);

use Tharn qw( sleep snooze);
use NATE::Log qw(log_global);
my $Log = log_global();
use Params::Validate qw(validate validate_with validate_pos :types);
use NATE::Exceptions::Argument qw(:try);
use NACL::APISet::Exceptions::ResponseException ();
use NATE::BaseException                         ();
use NACL::Exceptions::VerifyFailure             ();
use NACL::APISet::Exceptions::TimeoutException  ();
use NACL::CS::ClusterIdentity;
use NACL::CS::ClusterPeer;
use NACL::STask::SystemLicense;
use NACL::STask::Volume;
use NACL::STask::VolumeSnapshot;
use NACL::C::Options;
use NACL::C::Cluster;
use NACL::C::Node;
use NACL::ComponentUtils qw(_optional_scalars);
use NACL::Exceptions::DataInconsistent;
use NACL::CS::SnapmirrorListDestinations;
use NACL::C::UnitNormalization;
use NACL::Exceptions::LunSetFailure;
use NACL::C::Exceptions::Volume::AlreadyRestricted;
use NACL::C::Exceptions::Snapmirror::NoVserverPeerRelationship;
use NACL::C::Exceptions::Snapmirror::NoDrRelationship;
use NACL::C::Exceptions::Snapmirror::BreakNotAllowedForLS;
use NACL::CS::SnapmirrorPolicy;
use NACL::GeneralUtils qw(nacl_method_retry);
use NACL::STask::VserverPeer;
use NACL::Exceptions::RetriesExceeded;
use NACL::STask::Exceptions::SnapshotMismatch;
use NATE::Time qw(timeout2time);
use NACL::C::Exceptions::Snapmirror::RelationshipNotInitialized;

# define task specific attributes
use Class::MethodMaker [
    scalar => [ { -type => 'NACL::C::Job' }, 'last_job' ],
    scalar => [ { -type => 'NACL::C::Job' }, 'job_component_check' ],
    scalar => 'sync_type'
];

# constants
use constant SNAPMIRROR_TIMEOUT => 1200;
use constant POLL_DELTA         => 10;
Readonly my $VOLSNAPRETRYCOUNT  => 10;

=head1 NAME

NACL::STask::Snapmirror

=head1 DESCRIPTION

C<NACL::STask::Snapmirror> provides a number of well-defined but
potentially complex or multi-step methods related to Snapmirror in
ONTAP.

It builds on top of, and is a derived class of "NACL::C::Snapmirror", and
so it also provides methods that are more in the scope of individual
Snapmirror-related commands. See "NACL::C::Snapmirror" for details. This
also means that a "NACL::STask::Snapmirror" object may generally be used
in place of a component object.

=head1 ATTRIBUTES

=head2 command_interface

(Required) As "NACL::C::Snapmirror". A component object that represents
the host to which to send commands. For Snapmirror, it represents
the host in which the destination volume or qtree resides.

=head2 destination-path

(Require) Destination path of the snapmirror relationship

=head2 source-path

(Optional) Source path of the snapmirror relationship

=head2 last_job

(Optional) A job component (NACL::C::Job) representing the last
operation performed on the snapmirror relationship.

=head2 sync_type

(Only applicable in 7Mode)
The sync-type of the Snapmirror relationship. This can be either of
"async" or "sync".

=head2 source_ci

(Optional)
Represents command interface of the source cluster or node.
This option is required if you are having intercluster snapmirror relationship
and if volume object is not passed for source-path or when the vserver
command_interface is passed to the method rather than node ci.
NOTE: source_ci is required in places where source-volume, source-vserver and
source-cluster parameters are passed to method call rather than source-path. This
is required since construction of source-path needs source node's version to
determine if the path should be in RR or SN style.

See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>.

=head1 LIMITATIONS

=head2 CMode support

1. If you are creating a DP relationship in a cross cluster where the
   source and destination resides in different cluster and if
   you are not passing volume C or STask objects as inputs to
   source-path and destination-path, then to license DP in
   source, the tharnhost file for source cluster should exist. Basically
   we look for a tharnhost file whose filename is same as the name of source
   cluster.

2. If you are using DP in a cross-cluster scenario, then you have to
   handle the step of peering the clusters by yourself.

=head1 Specifying source-path and destination-path

For all methods in this STask, the source-path and destination-path
arguments can be provided in a number of different ways which are explained
below.

=over

=item CMode

=over

=item * path

A string specifying the full path to the source volume. In CMode, the
path is of the form
cluster_name://vserver_name/volume_name
(or)
vserver_name:volume_name

=item * Volume C/STask object

An object of type NACL::C::Volume or NACL::STask::Volume. Note that the
if the command_interface of the specified object is a NACL::C::Vserver
object, then the C<source-cluster> needs to be specified (if it's the
C<source-path> argument; C<destination-cluster> needs to be specified
if it's the C<destination-path> argument).
In case of CMode intercluster relationship, if source volume object is not
passed, then source_ci should be passed which is required in certain methods
where accessing source cluster is required.
When destination-path is provided as a Volume object, then command_interface
is optional.

=item * Vserver C/STask object

An object of type NACL::C::Vserver or NACL::STask::Vserver.

=back

=item 7Mode

=over

=item * path

A string specifying the full path to the source volume. In 7Mode,
the path for a volume is of the form:
node_name:/vol/volume_name

The path for a qtree is of the form:
node_name:/vol/volume_name/qtree_name

=item * Volume C/STask object

For VSM/SSM.

=item * Qtree C/STask object

For QSM only.

=back

=back

=head1 METHODS

=head2 create

=head3 CMode

  # Specifying the direct path information as string to
  # source-path and destination-path inputs.
  my $snapmirror_object  = NACL::STask::Snapmirror->create(
      command_interface  =>  $command_interface,
      'source-path'      =>  $source_path,                   # Eg: 'cluster_name://vserver_name/volume_name'
      'destination-path' =>  $destination_path,              # Eg: 'cluster_name://vserver_name/volume_name'
      'type'             =>  $type,                          # LS/DP
      if_exists          =>  $action,                        # default 'die'
      nacltask_wait      =>  $number,                        # default '1'
      do_initialize      =>  $initialize_boolean,            # default '1'
      polling_interval   =>  $delta_value,                   # default 10 sec
      %other_options
  );

  # Specifying the volume component/STask object to
  # source-path and destination-path inputs. The command interface
  # in the volume object is not of type NACL::C::Vserver
  my $snapmirror_object  = NACL::STask::Snapmirror->create(
      command_interface  =>  $command_interface,
      'source-path'      =>  $source_volume_object,
      'destination-path' =>  $destination_volume_object,
      %other_task_overrides,
      %other_options
  );

  # Specifying the vserver component/STask object to
  # source-path and destination-path inputs.
  my $snapmirror_object  = NACL::STask::Snapmirror->create(
      command_interface  =>  $command_interface,
      'source-path'      =>  $source_vserver_object,
      'destination-path' =>  $destination_vserver_object,
      %other_task_overrides,
      %other_options
  );

  # Specifying the volume component/STask object to
  # source-path and destination-path inputs. The command interface
  # in the volume object is of type NACL::C::Vserver and hence specifying
  # source-cluster and destination-cluster is required
  my $snapmirror_object  = NACL::STask::Snapmirror->create(
      command_interface     =>  $command_interface,
      'source-path'         =>  $source_volume_object,
      'destination-path'    =>  $destination_volume_object,
      'source-cluster'      =>  $source_cluster_name,
      'destination-cluster' =>  $destination_cluster_name,
      %other_task_overrides,
      %other_options
  );

  # Specifing source-cluster, source-vserver, source-volume instead of
  # source-path. Destination is also specified as destination-cluster,
  # destination-vserver and destination-volume instead of destination-path
  my $snapmirror_object  = NACL::STask::Snapmirror->create(
      command_interface     =>  $command_interface,
      'source-cluster'      =>  $source_cluster_name,
      'destination-cluster' =>  $destination_cluster_name,
      'source-vserver'      =>  $source_vserver_name,
      'destination-vserver' =>  $destination_vserver_name,
      'source-volume'       =>  $source_volume_name,
      'destination-volume'  =>  $destination_volume_name,
      %other_task_overrides,
      %other_options
  );

  # Specifing source-path as volume object, in this case destination-path
  # will be picked by automatically and this is arrived only for volume
  # not for qtree or vserver
  my $snapmirror_object  = NACL::STask::Snapmirror->create(
      'source-path'      =>  $source_volume_object,
  );

=head3 7Mode

In 7Mode, the functionality provided by "create" is the same as that
for "initialize" (there is no "snapmirror create" in 7Mode). For 7Mode
examples, see L<initialize|lib-NACL-STask-Snapmirror-pm/initialize>.

(Class method) Creates and initializes a Snapmirror. This method
provides additional services beyond what's inherent to the product's
snapmirror creation commands, as controlled and described by the new
"if_exists" and "nacltask_wait" options. It also takes care of the
following dependencies:

CMode:

1) Licensing DP Snapmirror in both source and destination cluster

7Mode:

See the documentation of the L<initialize|lib-NACL-STask-Snapmirror-pm/initialize>
method.

Supports CMode CLI/ZAPI and 7Mode CLI.

=head3 NOTES

If the creation of snapmirror relationship fails for any of the following
reasons and if user hasn't overidden nacltask_wait to 0, then the task would
take care of purging and cleaning up the snapmirror relationship created,
if any.

* The job id returned by create/initialize has failed (applicable only for
  CMode)
* The state or status of snapmirror has reached an unexpected state
* Timeout is hit while waiting for snapmirror creation or initialization
  to complete

=over

=item Options

=over

=item C<< "sync_type => $sync_type" >>

(Optional, defaults to 'async')

To specify whether you need a synchronous or asynchronous
snapmirror.  Possible values are async, sync, semi-sync.
This input applies only for 7Mode.

=item C<< "if_exists=>$action" >>

B<DEPRECATED>, use C<nacltask_if_exists> instead. (This has been
retained only for backwards compatibility).

=item C<< nacltask_if_exists=>$action >>

(Optional) What to do if the Snapmirror relationship already exists
between the source and destination specified.  If $action is "die"
(the default), then fail with an exception (in the same way that
the Snapmirror component would have: by trying the Snapmirror create
and letting the product complain about the Snapmirror relationship
already existing).  If action is "reuse", then do nothing (return
a task object referring to the existing Snapmirror). If action is
"purge", then purge the relationship and create it newly.

=item C<< nacltask_wait => 0|1 >>

(Optional, defaults to 1)
Specifies whether to wait till the data transfer completes.
1 means to return after the initial transfer has completed.
0 means do not wait any longer than necessary (any longer
than the Snapmirror initialize command waits).
This option is applicable only when do_initialize option is set to 1.

Note that specifying a value between 0 & 1 is deprecated and is no longer
supported because of removal of support for real-time data from snapmirror show
(see burt 694025).
NOTE: In 7 Mode, nacltask_wait is a boolean and accepts 0 or 1. Support for
semi-background initialize is not provided as of now.

=item C<< "nacltask_verify_health => 0|1" >>

(Optional, defaults to '0')
If set to '1', verifies the health of the snapmirror relation.

=item C<< "do_initialize =>  0|1" >>

(Optional) If true (the default), a snapmirror initialize will be
performed on the snapmirror element after the relationship has
been made.  If FALSE is passed, the SM relationship will be made
between the volumes, but no initialize will be performed.  I.e.,
if do_initialize => 0 the dst volume will NOT mirror the src volume
until an initialize is explicitly performed at some later point.

This option is applicable only for CMode.

=item C<< "polling_interval =>  $polling_value" >>

(Optional) This is the interval (in seconds) after which the value of
the attributes will be polled. If not specified it defaults to 10 seconds.

=item C<< 'source-path' => $path_specification >>

(Required)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< "destination-path =>  $volume_specification" >>

(Optional)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< "source-cluster =>  $string" >>

(Optional) Name of cluster where the source volume resides.
This input applies only for CMode.


=item C<< "destination-cluster =>  $string" >>

(Optional) Name of cluster where the destination volume resides.
This input applies only for CMode.

=item C<< source_ci => $source_ci >>

(Optional)
See documentation at the beginning of this page

=item command_interface, apiset_must, apiset_should, method-timeout, etc.

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, defaults to 0(no to cleanup))
Flag indicating if this operation is to be cleaned up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to use for registering.
Default : Will use the default cleanup manager.

=item C<< nacltask_retry_tries_count => $tries_count >>

(Optional, defaults to 20)
The number of times to retry the resync command if it fails because
the snapshot is temporarily busy or stale catch entry.

Note that it is possible to make it not retry the delete command by setting
the value of this argument to 1.

=item C<< nacltask_retry_sleep_time => $sleep_time >>

(Optional, defaults to 10 seconds)
The time to sleep between retries of the resync command.

=item C<< nacltask_retry_num_times_code_ran => \$variable >>

(Optional)
A reference to a scalar variable which gets filled in with the number
of times the resync command was invoked.

=item C<< nacltask_config_diff => 0|1 >>

(Optional, defaults to 0, applicable only for VDR relationships and only
if do_initialize is set to 1)
If set to 1, perform a config diff check. Internally uses L<config_diff>.

All other options supported by
L<< NACL::C::Snapmirror->create|lib-NACL-C-Snapmirror-pm/create >>
are supported by NACL::STask::Snapmirror->create as well.

=back

=item Exceptions

=over

=item C<NATE::BaseException>

This type of exception is thrown when filer mode is not supported.

=item C<NACL::C::Exceptions::Snapmirror::NoVserverPeerRelationship>

This exception is raised when snapmirror is created without the vserver peering done for the
source and destination vservers.

=item C<NACL::C::Exceptions::Snapmirror::SubsystemNotFullyOperational>

This exception is thrown when an attempt is made to perform snapmirror
create operation when SnapMirror subsystem is not fully operational.

=back

=back

=cut

sub create {
    $Log->enter();
    my ( $pkg, %args ) = @_;

    my @tmp_args = %args;
    $pkg->_check_ava_perform_actions( 'arglist' => \@tmp_args );
    %args = @tmp_args;

    #delete 'destination-volume' from input argument list if present,
    #since by now 'destination-path' will already be formed for ONTAP-AVA scenario and
    #also field "-destination-path" and "-destination-volume" could not be used together
    #for snapmirror create
    delete $args{'destination-volume'} if ( $args{destination_ci} );


    my @args = %args;

    # We cannot directly use _if_exists_validate_spec since it defaults
    # the value to "die". This would mean that we would not know whether
    # if_exists or nacltask_if_exists was provided.
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {

            _was_created => { type => SCALARREF, optional => 1 },

            # component options which are overloaded
            'source-path'      => { type => OBJECT | SCALAR, optional => 1 },
            'destination-path' => { type => OBJECT | SCALAR, optional => 1 },

            # options which would be consumed by stask method itself apart
            # from passing the information to components.
            _optional_scalars(qw(source-cluster destination-cluster)),

            # stask specific options
            'sync_type'     => { type => SCALAR, default => 'async' },
            'do_initialize' => { type => SCALAR, default => 1 },

            # Kept for backwards compatibility
            if_exists => {
                type      => SCALAR,
                callbacks => $pkg->_if_exists_validate_callback(),
                optional  => 1
            },

            nacltask_if_exists => {
                type      => SCALAR,
                callbacks => $pkg->_if_exists_validate_callback(),
                optional  => 1
            },
            'nacltask_wait'                     => { type => SCALAR,    default  => 1 },
            'nacltask_verify_health'            => { type => SCALAR,    default  => 0 },
            'nacltask_retry_tries_count'        => { type => SCALAR,    default  => 20 },
            'nacltask_retry_sleep_time'         => { type => SCALAR,    default  => 10 },
            'nacltask_retry_num_times_code_ran' => { type => SCALARREF, optional => 1 },
            $pkg->_cleanup_validate_spec(),
        },

        allow_extra => 1,
    );

    if ( !$opts{'destination-path'} ) {

        # Check if destination-path can be constructed
        $pkg->_construct_single_path(
            params => \%opts,
            type   => 'destination'
         );

        if ( !$opts{'destination-path'} ) {
            # Create the destination volume
            $opts{'destination-path'} = $pkg->_create_dest_volume(%opts);
        }
    }

    # After the call to _common_validate_with(), %opts will have common params
    # like method-timeout, apiset_must, apiset_should, command_interface,
    # primary keys, polling_interval, stask specific params, other options
    # taken by NACL::C::Snapmirror->create()
    my $ci        = $opts{command_interface};
    my $stask_obj = undef;
    $opts{'method-timeout'} ||= SNAPMIRROR_TIMEOUT;

    # The dependencies which need to be handled are different between 7M and CM
    # Hence we will have two backends for 7M and CM separately.
    try {
        if ( $ci->mode() eq '7Mode' ) {

            # doesn't apply for 7Mode
            my %retry_opts;
            $pkg->_move_retry_params(
                source => \%opts,
                target => \%retry_opts
            );

            # Invoke 7M backend
            $stask_obj = $pkg->_create_7M(%opts);
        } elsif ( $ci->mode() eq 'CMode' ) {

            # Invoke CM backend
            $stask_obj = $pkg->_create_CM(%opts);
        } else {

            # Just to avoid open if-elsif
            NATE::BaseException->throw( "Filer mode '" . $ci->mode() . "' not supported." );
        }

        # Cleanup parameters
        my ( %opts_for_cleanup, %opts_for_register, $nacltask_to_cleanup );
        $pkg->_move_common_cleanup_opts(
            source => \%opts,
            target => \%opts_for_cleanup,
        );

        # Register this resource with the Cleanup Manager for cleanup
        $pkg->_copy_common_opts_for_cleanup(
            'source'              => { %opts, %opts_for_cleanup },
            'target'              => \%opts_for_register,
            'nacltask_to_cleanup' => \$nacltask_to_cleanup,
            'to_cleanup'          => 'purge'
        );

        if ( $ci->is_7mode() ) {
            $opts_for_register{'other_opts'} = { 'sync_type' => $opts{'sync_type'} };
        } elsif ( defined $opts{source_ci} ) {
            my %additional_opts;
            %additional_opts = ( destination_ci => $opts{destination_ci} ) if ( defined $opts{destination_ci} );
            $opts_for_register{'other_opts'} = { source_ci => $opts{source_ci}, %additional_opts };
        }

        $pkg->_register_for_cleanup(%opts_for_register)
            if ($nacltask_to_cleanup);
    }
    catch NATE::BaseException with {
        my $ex = shift;
        $pkg->_handle_create_failure(
            exception_obj => $ex,
            options       => \%opts
        );
        $ex->throw();
    };
    $Log->exit();
    return $stask_obj;
}

# end sub create

sub _handle_create_failure {
    my ( $pkg_or_obj, %args ) = @_;
    $Log->enter();
    my $ex   = $args{exception_obj};
    my %opts = %{ $args{options} };
    my $do_proceed
        = $ex->isa('NACL::APISet::Exceptions::TimeoutException')
        || $ex->isa('NACL::Exceptions::UnexpectedState')
        || $ex->isa('NACL::Exceptions::JobFailed');
    if ( !$do_proceed ) {
        $Log->exit();
        return;
    }
    my %common_options = ();
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_options
    );
    my $pri_keys = $pkg_or_obj->get_primary_keys_options(%opts);
    try {
        $Log->comment( "Snapmirror creation has failed with the exception " . ref($ex) . ". Cleanup would be performed by the task." );
        $Log->debug( "Error reason is\n" . $ex->text );
        $pkg_or_obj->purge(
            %$pri_keys,
            nacltask_if_purged => 'pass',
            %common_options
        );
    }
    otherwise {
        my $e = shift;
        $Log->trace( "Purge has failed because,\n" . $e->text() );
    };
    $Log->exit();
} ## end sub _handle_create_failure

=head2 monitor


  NACL::STask::Snapmirror->monitor(command_interface=>$ci,
                             'source-path'=>$source_path,
                             'destination-path'=>$destination_path,
                             'method-timeout'  =>  $timeout,
                             'polling_interval'=>  $polling_interval
                             'nacltask_wait_for_state'  => $expected_state,
                             'nacltask_wait_for_status' => $expected_status);

  (or)

  $snapmirror->monitor('method-timeout'  =>  $timeout,'polling_interval'=>  $polling_interval);

  (or)

  $snapmirror->monitor(...);

  (Class or instance method) The method monitors the snapmirror action by checking until the snapmirror reaches a specific state/status combination.
  Uses a CMode-CLI or CMode-ZAPI.

=over

=item Options

=over

=item C<< "source-path =>  $source_path" >>

(Optional)$volume_specification may be either a string specifying the full path
to the source volume, or it may be a NACL::(C|STask)::Volume with the
caveat that source-cluster must also be passed as an argument if the
Volume object's command_interface is a NACL::C::Vserver.

=item C<< "destination-path =>  $destination_path" >>

(Optional) destination-path takes the same form as the source-path argument, but
for the destination volume (and possibly requiring destination-cluster.)

=item C<< "source-cluster =>  $string" >>

(Optional) Name of cluster where the source volume resides.
This input applies only for CMode.


=item C<< "destination-cluster =>  $string" >>

(Optional) Name of cluster where the destination volume resides.
This input applies only for CMode.

=item C<< "polling_interval =>  $polling_value" >>

(Optional) This is the interval (in seconds) after which the value of
the attributes will be polled. If not specified it defaults to 10 seconds.
Disclaimer: User need to supply appropriate polling_interval. Larger value of
polling_interval might end up missing certain state or status that is being monitored if
transfer size is small.

=item C<< "nacltask_wait_for_state  =>  $expected_state | \@expected_states" >>

(Optional) State or states to monitor mirror for. Default: snapmirrored
This parameter can be specified as a scalar or as an array-reference.
If specified as a scalar, it is the value of the state which denotes
the completion of the operation.

In SFR operation where the state value is '-', it is recommended that the
user pass nacltask_wait_for_state => '-'.

If specified as an array-reference, it means that when the state reaches
any of the values specified, then the operation has completed.

=item C<< "nacltask_wait_for_status =>  $expected_status | \@expected_status" >>

(Optional) Status to monitor mirror for. Default: idle
This parameter can be specified as a scalar or as an array-reference.
If specified as a scalar, it is the value of the status which denotes
the completion of the operation.

If specified as an array-reference, it means that when the state reaches
any of the values specified, then the operation has completed.
If nacltask_wait_for_progress is specified and nacltask_wait_for_status
is not passed, nacltask_wait_for_status will be [ 'Transfer']

=item C<< "nacltask_wait_for_progress =>  $expected_transfer_progress" >>

(Optional) Transfer Progress to monitor mirror for. Unit is MB's. Value is expected without suffix as "100" rather than "100MB"
This option is deprecated as it no longer represents real time progress.
Use of nacltask_wait_for_status or nacltask_wait_for_state is encouraged.

=item C<< "nacltask_verify_health =>  $boolean" >>

(Optional) The default is 0. This is to check the health of the snapmirror
relationship. 1 means to check the health and last transfer error, throws
exception if the health is false. 0 means no health check after monitor.

=item C<< "set_flag =>  $boolean" >>

(Optional) When set_flag is 1, there will not be a default 10 second snooze
before polling for the snapmirror state/status. It is not recommended to use
this flag during STask monitor call. When a number of sm relations are monitored
at the same time, then this can be set to avoid the latency in getting the state/status.
If not specified it defaults to 0.

=back

=item Exceptions

=over

=item C<NATE::BaseException>

This type of exception is thrown when filer mode is not supported or when snapmirror status is not 'Transferring'.

=item C<NACL::Exceptions::InvalidChoice>

This type of exception is thrown when the command does not support current mode of filer.

=item C<NACL::Exceptions::UnexpectedState>

This type of exception is thrown when snapmirror status does not match the expected status.

=back

=back

=cut

sub monitor {
    $Log->enter();
    my ( $pkg, @args ) = @_;
    $pkg->_check_ava_perform_actions( 'arglist' => \@args );
    my %opts = $pkg->_common_validate_with(
        params          => \@args,
        additional_spec => {

            # component options which are overloaded
            'source-path'      => { type => OBJECT | SCALAR, optional => 1 },
            'destination-path' => { type => OBJECT | SCALAR, optional => 1 },

            # options which would be consumed by stask method itself apart
            # from passing the information to components.
            _optional_scalars(qw(source-cluster destination-cluster)),

            # stask specific options
            'nacltask_wait_for_state'    => { type => SCALAR | ARRAYREF, default  => 'snapmirrored' },
            'nacltask_wait_for_status'   => { type => SCALAR | ARRAYREF, optional => 1 },
            'nacltask_wait_for_progress' => { type => SCALAR,            optional => 1 },
            'set_flag'               => { type => SCALAR, optional => 1, default => 0 },
            'nacltask_verify_health' => { type => SCALAR, default  => 0 },

            #Burt845579. Kept for the backward compatiblity.
            'nacltask_health_check' => { type => SCALAR, default => 0 },
        },
        allow_extra => 1,
    );

    # After the call to _common_validate_with(), %opts will have common params
    # like method-timeout, apiset_must, apiset_should, command_interface,
    # primary keys, polling_interval and stask specific params
    my $ci = $opts{command_interface};

    $opts{'method-timeout'}   ||= SNAPMIRROR_TIMEOUT;
    $opts{'polling_interval'} ||= POLL_DELTA;

    # The dependencies which need to be handled are different between 7M and CM
    # Hence we will have two backends for 7M and CM separately.
    if ( $ci->mode() eq '7Mode' ) {

        # 7mode support
        my $msg = '7Mode not supported currently';
        $Log->exit();
        NACL::Exceptions::InvalidChoice->throw($msg);
    } elsif ( $ci->mode() eq 'CMode' ) {

        # Invoke CM backend
        $pkg->_monitor_CM(%opts);
    } else {

        # Just to avoid open if-elsif
        $Log->exit();
        NATE::BaseException->throw( "Filer mode '" . $ci->mode() . "' not supported." );
    }

    $Log->exit();

} ## end sub monitor

# CMode backend for monitor()
sub _monitor_CM {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;
    my $expected_state = $opts{nacltask_wait_for_state}
        if defined $opts{nacltask_wait_for_state};
    my $expected_status;
    my $expected_progress = $opts{nacltask_wait_for_progress}
        if defined $opts{nacltask_wait_for_progress};

    #Checking 'nacltask_health_check' for backward compatiblity.
    my $health_check = delete $opts{nacltask_verify_health}
        || delete $opts{nacltask_health_check};

    my $timeout = $opts{'method-timeout'};
    my $sleep   = $opts{polling_interval};
    my $ci      = $opts{command_interface};

    # By default set_flag will be '0'
    my $set_flag = delete $opts{set_flag};

    $expected_status = $opts{nacltask_wait_for_status}
        if defined $opts{nacltask_wait_for_status};

    if ( defined $expected_progress ) {

        $Log->warn( "Use of 'nacltask_wait_for_progress' is deprecated"
                . "\nas it no longer represents real time progress."
                . "\n'nacltask_wait_for_status' or 'nacltask_wait_for_state' "
                . "\nSHOULD be used to monitor the snapmirroring." );

        if ( !$expected_status ) {
            $expected_status = qr/Transfer/i;
        }
    }
    $expected_status = 'Idle' if ( !$expected_status );
    my $status_scalar = $expected_status;

    if (   defined $expected_progress
        && defined $expected_status )
    {
        if ( ref($expected_status) =~ /ARRAY/i ) {
            $status_scalar = join( '|', @{$expected_status} );
        }
        if ( $status_scalar !~ /preparing/i && $status_scalar !~ /Transfer/i && $status_scalar !~ /finalizing/i ) {

            $Log->exit();
            NATE::BaseException->throw( "Status has to be 'Transfer' if we were to monitor for the Snapmirror transfer progress" );
        }
    }

    #Skip check for progress. re: Burt 694025
    $expected_progress = undef;

    # Always make class method calls so that we don't have to bother
    # about skipping primary keys for instance method call
    my $pkg = ref $pkg_or_obj || $pkg_or_obj;

    my %apiset_must_should;
    $pkg->_hash_copy(
        source => \%opts,
        target => \%apiset_must_should,
        copy   => [qw(apiset_must apiset_should)]
    );

    # Delete the Job component as we have already waited for it.
    # Passing it again is unnecessary.Also delete Task specific options.
    $pkg_or_obj->_hash_move(
        source => \%opts,
        target => {},
        move   => [ qw(job_component nacltask_wait_for_status nacltask_wait_for_state nacltask_wait_for_progress) ]
    );

    # If there is no sleep it fails with "No matching snapmirror(s) found"
    # Sleep only for the first snapmirror instance if
    # STask::Snapmirror->monitor() is called by MTask::SnapmirrorSet->monitor().
    Tharn::sleep(10) if ( !$set_flag );

    my @attributes;
    my $populate_attribute = sub {
        my %args = @_;
        my $hash;
        if ( ref( $args{till_value} ) =~ /ARRAY/i ) {
            foreach my $state ( @{ $args{till_value} } ) {
                if ( ref($state) !~ /Regexp/i ) {
                    $state = qr/$state/i;
                }
            }
            $hash = {
                attribute_to_check => $args{attribute},
                till_value         => $args{till_value},
            };
        } else {
            $hash = {
                attribute_to_check => $args{attribute},
                till_value         => qr/$args{till_value}/i
            };
        }
        push @attributes, $hash;
    };
    if ( defined $expected_status ) {
        $populate_attribute->( till_value => $expected_status, attribute => 'status' );
    }
    if ( defined $expected_state ) {
        $populate_attribute->( till_value => $expected_state, attribute => 'state' );
    }

    if (@attributes) {
        $pkg->wait_for_completion( %opts, attributes => \@attributes, show_all_fields => 1, );
    }

    # This block is going to be skipped always.
    # Burt: 700633. see also Burt 694025
    # Checking for progress is not reliable as this field is not updated in real time.
    # This block is to be deleted eventually. Kept till above burts are well tested.
    if ( defined $expected_progress ) {
        my $expected_progress_in_mb = $expected_progress . "M";
        my $expected_progress_in_b = NACL::C::UnitNormalization::to_unit( $expected_progress_in_mb, 'b' );
        $expected_progress_in_b =~ s/b$//;

        my $start_time = time();
        my $prog_flag;
        my ( $cs, $total_progress );
    PROGRESS:
        {
            $cs = NACL::CS::Snapmirror->fetch(
                command_interface => $ci,
                filter            => { 'destination-path' => $opts{'destination-path'} },
                %apiset_must_should
            );
            $total_progress = $cs->total_progress();
            if ( $total_progress =~ /-/ || !defined($total_progress) ) {
                if ( !$prog_flag ) {
                    $prog_flag = 1;
                    Tharn::sleep(5);
                    redo PROGRESS;
                }
                $Log->exit();
                NACL::Exceptions::UnexpectedState->throw(
                    "Transfer is already Complete",
                    monitoring_fields => ['state'],
                    state             => $expected_state
                );
            }
        }
        my $end_time = time();
        my $lag      = $end_time - $start_time;

        # Take into account the time needed for retreiving the status
        $timeout = $timeout + $lag;

        $total_progress =~ s/b$//;
        while ( $total_progress < $expected_progress_in_b && $timeout > 0 ) {
            sleep $sleep;
            $timeout = $timeout - $sleep;

            $start_time = time();
            $cs         = NACL::CS::Snapmirror->fetch(
                command_interface => $ci,
                filter            => { 'destination-path' => $opts{'destination-path'} },
                %apiset_must_should
            );
            $total_progress = $cs->total_progress();
            if ( $total_progress =~ /-/ || !defined($total_progress) ) {
                $Log->exit();
                NACL::Exceptions::UnexpectedState->throw(
                    "Transfer is already Complete",
                    monitoring_fields => ['state'],
                    state             => $expected_state
                );
            }
            $total_progress =~ s/b$//;
            my $total_progress_in_mb = NACL::C::UnitNormalization::to_unit( $total_progress, 'M' );
            $total_progress_in_mb =~ s/m$//;
            my $total_progress_string = sprintf( "%.2f", $total_progress_in_mb );
            $Log->comment("Transferred $total_progress_string MB of data");
            $end_time = time();

            $lag = $end_time - $start_time;

            # Take into account the time needed for retreiving the status
            $timeout = $timeout + $lag;
        }
    }

    # Check for health and last transfer error if state is snapmirrored and
    # status is idle
    if ($health_check) {
        $pkg->verify_health(%opts);
    }
    $Log->exit();

} ## end sub _monitor_CM

# CMode backend for create()
sub _create_CM {
    $Log->enter();
    my ( $pkg, %all_opts ) = @_;

    # This is %opts for snapmirror_create
    my %opts = %all_opts;

    my $was_created = delete $opts{_was_created};
    $$was_created = 0;

    # Get the source-path and destination-path. Rest of the code
    # understands only these two inputs.
    # Also store the volume objects if passed.
    my ( $src_vol_obj, $dst_vol_obj, $ls_snapmirror );

    my ( %common_options, %task_opts );

    # Get the common component params: apiset_must, apiset_should,
    # method_timeout
    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%common_options
    );

    $pkg->_move_common_cleanup_opts(
        source => \%opts,
        target => \%task_opts,
    );

    # source-path could be either object or path. if user had specified
    # volume, vserver, cluster names separately, then _common_validate_with()
    # which is overloaded in NACL::C::Snapmirror would have constructed a
    # path out of it.
    if ( ref $opts{'source-path'} ) {
        $src_vol_obj = $opts{'source-path'};
        $opts{'source-path'} = $pkg->construct_volpath(
            vol_obj_or_volpath => $src_vol_obj,
            cluster_name       => delete( $opts{'source-cluster'} ),
            %common_options
        );
    }
    if ( ref $opts{'destination-path'} ) {
        $dst_vol_obj = $opts{'destination-path'};
        $opts{'destination-path'} = $pkg->construct_volpath(
            vol_obj_or_volpath => $dst_vol_obj,
            cluster_name       => delete( $opts{'destination-cluster'} ),
            %common_options
        );
    }
    my %path_inputs = (
        'source-path'      => $opts{'source-path'},
        'destination-path' => $opts{'destination-path'}
    );
    $Log->debug( "Path inputs:\n" . Dumper( \%path_inputs ) );

    # ignore sync_type option for cmode
    delete( $opts{'sync_type'} );

    # save these for use after component create()
    my $initialize_mirror   = delete( $opts{'do_initialize'} );
    my $wait_for_initialize = delete( $opts{'nacltask_wait'} );
    my $polling_interval    = delete( $opts{'polling_interval'} );
    my $verify_health       = delete( $opts{'nacltask_verify_health'} );
    my $config_diff         = delete $opts{nacltask_config_diff};

    my $command_interface = $opts{command_interface};

    # deal with if_exists
    my $if_exists
        = delete $opts{nacltask_if_exists}
        || $opts{if_exists}
        || 'die';

    # Handle Job ID
    my $job_component = delete $opts{job_component};
    {

        # We want to know the job ID of the job performed by
        # SUPER::create/initialize, whether or not our caller does.
        my $dummy;
        $job_component ||= \$dummy;
    }

    # checking the type of the snapmirror created
    my $type = $opts{type} || 'DP';
    if ( $type =~ /LS/i && ( $initialize_mirror == 1 ) ) {
        $ls_snapmirror = $pkg->find(
            command_interface => $opts{command_interface},
            filter            => {
                "source-path" => $path_inputs{"source-path"},
                'type'        => 'LS',
            },
            allow_empty => 1,
            %common_options,
        );
    }

    my %retry_opts;
    $pkg->_move_retry_params(
        source => \%opts,
        target => \%retry_opts
    );

    my ( $self, $reused );

CREATE: {
        use warnings;
        try {
            nacl_method_retry(
                code => sub {
                    $self = $pkg->SUPER::create( %opts, job_component => $job_component );
                },
                exceptions => ['NACL::C::Exceptions::Snapmirror::SubsystemNotFullyOperational'],
                %retry_opts
            );
            $$was_created = 1;
        }
        catch NACL::C::Exceptions::Snapmirror::DestinationAlreadyExists with {
            my $exception           = shift;
            my $existing_snapmirror = $pkg->find(
                command_interface => $opts{command_interface},
                filter            => \%path_inputs,
                allow_empty       => 1,
                %common_options,
            );
            if ( defined $existing_snapmirror ) {
                $self = $pkg->_element_exists_handler(
                    create_opts        => \%opts,
                    nacltask_if_exists => $if_exists,
                    exception          => $exception
                );
                if ($self) {
                    $reused = 1;
                } else {
                    no warnings qw(exiting);
                    redo CREATE;
                }
            } else {

                # This means that the snapmirror doesn't exist but the destination is already used.
                # Add new text to existing error text
                $exception->set_text( "destination-path is already used but is used with a different source-path\n" . $exception->text() );
                $Log->exit();
                $exception->throw();
            }
        }
        catch NACL::C::Exceptions::Snapmirror::SnapmirrorDpLicenseCheckFailed with {
            my $e = shift;
            if ( $e->text() =~ /SnapMirror_DP license check failed for cluster ([\w-]+)/i ) {
                my $cluster_reported  = $1;
                my $dest_cluster_name = $pkg->destination_cluster(
                    command_interface  => $command_interface,
                    'destination-path' => $opts{'destination-path'},
                    %common_options
                );

                # check destination first as we have command interface to
                # access the destination cluster. Don't change this order.
                if ( $cluster_reported eq $dest_cluster_name ) {
                    $Log->trace( "Snapmirror DP needs to be installed" . " in the destination cluster $dest_cluster_name" );
                    $pkg->_install_license(
                        volume_obj => $dst_vol_obj,
                        path       => $opts{'destination-path'},

                        # pass the obtained command_interface
                        command_interface => $command_interface,
                        %common_options,
                        %task_opts
                    );
                } else {
                    $Log->trace( "Snapmirror DP needs to be installed" . " in the source cluster" );

                    unless ( defined $opts{source_ci} ) {
                        $Log->exit();
                        NATE::Exceptions::Argument->throw(
                            "source_ci must be passed by user" . " since this is an inter-cluster snapmirror relationship" );
                    }

                    $pkg->_install_license(
                        volume_obj        => $src_vol_obj,
                        path              => $opts{'source-path'},
                        command_interface => $opts{source_ci},
                        destination_ci    => $command_interface,
                        %common_options,
                        %task_opts
                    );
                }
                no warnings;
                redo CREATE;
            }
        }
        catch NACL::C::Exceptions::Snapmirror::NoVserverPeerRelationship with {

            # Below regex fetch the vserver name when path is in any of below format-
            # 1- node_name://vserver/volume
            # 2- vserver:volume
            my $src_vserver = $2 || $4
                if ( $opts{'source-path'} =~ m[(.*://(.*?)/)|((.*?):)] );
            my $remote_vserver = $2 || $4
                if ( $opts{'destination-path'} =~ m[(.*://(.*?)/)|((.*?):)] );

            if ( $pkg->check_if_intercluster(%opts) ) {
                NACL::STask::VserverPeer->create(
                    command_interface  => $opts{source_ci},
                    vserver            => $src_vserver,
                    'peer-vserver'     => $remote_vserver,
                    remote_interface   => $command_interface,
                    nacltask_if_exists => 'reuse',
                    %common_options,
                    %task_opts
                );
            } else {
                NACL::STask::VserverPeer->create(
                    command_interface  => $command_interface,
                    vserver            => $src_vserver,
                    'peer-vserver'     => $remote_vserver,
                    nacltask_if_exists => 'reuse',
                    %common_options,
                    %task_opts
                );
            }

            no warnings;
            redo CREATE;
        };

        #If 'if_exists=reuse', then return without re-initialization
        if ($reused) {
            $Log->exit();
            return $self;
        }

        # Even create() returns a job ID
        $self->last_job( ${$job_component} );
        $self->source_ci( $opts{source_ci} );
        unless ($initialize_mirror) {

            # initialize is not required
            if ( !$wait_for_initialize ) {

                # non-blocking behavior is requested.So return immediately now.
                $Log->debug("Create without initialize -non blocking behavior");
                $Log->exit();
                return $self;
            }
        }

        # BTW, wait_on_job also takes polling_interval
        $common_options{polling_interval} = $polling_interval
            if ( defined $polling_interval );
        if ( $wait_for_initialize || $initialize_mirror ) {

            # blocking behavior is requested. So wait till create is completed
            if ( defined $self->last_job() ) {
                $self->wait_on_job(
                    job_component => $self->last_job(),
                    %common_options
                );
            }

            #Here snapmirror state is checked for either Uninitialized or
            # Snapmirrored. After deleting an existing snapmirror relation
            # and doing snapmirror create on the same volume pair results
            # in the state to be snapmirrored. Snapmirror delete does not
            # completely destroy the relationship. Release is required to
            # delete base snapshot from source, which is not available
            # currently. Until release is available, we can add this check
            # for snapmirrored state. Placeholder burt for this 528166
            my %opts_for_wait = ( %common_options, %{ $pkg->get_primary_keys_options(%opts) }, command_interface => $command_interface );
            my @status_values = ( qr/^Idle$/i, '-' );

            # We know that $self is an object, so we aren't calling is_granular
            # statically in this case.
            if ( $self->is_granular ) {
                push @status_values, qr/^OutOfSync$/i;
            }
            $pkg->wait_for_completion(
                %opts_for_wait,
                attributes => [
                    {   attribute_to_check => 'state',
                        till_value         => [ qr/Uninitialized/i, qr/Snapmirrored/i, qr/Broken-off/i, qr/^-$/ ]
                    },
                    {   attribute_to_check => 'status',
                        till_value         => \@status_values,
                    },
                ],
                show_all_fields => 1,
            );
            $pkg->verify_health( %opts_for_wait, job_component => $self->last_job(), )
                if ($verify_health);
        }
        if ($initialize_mirror) {

            # Perform initialize
            # We should make the snapmirror initialize to run in the background
            # since user had asked for a non-blocking behavior.
            my ( $method_name, %job );
            if (   ( $type !~ /LS/i )
                || ( defined($ls_snapmirror) ) )
            {
                $method_name                       = "initialize";
                $job{job_component}                = $job_component;
                $task_opts{nacltask_verify_health} = $verify_health;
            } else {
                $method_name = "initialize_ls_set";
            }
            my $apiset  = $command_interface->apiset( interface => 'CLI' );
            my $api     = "snapmirror_" . $method_name;
            my @options = $apiset->get_allowed_options_for_command( command => $api );

            # as per burt 818898 source_ci was missed, hence added the same.
            my @other_options = qw(polling_interval nacltask_wait source_ci
                nacltask_config_diff);
            @options = ( @options, @other_options );
            my %common_opts;
            $pkg->_copy_common_component_params_with_ci(
                source => \%all_opts,
                target => \%common_opts
            );
            my $target = $pkg->_hash_copy(
                source => \%all_opts,
                copy   => \@options
            );

            ##burt934176 Extra input: vserver-type is removed.

            delete $$target{type} if ( defined $$target{type} );

            $self = $pkg->$method_name( %{$target}, %job, %task_opts, %common_opts );
        }
    }

        # add destination_ci to created stask object
        $self->{destination_ci} = $opts{destination_ci} if ( defined $opts{destination_ci} );

    $Log->exit();
    return $self;
} ## end sub _create_CM

# 7Mode backend for create().
sub _create_7M {
    $Log->enter();
    my $self      = shift;
    my %opts      = @_;
    my $stask_obj = $self->_initialize_7M(%opts);
    $Log->exit();
    return $stask_obj;
} ## end sub _create_7M

=head2 initialize

  my $snapmirror_object  = NACL::STask::Snapmirror->initialize(
      command_interface  =>  $command_interface,
      'source-path'      =>  $source_specification,
      'destination-path' =>  $destination_specification,
      'sync_type'        =>  $sync_type,                     # default is async
      if_exists          =>  $action,                        # default 'die'
      nacltask_wait      =>  $number,                        # default '1'
      polling_interval   =>  $delta_value,                   # default 10 sec
      %other_options
  );

  (or)

  $snapmirror_object->initialize(nacltask_wait => '0',
                 %options );

  $snapmirror_object->abort();                               #abort the transfer in progress

=head3 7Mode examples

=over

=item SSM (Synchronous Snapmirror)

=over

=item Providing source-path and destination-path as Volume C/STask objects

    my $snapmirror_object  = NACL::STask::Snapmirror->initialize(
        command_interface  =>  $command_interface,
        'source-path'      =>  $source_volume_object,
        'destination-path' =>  $destination_volume_object,
        'sync_type'        =>  'sync',                         # default 'async'
        if_exists          =>  $action,                        # default 'die'
        nacltask_wait      =>  $boolean,                       # default '1'
        polling_interval   =>  $delta_value,                   # default 10 sec
        %other_options
    );

=item Providing source-path and destination-path as paths

    my $snapmirror_object  = NACL::STask::Snapmirror->initialize(
        command_interface  =>  $command_interface,
        'source-path'      =>  $source_path,                   # Example: 'filer_name:vol_name'
        'destination-path' =>  $destination_path,              # Example: 'filer_name:vol_name'
        'sync_type'        =>  'sync',                         # default 'async'
        if_exists          =>  $action,                        # default 'die'
        nacltask_wait      =>  $boolean,                       # default '1'
        polling_interval   =>  $delta_value,                   # default 10 sec
        %other_options
    );

=back

=back

=over

=item VSM (Volume Snapmirror)

=over

=item Providing source-path and destination-path as Volume C/STask objects

    my $snapmirror_object  = NACL::STask::Snapmirror->initialize(
        command_interface  =>  $command_interface,
        'source-path'      =>  $source_volume_object,
        'destination-path' =>  $destination_volume_object,
        if_exists          =>  $action,                        # default 'die'
        nacltask_wait      =>  $boolean,                       # default '1'
        polling_interval   =>  $delta_value,                   # default 10 sec
        %other_options
    );

=item Providing source-path and destination-path as paths

    my $snapmirror_object  = NACL::STask::Snapmirror->initialize(
        command_interface  =>  $command_interface,
        'source-path'      =>  $source_path,                   # Example: 'filer_name:vol_name'
        'destination-path' =>  $destination_path,              # Example: 'filer_name:vol_name'
        if_exists          =>  $action,                        # default 'die'
        nacltask_wait      =>  $boolean,                       # default '1'
        polling_interval   =>  $delta_value,                   # default 10 sec
        %other_options
    );

=back

=back

=over

=item QSM (Qtree Snapmirror)

Note that the destination-path has to be given as a path since for QSM,
the destination qtree gets created as a part of the snapmirror initialize.

=over

=item Providing source-path as VolumeQtree C/STask object

    my $snapmirror_object  = NACL::STask::Snapmirror->initialize(
        command_interface  =>  $command_interface,
        'source-path'      =>  $source_qtree_obj,
        'destination-path' =>  $destination_qtree_path,        # Example: 'filer_name:/vol/vol_name/qtree_name'
        if_exists          =>  $action,                        # default 'die'
        nacltask_wait      =>  $boolean,                       # default '1'
        polling_interval   =>  $delta_value,                   # default 10 sec
        %other_options
    );

=item Providing source-path as a path

    my $snapmirror_object  = NACL::STask::Snapmirror->initialize(
        command_interface  =>  $command_interface,
        'source-path'      =>  $source_qtree_path,             # Example: 'filer_name:/vol/vol_name/qtree_name'
        'destination-path' =>  $destination_qtree_path,        # Example: 'filer_name:/vol/vol_name/qtree_name'
        if_exists          =>  $action,                        # default 'die'
        nacltask_wait      =>  $boolean,                       # default '1'
        polling_interval   =>  $delta_value,                   # default 10 sec
        %other_options
    );

=back

=back

(Class or Instance method) Initializes a Snapmirror. This method
provides additional services beyond what's inherent to the product's
snapmirror initialize command, as controlled and described by the new
"if_exists" and "nacltask_wait" options. It also takes care of the
following dependencies.

In case of 7Mode/7G systems, it handles
1) Licensing sync snapmirror in source filer, if necessary
2) Switching on snapmirror, if necessary
2) Restricting the destination volume (not applicable for QSM)
3) Adding entries to the registry to set up /etc/snapmirror.conf file
(applicable only for SSM)

In case of CMode, it handles
1) Licensing DP Snapmirror in both source and destination cluster

Supports CMode CLI/ZAPI and 7Mode CLI.

=over

=item Options

=over

=item C<< "sync_type => $sync_type" >>

(Optional) To specify whether you need a synchronous or asynchronous
snapmirror.  Possible values are async, sync, semi-sync.
This input applies only for 7Mode.

=item C<< "if_exists=>$action" >>

(Optional) What to do if the Snapmirror relationship already exists
between the source and destination specified.  If $action is "die"
(the default), then fail with an exception (in the same way that
the Snapmirror component would have: by trying the Snapmirror create
and letting the product complain about the Snapmirror relationship
already existing).  If action is "reuse", then do nothing (return
a task object referring to the existing Snapmirror).

=item C<< nacltask_wait => 0|1 >>

=over

=item CMode

(Optional, defaults to 1)
Specifies whether to wait till the data transfer completes.
1 means to return after the initial transfer has completed.
0 means do not wait any longer than necessary (any longer
than the Snapmirror initialize command waits).

Note that specifying a value between 0 & 1 is deprecated and is no longer
supported because of removal of support for real-time data from snapmirror show
(see burt 694025).

=item 7Mode

(Optional, defaults to 1)
Is a boolean and accepts 0 or 1. Support for semi-background initialization
is not provided yet.

Note that if a decimal number is provided (as might be necessary for CMode)
then it will be taken to mean to wait in 7Mode.

=back

=item C<< nacltask_retry_tries_count => $tries_count >>

(Optional, Only for Cmode defaults to 30)
The number of times to retry the initialize command if it fails because of
the system initialization exception.

Note that it is possible to make it not to retry the delete command by setting
the value of this argument to 1.

=item C<< nacltask_retry_sleep_time => $sleep_time >>

(Optional, Only for Cmode defaults to 10 seconds)
The time to sleep between retries of the initialization command.

=item C<< nacltask_verify_health => 0|1 >>

(Optional, defaults to '1', Only for CMode)
If set to '1', verifies the health of the snapmirror.

=item C<< nacltask_config_diff => 0|1 >>

(Optional, defaults to 0, applicable only for VDR relationships)
If set to 1, perform a config diff check. Internally uses L<config_diff>.

=item C<< "polling_interval =>  $polling_value" >>

(Optional) This is the interval (in seconds) after which the value of
the attributes will be polled. If not specified it defaults to 10 seconds.

=item C<< 'source-path' => $path_specification >>

=item C<< nacltask_to_cleanup => 0|1 >>

(Optional, defaults to 0(no to cleanup))
Flag indicating license install needs to be cleaned up or not.

=item C<< nacltask_cleanup_manager >>

Cleanup manager to use for registering.
Default : Will use the default cleanup manager.

=over

=item CMode

Required when $nacltask_wait value is > 0 and < 1. Source information is
required to compute difference in vol sizes so that specified percentage
of transfer can be done in background.

$volume_specification may be either a string specifying the full path
to the source volume, or it may be a NACL::(C|STask)::Volume with the
caveat that source-cluster must also be passed as an argument if the
Volume object's command_interface is a NACL::C::Vserver.

=item 7Mode

(Required)

This can be in any of the following forms:

=over

=item * path

The path is provided as a string. This will need to include the source filer
name if the source volume is on a different filer from the destination
volume.

=item * Volume C/STask object

For VSM/SSM. Given a Volume C/STask object, the task method will construct
the path.

=item * Qtree C/STask object

For QSM only. Given a Qtree C/STask object, the task method will construct
the path.

=back

=back

=item C<< 'destination-path' =>  $path_specification >>

=over

=item CMode

(Optional) destination-path takes the same form as the source-path argument,
but for the destination volume (and possibly requiring destination-cluster.)

=item 7Mode

(Required)

Takes the same form as the C<source-path> argument.

=back

=item C<< "source-cluster =>  $string" >>

(Optional) Name of cluster where the source volume resides.
This input applies only for CMode.


=item C<< "destination-cluster =>  $string" >>

(Optional) Name of cluster where the destination volume resides.
This input applies only for CMode.

=item command_interface, apiset_must, apiset_should, method-timeout, etc.

All other options supported by
L<< NACL::C::Snapmirror->initialize|lib-NACL-C-Snapmirror-pm/initialize >>
are supported by NACL::STask::Snapmirror->initialize as well.

=back

=item Exceptions

=over

=item C<NATE::Exceptions::Argument>

This type of exception is thrown when 'destination-path' is neither a path to a volume nor a Volume Component/STask object.

or

This type of exception is thrown when either of the options (path or volume_obj or command_interface) are missing.

=item C<NATE::BaseException>

This type of exception is thrown when a tharnhost file in the name of cluster doesn't exist.

=item C<NACL::C::Exceptions::Volume::AlreadyRestricted>

This type of exception is thrown when an attempt is made to restict a volume which is already in restricted state.

=item C<NACL::C::Exceptions::Snapmirror::NoDrRelationship>

This exception is raised when an attempt is made to initialize the snapmirror without
the vserver peering done for source and destination vservers.

=back

=back

=cut

sub initialize {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {

            # stask specific options
            'sync_type' => {
                type      => SCALAR,
                default   => 'async',
                callbacks => {
                    "sync_type must be 'sync' or 'async'" => sub {
                        $_[0] =~ /(sync)|(async)/i;
                        }
                }
            },

            # Kept for backwards compatibility
            if_exists => {
                type      => SCALAR,
                callbacks => $pkg_or_obj->_if_exists_validate_callback(),
                optional  => 1
            },

            nacltask_if_exists => {
                type      => SCALAR,
                callbacks => $pkg_or_obj->_if_exists_validate_callback(),
                optional  => 1
            },

            'nacltask_wait'              => { type => SCALAR, default => 1 },
            'nacltask_verify_health'     => { type => SCALAR, default => 1 },
            'nacltask_retry_tries_count' => { type => SCALAR, default => 30 },
            'nacltask_retry_sleep_time'  => { type => SCALAR, default => 10 },
            nacltask_config_diff         => { type => SCALAR, default => 0 },
            $pkg_or_obj->_cleanup_validate_spec(),
        },
    );

    my $ci        = $opts{command_interface};
    my $stask_obj = undef;

    # The dependencies which need to be handled are different between 7M and CM
    # Hence we will have two backends for 7M and CM separately.
    if ( $ci->mode() eq '7Mode' ) {

        # This input doesn't apply for 7Mode
        delete $opts{do_initialize};
        delete $opts{nacltask_verify_health};
        delete $opts{nacltask_config_diff};

        # doesn't apply for 7Mode
        my %retry_opts;
        $pkg_or_obj->_move_retry_params(
            source => \%opts,
            target => \%retry_opts
        );

        # Invoke 7M backend
        $stask_obj = $pkg_or_obj->_initialize_7M(%opts);
    } elsif ( $ci->mode() eq 'CMode' ) {

        # sync_type not applicable for CMode
        delete $opts{sync_type};

        # Invoke CM backend
        $stask_obj = $pkg_or_obj->_initialize_CM(%opts);
    }

    $Log->exit();
    return $stask_obj;
} ## end sub initialize

# CMode backend for initialize()
sub _initialize_CM {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;

    # Get the source-path and destination-path. Rest of the code
    # understands only these two inputs.
    # Also store the volume objects if passed.
    my ( $src_vol_obj, $dst_vol_obj );

    my ( %common_options, %opts_for_cleanup );

    # Get the common component params: apiset_must, apiset_should,
    # method_timeout
    $pkg_or_obj->_copy_common_component_params(
        source => \%opts,
        target => \%common_options
    );

    $pkg_or_obj->_move_common_cleanup_opts(
        source => \%opts,
        target => \%opts_for_cleanup,
    );

    my %path_inputs = (
        'source-path'      => $opts{'source-path'},
        'destination-path' => $opts{'destination-path'}
    );
    $Log->debug( sub { "Path inputs:\n" . Dumper( \%path_inputs ) } );

    # save these for use after component initialize()
    my $wait_for_initialize = delete( $opts{'nacltask_wait'} );
    my $polling_interval    = delete( $opts{'polling_interval'} );
    my $verify_health       = delete( $opts{'nacltask_verify_health'} );
    my $config_diff         = delete $opts{nacltask_config_diff};

    # Should wait if config_diff is set to 1
    $wait_for_initialize ||= $config_diff;

    my %retry_opts;
    $pkg_or_obj->_move_retry_params(
        source => \%opts,
        target => \%retry_opts
    );

    my $command_interface = $opts{command_interface};

    delete $opts{'source-cluster'}      if defined $opts{'source-path'};
    delete $opts{'destination-cluster'} if defined $opts{'destination-path'};

    # deal with if_exists
    my $if_exists
        = delete $opts{nacltask_if_exists}
        || $opts{if_exists}
        || 'die';

    my $license;

    my $pkg = ref $pkg_or_obj || $pkg_or_obj;
    my $self = $pkg->_handle_if_exists_action(
        if_exists   => $if_exists,
        create_opts => {%opts}
    );

    # Handle Job ID
    my $job_component = delete $opts{job_component};
    {

        # We want to know the job ID of the job performed by
        # SUPER::initialize, whether or not our caller does.
        my $dummy;
        $job_component ||= \$dummy;
    }

    # The snapmirror relationship could exist already and initialized.
    # Check whether the state of the obtained object is in "Snapmirrored"
    # state and do not perform initialize if 'if_exists' is reuse.
    my $force_initialize = 1;
    if ($self) {
        if ( lc $if_exists eq 'reuse' ) {
            my $state = $self->get_one_state_attribute('state');
            if ( lc $state eq 'snapmirrored' ) {
                $force_initialize = 0;
            }
        }
    }
    if ( ref($pkg_or_obj) && $pkg_or_obj->source_ci() ) {
        $opts{source_ci} ||= $pkg_or_obj->source_ci();
    }

    # finally try to create a new element w/object if it doesn't exist (or, in
    # "die"'s case, if we are blindly trying to create a new element w/object,
    # no matter whether it exists already or not.)
    if ( !$self || $force_initialize ) {

        # We should make the snapmirror initialize to run in the background
        # since user had asked for a non-blocking behavior.
        if ( $wait_for_initialize != 0 && $wait_for_initialize < 1 ) {
            $Log->warn( 'Specifying a nacltask_wait value between 0 & 1 '
                    . 'is deprecated because it is no longer possible to get '
                    . 'real-time snapmirror data; see 694025. (the value provided '
                    . "was $wait_for_initialize) The only currently supported values "
                    . 'for nacltask_wait are 0 and 1' );

            # Set to 0 because the user would have wanted it to return
            # while data is still being transferred
            $wait_for_initialize = 0;
        }
    INITIALIZE: {
            use warnings;
            try {

                # Always make class method calls so that we don't have to bother
                # about skipping primary keys for instance method call
                # Perform initialize

                nacl_method_retry(
                    code => sub {
                        $self = $pkg->SUPER::initialize( %opts, job_component => $job_component );
                    },
                    exceptions => 'NACL::C::Exceptions::Snapmirror::SystemInitialization',
                    %retry_opts
                );

                # If this is invoked as an instance method, then we should not
                # return a new object.
                $self = $pkg_or_obj if ( ref $pkg_or_obj );
            }
            catch NACL::C::Exceptions::Snapmirror::SnapmirrorDpLicenseCheckFailed with {
                my $error = shift;

                # Error text is same between CLI and ZAPI when snapmirror is
                # not licensed
                $error->text() =~ /SnapMirror_DP\s+license\s+check\s+failed\s+for\s+cluster\s+([\w-]+)/i;
                my $cluster_reported  = $1;
                my $dest_cluster_name = $pkg_or_obj->destination_cluster(
                    command_interface  => $command_interface,
                    'destination-path' => $opts{'destination-path'},
                    %common_options,
                );

                # check destination first as we have command interface to
                # access the destination cluster. Don't change this order.
                if ( $cluster_reported eq $dest_cluster_name ) {
                    $Log->trace( "Snapmirror DP needs to be installed" . " in the destination cluster $dest_cluster_name" );
                    $pkg_or_obj->_install_license(
                        path => $opts{'destination-path'},

                        # pass the obtained command_interface
                        command_interface => $command_interface,
                        %common_options,
                        %opts_for_cleanup
                    );
                } else {
                    $Log->trace( "Snapmirror DP needs to be installed" . " in the source cluster" );
                    $pkg_or_obj->_install_license(
                        path              => $opts{'source-path'},
                        command_interface => $opts{source_ci},
                        destination_ci    => $command_interface,
                        %common_options,
                        %opts_for_cleanup
                    );
                }
                no warnings;
                redo INITIALIZE;
            }

            catch NACL::C::Exceptions::Snapmirror::NoDrRelationship with {

                # Need to perform vserver peer creation & snapmirror creation
                # which will happen in below call to create method
                $pkg_or_obj->create(
                    command_interface  => $command_interface,
                    "source_ci"        => $opts{'source_ci'},
                    'source-path'      => $opts{'source-path'},
                    "destination-path" => $opts{'destination-path'},
                    nacltask_wait      => 1,
                    do_initialize      => 0,
                    %common_options,
                );
                no warnings;
                redo INITIALIZE;
            };

            # In SN, if license for snapmirror is not installed, then the error is
            # reported in snapmirror show output. Add check here to see initialization
            # is successful.
            Tharn::snooze(2);
            $license = $pkg->_license_check( %opts, %opts_for_cleanup );
            if ( !$license ) {
                redo INITIALIZE;
            }
        }
        $self->last_job( ${$job_component} );
        $self->source_ci( $opts{source_ci} );

        # add destination_ci to created stask object
        $self->{destination_ci} = $opts{destination_ci} if ( defined $opts{destination_ci} );

        if ($wait_for_initialize) {
            my %polling_interval_hash;
            $polling_interval_hash{polling_interval} = $polling_interval
                if ( defined $polling_interval );

            # This is an instance method call and so we shouldn't pass
            # primary keys and command_interface. So we have to pass
            # polling_interval + apiset_(must|should), method_timeout
            Tharn::snooze(5);
            $self->wait_for_initialization(
                %common_options,
                %polling_interval_hash,
                nacltask_verify_health => $verify_health,
                'last-transfer-type'   => 'initialize'
            );
        }

        if ($config_diff) {
            $self->config_diff(%common_options);
        }
    }
    $Log->exit();
    return $self;
} ## end sub _initialize_CM

sub _license_check {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params      => \@args,
        allow_extra => 1,
    );
    my $installed = 1;
    my $is_xdp    = 0;

    # Check that the last snapmirror operation was successful by looking at the
    # "healthy" and "last_transfer_error" fields of the current state:
    my ( %common_opts, %opts_for_cleanup );
    $pkg_or_obj->_copy_common_component_params(
        source => \%opts,
        target => \%common_opts
    );

    $pkg_or_obj->_move_common_cleanup_opts(
        source => \%opts,
        target => \%opts_for_cleanup,
    );
    my $state = NACL::CS::Snapmirror->fetch(
        command_interface => $opts{command_interface},
        filter            => { 'destination-path' => $opts{'destination-path'} },
        %common_opts,
    );

    if ( ( lc $state->healthy eq 'false' ) && ( $state->status =~ /Idle/i ) ) {

        # For some reason, the error message is coming back from ONTAP
        # braced within "{" and "}" characters. Do not want.
        ( my $errtext = $state->last_transfer_error ) =~ s/^{(.*)}$/$1/;
        if ( $errtext =~ /SnapMirror_DP license check failed/i ) {
            $installed = 0;
            $is_xdp    = 0;
        } elsif ( $errtext =~ /SnapVault license check failed/i ) {
            $installed = 0;
            $is_xdp    = 1;
        }
        if ( !$installed ) {
            if ( $errtext =~ /DST Cluster/i ) {
                $Log->trace( "Snapmirror DP/SnapVault needs to be installed" . " in the destination cluster " );
                $pkg_or_obj->_install_license(
                    path              => $opts{'destination-path'},
                    command_interface => $opts{command_interface},
                    'is_xdp'          => $is_xdp,
                    %common_opts,
                    %opts_for_cleanup
                );
            } else {
                $Log->trace( "Snapmirror DP/SnapVault needs to be installed" . " in the source cluster" );
                $pkg_or_obj->_install_license(
                    path              => $opts{'source-path'},
                    command_interface => $opts{source_ci},
                    destination_ci    => $opts{command_interface},
                    'is_xdp'          => $is_xdp,
                    %common_opts,
                    %opts_for_cleanup
                );
            }

        }
    }
    return $installed;
}

# Initialize 7M backend
sub _initialize_7M {
    $Log->enter();
    my ( $pkg, %opts ) = @_;

    my %opts_for_cleanup;
    $pkg->_move_common_cleanup_opts(
        source => \%opts,
        target => \%opts_for_cleanup,
    );

    my $mode = delete $opts{sync_type};

    my $if_exists
        = delete $opts{nacltask_if_exists}
        || $opts{if_exists}
        || 'die';

    # deal with if_exists
    my $snapmirror_obj = $pkg->_handle_if_exists_action(
        if_exists   => $if_exists,
        create_opts => \%opts
    );

    # At this point, if we still have some object, then we don't have
    # to create. Simply return the object.
    if ($snapmirror_obj) {
        $Log->exit();
        return $snapmirror_obj;
    }

    my $isa_vol = $pkg->is_operated_on_volume(%opts);
    if ( $mode =~ /^sync$/i ) {
        if ($isa_vol) {
            $snapmirror_obj = $pkg->_initialize_7m_ssm( %opts, %opts_for_cleanup );
        } else {
            $Log->exit();
            NATE::Exceptions::Argument->throw( "The 'destination-path' "
                    . 'parameter must be either the path to a volume or must '
                    . 'be a Volume Component/STask object when sync_type is '
                    . "provided as 'sync' (this is because SSM is only possible "
                    . 'for Volumes), but instead it was provided as a Qtree' );
        }
    } elsif ($isa_vol) {
        $snapmirror_obj = $pkg->_initialize_7m_vsm( %opts, %opts_for_cleanup );
    } else {
        $snapmirror_obj = $pkg->_initialize_7m_qsm( %opts, %opts_for_cleanup );
    }

    $Log->exit();
    return $snapmirror_obj;
} ## end sub _initialize_7M

# Handle the steps for initializing synchronous snapmirror for 7M
sub _initialize_7m_ssm {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;

    my %opts_for_cleanup;
    $pkg_or_obj->_move_common_cleanup_opts(
        source => \%opts,
        target => \%opts_for_cleanup,
    );

    # Delete the task specific options before they could be passed
    # to the corresponding component method.
    my $wait             = delete $opts{nacltask_wait};
    my $polling_interval = delete( $opts{'polling_interval'} );
    my $destination_obj  = $pkg_or_obj->get_destination_volume(%opts);

    my $source_path      = $opts{'source-path'};
    my $destination_path = $opts{'destination-path'};
    my $ci               = $opts{command_interface};

    # Get the common component params: apiset_must, apiset_should,
    # method_timeout
    my %common_options = ();
    $pkg_or_obj->_copy_common_component_params(
        source => \%opts,
        target => \%common_options
    );

    $pkg_or_obj->_restrict_volume(
        destination_obj => $destination_obj,
        %common_options
    );

    # Update the /etc/snapmirror.conf file for source and destination
    # registry key. This step will always be performed as SSM
    # should be created ONLY after setting these entries.
    $Log->trace("Adding entries to registry");

    # Registry can only be updated using CLI interface. So we are
    # ignoring user prefernces in common component params here.
    $pkg_or_obj->_update_registry_ssm(
        command_interface  => $ci,
        'source-path'      => $source_path,
        'destination-path' => $destination_path
    );

    my $snapmirror_obj = $pkg_or_obj->_call_initialize_and_handle_errors( %opts, %opts_for_cleanup );
    $snapmirror_obj->sync_type('sync');

    # Now wait for initialize to complete if a blocking behavior is requested
    if ($wait) {
        $common_options{polling_interval} = $polling_interval
            if ( defined $polling_interval );
        $snapmirror_obj->wait_for_initialization(%common_options);
    }
    $Log->exit();
    return $snapmirror_obj;
} ## end sub _initialize_7m_ssm

sub _initialize_7m_vsm {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;

    my %opts_for_cleanup;
    $pkg_or_obj->_move_common_cleanup_opts(
        source => \%opts,
        target => \%opts_for_cleanup,
    );

    # Delete the task specific options before they could be passed
    # to the corresponding component method.
    my $wait             = delete $opts{nacltask_wait};
    my $polling_interval = delete( $opts{'polling_interval'} );
    my $destination_obj  = $pkg_or_obj->get_destination_volume(%opts);

    # Get the common component params: apiset_must, apiset_should,
    # method_timeout
    my %common_options = ();
    $pkg_or_obj->_copy_common_component_params(
        source => \%opts,
        target => \%common_options
    );

    $pkg_or_obj->_restrict_volume(
        destination_obj => $destination_obj,
        %common_options
    );

    my $snapmirror_obj = $pkg_or_obj->_call_initialize_and_handle_errors( %opts, %opts_for_cleanup );
    $snapmirror_obj->sync_type('async');

    # Now wait for initialize to complete if a blocking behavior is requested
    if ($wait) {
        $common_options{polling_interval} = $polling_interval
            if ( defined $polling_interval );
        $snapmirror_obj->wait_for_initialization(%common_options);
    }
    $Log->exit();
    return $snapmirror_obj;
} ## end sub _initialize_7m_vsm

sub _initialize_7m_qsm {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;

    my %opts_for_cleanup;
    $pkg_or_obj->_move_common_cleanup_opts(
        source => \%opts,
        target => \%opts_for_cleanup,
    );

    # Delete the task specific options before they could be passed
    # to the corresponding component method.
    my $wait             = delete $opts{nacltask_wait};
    my $polling_interval = delete( $opts{'polling_interval'} );

    my $snapmirror_obj = $pkg_or_obj->_call_initialize_and_handle_errors( %opts, %opts_for_cleanup );
    $snapmirror_obj->sync_type('async');

    # Now wait for initialize to complete if a blocking behavior is requested
    if ($wait) {

        # Get the common component params: apiset_must, apiset_should,
        # method_timeout
        my %common_options = ();
        $pkg_or_obj->_copy_common_component_params(
            source => \%opts,
            target => \%common_options
        );
        $common_options{polling_interval} = $polling_interval
            if ( defined $polling_interval );
        $snapmirror_obj->wait_for_initialization(%common_options);
    }
    $Log->exit();
    return $snapmirror_obj;
} ## end sub _initialize_7m_qsm

# Code to restrict a destination volume and ignore the error if it's already
# restricted. Moved to this routine since it's necessary for both SSM and VSM
# Requires a "destination_obj" parameter (which is a NACL::(C|STask)::Volume
# object to be passed to it.
sub _restrict_volume {
    my $pkg  = shift;
    my %opts = validate_with(
        params => \@_,
        spec   => {
            'destination_obj' => { type => OBJECT, isa => 'NACL::C::Volume' },
            %{ $pkg->_common_validate_spec_without_ci() }
        },
    );

    my $destination_obj = delete $opts{'destination_obj'};

    # In a normal usecase scenario, the destination volume would not be in a
    # restricted state and so we are going to perform this step always and
    # catch the exception if its already restricted (least common usecase
    # scenario)
    try {
        $Log->trace("Restricting the destination volume");
        $destination_obj->restrict(%opts);
    }
    catch NACL::C::Exceptions::Volume::AlreadyRestricted with {
        my $e = shift;
        if ( !$e->is_in_snapmirror_relationship() ) {

            # volume might be in snapmirror relationship and we would
            # have reached this point because if_exists => die
            $Log->exit();
            $e->throw();
        }
    };

    $Log->exit();
} ## end sub _restrict_volume

# 7M only helper to call initialize and handle any errors that might occur
# and retry.
sub _call_initialize_and_handle_errors {
    my ( $pkg, %opts ) = @_;
    my $ci = $opts{command_interface};

    my ( %common_options, %opts_for_cleanup );
    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%common_options
    );

    $pkg->_move_common_cleanup_opts(
        source => \%opts,
        target => \%opts_for_cleanup,
    );

    # Initialize the snapmirror.
    my $snapmirror_obj;
    my $option_not_set_not_licensed;
INIT_7M: {
        use warnings qw(exiting);
        try {
            $snapmirror_obj = $pkg->SUPER::create(%opts);
        }
        catch NACL::C::Exceptions::Snapmirror::SnapmirrorIsTurnedOff with {
            $Log->trace('Snapmirror is off, switch it on');
            NACL::C::Options->option(
                command_interface => $ci,
                'option-name'     => 'snapmirror.enable',
                'option-value'    => 'on',
                %common_options
            );
            no warnings qw(exiting);
            redo INIT_7M;
        }
        catch NACL::C::Exceptions::Snapmirror::CannotConnectToSourceFiler with {

            $Log->trace( 'Cannot connect to source filer either because '
                    . 'snapmirror is not enabled or because an /etc/hosts '
                    . 'entry is missing for the source' );

            # Rather than check which of the two need to be rectified,
            # we will just blindly enable snapmirror and add an entry
            # to /etc/hosts. Enabled an option if already enabled does not
            # result in failure, and adding an already existing entry
            # to /etc/hosts will not cause any failure because the first
            # such entry in /etc/hosts is read.
            # Checking first will in the best case be 3 commands (2 checks
            # and 1 command to rectify), and 4 in the worst case (2 checks
            # and 2 commands to rectify). By blindly rectifying both, we're
            # only running two commands, with no adverse effect.

            my $src_node_name = _get_node_name_from_path( path => $opts{'source-path'} );
            $src_node_name ||= $opts{command_interface}->node();

            # We assume there's a tharnhost file for the source node
            my $src_node_obj = NACL::C::Node->new( node => $src_node_name );
            my $src_ip = $src_node_obj->hostrec()->default_ip();

            $Log->trace( 'Adding entry for the source into the /etc/hosts ' . 'file of the destination node' );
            my $apiset = $ci->get_7m_or_nodescope_apiset();
            $apiset->wrfile(
                append  => 1,
                content => "$src_ip $src_node_name",
                file    => '/etc/hosts'
            );

            $Log->trace('Enabling snapmirror on the source node');
            try {
                NACL::C::Options->option(
                    command_interface => $src_node_obj,
                    'option-name'     => 'snapmirror.enable',
                    'option-value'    => 'on',
                    %common_options
                );
            }
            catch NACL::C::Exceptions::Options::SnapmirrorNotLicensed with {
                $_[0]->license_feature(%common_options);
            };
            no warnings qw(exiting);
            redo INIT_7M;
        }
        catch NACL::C::Exceptions::Snapmirror::ServiceNotEnabledOnSource with {
            $Log->trace('Need to license snapmirror on the source node');
            my $src_node_name = _get_node_name_from_path( path => $opts{'source-path'} );

            # We assume there's a tharnhost file for the source node
            my $src_node_obj = NACL::C::Node->new( node => $src_node_name );
            NACL::STask::SystemLicense->add(
                command_interface => $src_node_obj,
                feature           => 'snapmirror',
                %common_options,
                %opts_for_cleanup
            );
            no warnings qw(exiting);
            redo INIT_7M;
        }
        catch NACL::C::Exceptions::Snapmirror::TransferFromSourceNotPossible with {
            $Log->trace('snapmirror.access needs to be set to all');
            my $src_node_name = _get_node_name_from_path( path => $opts{'source-path'} );

            # We assume there's a tharnhost file for the source node
            my $src_node_obj = NACL::C::Node->new( node => $src_node_name );
            try {
                NACL::C::Options->option(
                    command_interface => $src_node_obj,
                    'option-name'     => 'snapmirror.access',
                    'option-value'    => 'all',
                    %common_options
                );
            }
            catch NACL::C::Exceptions::Options::SnapmirrorNotLicensed with {
                $_[0]->license_feature( %common_options, feature => "snapmirror" );
            };
            no warnings qw(exiting);
            redo INIT_7M;
        }
        catch NACL::APISet::Exceptions::ResponseException with {

            # Catching all ResponseException here because when 7M ZAPI
            # support is added to the underlying component, we would have
            # to catch ResponseException and then check for the error text
            # or error number.
            my $e    = shift;
            my $text = $e->text();

            if (( $e->isa( 'NACL::APISet::Exceptions::FeatureNotLicensedException' ) )
                || (   ( $e->isa( 'NACL::APISet::Exceptions::CommandFailedException' ) )
                    && ( $text =~ /No such command exist/ ) )
                || ( $text =~ /requires license for snapmirror/i )
                )
            {

                # snapmirror is not licensed in the destination
                # INFO: "snapmirror_sync" is an included feature of "snapmirror"
                $Log->trace( "Licensing snapmirror in " . $ci->name() );
                NACL::STask::SystemLicense->add(
                    command_interface => $ci,
                    feature           => 'snapmirror',
                    %common_options,
                    %opts_for_cleanup
                );
            }
            no warnings qw(exiting);
            redo INIT_7M;
        };
    }

    $Log->exit();
    return $snapmirror_obj;
} ## end sub _call_initialize_and_handle_errors

# 7M only - given a path return the node name
# Path is of the form <node-name>:<vol_or_qtree_path>
sub _get_node_name_from_path {
    $Log->enter();
    my %opts = validate( @_, { path => { type => SCALAR } } );
    my $path = $opts{path};
    $path =~ /(.*):.*/;
    my $node_name = $1;
    $Log->exit();
    return $node_name;
} ## end sub _get_node_name_from_path

# Update registry for Synchronous Snapmirror (7Mode)
sub _update_registry_ssm {
    $Log->enter();
    my ( $self, %args ) = @_;
    my $destination_ci   = $args{command_interface};
    my $destination_path = $args{'destination-path'};
    my $source_path      = $args{'source-path'};
    my $dst_Api_Set_Obj  = $destination_ci->apiset(
        category  => 'Node',
        interface => 'CLI',
        set       => '7Mode'
    );

    my $reg_key = "options.files.snapmirror.conf.$destination_path.dest";
    my $value   = '"' . $destination_path . '"';
    $self->_registry_set( $dst_Api_Set_Obj, $reg_key, $value );

    $reg_key = "options.files.snapmirror.conf.$destination_path.source";
    $value   = '"' . $source_path . '"';
    $self->_registry_set( $dst_Api_Set_Obj, $reg_key, $value );

    $reg_key = "options.files.snapmirror.conf.$destination_path.args";
    $value   = '"' . '-' . '"';
    $self->_registry_set( $dst_Api_Set_Obj, $reg_key, $value );

    $reg_key = "options.files.snapmirror.conf.$destination_path.sched";
    $value   = '"sync"';
    $self->_registry_set( $dst_Api_Set_Obj, $reg_key, $value );

    if ( $Log->may_debug() ) {
        my $raw_output = $dst_Api_Set_Obj->execute_raw_command( command => 'rdfile /etc/snapmirror.conf' );
        $Log->debug("The contents of /etc/snapmirror.conf is:\n $raw_output");
    }

    $Log->exit();
} ## end sub _update_registry_ssm

sub _install_DP_license {
    $Log->enter();
    my $self = shift;
    $Log->warn("_install_DP_license is deprecated. Kindly use - _install_license");
    $Log->exit();
    return $self->_install_license(@_);
} ## end sub _install_DP_license

# Install the Snapmirror DP or SnapVault license.
# Accepts either a command_interface or a volume object or a path input
# like clustername://vservername/volumename
# and of course the common component params.
sub _install_license {
    $Log->enter();
    my ( $self, %opts ) = @_;
    my $vol_obj        = delete $opts{volume_obj};
    my $path           = delete $opts{path};
    my $destination_ci = delete $opts{destination_ci};
    my $XDP_type       = delete $opts{'is_xdp'};
    my $vserver_obj    = delete $opts{'vserver_obj'}, my $intra;

    my %opts_for_cleanup;
    $self->_move_common_cleanup_opts(
        source => \%opts,
        target => \%opts_for_cleanup,
    );

    if ( defined $opts{command_interface} ) {

        # do nothing
    } elsif ( defined $vserver_obj ) {
        $opts{command_interface} = $vserver_obj->command_interface();
    } elsif ( defined $vol_obj ) {
        $opts{command_interface} = $vol_obj->command_interface();
    } elsif ( defined $path ) {

        # This part of the code is executed in case of licensing source cluster
        my $cluster_name = $self->_get_cluster_name(
            command_interface => $destination_ci,
            path              => $path,
            is_source         => 1,
            intra             => \$intra,
        );
        my $flag;

        # $intra will be 1, if the source is in the same cluster as that of
        # destination. In this case, destiantion's ci can be used for source also.
        if ( defined $intra ) {
            if ($intra) {
                $opts{command_interface} = $destination_ci;
                $flag = 1;
            }
        } else {

            # In case of RR style path, intra will be undef. Just check if the
            # local cluster and source cluster name obtained are the same. If yes,
            # use destiantion's ci for source also.
            my $cluster_cs_obj = NACL::CS::ClusterIdentity->fetch(
                command_interface => $destination_ci,
                requested_fields  => ['name'],
            );
            my $d_cluster_name = $cluster_cs_obj->name();
            if ( $d_cluster_name eq $cluster_name ) {
                $opts{command_interface} = $destination_ci;
                $flag = 1;
            }
        }
        if ( !$flag ) {
            $opts{command_interface} = NACL::C::Cluster->new( name => $cluster_name );
            unless ( $opts{command_interface}->can_use_as_command_interface ) {
                $Log->exit();
                NATE::BaseException->throw( "A tharnhost file in the name of cluster"
                        . " '$cluster_name' doesn't exist. Either pass source_ci"
                        . " parameter or create tharnhost file with the name of the cluster" );
            }
        }
    } else {
        NATE::Exceptions::Argument->throw( "Internal error: Either of path," . " volume_obj or command_interface should be specified" );
    }

    if ($XDP_type) {
        NACL::STask::SystemLicense->add(
            %opts,
            feature => 'SnapVault',
            %opts_for_cleanup
        );
    } else {
        NACL::STask::SystemLicense->add(
            %opts,
            feature => 'snapmirror_dp',
            %opts_for_cleanup
        );
    }

    $Log->exit();
} ## end sub _install_license

# To set the given key value pair in the registry
sub _registry_set {
    $Log->enter();
    my $self = shift;
    my ( $dst_Api_Set_Obj, $reg_key, $value ) = @_;

    try {
        my $response = $dst_Api_Set_Obj->registry_set(
            key   => $reg_key,
            value => $value
        );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $exception_object = shift;
        my $resp             = $exception_object->get_response_object();
        $Log->debug( "Raw output was:\n" . $resp->get_raw_output() );
    };

    $Log->exit();
} ## end sub _registry_set

=head2 wait_for_initialization

    $snapmirror->wait_for_initialization(
      'method-timeout'            =>  $timeout,
      'polling_interval'          =>  $polling_interval
    );


    NACL::STask::Snapmirror->wait_for_initialization(
      'method-timeout'   =>  $timeout,
      command_interface  =>  $command_interface,
      'destination-path' =>  $destination_path,
      job                =>  $job_component,
      'polling_interval' =>  $polling_interval
    );

Wait for the Snapmirror initialization to complete. If a job
for initializing this Snapmirror is known (because the mode
is C-mode and NACL::STask::Snapmirror->create or
NACL::STask::Snapmirror->initialize was used to
initialize this object, or because it was passed in using
the "job" option) then this waits for that job.

Otherwise, it polls until the "state" of the Snapmirror is
"Snapmirrored" and the "status" is "Idle"  or "In-sync" (for SSM).
If the state and status of the Snapmirror becomes something
unexpected (it will not fail for valid intermediate state and status
combination, if any) then a NACL::Exceptions::UnexpectedState
exception is thrown.
With CMode, health of the relationship is verified at the end of transfer and an
exception is thrown if the health is false.

Supports CMode CLI/ZAPI and 7Mode CLI.

=over

=item Options

=over

=item C<< "command_interface=>$command_interface" >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< "destination-path" => $destination_path >>

(Required for class method, Not Applicable for instance method)
The destination path of the snapmirror relationship.

=item C<< nacltask_verify_health => 0|1 >>

(Optional, defaults to '1', Only for CMode)
If set to '1', verifies the health of the snapmirror.

=item C<< "last-transfer-type=>last transfer option" >>

(Optional, defaults to "initialize")

If we have to verify the health only after having some particular operation.
Ex: if we have to perform verify_health after initialize operation, pass this option as follows:
'last-transfer-type' => initialize.
verify_health() should do nothing if the last-transfer-type from the CS call is NOT initialize.

=item C<< "method-timeout=>$timeout" >>

(Optional) How long in seconds to wait for the job to finish
or for the state to become "snapmirrored". Defaults to 1200 (SNAPMIRROR_TIMEOUT).
method-timout=0, will wait indefinitely.

=item C<< "polling_interval =>  $polling_value" >>

(Optional) This is the interval (in seconds) after which the value of
the attributes will be polled. If not specified it defaults to 10 seconds.

=item C<< "job=>$job_component" >>

(Optional for class method, Not Applicable for instance method) A
NACL::C::Job object representing the C-mode Job used to initialize
this Snapmirror.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub wait_for_initialization {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job                    => { type => OBJECT | UNDEF, optional => 1 },
            polling_interval       => { type => SCALAR,         default  => POLL_DELTA },
            nacltask_verify_health => { type => SCALAR,         default  => 1 },
            'last-transfer-type'   => { type => SCALAR | UNDEF, default  => 'initialize' },
        },
        allow_extra => 1,
    );

    my $verify_health      = delete $opts{nacltask_verify_health};
    my $last_transfer_type = delete $opts{'last-transfer-type'};

    #Check if 'last-transfer-type' already has value 'initialize' w.r.t burt986660
    my $state_obj = $pkg_or_obj->_get_cs_objects( %opts, 'requested-field-name' => 'last-transfer-type' );
    my %check_hash = ();
    if ( defined $state_obj ) {
        if ( $state_obj->last_transfer_type() !~ /initialize/ ) {
            $check_hash{'last-transfer-type'} = 'initialize';
        }
    }

    my @status_till_value = (qr/Idle/i);
    if ( $opts{command_interface}->mode() eq 'CMode' ) {
        my $job = $pkg_or_obj->last_job() if ( ref $pkg_or_obj );

        # job_component applicable only for CMode
        $opts{job_component} = delete $opts{job} || $job;

        # Because this can be called statically, pass %opts in to the call to
        # is_granular. If the $pkg_or_obj is just the package name, then the
        # paths in %opts will be tested.
        if ( $pkg_or_obj->is_granular(%opts) ) {
            push @status_till_value, qr/^InSync$/i;
        }
    } else {

        # A final value of "In-sync" for status seems to be applicable
        # only for 7Mode
        push @status_till_value, qr/In\-sync/i;
    }

    # Always make class method calls so that we don't have to bother
    # about skipping primary keys for instance method call
    my $pkg = $pkg_or_obj->get_package_name();
    my %common_opts;

    $pkg->_hash_copy(
        source => \%opts,
        target => \%common_opts,
        copy =>
            [ qw(apiset_must apiset_should method-timeout job_component polling_interval command_interface source-path destination-path) ]
    );

    $pkg->wait_for_completion(
        %common_opts,
        attributes => [
            {   attribute_to_check => 'state',
                till_value         => qr/Snapmirrored/i
            },
            {   attribute_to_check => 'status',
                till_value         => \@status_till_value,
            },
        ],
        check_error_in_wait_for_initialization => \%check_hash,
        show_all_fields                        => 1,
    );

    if ( $verify_health && $opts{command_interface}->is_cmode() ) {
        $pkg->verify_health( %common_opts, 'last-transfer-type' => $last_transfer_type );
    }
    $Log->exit();
} ## end sub wait_for_initialization

sub _get_cs_objects {
    my ( $pkg_or_obj, %opts ) = @_;
    my $filter = $pkg_or_obj->get_primary_keys_options(%opts);
    my $CS_pkg = $pkg_or_obj->get_CS_package_name();
    my %common_options;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_options
    );
    my %fetch_opts = %common_options;
    $fetch_opts{filter}           = $filter;
    $fetch_opts{requested_fields} = [ $opts{'requested-field-name'} ];
    my $state_obj = $CS_pkg->fetch( %fetch_opts, allow_empty => 1 );
    return $state_obj;
}

sub get_SN_style_path {
    $Log->enter();

    my $pkg_or_obj = shift;
    my %opts       = NACL::C::Component->_common_validate_with(
        params          => \@_,
        additional_spec => {
            'path'              => { type => SCALAR, optional => 0 },
            'command_interface' => {
                isa      => 'NACL::C::CommandInterface::ONTAP',
                type     => OBJECT,
                optional => 1,
            },
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );

    my $path = delete( $opts{path} );
    my %result;

    if ( $path =~ /(.+):\/\/(.+)\/(.+)/ ) {
        $result{SN_path} = $2 . ':' . $3;
        $result{cluster} = $1;
    } elsif ( $path =~ /(.+):(.+)/ ) {
        $result{SN_path} = $path;
        if (  !( $opts{command_interface} && $opts{command_interface}->isa('NACL::C::Vserver') )
            && ( ref($pkg_or_obj) && $pkg_or_obj->command_interface()->isa('NACL::C::Vserver') ) )
        {
            $result{cluster} = $pkg_or_obj->_get_cluster_name( %opts, path => $path, );
        }
    } else {
        $Log->exit();
        NATE::Exceptions::Argument->throw( "Invalid path format passed. Expected format: vserver_name:volume_name"
                . " or cluster_name://vserver_name/volume_name. But passed $path " );
    }

    $Log->exit();
    return %result;
} ## end sub get_SN_style_path

=head2 release

    NACL::STask::Snapmirror->release(
        command_interface    => $ci,
        'source-path'        => $source_path,
        'destination-path'   => $destination_path,
        ...
    );

                             (or)

    $snapmirror->release( source_ci => $ci );

(Class or instance method)

This method does snapmirror list-destinations and release on the source node.
Supports CMode CLI/ZAPI and 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>
This represents the source end point (source cluster /source node / source
vserver) of the snapmirror relationship.

=item C<< source_ci => $source_ci >>

(Optional)
See documentation at the beginning of this page
This method tries to create source_ci with the cluster name obtained from
source-path in case of RR style path (if both source_ci and src vol obj are not
passed). If tharnhost file with the name of source cluster is not present, NATE::BaseException is thrown.
This command interface would be used to perform snapmirror release on the source.

=item C<< 'source-path' => $source_path >>

(Optional for class method, Not Applicable for instance method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< 'destination-path' => $path_specification >>

(Required for class method, Not Applicable for instance method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< 'source-cluster' => $cluster_name >>

(Optional)
Name of cluster where the source volume resides. This input applies only for
CMode.

=item C<< 'destination-cluster' => $cluster_name >>

(Optional)
Name of cluster where the destination volume resides. This input applies only
for CMode.

=item C<< "nacltask_wait =>  $boolean" >>

(Optional) A number 0 or 1 that determines if the operation has to happen in the
background.  1 means to return after release has completed and 0 means to
return immediately. The default is 1.

=item C<< nacltask_verify => 0 | 1 >>

(Optional, defaults to 0)

A boolean value which specifies whether to verify whether the snapmirror
relationship does not exist after the release has completed. A
L<NACL::Exceptions::VerifyFailure|lib-NACL-Exceptions-VerifyFailure-pm>
exception is thrown if the entry still exists.

=item C<< nacltask_retry_tries_count => $tries_count >>

(Optional, defaults to 30)
The number of times to retry the release command if it fails because
the snapshot is temporarily busy or stale catch entry.

Note that it is possible to make it not retry the delete command by setting
the value of this argument to 1.

=item C<< nacltask_retry_sleep_time => $sleep_time >>

(Optional, defaults to 10 seconds)
The time to sleep between retries of the release command.

=item C<< nacltask_retry_num_times_code_ran => \$variable >>

(Optional)
A reference to a scalar variable which gets filled in with the number
of times the release command was invoked.

=back

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when snapmirror list-destinations shows the relationship entry even after snapmirror release.

=item C<NATE::BaseException>

This type of exception is thrown when filer mode is not supported or  when a tharnhost file in the name of cluster doesn't exist.

=item C<NACL::C::Exceptions::Snapmirror::DeleteIsPending>

This exception is thrown when an attempt is made to perform snapmirror operation when delete is in progress.

=back

=back

=cut

sub release {

    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {

            # component options which are overloaded
            'source-path'      => { type => OBJECT | SCALAR, optional => 1 },
            'destination-path' => { type => OBJECT | SCALAR, optional => 1 },

            # options which would be consumed by stask method itself apart
            # from passing the information to components.
            'source-cluster'                    => { type => SCALAR,    optional => 1 },
            'destination-cluster'               => { type => SCALAR,    optional => 1 },
            'nacltask_verify'                   => { type => BOOLEAN,   default  => 0 },
            'nacltask_wait'                     => { type => SCALAR,    default  => 1 },
            'nacltask_retry_tries_count'        => { type => SCALAR,    default  => 30 },
            'nacltask_retry_sleep_time'         => { type => SCALAR,    default  => 10 },
            'nacltask_retry_num_times_code_ran' => { type => SCALARREF, optional => 1 },
        },
        allow_extra => 1,
    );

    # After the call to _common_validate_with(), %opts will have common params
    # like method-timeout, apiset_must, apiset_should, command_interface,
    # primary keys, polling_interval, stask specific params, other options
    # taken by NACL::C::Snapmirror->release()
    my $ci = $opts{command_interface};

    $opts{'method-timeout'} ||= SNAPMIRROR_TIMEOUT;

    # The dependencies which need to be handled are different between 7M and CM
    # Hence we will have two backends for 7M and CM separately.
    if ( $ci->mode() eq '7Mode' ) {

        # Invoke 7M backend
        $pkg_or_obj->SUPER::release(%opts);

    } elsif ( $ci->mode() eq 'CMode' ) {

        # Invoke CM backend
        $pkg_or_obj->_release_CM(%opts);
    } else {

        # Just to avoid open if-elsif
        $Log->exit();
        NATE::BaseException->throw( "Filer mode '" . $ci->mode() . "' not supported." );
    }
    $Log->exit();
} ## end sub release

sub _release_CM {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;
    my ( $dst_path, $src_path );
    my $verify           = delete $opts{nacltask_verify};
    my $wait_for_release = delete $opts{nacltask_wait};
    my $flag;

    my %retry_opts;
    $pkg_or_obj->_move_retry_params(
        source => \%opts,
        target => \%retry_opts
    );

    # Handle Job ID
    my $job_component = delete $opts{job_component};
    {

        # We want to know the job ID of the job performed by
        # SUPER::release, whether or not our caller does.
        my $dummy;
        $job_component ||= \$dummy;
    }
    my $pkg = ref $pkg_or_obj || $pkg_or_obj;

    my %common_opts;
    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%common_opts
    );

    # Get source path if it is not present
    if ( !$opts{'source-path'} ) {
        my $sm = $pkg->find(
            %common_opts,
            command_interface => $opts{command_interface},
            filter            => { 'destination-path' => $opts{'destination-path'} },
            allow_empty       => 1,
        );
        if ($sm) {
            $opts{'src-path'} = $sm->source_path();
        }
    }
    $opts{'src-path'} = $opts{'source-path'} if ( !$opts{'src-path'} );

    if ( ref($pkg_or_obj) && $pkg_or_obj->source_ci() ) {
        $opts{source_ci} ||= $pkg_or_obj->source_ci();
    }
    if ( !( ref $pkg_or_obj ) ) {
        $opts{source_ci} ||= $opts{command_interface};
    }

    if ( !$opts{source_ci} ) {
        my $cluster;

        if ( $opts{'src-path'} ) {
            my $src_flag;
            if ( !$opts{'source-path'} ) {
                $opts{'source-path'} = $opts{'src-path'};
                $src_flag = 1;
            }
            $cluster = $pkg->_check_and_get_intercluster(%opts);
            if ($src_flag) {
                delete $opts{'source-path'};
            }

            # Check here if source and destination belong to different cluster
            if ( $cluster->{intercluster} ) {
                if ( $cluster->{source} ) {
                    $opts{'source_ci'} = NACL::C::Cluster->new( name => $cluster->{source} );
                    unless ( $opts{'source_ci'}->can_use_as_command_interface ) {
                        $Log->exit();
                        NATE::BaseException->throw( "A tharnhost file in the name of"
                                . " cluster '"
                                . $cluster->{source}
                                . "' doesn't exist. Pass source_ci or"
                                . " create a tharnhost file in the name of cluster to"
                                . " execute Snapmirror release on source node" );
                    }
                } else {
                    $Log->exit();
                    NATE::BaseException->throw( "Pass source_ci to" . " execute Snapmirror release on source node/vserver" );
                }

            }
        }
    }

    # Swap command interface of source and destination (list-destination and
    # release requires source ci).
    ( $opts{'command_interface'}, $opts{'source_ci'} ) = ( $opts{'source_ci'}, $opts{'command_interface'} )
        if ( $opts{'source_ci'} );

    my %common_options = ();
    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_options
    );
    my $SMListDestinations = NACL::CS::SnapmirrorListDestinations->fetch(
        filter => {
            'destination-path' => $opts{'destination-path'},
            'source-path'      => $opts{'src-path'}
        },
        allow_empty => 1,
        %common_options,
    );

    # If none of source_ci and list-destinations are defined, then log
    # message to check if it is an intercluster relationship.
    if ( !( $opts{'source_ci'} || $SMListDestinations ) && $flag ) {
        $Log->warn( "Check if the source vol resides on the same cluster."
                . "\nIf intercluster relationship, pass source_ci of source"
                . " node since sm list-destinations and snapmirror release "
                . "SHOULD be executed on source cluster." );
    }

    # Perform snapmirror release only if 'list-destinations' returns the entry.
    if ($SMListDestinations) {

        # Perform release
        # Always make class method calls so that we don't have to bother
        # about skipping primary keys for instance method call

        my %all_retry_opts = (
            code => sub { $pkg->SUPER::release( %opts, job_component => $job_component, ) },
            exceptions => [ 'NACL::C::Exceptions::Snapmirror::SnapshotInUse', 'NACL::C::Exceptions::Snapmirror::DeleteIsPending' ],
            %retry_opts
        );

        # As per the burt#856182
        try {
            nacl_method_retry(%all_retry_opts);
        }
        catch NACL::C::Exceptions::Snapmirror::StaleCache with {
            try {
                $pkg->SUPER::release(
                    %opts,
                    job_component            => $job_component,
                    'relationship-info-only' => 'true'
                );
            }
            otherwise {
                my $if_exists = NACL::CS::SnapmirrorListDestinations->fetch(
                    filter => {
                        'destination-path' => $opts{'destination-path'},
                        'source-path'      => $opts{'src-path'}
                    },
                    allow_empty => 1,
                    %common_options,
                );
                if ($if_exists) {
                    push @{ $all_retry_opts{exceptions} }, 'NACL::C::Exceptions::Snapmirror::StaleCache';
                    nacl_method_retry(%all_retry_opts);
                }
            };
        };

        $pkg_or_obj->last_job( ${$job_component} ) if ( ref $pkg_or_obj );

        if ($wait_for_release) {
            my %job_args;
            if ( defined $$job_component ) {
                $job_args{job_component} = $$job_component;
            }
            $pkg->wait_for_release( %common_options, %job_args, 'destination-path' => $opts{'destination-path'} );
        }

        if ($verify) {
            my $flag;

            # Add sleep to make sure that the entry goes after release.
            Tharn::sleep(2);
        VERIFY: {
                $SMListDestinations = NACL::CS::SnapmirrorListDestinations->fetch(
                    filter      => { 'destination-path' => $opts{'destination-path'}, },
                    allow_empty => 1,
                    %common_options,
                );
                if ( $SMListDestinations && $flag ) {

                    # Throw exception if the relationship persists even after
                    # 5 seconds of sleep.
                    $Log->exit();
                    NACL::Exceptions::VerifyFailure->throw( 'Snapmirror ' . 'list-destinations still shows the relationship entry' );
                } elsif ( $SMListDestinations && !$flag ) {

                    # Check after sometime if the relationship is still
                    # present in the filer.
                    $flag = 1;
                    Tharn::sleep(3);
                    redo VERIFY;
                }
            }
        }
    }
    $Log->exit();
} ## end sub _release_CM

=head2 wait_for_release

    # Package call
    NACL::STask::Snapmirror->wait_for_release(
        command_interface  => $src_ci,
        'destination-path' => $dst_path,
        job_component      => $job_obj,
    );

    # Instance call
    $snapmirror_obj->wait_for_release();

Wait for the snapmirror operation to be released. Note that this operation
is driven from the source, so the command_interface of the source should
be provided.

If the call does not complete within the timeout value specified, then
a C<NACL::Exceptions::Timeout> exception is thrown.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
This should be the command_interface of the source.
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< "destination-path" =>$destination_path >>

(Required for class method, Not Applicable for instance method)
The destination path of the snapmirror relationship.

=item C<< "method-timeout" =>$timeout >>

(Optional) How long in seconds to wait. Defaults to 1200 seconds.

=item C<< job_component => $job_component >>

(Optional for class method, Not Applicable for instance method) A
NACL::C::Job object representing the C-mode Job used to initialize
this Snapmirror operation.

=back

=back

=cut

sub wait_for_release {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => { job_component => { type => OBJECT | UNDEF, optional => 1 }, },
    );
    my $job_component = delete $opts{job_component} // $pkg_or_obj->last_job();
    my %common_params_with_ci;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_params_with_ci
    );

    # Needs to be invoked on source_ci
    $common_params_with_ci{command_interface} = $opts{source_ci}
        || $opts{command_interface};

    if ( defined $job_component ) {
        my %common_params_without_ci = %common_params_with_ci;
        delete $common_params_without_ci{command_interface};
        $job_component->wait_on_job(%common_params_without_ci);
    } else {

        # ZAPI does not return a job-id, so need to poll manually
        my $scaled_timeout = $opts{command_interface}->scale_timeout( 'method-timeout' => $opts{'method-timeout'} );
        my $end_time = timeout2time($scaled_timeout);
        while (1) {
            my $destination_cs = NACL::CS::SnapmirrorListDestinations->fetch(
                %common_params_with_ci,
                filter      => { 'destination-path' => $opts{'destination-path'}, },
                allow_empty => 1
            );
            if ($destination_cs) {
                if ( time() < $end_time ) {

                    # Doesn't really take very long for the release
                    # to complete, so set a low snooze time
                    Tharn::snooze(2);
                } else {
                    $Log->exit();
                    NACL::Exceptions::Timeout->throw(
                        'The ' . 'snapmirror release did not complete ' . "after waiting for $scaled_timeout seconds" );
                }
            } else {
                last;
            }
        }
    }
    $Log->exit();
}

=head2 purge

    NACL::STask::Snapmirror->purge(
        command_interface    => $ci,
        'source-path'        => $source_path,
        'destination-path'   => $destination_path,
        ...
    );

                             (or)

    $snapmirror->purge(
        nacltask_release => 1,
        source_ci        => $source_node,       #Pass source_ci for intercluster relationship
        ...
    );

(Class or instance method)

This method does snapmirror release on the source (for CMode) and deletes
snapmirror relationship from the destination, as well as aborts any existing
relationship that may be preventing from deleting the relationship entry. For
SSM, the relevant entry in /etc/snapmirror.conf is removed as well.

This call waits till purge is complete. Default wait is 300 seconds. This can
be overridden by passing 'method-timeout' parameter. No exception is thrown
when timeout expires.  Use nacltask_verify parameter for verification.

Purge() task handles only basic release() scenario. If user is interested in
calling release() with specific options, it is recommended to call the release()
task prior/after purge( nacltask_release => 0).

Supports CMode CLI/ZAPI and 7Mode CLI.

=over

=item Options

=over

=item C<< source_ci => $source_ci >>

(Required in inter-cluster or inter-vserver scenarios)
See documentation at the beginning of this page.
This command interface would be used to perform snapmirror release on the source.

This method tries to create source_ci with the cluster name obtained from
source-path (if both source_ci and src vol obj are not passed). If tharnhost file
with the name of source cluster is not present, NATE::BaseException is thrown.

=item C<< 'source-path' => $source_path >>

(Optional for class method, Not Applicable for instance method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< 'destination-path' => $path_specification >>

(Required for class method, Not Applicable for instance method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< 'source-cluster' => $cluster_name >>

(Optional)
Name of cluster where the source volume resides. This input applies only for
CMode.

=item C<< 'destination-cluster' => $cluster_name >>

(Optional)
Name of cluster where the destination volume resides. This input applies only
for CMode.

=item C<< nacltask_if_purged => "die" | "pass" >>

(Optional, defaults to "die")

If "die" (default), exception is thrown when snapmirror relationship to be
deleted does not exist.

If 'pass', exits quietly when snapmirror relationship to be deleted does not
exist.

Note that currently no error is thrown by ontap if a deleted relationship is
deleted again.

=item C<< nacltask_release => "0 | 1" >>

(Optional, defaults to "1")

If "1" (default), snapmirror release will be done on the source side as part of
purge method call. Pass '0' in case it is not required to release the relationship.
It is mandatory to pass source_ci when nacltask_release is 1 and if it is an
intercluster relationship. In case of snapmirror type TDP, nacltask_release
should be 0.

=item C<< nacltask_verify => 0 | 1 >>

(Optional, defaults to 0)

A boolean value which specifies whether to verify whether the snapmirror
relationship does not exist after the purge has completed. A
L<NACL::Exceptions::VerifyFailure|lib-NACL-Exceptions-VerifyFailure-pm>
exception is thrown if the entry still exists.

In 7Mode it takes some time for the entry to get removed. The verification
here is to poll for 2 minutes checking whether the relationship entry still
exists. If after two minutes the entry still exists, then it fails. The
assumption here is that if it has taken more than 2 minutes for the entry
to get removed, then something must have gone wrong.

(This means that in 7Mode, if nacltask_verify is not set to 1 then the method
returns immediately without waiting for removal of the entry)

=item C<< nacltask_retry_tries_count => $tries_count >>

(Optional, defaults to 3)
The number of times to retry the delete command if it fails because
the snapshot is temporarily busy.

Note that it is possible to make it not retry the delete command by setting
the value of this argument to 1.

=item C<< nacltask_retry_sleep_time => $sleep_time >>

(Optional, defaults to 10 seconds)
The time to sleep between retries of the delete command.

=item C<< nacltask_retry_num_times_code_ran => \$variable >>

(Optional)
A reference to a scalar variable which gets filled in with the number
of times the delete command was invoked.

=back

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for the deleted snapmirror relationship.

=item C<NATE::Exceptions::Argument>

This type of exception is thrown when 'sync_type' is provided as an argument to the 'purge' method or may be present in the object.

=item C<NACL::APISet::Exceptions::CommandFailedException>

This type of exception is thrown when there is failure during registry delete.

=back

=back

=cut

sub purge {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {

            # stask specific options
            'nacltask_verify'                   => { type => BOOLEAN,   default  => 0 },
            'nacltask_release'                  => { type => BOOLEAN,   default  => 1 },
            'nacltask_if_purged'                => $pkg_or_obj->_if_action_completed_already_validate_spec(),
            'nacltask_retry_tries_count'        => { type => SCALAR,    default  => 20 },
            'nacltask_retry_sleep_time'         => { type => SCALAR,    default  => 30 },
            'nacltask_retry_num_times_code_ran' => { type => SCALARREF, optional => 1 },
            _optional_scalars(
                qw(sync_type source-cluster destination-cluster
                    source-vserver destination-vserver
                    source-volume destination-volume source-cg
                    destination-cg)
            ),
            'cg-item-mappings' => { type => HASHREF | ARRAYREF, optional => 1 },
        },
        allow_extra => 1
    );

    $opts{destination_ci} = $pkg_or_obj->{destination_ci} if ( ( ref $pkg_or_obj ) && ( defined $pkg_or_obj->{destination_ci} ) );

    # After the call to _common_validate_with(), %opts will have common params
    # like method-timeout, apiset_must, apiset_should, command_interface,
    # primary keys, polling_interval and stask specific params
    my $ci = $opts{command_interface};

    # The dependencies which need to be handled are different between 7M and CM
    # Hence we will have two backends for 7M and CM separately.
    my $pkg = $pkg_or_obj->get_package_name();
    if ( $ci->mode() eq '7Mode' ) {
        $opts{sync_type} ||= $pkg_or_obj->sync_type();
        if ( !defined $opts{sync_type} ) {
            $Log->exit();
            NATE::Exceptions::Argument->throw(
                "'sync_type' must either be " . "provided as an argument to the 'purge' method or must be " . 'present in the object' );
        }
        $pkg->_purge_7M(%opts);
    } elsif ( $ci->mode() eq 'CMode' ) {

        # sync_type not applicable for CMode
        delete $opts{sync_type};
        if ( ref($pkg_or_obj) && $pkg_or_obj->source_ci() ) {
            $opts{source_ci} ||= $pkg_or_obj->source_ci();
        }
        $pkg->_purge_CM(%opts);
    }
    $Log->exit();
} ## end sub purge

# CMode backend for purge()
sub _purge_CM {
    $Log->enter();
    my ( $pkg, %opts ) = @_;
    my $pass = undef;

    my $if_purged = delete $opts{nacltask_if_purged};
    my $verify    = delete $opts{nacltask_verify};
    my $release   = delete $opts{nacltask_release};
    my ( $src_cluster, $dst_cluster, $r_error, $d_error, $src_path );

    my %common_opts;
    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%common_opts
    );
    my %retry_opts;
    $pkg->_move_retry_params(
        source => \%opts,
        target => \%retry_opts
    );

    # The cg-item-mappings option is only applicable to two of the methods
    # that we later call. So remove it from %opts if it exists, and preserve
    # any value in the hash %cg_item_mappings.
    my %cg_item_mappings;
    if ( my $cg_item_mappings = delete $opts{'cg-item-mappings'} ) {

        # This will be used in the calls to delete() and release()
        $cg_item_mappings{'cg-item-mappings'} = $cg_item_mappings;
    }

    # Get source path if it is not present
    if ( !$opts{'source-path'} ) {

        my $sm = $pkg->find(
            %common_opts,
            command_interface => $opts{command_interface},
            filter            => { 'destination-path' => $opts{'destination-path'} },
            allow_empty       => 1,
        );

        if ($sm) {

            # Using a different opts name, in case user does not want to pass
            # source-path during command execution.
            $src_path = $sm->source_path();
        }
    }

    my $apiset;

    try {
        my $is_granular = $pkg->is_granular(%opts);

        # Perform break and then delete.
        nacl_method_retry(
            code => sub {

                # TODO: Fix this when "break" is delivered for SMGR
                #
                # This is a temporary work-around for the fact that "break"
                # is currently a no-op for granular relationships. The break
                # command doesn't return any error, but the state never goes
                # from "Snapmirrored" to "Broken-off".
                #since break is not supported yet for ONTA-AVA scenario hence
                #do not call break if destination_ci is present
                if ( ( !$is_granular ) && ( !defined $opts{destination_ci} ) ) {
                    $pkg->break(
                        %opts,
                        nacltask_if_transferring  => 'abort',
                        nacltask_if_broken        => 'pass',
                        nacltask_if_uninitialized => 'pass',
                        nacltask_if_ls            => 'pass',
                        nacltask_verify           => 1,
                    );
                }
                $pkg->delete( %opts, %cg_item_mappings, apiset_chosen => \$apiset );
            },
            exceptions => [
                qw(NACL::C::Exceptions::Snapmirror::SnapshotInUse
                    NACL::C::Exceptions::Snapmirror::RelationshipInfoBeingUpdated)
            ],
            %retry_opts
        );
    }
    catch NACL::C::Exceptions::Snapmirror::TransferInProgress with {
        my $error    = shift;
        my %pri_opts = %{ $pkg->get_primary_keys_options(%opts) };
        $pkg->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%pri_opts,
        );
        try {
            $pkg->abort( %pri_opts, nacltask_verify => 1 );
        }
        catch NACL::C::Exceptions::Snapmirror::NoTransferToAbort with {
            my $e = shift;

            # Sometimes tranfer completes by the time abort is called. This
            # catch block is to avoid such failure during purge call.
            $Log->trace( 'No transfer in progress, not failing inside ' . "purge" );
        };
        $Log->debug("\n Inside Purge : Calling snapmirror delete after abort");
        $pkg->delete(%pri_opts);
    }
    catch NACL::C::Exceptions::Snapmirror::NotQuiescedOrQuiescing with {
        my $error    = shift;
        my %pri_opts = %{ $pkg->get_primary_keys_options(%opts) };
        $pkg->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%pri_opts,
        );
        $Log->debug( "\n Inside Purge : Quiescing is in progress shall execute delete command in a while" );
        $pkg->wait_for_completion(
            attribute_to_check => 'state',
            till_value         => qr/'quiesced'/i,    #Case insensitive match
            show_all_fields    => 1,
            %pri_opts
        );
        $Log->debug( "\n Inside Purge : Calling snapmirror delete in quiesced state" );
        $pkg->delete(%pri_opts);
    }
    catch NACL::C::Exceptions::Snapmirror::DoesNotExist with {
        my $error = shift;
        $Log->debug("snapmirror relationship does not exist.");
        if ( $if_purged =~ /die/i ) {
            $d_error = $error;
        }
    }
    catch NACL::Exceptions::Timeout with {

        # Ignore timeout in delete.
        # verify delete only if  nacltask_verify parameter is set.
    };

    delete $opts{'force'};
    delete $opts{'foreground'};
    my $src_ci = delete $opts{source_ci};

    if ( $verify && !$d_error ) {
        Tharn::sleep(2) if ( $apiset->{interface} eq 'ZAPI' );

        # At times, snapmirror show soon after delete shows the entry. To avoid
        # failure, a 2 sec sleep is introduced before verification.
        $pkg->_generic_purge_verify(%opts);
    }

    $opts{nacltask_verify} = $verify;
    try {
        if ($release) {
            $retry_opts{nacltask_retry_sleep_time}         = delete $retry_opts{sleep_time};
            $retry_opts{nacltask_retry_tries_count}        = delete $retry_opts{tries_count};
            $retry_opts{nacltask_retry_num_times_code_ran} = delete $retry_opts{num_times_code_ran} if ( $retry_opts{num_times_code_ran} );

            $opts{source_ci} = $src_ci if ($src_ci);
            $opts{'src-path'} = $src_path;
            $pkg->release( %opts, %cg_item_mappings, %retry_opts );
        }
    }
    catch NATE::BaseException with {
        $r_error = shift;
        if ($d_error) {
            $d_error->set_text( $d_error->text . "\nRelease also failed due to : " . $r_error->text );
        } else {
            $d_error = $r_error;
        }
    };

    # Throw exception if delete or release has failed
    if ($d_error) {
        $Log->exit();
        $d_error->throw();
    }
    $Log->exit();
} ## end sub _purge_CM

sub _purge_7M {
    my ( $pkg, %opts ) = @_;

    my $sync_type = delete $opts{sync_type};
    my %task_options;
    $pkg->_move_nacltask_options(
        source => \%opts,
        target => \%task_options
    );
    my $if_purged = $task_options{nacltask_if_purged};

    my $verify = $task_options{nacltask_verify};

    # "break" throws the same error message both when the relationship
    # is broken and even after it's been purged, so if we find the
    # relationship to be broken we check whether the relationship exists
    # or not. If it doesn't exist, we can return early (skip needing to
    # perform all of the following steps: release, purge snapshot, remove
    # snapmirror.conf entry)
    my $need_to_return;

    my %common_opts;
    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%common_opts
    );

    try {
        $pkg->break( %opts, nacltask_if_transferring => 'abort' );
    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $e     = shift;
        my $throw = sub {
            $Log->exit();
            $e->throw();
        };

        if ( $pkg->_is_broken_7M( text => $e->text() ) ) {
            $Log->trace( 'Relationship is broken, check if it is because ' . 'the relationship has already been purged' );
            my $snapmirror = $pkg->find(
                %common_opts,
                command_interface => $opts{command_interface},
                filter            => { 'destination-path' => $opts{'destination-path'} },
                allow_empty       => 1
            );
            if ( !$snapmirror ) {
                if ( $if_purged =~ /pass/i ) {
                    $need_to_return = 1;
                } else {
                    $throw->();
                }
            }
        } else {
            $throw->();
        }
    };

    if ($need_to_return) {
        $Log->exit();
        return;
    }

    # If 'source-path' wasn't provided by the user, then we get it
    # by doing a find() and extracting it from the returned object.
    # We need the source-path argument for release to work
    $opts{'source-path'} ||= $pkg->find(
        %common_opts,
        command_interface => $opts{command_interface},
        filter            => { 'destination-path' => $opts{'destination-path'} }
    )->source_path();

    # "release" needs to be invoked from the source-filer. Further, the
    # source path should not contain the filer name, while the
    # destination path has to contain the filer name.
    my $source_ci;
    my $source_path = $opts{'source-path'};
    if ( $opts{'source-path'} =~ /(.*):(.*)/ ) {
        my $source_node = $1;
        $source_path = $2;
        if ( $source_node eq $opts{command_interface}->node() ) {
            $source_ci = $opts{command_interface};
        } else {
            $source_ci = NACL::C::Node->new( node => $source_node );
        }
    }

    try {
        $pkg->release(
            %opts,
            command_interface => $source_ci,
            'source-path'     => $source_path
        );
    }
    catch NACL::C::Exceptions::Snapmirror::NoReleasableDestinationFound with {
        my $e = shift;
        $Log->trace('Snapmirror relationship has already been released');
    };

    # Find the snapmirror common snapshot and purge it
    my ( $vol_filter, $vol_or_qtree );
    if ( $pkg->is_operated_on_volume(%opts) ) {
        my $vol_obj = $pkg->get_destination_volume(%opts);
        $vol_filter   = $vol_obj->volume();
        $vol_or_qtree = $vol_filter;
    } else {
        my $qtree_obj = $pkg->get_destination_qtree(%opts);
        $vol_filter   = $qtree_obj->volume();
        $vol_or_qtree = $vol_filter . '_' . $qtree_obj->qtree() . '-dst';
    }

    my @snapshots = NACL::STask::VolumeSnapshot->find(
        command_interface => $opts{command_interface},
        filter            => { volume => $vol_filter },
        %common_opts
    );

    my $dest_node_name = $opts{command_interface}->name();

    foreach my $snapshot (@snapshots) {

        # The snapmirror common snapshot is of this form for volumes:
        # <node-name>(<number>)_<vol-name>.<index>
        # (Note that <number> could contain hyphens as well)
        # And of this form for qtrees:
        # <node-name>(<number>)_<vol-name>_<qtree-name>-dst.<index>
        # Iterate through all snapshots, find the one that matches the pattern
        # and purge it
        if ( $snapshot->snapshot() =~ /$dest_node_name\(.*?\)_$vol_or_qtree\.\d+/ ) {

            # Not waiting, so that we can work around burt 534575
            $snapshot->purge( nacltask_wait => 0 );
            last;
        }
    }

    if ( $sync_type =~ /^sync$/i ) {
        my $apiset    = $opts{command_interface}->get_7m_or_nodescope_apiset();
        my $dest_path = $opts{'destination-path'};
        my $reg_key   = "options.files.snapmirror.conf.$dest_path.dest";
        try {
            $apiset->registry_delete( key => $reg_key );
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {
            my $e = shift;

            my $throw = sub {
                $Log->exit();
                $e->throw();
            };

            if ( $e->text() =~ /An error occurred while trying to commit the registry/i ) {
                my $response = $apiset->rdfile( file => '/etc/snapmirror.conf' );
                my $output   = $response->get_raw_output();
                my @lines    = split /\n/, $output;

                foreach my $line (@lines) {
                    if ( $output =~ /$dest_path/ && $output !~ /^\s*#/ ) {

                        # /etc/snapmirror.conf contains an (uncommented)
                        # entry for this destination
                        $throw->();
                    }
                }
            } else {
                $throw->();
            }
        };
    }

    if ($verify) {

        # Even though the relationship has been well and truly purged by this
        # point, it takes some time for the entry to get removed. We poll
        # for two minutes and fail if it hasn't been removed by then, the
        # assumption being that if it's taking longer than that then
        # something is not right.
        try {
            $pkg->wait_for_deletion( %opts, 'method-timeout' => 120 );
        }
        catch NACL::APISet::Exceptions::TimeoutException with {
            $Log->exit();
            NACL::Exceptions::VerifyFailure->throw( 'After waiting for 2 '
                    . 'minutes the snapmirror entry for the relationship '
                    . 'between source '
                    . $opts{'source-path'}
                    . ' and destination '
                    . $opts{'destination-path'}
                    . ' is still present' );
        };
    }

    $Log->exit();
} ## end sub _purge_7M

=head2 quiesce

    NACL::STask::Snapmirror->quiesce(
        command_interface   => $command_interface,
        'source-path'       => $src_path_specification
        'destination-path'  => $dst_path_specification
        %other_opts
    );


                           (or)

    $snapmirror->quiesce(%other_opts);

(Class or instance method)

Performs a 'snapmirror quiesce' to disable the future transfers for snapmirror
relationship and verifies the state of the relationship if 'nacltask_verify'
task specific parameter is set.

Supports CMode CLI/ZAPI and 7Mode CLI.

=over

=item Options

=over

=item C<< 'source-path' => $source_path >>

(Optional for class method, Not Applicable for instance method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< 'destination-path' => $destion_path >>

(Required for class method, Not Applicable for instance method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< source-cluster => $src_cluster_name >>

(Optional) Name of cluster where the source volume resides.
This input applies only for CMode.


=item C<< destination-cluster => $dst_cluster_name >>

(Optional) Name of cluster where the destination volume resides.
This input applies only for CMode.


=item C<< nacltask_verify => 0 | 1 >>

(Optional, defaults to 0)

If set to 1, it is verified that the snapmirror relationship is in
"Quiesced" state.

If it is not, then it throws a
L<NACL::Exceptions::VerifyFailure|lib-NACL-Exceptions-VerifyFailure-pm>
exception.

Note that verification can also be performed later by calling the
L<taskverify_quiesce|lib-NACL-STask-pm/taskverify_quiesce> method.

=item Other options

All other options supported by
L<< NACL::C::Snapmirror->quiesce|lib-NACL-C-Snapmirror-pm/quiesce >>
are supported by NACL::STask::Snapmirror->quiesce as well.

=back

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification snapmirror quiesce has failed.

=back

=back

=cut

sub quiesce {

    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => { 'nacltask_verify' => { type => BOOLEAN, default => 0 }, },
    );

    # After the call to _common_validate_with(), %opts will have common params
    # like method-timeout, apiset_must, apiset_should, command_interface,
    # primary keys, polling_interval, stask specific params, other options
    # taken by NACL::C::Snapmirror->quiesce()
    my $verify = delete $opts{nacltask_verify};

    # Perform quiesce
    # Always make class method calls so that we don't have to bother
    # about skipping primary keys for instance method call
    my $pkg = $pkg_or_obj->get_package_name();

    # Call the component method from the base class
    $pkg->SUPER::quiesce(%opts);

    # Verify if the snapmirror state is 'quiesced' after quiesce
    # This action is performed only if nacltask_verify is set to 1.
    if ($verify) {
        my %common_comp_opts_with_ci;
        $pkg->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_comp_opts_with_ci,
        );

        $pkg->taskverify_quiesce( %common_comp_opts_with_ci, 'destination-path' => $opts{'destination-path'} );
    }

    $Log->exit();
} ## end sub quiesce

=head2 taskverify_quiesce

    NACL::STask::Snapmirror->taskverify_quiesce(
        command_interface  => $cmd_int,
        'source-path'      => $source_path,
        'destination-path' => $destination_path
    );

                           or

    $snapmirror_obj->taskverify_quiesce();

(Class or Instance method)

This method verifies that the snapmirror relationship is in "Quiesced"
state. If it is not, then it throws a
L<NACL::Exceptions::VerifyFailure|lib-NACL-Exceptions-VerifyFailure-pm>
exception.

=over

=item Options

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for Instance method)

See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)

=item C<< apiset_should => $ruleset >>

(Optional)

=item C<< 'method-timeout' => $timeout >>

(Optional) default timeout is 600.

=item C<< 'source-path' => $path_specification >>

(Optional for class method, Not Applicable for Instance method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< 'destination-path' => $path_specification >>

(Required for class method, Not Applicable for Instance method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< "source-cluster =>  $string" >>

(Optional) Name of cluster where the source volume resides.
This input applies only for CMode.

=item C<< "destination-cluster =>  $string" >>

(Optional) Name of cluster where the destination volume resides.
This input applies only for CMode.

=back

=cut

sub taskverify_quiesce {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );
    my %opts = $pkg_or_obj->_common_validate_with( params => \@args );

    my $pkg = $pkg_or_obj->get_package_name();

    # In CMode, the "status" field will be 'Quiesced', in 7mode the
    # "state" field will be 'Quiesced'
    if ( $opts{command_interface}->mode() eq 'CMode' ) {

        # In CMode, if transfer is in progress and quiesce is issued, the status
        # moves to quiescing until the transfer completes and then moves to 'quiesced'.
        $pkg->wait_on_attribute(
            %opts,
            attribute_to_check        => 'status',
            till_value                => qr/Quiesced/i,
            valid_intermediate_states => [qr/Quiescing/i],
            'method-timeout'          => $opts{'method-timeout'} || 600,
        );
    } else {
        $pkg->verify_state( %opts, state => 'Quiesced' );
    }

    $Log->exit();
} ## end sub taskverify_quiesce

=head2 break


    NACL::STask::Snapmirror->break(
        command_interface   => $command_interface,
        'source-path'       => $src_path_specification,
        'destination-path'  => $dst_path_specification,
        %other_opts
    );

                           (or)

  $snapmirror->break(%other_opts);

(Class or instance method)

Performs a 'snapmirror break' to break a source/destination relationship and
makes destination writable. Verifies that destination is writable and snapmirror
state is Broken-off after break if 'nacltask_verify' task specific parameter is set.

Supports CMode CLI/ZAPI and 7Mode CLI.

=over

=item Options

=over

=item C<< 'source-path' => $source_path >>

(Optional for class method, Not Applicable for instance method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< 'destination-path' => $path_specification >>

(Required for class method, Not Applicable for instance method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< 'source-cluster' =>  $string >>

(Optional) Name of cluster where the source volume resides.
This input applies only for CMode.

=item C<< 'destination-cluster' =>  $string >>

(Optional) Name of cluster where the destination volume resides.
This input applies only for CMode.

=item C<< nacltask_verify =>  0 | 1 >>

(Optional, defaults to 0)

If 1, verify if the snapmirror state is Broken-off and also check
if the destination volume type is writable. (note that for QSM this second
step is not performed)

If it is not, then it throws a
L<NACL::Exceptions::VerifyFailure|lib-NACL-Exceptions-VerifyFailure-pm>
exception.

It is also possible to verify separately that the relationship has been
broken by using the
L<taskverify_break|lib-NACL-STask-Snapmirror-pm/taskverify_break>
method.

=item C<< nacltask_if_broken => $action >>

(Optional, defaults to 'die')

What to do if the Snapmirror relationship is already broken-off
between the source and destination specified.  If $action is "die"
(the default), then fail with an exception (in the same way that
the Snapmirror component would have: by trying the Snapmirror break
and letting the product complain about the Snapmirror relationship
already broken).  If action is "pass", then do not fail

=item C<< nacltask_if_transferring => $action >>

(Optional, defaults to "wait")

What should be done if a snapmirror transfer is in progress when we attempt
to break. This argument can take any of three values: wait|abort|die
if $action is "wait", then it will wait till the transfer is not completed.
if $action is "abort", then it will abort the current transfer process.
if $action is "die", then fail with an exception.

=item C<< nacltask_if_uninitialized => $action >>

(Optional, defaults to 'die')

What to do if the snapmirror break is attempted on an uninitialized
volume snapmirror. If $action is "die" (the default), then fail with
an exception (in the same way that the Snapmirror component would
have : by trying the volume snapmirror break and letting the product complain
about the volume not initialized. If action is "pass", then do not fail.

=item C<< nacltask_if_ls => $action >>

(Optional, defaults to 'die')

What to do if the snapmirror break is attempted on a LS snapmirror
where the operation is not allowed.
If $action is "die" (the default), then fail with
an exception (in the same way that the Snapmirror component would
have : by trying the volume snapmirror break and letting the product complain
about the volume not initialized. If action is "pass", then do not fail.

=over

=item C< wait >

Wait until the transfer is complete (this is the default).

=item C< abort >

Abort the transfer

=item C< die >

Snapmirror break will fail with an exception.

=back

=item C<< "nacltask_wait =>  $boolean" >>

(Optional) A number 0 or 1 that determines if the method has to wait till the
operation is complete.  1 means to return after break is complete and 0 means to
return immediately. The default is 1.

=item command_interface, apiset_must, apiset_should, method-timeout, etc.

All other options supported by
L<< NACL::C::Snapmirror->break|lib-NACL-C-Snapmirror-pm/break >>
are supported by NACL::STask::Snapmirror->break as well.

=back

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when destination volume is not writable after performing snapmirror break.

=item C<NACL::C::Exceptions::Snapmirror::TransferInProgress>

This type of exception is thrown when an attempt is made to break snapmirror relationship when a snapmirror trnasfer is in progress.

=item C<NACL::APISet::Exceptions::ResponseException>

This type of exception is thrown when an attempt is made to break snapmirror relationship when it is already broken or
when there is any failure while breaking snapmirror relationship.

=back

=back

=cut

sub break {

    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'nacltask_verify'    => { type                                                        => SCALAR, default => 0 },
            'nacltask_if_broken' => $pkg_or_obj->_if_action_completed_already_validate_spec( name => 'nacltask_if_broken' ),
            'nacltask_if_uninitialized' => $pkg_or_obj->_if_action_completed_already_validate_spec( name => 'nacltask_if_uninitialized' ),
            'nacltask_if_ls'            => $pkg_or_obj->_if_action_completed_already_validate_spec( name => 'nacltask_if_ls' ),
            'nacltask_if_transferring'  => {
                type      => SCALAR,
                default   => 'wait',
                callbacks => {
                    "Value of 'nacltask_if_transferring' must be 'wait' " . "or 'abort' or 'die'" => sub { $_[0] =~ /^(?:wait|abort|die)$/ }
                }
            },
            'nacltask_wait' => { type => SCALAR, default => 1 },
        },
    );

    # After the call to _common_validate_with(), %opts will have common params
    # like method-timeout, apiset_must, apiset_should, command_interface,
    # primary keys, polling_interval, stask specific params, other options
    # taken by NACL::C::Snapmirror->break()
    my %task_options;
    $pkg_or_obj->_move_nacltask_options(
        source => \%opts,
        target => \%task_options
    );
    my $verify           = $task_options{nacltask_verify};
    my $broken           = $task_options{nacltask_if_broken};
    my $if_transferring  = $task_options{nacltask_if_transferring};
    my $nacltask_wait    = $task_options{nacltask_wait};
    my $if_uninitialized = $task_options{nacltask_if_uninitialized};
    my $if_ls            = $task_options{nacltask_if_ls};

    # Always make class method calls so that we don't have to bother
    # about skipping primary keys for instance method call
    my $pkg = $pkg_or_obj->get_package_name();

    # This will contain the common Component params along with
    # destination-path
    my %common_params;

    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_params
    );

    # This will contain the common Component params only
    my %common_params_only = %common_params;

    $common_params{'destination-path'} = $opts{'destination-path'};

    # Handle Job ID
    my $job_component = delete $opts{job_component};
    {

        # We want to know the job ID of the job performed by
        # SUPER::update, whether or not our caller does.
        my $dummy;
        $job_component ||= \$dummy;
    }

    my $need_to_return = 0;

BREAK: {
        use warnings qw(exiting);
        try {
            $pkg->SUPER::break( %opts, job_component => $job_component );
        }
        catch NACL::C::Exceptions::Snapmirror::TransferInProgress with {
            my $e = shift;
            $Log->trace('Snapmirror transfer in process');
            if ( $if_transferring eq 'abort' ) {
                $pkg->abort(
                    %common_params,
                    nacltask_if_aborted => 'pass',
                    nacltask_verify     => 1
                );
            } elsif ( $if_transferring eq 'die' ) {
                $Log->exit();
                $e->throw();
            } else {
                $pkg->wait_for_transfer_to_complete(%common_params);
            }
            my $redo = sub {
                no warnings qw(exiting);
                redo BREAK;
            };
            $redo->();

        }
        catch NACL::C::Exceptions::Snapmirror::BreakNotAllowedForVault with {
            $need_to_return = 1;
        }
        catch NACL::C::Exceptions::Snapmirror::RelationshipNotInitialized with {
            $need_to_return = 1;
        }
        catch NACL::APISet::Exceptions::ResponseException with {
            my $e    = shift;
            my $text = $e->text();

            my $throw_exception = sub {
                $Log->exit();
                $e->throw();
            };

            my $redo = sub {
                no warnings qw(exiting);
                redo BREAK;
            };

            if ($e->isa('NACL::C::Exceptions::Snapmirror::DataProtectionVolume')    # CMode
                || $pkg->_is_broken_7M( text => $text )                             # 7Mode
                )
            {
                $Log->trace('Snapmirror relationship has already been broken');
                $throw_exception->() if ( $broken =~ /die/i );
                $need_to_return = 1;
            } elsif ( $e->isa('NACL::C::Exceptions::Snapmirror::NotQuiescedOrQuiescing') ) {
                if ( $e->text() =~ /not quiesced/i ) {
                    $Log->trace('Need to quiesce the destination first');
                    $pkg->quiesce(%common_params);
                } else {

                    # Must be in Quiescing state
                    $pkg->taskverify_quiesce(%common_params);
                }
                $redo->();
            } elsif ( $e->isa('NACL::C::Exceptions::Snapmirror::VolumeNotInitializedorRestricted') ) {
                $throw_exception->() if ( $if_uninitialized =~ /die/i );
                $need_to_return = 1;
            } elsif ( $e->isa('NACL::C::Exceptions::Snapmirror::BreakNotAllowedForLS') ) {

                # Might be case when snapmirror of LS relationship is purged.
                # Ignoring it as break is not allowed on LS relationship.
                $throw_exception->() if ( $if_ls =~ /die/i );
                $need_to_return = 1;
            } else {

                # Unknown error
                $throw_exception->();
            }
        };
    }

    # A call to break on an already deleted SM relationship,
    # by just passing 'destination-path' will not throw in
    # any exception but will successfully delete the SM
    # from the destination . Need to return from here in
    # such scenario as there will not be any SM relationship
    # to be verified.

    if ( !$need_to_return ) {
        my $sm = $pkg->find(
            %common_params_only,
            filter      => { 'destination-path' => $opts{'destination-path'} },
            allow_empty => 1,
        );

        unless ($sm) {

            # There is no relationship to be queried.
            $need_to_return = 1;
        }
    }

    if ($need_to_return) {
        $Log->exit();
        return;
    }
    $pkg_or_obj->last_job( ${$job_component} ) if ( ref $pkg_or_obj );

    if ($nacltask_wait) {
        my %job_args;
        if ( defined $job_component ) {
            $job_args{job_component} = $$job_component;
        }
        $pkg->wait_for_break( %common_params, %job_args );
    }

    $pkg->taskverify_break(%common_params) if ($verify);

    $Log->exit();
} ## end sub break

=head2 wait_for_break

    # Package call
    NACL::STask::Snapmirror->wait_for_break(
        command_interface  => $ci,
        'destination-path' => $dst_path,
        ...
    );

    # Instance call
    $snapmirror_obj->wait_for_break(...);

Wait for the snapmirror break operation to complete.

If the call does not complete within the timeout value specified, then
a C<NACL::Exceptions::Timeout> exception is thrown.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< "destination-path" =>$destination_path >>

(Required for class method, Not Applicable for instance method)
The destination path of the snapmirror relationship.

=item C<< "method-timeout" =>$timeout >>

(Optional) How long in seconds to wait. Defaults to 1200 seconds.

=item C<< job_component => $job_component >>

(Optional for class method, Not Applicable for instance method) A
NACL::C::Job object representing the C-mode Job used to initialize
this Snapmirror operation.

=item C<< polling_interval => $polling_interval >>

(Optional, defaults to 10s)
The time between polls to check if the break operation is complete.

=back

=back

=cut

sub wait_for_break {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    $pkg_or_obj->wait_for_completion(
        @args,
        attribute_to_check        => 'status',
        till_value                => [ qr/^Idle$/i, qr/^OutOfSync$/i ],
        valid_intermediate_states => [qr/Breaking/i, qr/-/i],
        show_all_fields           => 1,
    );
    $Log->exit();
}

# Helper routine to check whether relationship is already broken or not
# Used by break() and purge()
sub _is_broken_7M {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts = validate( @_, { text => { type => SCALAR } } );
    my $val;
    if (   $opts{text} =~ /destination is not in snapmirrored state/
        || $opts{text} =~ /qtree is not a snapmirror qtree replica/ )
    {
        $val = 1;
    } else {
        $val = 0;
    }
    $Log->exit("Returning $val");
    return $val;
} ## end sub _is_broken_7M

=head2 taskverify_break

    NACL::STask::Snapmirror->taskverify_break(
        command_interface  => $cmd_int,
        'source-path'      => $source_path,
        'destination-path' => $destination_path
    );

                           or

    $snapmirror_obj->taskverify_break();

(Class or Instance method)

This method verifies that the snapmirror relationship has been broken.
It does so by checking that the state of the relationship is "Broken-off"
and for non-QSM relationships also checks that the volume is writable.
If it is not, then it throws a
L<NACL::Exceptions::VerifyFailure|lib-NACL-Exceptions-VerifyFailure-pm>
exception.

Supports CMode CLI/ZAPI and 7Mode CLI.

=over

=item Options

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for Instance method)

See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)

=item C<< apiset_should => $ruleset >>

(Optional)

=item C<< 'method-timeout' => $timeout >>

(Optional)

=item C<< 'source-path' => $path_specification >>

(Optional for class method, Not Applicable for Instance method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< 'destination-path' => $path_specification >>

(Required for class method, Not Applicable for Instance method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< "source-cluster =>  $string" >>

(Optional) Name of cluster where the source volume resides.
This input applies only for CMode.

=item C<< "destination-cluster =>  $string" >>

(Optional) Name of cluster where the destination volume resides.
This input applies only for CMode.

=back

=cut

sub taskverify_break {
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts       = $pkg_or_obj->_common_validate_with( params => \@args );
    my $pkg        = $pkg_or_obj->get_package_name();

    # Verify status and destination Vserver subtype for
    # VDR relationship.
    my %extra_verify_opts;
    if ( $pkg->is_operated_on_vserver(%opts) ) {
        $extra_verify_opts{status} = 'Idle';
    }

    # remove 'destination_ci' and 'source_ci' from argument list before calling
    # verify_state() as these arguments are not listed in validations options
    # and are not required to be passed
    my %tmp_opts = %opts;
    delete $tmp_opts{'destination_ci'};
    delete $tmp_opts{'source_ci'};

    $pkg->verify_state(
        'state' => 'Broken-off',
        %extra_verify_opts,
        %tmp_opts,
    );

    $pkg->_verify_vserver_subtype(%opts);

    if ( $pkg->is_operated_on_volume(%opts) ) {
        my $dst_vol = $pkg->get_destination_volume(%opts);

        my %common_comp_opts;
        $pkg_or_obj->_copy_common_component_params(
            source => \%opts,
            target => \%common_comp_opts
        );

        # CMode has a "type" field for NACL::CS::Volume which describes
        # whether the volume is RW (read-write) or not. 7Mode doesn't
        # populate this field. The best way to determine whether a volume is
        # writable is to look for the absence of "read-only" in the
        # raidstatus field.
        if ( $opts{command_interface}->is_cmode() ) {
            $dst_vol->verify_state(
                'type' => 'RW',
                %common_comp_opts
            );
        } else {
            my $raidstatus = $dst_vol->get_one_state_attribute( 'raidstatus', %common_comp_opts );
            if ( $raidstatus =~ /read-only/ ) {
                $Log->exit();
                NACL::Exceptions::VerifyFailure->throw(
                    'Volume ' . $dst_vol->volume() . 'should have been writable after break but was not' );
            }
        }
    }

    $Log->exit();
} ## end sub taskverify_break

=head2 wait_for_transfer_to_complete

    NACL::STask::Snapmirror->wait_for_transfer_to_complete(
        command_interface   => $command_interface,
        'source-path'       => $src_path_specification,
        'destination-path'  => $dest_path_specification,
        %other_opts
    );

                           or

    $snapmirror_obj->wait_for_transfer_to_complete(%other_opts);

(Class or Instance method, currently 7Mode only)

Many snapmirror operations are possible only if there is no transfer
in progress. This method provides a way to wait for the transfer to complete.

This method operates by polling the amount of data transferred as long as
there is progress being made until it reaches the max_iterations. If it is found
for 3 successive iterations that the amount of data transferred has not increased,
then this method times out and throws an exception. If the number of iterations
done by this method exceeds the max_iterations and the status of transfer is still
in progress, then it will throw the timeout exception.

=cut

sub wait_for_transfer_to_complete {
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            polling_interval => { type => SCALAR, default => POLL_DELTA },
            max_iterations   => { type => SCALAR, default => 3 },
        }
    );
    my $interval       = delete $opts{polling_interval};
    my $max_iterations = delete $opts{max_iterations};
    my $filter         = $pkg_or_obj->_hash_move(
        source => \%opts,
        move   => [ $pkg_or_obj->get_primary_keys() ]
    );

    my $previous_size             = 0;
    my $no_of_repeated_iterations = 0;

    while (1) {
        my $cs = NACL::CS::Snapmirror->fetch(
            %opts,
            filter           => $filter,
            requested_fields => [qw(status data-transferred)]
        );
        my $status = $cs->status();
        last if ( $status =~ /(Idle|InSync|OutOfSync|Quiesced)/i );

        if ( $status !~ /Transferring|Preparing|Finalizing|-/i ) {

            # Invalid state
            $Log->exit();
            NATE::BaseException->throw("status of snapmirror $status is not a valid state");
        }

        my $data_transferred = $cs->data_transferred();
        $data_transferred =~ /(\d+)/ if ( defined $data_transferred );
        my $size_transferred = $1;

        if (   defined $size_transferred
            && defined $previous_size
            && $size_transferred == $previous_size )
        {
            $no_of_repeated_iterations++;
        } else {
            $no_of_repeated_iterations = 0;
        }

        if ( $no_of_repeated_iterations == $max_iterations ) {
            $Log->exit();
            NACL::Exceptions::Timeout->throw( "Snapmirror is still in transfer state" . " and maximum time limit is crossed" );
        } else {
            $previous_size = $size_transferred
                if ( defined $size_transferred );
            Tharn::snooze($interval);
        }
    }

    $Log->exit();
} ## end sub wait_for_transfer_to_complete

=head2 update

  NACL::STask::Snapmirror->update(
      command_interface     =>  $command_interface,
      'source-path'         =>  $source_volume_object,
      'destination-path'    =>  $destination_volume_object,
      'nacltask_wait'       =>  $number,
      %other_options
  );

  (or)

  $snapmirror_object->update(nacltask_wait  =>  0 );

  (or)

  NACL::STask::Snapmirror->update(
      command_interface     =>  $command_interface,
      'source-path'         =>  $source_volume_object,
      'destination-path'    =>  $destination_volume_object,
      %other_options
  );


(Class or Instance method) Performs a 'snapmirror update'.
This method provides additional services like wait till a specific percentage of
transfer happens after update.

=over

=item Options

=over


=item C<< "source-path =>  $src_volume_specification" >>

(Optional for class method, Not Applicable for instance method) Required when
$nacltask_wait value is > 0 and < 1. Source information is required to compute
difference in vol sizes so that specified percentage of transfer can be done in
background. $src_volume_specification may be either a string specifying the full
path to the source volume, or it may be a NACL::(C|STask)::Volume with the
caveat that source-cluster must also be passed as an argument if the
Volume object's command_interface is a NACL::C::Vserver.


=item C<< "destination-path =>  $dst_volume_specification" >>

(Required for class method, Not Applicable for instance method)
destination-path takes the same form as the source-path argument, but
for the destination volume (and possibly requiring destination-cluster.)


=item C<< "source-cluster =>  $string" >>

(Optional) Name of cluster where the source volume resides.
This input applies only for CMode.


=item C<< "destination-cluster =>  $string" >>

(Optional) Name of cluster where the destination volume resides.
This input applies only for CMode.


=item C<< nacltask_retry_tries_count => $tries_count >>

(Optional, defaults to 3)
The number of times to retry the update command if it fails because
the snapmirror is temporarily busy.

Note that it is possible to make it not retry the update command by setting
the value of this argument to 1.


=item C<< nacltask_retry_sleep_time => $sleep_time >>

(Optional, defaults to 10 seconds)
The time to sleep between retries of the update command.


=item C<< nacltask_wait => 0|1 >>

(Optional, defaults to 1)
Specifies whether to wait till the data transfer completes.
1 means to return after the initial transfer has completed.
0 means do not wait any longer than necessary (any longer
than the Snapmirror initialize command waits).

Note that specifying a value between 0 & 1 is deprecated and is no longer
supported because of removal of support for real-time data from snapmirror show
(see burt 694025).nacltask_wait is based on method-timeout.

=item C<< nacltask_verify_health => 0|1 >>

(Optional, defaults to '1', Only for CMode)
If set to '1', verifies the health of the snapmirror.

=item command_interface, apiset_must, apiset_should, method-timeout, etc.

All other options supported by
L<< NACL::C::Snapmirror->update|lib-NACL-C-Snapmirror-pm/update >>
are supported by NACL::STask::Snapmirror->update as well.

=back

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

If the update fails, an exception of type B<NACL::Exceptions::VerifyFailure>
is thrown with the content of the C<Last Transfer Error> field from the
snapmirror relationship's state.

=item C<NACL::Exceptions::InvalidChoice>

This type of exception is thrown when update command is executed in 7mode.

=item C<NATE::Exceptions::Argument>

This type of exception is thrown when source command interface is missing.

=back

=back

=cut

sub update {

    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {

            # component options which are overloaded
            'source-path'      => { type => OBJECT | SCALAR, optional => 1 },
            'destination-path' => { type => OBJECT | SCALAR, optional => 1 },

            # options which would be consumed by stask method itself apart
            # from passing the information to components.
            'source-cluster'         => { type => SCALAR, optional => 1 },
            'destination-cluster'    => { type => SCALAR, optional => 1 },
            'nacltask_wait'          => { type => SCALAR, default  => 1 },
            'nacltask_verify_health' => { type => SCALAR, default  => 1 },
            'nacltask_verify'        => { type => SCALAR, default  => 0 },
            nacltask_config_diff     => { type => SCALAR, default  => 0 },
	    'nacltask_retry_tries_count' => { type => SCALAR, default  => 3 },
	    'nacltask_retry_sleep_time'  => { type => SCALAR, default  => 10 },
        },
        allow_extra => 1,
    );

    # After the call to _common_validate_with(), %opts will have common params
    # like method-timeout, apiset_must, apiset_should, command_interface,
    # primary keys, polling_interval, stask specific params, other options
    # taken by NACL::C::Snapmirror->update()
    my $ci = $opts{command_interface};

    $opts{'method-timeout'} ||= SNAPMIRROR_TIMEOUT;

    # The dependencies which need to be handled are different between 7M and CM
    # Hence we will have two backends for 7M and CM separately.
    if ( $ci->mode() eq '7Mode' ) {

        # Invoke 7M backend
        $Log->trace("7Mode update not supported");
        $Log->exit();
        NACL::Exceptions::InvalidChoice->throw("7Mode not supported currently.");
    } elsif ( $ci->mode() eq 'CMode' ) {

        # Invoke CM backend
        $pkg_or_obj->_update_CM(%opts);
    } else {

        # Just to avoid open if-elsif
        NATE::BaseException->throw( "Filer mode '" . $ci->mode() . "' not supported." );
    }
    $Log->exit();
} ## end sub update

sub _update_CM {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;
    my $nacltask_wait    = delete $opts{nacltask_wait};
    my $polling_interval = delete( $opts{'polling_interval'} );
    my $verify_health    = delete( $opts{'nacltask_verify_health'} );
    my $config_diff      = delete $opts{nacltask_config_diff};
    my $tries_count	 = delete $opts{'nacltask_retry_tries_count'};
    my $sleep_time	 = delete $opts{'nacltask_retry_sleep_time'};

    # Should wait if config_diff needs to be checked
    $nacltask_wait ||= $config_diff;

    # Get the source-path and destination-path.
    # Also store the volume objects if passed.
    my ( $src_vol_obj, $dst_vol_obj );

    # source-path could be either object or path. if user had specified
    # volume, vserver, cluster names separately, then _common_validate_with()
    # which is overloaded in NACL::C::Snapmirror would have constructed a
    # path out of it.
    if ( ref $opts{'source-path'} ) {
        $src_vol_obj = $opts{'source-path'};
        $opts{'source-path'} = $pkg_or_obj->construct_volpath(
            vol_obj_or_volpath => $src_vol_obj,
            cluster_name       => delete( $opts{'source-cluster'} ),
            %opts
        );
    }
    if ( ref $opts{'destination-path'} ) {
        $dst_vol_obj = $opts{'destination-path'};
        $opts{'destination-path'} = $pkg_or_obj->construct_volpath(
            vol_obj_or_volpath => $dst_vol_obj,
            cluster_name       => delete( $opts{'destination-cluster'} ),
            %opts
        );
    }

    my $command_interface = $opts{command_interface};

    # Handle Job ID
    my $job_component = delete $opts{job_component};
    {

        # We want to know the job ID of the job performed by
        # SUPER::update, whether or not our caller does.
        my $dummy;
        $job_component ||= \$dummy;
    }

    #calculate difference in used size between source and destination volumes
    if ( $nacltask_wait != 0 && $nacltask_wait < 1 ) {
        $Log->warn( 'Specifying a nacltask_wait value between 0 & 1 '
                . 'is deprecated because it is no longer possible to get '
                . 'real-time snapmirror data; see 694025. (the value provided '
                . "was $nacltask_wait) The only currently supported values "
                . 'for nacltask_wait are 0 and 1' );

        # Set to 0 because the user would have wanted it to return
        # while data is still being transferred
        $nacltask_wait = 0;
    }

    # Perform update
    # Always make class method calls so that we don't have to bother
    # about skipping primary keys for instance method call
    my $pkg = ref $pkg_or_obj || $pkg_or_obj;
    nacl_method_retry(
	code => sub { $pkg->SUPER::update( %opts, job_component => $job_component ); },
	exceptions => 'NACL::C::Exceptions::Snapmirror::TransferInProgress',
	tries_count => $tries_count,
	sleep_time  => $sleep_time,
    );

    delete $opts{foreground};
    $pkg_or_obj->last_job( ${$job_component} ) if ( ref $pkg_or_obj );

    $opts{polling_interval} = $polling_interval
        if ( defined $polling_interval );

    #Call monitor only when there is some data to be transferred
    if ( $nacltask_wait == 1 ) {

        # Adding snooze of 5 secs. This is added as fix for Burt624613 where snapmirror
        # show takes few seconds to catch change in state of snapmirror relationship.
        Tharn::snooze(5);

        $pkg->wait_for_update(
            %opts,
            job                    => $$job_component,
            nacltask_verify_health => $verify_health,
        );
    }

    if ( $opts{nacltask_verify} ) {
        $pkg->_verify_states(%opts);
    }

    if ($config_diff) {
        my ( %primary_keys, %common_opts );
        $pkg->_copy_primary_keys(
            source => \%opts,
            target => \%primary_keys,
        );
        $pkg->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_opts,
        );
        $pkg->config_diff( %primary_keys, %common_opts );
    }

    $Log->exit();
    return;
} ## end sub _update_CM

=head2 wait_for_update

Waits for the snapmirror update operation to complete. This method accepts the
same arguments as L<wait_for_initialization>.

=cut

sub wait_for_update {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->wait_for_initialization( @args, 'last-transfer-type' => 'update' );
    $Log->exit();
}

# Sub to calculate snap delta
sub _snap_delta {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;
    my $ci = $opts{source_volume}->command_interface();
    my $size;

    my $Api_Set_Obj = $ci->apiset( category => "Node", interface => "CLI", set => "Nodescope", );
    my $response = $Api_Set_Obj->execute_command(
        'command' => 'snap delta ' . $opts{source_volume}->volume() . ' ' . $opts{snap1} . ' ' . $opts{snap2} );
    my @lines = split /\n/,    $response;
    my @words = split /(\w+)/, $lines[$#lines];
    if ( $words[ $#words - 5 ] =~ /:/ ) {
        $size = $words[ $#words - 10 ];
    } else {
        $size = $words[ $#words - 6 ];
    }

    $Log->exit();
    return $size;
}

sub _newest_common_snap_index {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;
    my @s1    = @{ $opts{all_snaps} };
    my @s2    = @{ $opts{snap_to_compare} };
    my $final = 0;
    if (@s2) {
        foreach my $dd (@s2) {
            my ($index) = grep { $s1[$_] =~ $dd } 0 .. $#s1;
            $final = $index if ( $index > $final );
        }
    } else {

        # If destination snap list is null, then there is no common snap.
        $final = -1;
    }
    $Log->exit();
    return $final;
}

=head2 resume



  NACL::STask::Snapmirror->resume(command_interface =>  $ci,
                            'source-path'           =>  $source_volume_object,
                            'destination-path'      =>  $destination_volume_object,
                            'source-cluster'        =>  $Cluster_name,          #clustername is passed if the Volume object's command_interface is a NACL::C::Vserver
                            'destination-cluster'   =>  $Cluster_name,
                            'nacltask_verify'       =>  $nacltask_verify_boolean,
                            'nacltask_if_resumed'   =>  $action,
                            %other_options);

  (or)

  NACL::STask::Snapmirror->resume(command_interface  =>$ci,
                             'source-path'          =>$source_path,
                             'destination-path'     =>$destination_path);


   (or)

   $snapmirror->resume;

  (Class or instance method) The method resumes a snapmirror relationship after quiesce
   Uses a CMode-CLI or CMode-ZAPI.

Performs a 'snapmirror resume' to resume a snapmirror relationship  which is in
quiesced state. Verifies that transfer starts again after resume.

=over

=item Options

=over


=item C<< "source-path =>  $source_volume_object" >>

(Optional for class method, Not Applicable for instance method)
$volume_specification may be either a string specifying the full path
to the source volume, or it may be a NACL::(C|STask)::Volume with the
caveat that source-cluster must also be passed as an argument if the
Volume object's command_interface is a NACL::C::Vserver.


=item C<< "destination-path =>  $destination_volume_object" >>

(Required for class method, Not Applicable for instance method)
destination-path takes the same form as the source-path argument, but
for the destination volume (and possibly requiring destination-cluster.)

=item C<< "nacltask_verify => $nacltask_verify_boolean" >>

(Optional) If '0' (default), verification will not be performed.
 If '1', verification will be performed to ensure that resume happened successfully.

=item C<< "nacltask_if_resumed=> $action " >>

(Optional) If 'die' (default), fails if snapmirror is not in quiesced state and resume is tried
 If set to 'pass' it will just log a debug message saying relationship was not quiesced.

=item C<< "source-cluster =>  $string" >>

(Optional) Name of cluster where the source volume resides.
This input applies only for CMode.

=item C<< nacltask_retry_tries_count => $tries_count >>

(Optional, defaults to 30)
The number of times to retry the resume command if it fails because
the snapshot is temporarily busy or stale catch entry.

Note that it is possible to make it not retry the delete command by setting
the value of this argument to 1.

=item C<< nacltask_retry_sleep_time => $sleep_time >>

(Optional, defaults to 10 seconds)
The time to sleep between retries of the resume command.

=item C<< nacltask_retry_num_times_code_ran => \$variable >>

(Optional)
A reference to a scalar variable which gets filled in with the number
of times the resume command was invoked.

=item C<< "destination-cluster =>  $string" >>

(Optional) Name of cluster where the destination volume resides.
This input applies only for CMode

All other options supported by
L<< NACL::C::Snapmirror->resume|lib-NACL-C-Snapmirror-pm/resume >>
are supported by NACL::STask::Snapmirror->resume as well.

=back

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification fails for snapmirror resume.

=item C<NACL::Exceptions::InvalidChoice>

This type of exception is thrown when resume command is executed in 7mode.

=item C<NACL::C::Exceptions::RelationshipInfoBeingPropagated>

This exception is thrown when we attempt to resume a snapmirror relation when the relationship information is being propagated

=back

=back

=cut

sub resume {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {

            # component options which are overloaded
            'source-path'      => { type => OBJECT | SCALAR, optional => 1 },
            'destination-path' => { type => OBJECT | SCALAR, optional => 1 },

            # options which would be consumed by stask method itself apart
            # from passing the information to components.
            'source-cluster'      => { type => SCALAR, optional => 1 },
            'destination-cluster' => { type => SCALAR, optional => 1 },
            'nacltask_verify'     => { type => SCALAR, default  => 0 },
            'nacltask_if_resumed' => $pkg_or_obj->_if_action_completed_already_validate_spec( name => 'nacltask_if_resumed' ),
            'nacltask_retry_tries_count'        => { type => SCALAR,    default  => 30 },
            'nacltask_retry_sleep_time'         => { type => SCALAR,    default  => 10 },
            'nacltask_retry_num_times_code_ran' => { type => SCALARREF, optional => 1 },
        },
        allow_extra => 1,
    );

    # After the call to _common_validate_with(), %opts will have common params
    # like method-timeout, apiset_must, apiset_should, command_interface,
    # primary keys, polling_interval, stask specific params, other options
    # taken by NACL::C::Snapmirror->resume()
    my $ci = $opts{command_interface};
    $opts{'method-timeout'} ||= SNAPMIRROR_TIMEOUT;

    # The dependencies which need to be handled are different between 7M and CM
    # Hence we will have two backends for 7M and CM separately.
    if ( $ci->mode() eq '7Mode' ) {

        # 7mode support
        my $msg = '7Mode not supported currently';
        $Log->exit();
        NACL::Exceptions::InvalidChoice->throw($msg);

    } elsif ( $ci->mode() eq 'CMode' ) {

        # Invoke CM backend
        $pkg_or_obj->_resume_CM(%opts);

    } else {

        # Just to avoid open if-elsif
        NATE::BaseException->throw( "Filer mode '" . $ci->mode() . "' not supported." );
        $Log->exit();
    }
    $Log->exit();

} ## end sub resume

sub _resume_CM {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;
    my $verify     = delete $opts{nacltask_verify};
    my $if_resumed = delete $opts{nacltask_if_resumed};

    # Get the source-path and destination-path. Rest of the code
    # understands only these two inputs.
    # Also store the volume objects if passed.
    my ( $src_vol_obj, $dst_vol_obj );

    # source-path could be either object or path. if user had specified
    # volume, vserver, cluster names separately, then _common_validate_with()
    # which is overloaded in NACL::C::Snapmirror would have constructed a
    # path out of it.
    if ( ref $opts{'source-path'} ) {
        $src_vol_obj = $opts{'source-path'};
        $opts{'source-path'} = $pkg_or_obj->construct_volpath(
            vol_obj_or_volpath => $src_vol_obj,
            cluster_name       => delete( $opts{'source-cluster'} ),
            %opts
        );
    }
    if ( ref $opts{'destination-path'} ) {
        $dst_vol_obj = $opts{'destination-path'};
        $opts{'destination-path'} = $pkg_or_obj->construct_volpath(
            vol_obj_or_volpath => $dst_vol_obj,
            cluster_name       => delete( $opts{'destination-cluster'} ),
            %opts
        );
    }

    my $command_interface = $opts{command_interface};

    my %retry_opts;
    $pkg_or_obj->_move_retry_params(
        source => \%opts,
        target => \%retry_opts
    );

    # Perform resume
    # Always make class method calls so that we don't have to bother
    # about skipping primary keys for instance method call
    my $pkg = ref $pkg_or_obj || $pkg_or_obj;
    try {
        nacl_method_retry(
            code       => sub { $pkg->SUPER::resume(%opts); },
            exceptions => [qw(NACL::C::Exceptions::RelationshipInfoBeingPropagated)],
            %retry_opts,
            )

    }
    catch NACL::C::Exceptions::Snapmirror::NotQuiescedOrQuiescing with {
        my $e = shift;
        if ( $if_resumed =~ /pass/i ) {
            $Log->debug( "Relationship is not in quiesced state, passing  as is preferred by user " );
        }
    };

    #Verify if the snapmirror state is Idle after resume
    #This is done if nacltask_verify is set to 1.
    if ($verify) {
        my $command_interface = $opts{command_interface};
        my %common_comp_opts;
        $pkg_or_obj->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_comp_opts
        );

        $pkg->verify_state(
            command_interface  => $command_interface,
            'destination-path' => $opts{'destination-path'},
            'state'            => qr/^(Snapmirrored|Insync)$/i,
            'status'           => 'idle',
            %common_comp_opts
        );
        $Log->debug("SnapMirror Relation has resumed succesfully ");
    }
    $Log->exit();
} ## end sub _resume_CM

=head2 modify



  NACL::STask::Snapmirror->modify(command_interface =>  $ci,
                            'destination-path'      =>  $destination_path,
                            'source-path'           =>  $source_path,
                            'tries'                 =>  $tries,
                            'throttle'              =>  $throttle,
                            'nacltask_verify'       =>  $nacltask_verify_boolean,
                            %other_options);

  (or)

  NACL::STask::Snapmirror->modify(command_interface  =>$ci,
                             'source-path'          =>$source_path,
                             'destination-path'     =>$destination_path,
                             'schedule'                 =>  $schedule,);


   (or)

   $sm->modify( 'throttle'   =>  $throttle,);

   If the source and destination are on same cluster then throttle change will not take place and the mthod will just print a log message.
   (or)

   $sm->modify(schedule => $schedule,
               tries => $tries );



  (Class or instance method) The method modifies a snapmirror relationship by altering its properties (like schedule,tries, throttle)
   Uses a CMode-CLI or CMode-ZAPI.


=over

=item Options

=over

=item C<< "source-path =>  $source_volume_object" >>

(Optional for class method, Not Applicable for instance method)
$volume_specification may be either a string specifying the full path
to the source volume, or it may be a NACL::(C|STask)::Volume with the
caveat that source-cluster must also be passed as an argument if the
Volume object's command_interface is a NACL::C::Vserver.


=item C<< "destination-path =>  $destination_volume_object" >>

(Required for class method, Not Applicable for instance method)
destination-path takes the same form as the source-path argument, but
for the destination volume (and possibly requiring destination-cluster.)

=item C<< "nacltask_verify => $nacltask_verify_boolean" >>

(Optional) If '0' (default), verification will not be performed.
 If '1', verification will be performed to ensure that property has changed succesfully.


=item C<< "source-cluster =>  $string" >>

(Optional) Name of cluster where the source volume resides.
This input applies only for CMode.


=item C<< "destination-cluster =>  $string" >>

(Optional) Name of cluster where the destination volume resides.
This input applies only for CMode


All other options supported by
L<< NACL::C::Snapmirror->modify|lib-NACL-C-Snapmirror-pm/modify >>
are supported by NACL::STask::Snapmirror->modify as well.
If Trying to Change throttle for relationship for which source and destination are on same cluster, then throttle will not change

=back

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification for snapmirror modify has failed.

=item C<NATE::BaseException>

This type of exception is thrown when filer mode is not supported.

=back

=back

=cut

sub modify {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {

            # component options which are overloaded
            'source-path'      => { type => OBJECT | SCALAR, optional => 1 },
            'destination-path' => { type => OBJECT | SCALAR, optional => 1 },

            # options which would be consumed by stask method itself apart
            # from passing the information to components.
            'source-cluster'      => { type => SCALAR, optional => 1 },
            'destination-cluster' => { type => SCALAR, optional => 1 },
            'nacltask_verify'     => { type => SCALAR, default  => 0 },
            $pkg_or_obj->_cleanup_validate_spec(),
        },
        allow_extra => 1,
    );

    # After the call to _common_validate_with(), %opts will have common params
    # like method-timeout, apiset_must, apiset_should, command_interface,
    # primary keys, polling_interval, stask specific params, other options
    # taken by NACL::C::Snapmirror->resume()
    my $ci = $opts{command_interface};
    $opts{'method-timeout'} ||= SNAPMIRROR_TIMEOUT;

    # Handle cleanup options
    my ( $opts_for_register, $register_for_cleanup, %opts_for_cleanup );
    $pkg_or_obj->_move_common_cleanup_opts(
        source              => \%opts,
        target              => \%opts_for_cleanup,
        nacltask_to_cleanup => \$register_for_cleanup
    );

    if ($register_for_cleanup) {
        $opts_for_register = $pkg_or_obj->_common_modify_for_cleanup( %opts_for_cleanup, %opts );
    }

    # The dependencies which need to be handled are different between 7M and CM
    # Hence we will have two backends for 7M and CM separately.
    if ( $ci->mode() eq '7Mode' ) {

        # 7mode support
        my $msg = '7Mode not supported currently';
        $Log->exit();
        NACL::Exceptions::InvalidChoice->throw($msg);

    } elsif ( $ci->mode() eq 'CMode' ) {

        # Invoke CM backend
        $pkg_or_obj->_modify_CM(%opts);

    } else {

        # Just to avoid open if-elsif
        NATE::BaseException->throw( "Filer mode '" . $ci->mode() . "' not supported." );
        $Log->exit();
    }

    # Register for cleanup if $nacltask_to_cleanup is set
    $pkg_or_obj->_register_for_cleanup( %{$opts_for_register} )
        if ($register_for_cleanup);

    $Log->exit();

} ## end sub modify

sub _modify_CM {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;
    my %modification_opts;
    my $verify = delete $opts{nacltask_verify};
    if ( exists $opts{throttle} ) {
        $modification_opts{throttle} = $opts{throttle};
    }
    if ( $opts{schedule} ) {
        $modification_opts{schedule} = $opts{schedule};
    }
    if ( exists $opts{tries} ) {
        $modification_opts{tries} = $opts{tries};
    }

    # Get the source-path and destination-path. Rest of the code
    # understands only these two inputs.
    # Also store the volume objects if passed.
    my ( $src_vol_obj, $dst_vol_obj );

    # source-path could be either object or path. if user had specified
    # volume, vserver, cluster names separately, then _common_validate_with()
    # which is overloaded in NACL::C::Snapmirror would have constructed a
    # path out of it.
    if ( ref $opts{'source-path'} ) {
        $src_vol_obj = $opts{'source-path'};
        $opts{'source-path'} = $pkg_or_obj->construct_volpath(
            vol_obj_or_volpath => $src_vol_obj,
            cluster_name       => delete( $opts{'source-cluster'} ),
            %opts
        );
    }

    if ( ref $opts{'destination-path'} ) {
        $dst_vol_obj = $opts{'destination-path'};
        $opts{'destination-path'} = $pkg_or_obj->construct_volpath(
            vol_obj_or_volpath => $dst_vol_obj,
            cluster_name       => delete( $opts{'destination-cluster'} ),
            %opts
        );
    }

    # Handle Job ID (In RR, job gets created for snapmirror modify)
    my $job_component = delete $opts{job_component};
    {

        # We want to know the job ID of the job performed by
        # SUPER::modify, whether or not our caller does.
        my $dummy;
        $job_component ||= \$dummy;
    }

    # Perform modify
    # Always make class method calls so that we don't have to bother
    # about skipping primary keys for instance method call
    my $pkg = ref $pkg_or_obj || $pkg_or_obj;
    try {
        $pkg->SUPER::modify( %opts, job_component => $job_component, );

    }
    catch NACL::APISet::Exceptions::ResponseException with {
        my $e = shift;

        # Unknown error
        $Log->exit();
        $e->throw();

    };
    $pkg_or_obj->last_job( ${$job_component} ) if ( ref $pkg_or_obj );

    #This is done if nacltask_verify is set to 1.
    if ($verify) {
        my %common_comp_opts;
        $pkg->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_comp_opts
        );
        $opts{'polling_interval'} ||= POLL_DELTA;

        if ( defined $$job_component ) {
            $$job_component->wait_on_job( %common_comp_opts );
        }

        try {
            $pkg->verify_state(
                command_interface  => $opts{command_interface},
                'destination-path' => $opts{'destination-path'},
                %modification_opts,
                %common_comp_opts
            );
        }
        catch NACL::Exceptions::VerifyFailure with {
            my $e = shift;
            my $line;
            my $error_string = $e->text;

            my $src_cluster = $pkg->source_cluster( %common_comp_opts, 'source-path' => $opts{'source-path'} );
            my $dst_cluster = $pkg->destination_cluster( %common_comp_opts, 'destination-path' => $opts{'destination-path'} );
            if (   ( $error_string =~ /throttle\s+$opts{throttle}\s+unlimited/ )
                && ( $src_cluster eq $dst_cluster ) )
            {
                $Log->debug(
                    "Warn : Trying to Change throttle for relationship for which source and destination are on same cluster. Throttle will not change"
                );
            } else {

                # Unknown error
                $Log->exit();
                $e->throw();
            }
        };

        $Log->debug("SnapMirror Modified Successfully ");
    }
    $Log->exit();

} ## end sub _modify_CM

=head2 resync

  NACL::STask::Snapmirror->resync(
      command_interface     =>  $command_interface,
      'source-path'         =>  $src_volume_specification,
      'destination-path'    =>  $dst_volume_specification,
      'nacltask_wait'       =>  $number,
      %other_options
  );

  (or)

  $snapmirror_object->resync(nacltask_wait  =>  0 );


(Class or Instance method) This method re-establishes a mirroring relationship between a source volume and a destination volume.
This method provides additional functionality like wait till a specified percentage of
transfer happens after resync.


=over

=item Options

=over


=item C<< "source-path =>  $src_volume_specification" >>

(Optional for class method, Not Applicable for instance method) Required in
class method call when $nacltask_wait value is > 0 and < 1. Source information
is required to compute difference in vol sizes so that specified percentage of
transfer can be done in background. $src_volume_specification may be either a
string specifying the full path to the source volume, or it may be a
NACL::(C|STask)::Volume with the caveat that source-cluster must also be passed
as an argument if the Volume object's command_interface is a NACL::C::Vserver.


=item C<< "destination-path =>  $dst_volume_specification" >>

(Required for class method, Not Applicable for instance method)
destination-path takes the same form as the source-path argument, but
for the destination volume (and possibly requiring destination-cluster.)


=item C<< "source-cluster =>  $string" >>

(Optional) Name of cluster where the source volume resides.
This input applies only for CMode.


=item C<< "destination-cluster =>  $string" >>

(Optional) Name of cluster where the destination volume resides.
This input applies only for CMode.

=item C<< nacltask_wait => 0|1 >>

(Optional, defaults to 1)
Specifies whether to wait till the data transfer completes.

Note that specifying a value between 0 & 1 is deprecated and is no longer
supported because of removal of support for real-time data from snapmirror show
(see burt 694025).

=item C<< nacltask_retry_tries_count => $tries_count >>

(Optional, defaults to 30)
The number of times to retry the resync command if it fails because
the snapshot is temporarily busy or stale catch entry.

Note that it is possible to make it not retry the delete command by setting
the value of this argument to 1.

=item C<< nacltask_retry_sleep_time => $sleep_time >>

(Optional, defaults to 10 seconds)
The time to sleep between retries of the resync command.

=item C<< nacltask_verify_health => 0|1 >>

(Optional, defaults to '1', Only for CMode)
If set to '1', verifies the health of the snapmirror.

=item C<< nacltask_retry_num_times_code_ran => \$variable >>

(Optional)
A reference to a scalar variable which gets filled in with the number
of times the resync command was invoked.

=item C<< nacltask_config_diff => 0|1 >>

(Optional, defaults to 0, applicable only for VDR relationships)
If set to 1, perform a config diff check. Internally uses L<config_diff>.

=item command_interface, apiset_must, apiset_should, method-timeout, etc.

All other options supported by
L<< NACL::C::Snapmirror->resync|lib-NACL-C-Snapmirror-pm/resync >>
are supported by NACL::STask::Snapmirror->resync as well.

=back

=item Exceptions

=over

=item C<NACL::Exceptions::InvalidChoice>

This type of exception is thrown when resync command is executed in 7mode.

=item C<NACL::C::Exceptions::Snapmirror::TransferInProgress>

This exception is thrown when an attempt is made to perform snapmirror operation when another operation is in progress.

=back

=back

=cut

sub resync {

    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {

            # component options which are overloaded
            'source-path'      => { type => OBJECT | SCALAR, optional => 1 },
            'destination-path' => { type => OBJECT | SCALAR, optional => 1 },

            # options which would be consumed by stask method itself apart
            # from passing the information to components.
            'source-cluster'                    => { type => SCALAR,    optional => 1 },
            'destination-cluster'               => { type => SCALAR,    optional => 1 },
            'nacltask_wait'                     => { type => SCALAR,    default  => 1 },
            'nacltask_retry_tries_count'        => { type => SCALAR,    default  => 30 },
            'nacltask_retry_sleep_time'         => { type => SCALAR,    default  => 10 },
            'nacltask_retry_num_times_code_ran' => { type => SCALARREF, optional => 1 },
            'nacltask_verify_health'            => { type => SCALAR,    default  => 1 },
            nacltask_config_diff                => { type => SCALAR,    default  => 0 },
        },
        allow_extra => 1,
    );

    # After the call to _common_validate_with(), %opts will have common params
    # like method-timeout, apiset_must, apiset_should, command_interface,
    # primary keys, polling_interval, stask specific params, other options
    # taken by NACL::C::Snapmirror->resync()
    my $ci = $opts{command_interface};

    $opts{'method-timeout'} ||= SNAPMIRROR_TIMEOUT;

    # The dependencies which need to be handled are different between 7M and CM
    # Hence we will have two backends for 7M and CM separately.
    my $obj;
    if ( $ci->mode() eq '7Mode' ) {

        # Invoke 7M backend
        $Log->trace("7Mode resync not supported");
        $Log->exit();
        NACL::Exceptions::InvalidChoice->throw("7Mode not supported currently.");
    } elsif ( $ci->mode() eq 'CMode' ) {

        # Invoke CM backend
        $obj = $pkg_or_obj->_resync_CM(%opts);
    } else {

        # Just to avoid open if-elsif
        $Log->exit();
        NATE::BaseException->throw( "Filer mode '" . $ci->mode() . "' not supported." );
    }
    $Log->exit();
    return $obj;
} ## end sub resync

sub _resync_CM {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;

    # save these for use after component resync()
    my $wait_for_resync  = delete( $opts{'nacltask_wait'} );
    my $polling_interval = delete( $opts{'polling_interval'} );
    my $verify_health    = delete( $opts{'nacltask_verify_health'} );
    my $config_diff      = delete $opts{nacltask_config_diff};

    # Need to wait if config diff needs to be checked
    $wait_for_resync ||= $config_diff;

    # Get the source-path and destination-path.
    # Also store the volume objects if passed.
    my ( $src_vol_obj, $dst_vol_obj );

    # source-path could be either object or path. if user had specified
    # volume, vserver, cluster names separately, then _common_validate_with()
    # which is overloaded in NACL::C::Snapmirror would have constructed a
    # path out of it.
    if ( ref $opts{'source-path'} ) {
        $src_vol_obj = $opts{'source-path'};
        $opts{'source-path'} = $pkg_or_obj->construct_volpath(
            vol_obj_or_volpath => $src_vol_obj,
            cluster_name       => delete( $opts{'source-cluster'} ),
            %opts
        );
    }
    if ( ref $opts{'destination-path'} ) {
        $dst_vol_obj = $opts{'destination-path'};
        $opts{'destination-path'} = $pkg_or_obj->construct_volpath(
            vol_obj_or_volpath => $dst_vol_obj,
            cluster_name       => delete( $opts{'destination-cluster'} ),
            %opts
        );
    }

    # Handle Job ID
    my $job_component = delete $opts{job_component};
    {

        # We want to know the job ID of the job performed by
        # SUPER::resync, whether or not our caller does.
        my $dummy;
        $job_component ||= \$dummy;
    }

    #calculate difference in used size between source and destination volumes
    if ( $wait_for_resync != 0 && $wait_for_resync < 1 ) {
        $Log->warn( 'Specifying a nacltask_wait value between 0 & 1 '
                . 'is deprecated because it is no longer possible to get '
                . 'real-time snapmirror data; see 694025. (the value provided '
                . "was $wait_for_resync) The only currently supported values "
                . 'for nacltask_wait are 0 and 1' );

        # Set to 0 because the user would have wanted it to return
        # while data is still being transferred
        $wait_for_resync = 0;
    }

    my %retry_opts;
    $pkg_or_obj->_move_retry_params(
        source => \%opts,
        target => \%retry_opts
    );

    # Perform resync
    # Always make class method calls so that we don't have to bother
    # about skipping primary keys for instance method call
    my $apiset;
    my $obj;
    my $pkg = ref $pkg_or_obj || $pkg_or_obj;
    nacl_method_retry(
        code => sub {
            $obj = $pkg->SUPER::resync(
                %opts,
                job_component => $job_component,
                apiset_chosen => \$apiset
            );
        },
        exceptions => [qw(NACL::C::Exceptions::TransferInProgress)],
        %retry_opts,
    );

    # If this is invoked as an instance method, then we should not return a new object.
    $obj = $pkg_or_obj if ( ref $pkg_or_obj );

    delete $opts{foreground};
    delete $opts{force};
    delete $opts{type};
    $pkg_or_obj->last_job( ${$job_component} ) if ( ref $pkg_or_obj );
    if ( $wait_for_resync == 1 ) {
        $opts{polling_interval} = $polling_interval
            if ( defined $polling_interval );
        Tharn::snooze(5);

        #Call wait_for_initialization() if nacltask_wait is set to 1
        $pkg->wait_for_resync(
            %opts,
            job                    => $$job_component,
            nacltask_verify_health => $verify_health,
        );
    }

    if ( $opts{nacltask_verify} ) {
        $pkg->_verify_states(%opts);
    }

    if ($config_diff) {
        my ( %primary_keys, %common_opts );
        $pkg->_copy_primary_keys(
            source => \%opts,
            target => \%primary_keys,
        );
        $pkg->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_opts,
        );
        $pkg->config_diff( %primary_keys, %common_opts );
    }

    #Add destination_ci to created stask object
    $obj->{destination_ci} = $opts{destination_ci} if ( defined $opts{destination_ci} );

    $Log->exit();
    return $obj;
} ## end sub _resync_CM

=head2 wait_for_resync

Waits for the snapmirror resync operation to complete. This method accepts the
same arguments as L<wait_for_initialization>.

=cut

sub wait_for_resync {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->wait_for_initialization( @args, 'last-transfer-type' => 'resync' );
    $Log->exit();
}

=head2 abort

    NACL::STask::Snapmirror->abort(
        command_interface       =>  $ci,
        'source-path'           =>  $src_path_specification,
        'destination-path'      =>  $dst_path_specification,
        %other_options);

                               (or)

    $snapmirror->abort(%other_opts);

Performs a 'snapmirror abort' to abort an active snapmirror transfer or to
remove previous checkpoint and verifies the state of the relationship if
'nacltask_verify' task specific parameter is set.

Supports CMode CLI/ZAPI and 7Mode CLI.

=over

=item Options

=over

=item C<< 'source-path' => $source_path >>

(Optional for class method, Not Applicable for instance method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< 'destination-path' => $path_specification >>

(Required for class method, Not Applicable for instance method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< nacltask_verify => 0 | 1 >>

(Optional defaults to 0)

If set as 1, then verify that the relationship has been broken. Throws a
L<NACL::Exceptions::VerifyFailure|lib-NACL-Exceptions-VerifyFailure-pm>
exception if it is not.

It is also possible to verify separately that the relationship has been
broken by using the
L<taskverify_abort|lib-NACL-STask-Snapmirror-pm/taskverify_abort>
method.

In SFR operation where the state value is '-', it is recommended not
to enable this option.

=item C<< nacltask_if_aborted=> $action  >>

(Optional) If 'die' (default), fails if snapmirror could not be aborted.
 If set to 'pass' it will just log a debug message saying relationship could not be aborted.

=item C<< nacltask_wait => 0 | 1 >>

(Optional, defaults to 1, is applicable only if the destination is an infinite volume)
If the destination is an infinite volume, then the abort operation is
performed through a background job. If this argument is set to 1 (the default),
it will wait till the job completes.

=item C<< 'source-cluster' =>  $string >>

(Optional) Name of cluster where the source volume resides.
This input applies only for CMode.


=item C<< 'destination-cluster' =>  $string >>

(Optional) Name of cluster where the destination volume resides.
This input applies only for CMode


All other options supported by
L<< NACL::C::Snapmirror->abort|lib-NACL-C-Snapmirror-pm/abort >>
are supported by NACL::STask::Snapmirror->abort as well.

=back

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification for abort snapmirror
relationship has failed ie. actual snamirror relationship state does not
match with the expected state after aborting snapmirror relationship.

=item C<NACL::C::Exceptions::Snapmirror::NoTransferToAbort>

This type of exception is thrown when there are no transfer in progress to abort.

=back

=back

=cut

sub abort {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            'nacltask_verify' => { type => SCALAR, default => 0 },
            'nacltask_if_aborted' => $pkg_or_obj->_if_action_completed_already_validate_spec( name => 'nacltask_if_aborted' ),
            nacltask_wait => { type => SCALAR, default => 1 },
        },
    );

    # After the call to _common_validate_with(), %opts will have common params
    # like method-timeout, apiset_must, apiset_should, command_interface,
    # primary keys, polling_interval, stask specific params, other options
    # taken by NACL::C::Snapmirror->abort()
    my $verify     = delete $opts{nacltask_verify};
    my $if_aborted = delete $opts{nacltask_if_aborted};
    my $wait       = delete $opts{nacltask_wait};

    # Perform abort
    # Always make class method calls so that we don't have to bother
    # about skipping primary keys for instance method call
    my $pkg = $pkg_or_obj->get_package_name();

    # Handle Job ID
    my $job_component = delete $opts{job_component};
    {

        # We want to know the job ID of the job performed by
        # SUPER::abort, whether or not our caller does.
        my $dummy;
        $job_component ||= \$dummy;
    }
    my $need_to_return;

    my %common_comp_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_comp_opts
    );

    try {
        $pkg->SUPER::abort( %opts, job_component => $job_component );

        if ( $wait || $verify ) {
            if ( defined $$job_component ) {
                my $job = ${$job_component};
                $pkg_or_obj->last_job($job) if ( ref $pkg_or_obj );

                try {
                    $job->wait_on_job(%common_comp_opts);
                }
                catch NACL::Exceptions::JobFailed with {
                    $pkg_or_obj->_abort_exception_handler(@_);
                };
            } else {
                my @target_states       = qw(idle quiesced outofsync);
                my @intermediate_states = qw(aborting transferring preparing
                    finalizing quiescing);
                local $" = q{|};
                my $target_states_re       = qr/^(?:@target_states)$/i;
                my $intermediate_states_re = qr/^(?:@intermediate_states)$/i;

                $pkg->wait_on_attribute(
                    "destination-path"        => $opts{"destination-path"},
                    attribute_to_check        => 'status',
                    till_value                => [$target_states_re],
                    valid_intermediate_states => [$intermediate_states_re],
                    %common_comp_opts,
                );
            }
        }
    }
    catch NACL::C::Exceptions::Snapmirror::NoTransferToAbort with {
        my $e = shift;
        if ( $if_aborted =~ /pass/i ) {
            $Log->trace( 'No transfer in progress, not failing since ' . "'if_aborted' was passed as 'pass'" );
            $need_to_return = 1;
        } else {
            $Log->exit();
            $e->throw();
        }
    };

    if ($need_to_return) {
        $Log->exit();
        return;
    }

    # Verify if the snapmirror state is Idle after abort
    # This is done if nacltask_verify is set to 1.
    if ($verify) {

        # Needed as it could take a couple of seconds before the transfer
        # gets aborted.
        Tharn::snooze(5);

        $pkg->taskverify_abort( %common_comp_opts, 'destination-path' => $opts{'destination-path'} );
    } ## end if ($verify)

    $Log->exit();
} ## end sub abort

=head2 taskverify_abort

    NACL::STask::Snapmirror->taskverify_abort(
        command_interface  => $cmd_int,
        'source-path'      => $source_path,
        'destination-path' => $destination_path
    );

                           or

    $snapmirror_obj->taskverify_abort();

(Class or Instance method)

This method verifies that the snapmirror transfer have been aborted.
If it is not, then it throws a
L<NACL::Exceptions::VerifyFailure|lib-NACL-Exceptions-VerifyFailure-pm>
exception.

Supports CMode CLI/ZAPI and 7Mode CLI.

=over

=item Options

=over

=item C<< command_interface => $command_interface >>

(Required for class method, Not Applicable for Instance method)

See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must => $ruleset >>

(Optional)

=item C<< apiset_should => $ruleset >>

(Optional)

=item C<< 'method-timeout' => $timeout >>

(Optional)

=item C<< 'source-path' => $path_specification >>

(Optional for class method, Not Applicable for Instance method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< 'destination-path' => $path_specification >>

(Required for class method, Not Applicable for Instance method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< 'source-cluster' =>  $string >>

(Optional) Name of cluster where the source volume resides.
This input applies only for CMode.

=item C<< 'destination-cluster' =>  $string >>

(Optional) Name of cluster where the destination volume resides.
This input applies only for CMode.

=item C<< "polling_interval =>  $polling_value" >>

(Optional) This is the interval (in seconds) after which the value of
the attributes will be polled. If not specified it defaults to 5 seconds.
This option is applicable only for CMode.

=back

=item Exceptions

=over

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification for snapmirror abort has failed or
when state or status of snapmirror has reached an unexpected state.

=back

=back

=cut

sub taskverify_abort {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );
    my %opts = $pkg_or_obj->_common_validate_with( params => \@args );

    if ( $opts{command_interface}->mode() eq 'CMode' ) {
        $opts{'method-timeout'}   ||= 300;
        $opts{'polling_interval'} ||= 5;

        $pkg_or_obj->monitor(
            nacltask_wait_for_state  => [ qr/Uninitialized/i, qr/Snapmirrored/i, ],
            nacltask_wait_for_status => [ qr/Idle/i,          qr/Quiesced/i ],
            %opts,
        );
        my %common_opts;
        $pkg_or_obj->_copy_common_component_params_with_ci(
            source => \%opts,
            target => \%common_opts
        );

        # check if last-transfer-error field is Transfer aborted/Transfer was aborted
        my $state = NACL::CS::Snapmirror->fetch(
            filter => { 'destination-path' => $opts{'destination-path'} },
            %common_opts,
        );

        if ( $state->last_transfer_error !~ /Transfer aborted|Transfer was aborted/i ) {
            $Log->exit();
            NACL::Exceptions::VerifyFailure->throw(
                "Snapmirror transfer not aborted and last-transfer-error is " . $state->last_transfer_error,
                unexpected_values => { last_transfer_error => $state->last_transfer_error }
            );
        }
    } else {

        my $filter = $pkg_or_obj->_hash_move(
            source => \%opts,
            move   => [ $pkg_or_obj->get_primary_keys() ]
        );

        # Not using verify_state at the moment because it currently doesn't
        # support special characters
        my $cs = NACL::CS::Snapmirror->fetch( %opts, filter => $filter, );

        my $state = $cs->state();
        if ( $state !~ /(Uninitialized)|(Snapmirrored)/i ) {
            $Log->exit();
            NACL::Exceptions::VerifyFailure->throw(
                "State should have been 'Uninitialized' or " . "'Snapmirrored' but was instead '$state'",
                unexpected_values => { state => $state } );
        }

        my $status = $cs->status();
        if ( $status !~ /Idle/i ) {
            $Log->exit();
            NACL::Exceptions::VerifyFailure->throw( "Status should have been 'Idle' but was instead '$status'",
                unexpected_values => { status => $status } );
        }
    }

    $Log->exit();
} ## end sub taskverify_abort

# Fix (Temporary?) to avoid any backward incompatible failures
# Burt642063
sub _ls_set {
    my ( $pkg_or_obj, @args ) = @_;
    $Log->enter();
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            _method       => { type => SCALAR },
            nacltask_wait => { type => SCALAR, default => 1 },
        },
        ignore_primary_keys => 1,
        allow_extra         => 1
    );
    my $method = delete $opts{_method};
    my $dummy;
    my %common_options = ();
    $pkg_or_obj->_copy_common_component_params(
        source => \%opts,
        target => \%common_options,
    );
    my $job_component = delete $opts{job_component} || \$dummy;
    delete $opts{'destination-path'};
    my $invoke_method = 'SUPER::' . $method;

    if ( ref $pkg_or_obj ) {
        $Log->warn( "$method method should be invoked "
                . "as class method, as it operates on multiple destinations. "
                . "Each snapmirror object models a relationship between a source "
                . "and unique destination, hence using instance method call is "
                . "not the right way of invocation" );
        my $src_path = $pkg_or_obj->source_path;
        my $ci       = $pkg_or_obj->command_interface;
        unless ( defined $src_path ) {
            $Log->exit();
            NATE::Exceptions::Argument->throw( 'source-path attribute is ' . 'not set in the object' );
        }

        # Make class method call
        my $pkg = $pkg_or_obj->get_package_name();
        $pkg->$invoke_method(
            'source-path'     => $src_path,
            command_interface => $ci,
            job_component     => $job_component,
            %opts
        );
    } else {
        $pkg_or_obj->$invoke_method( job_component => $job_component, %opts );
    }

    $Log->trace( "Wait for completion of the job as the initialize-ls-set will" . " run in background" );
    $$job_component->wait_on_job( %common_options, ) if ( $opts{nacltask_wait} );

    $Log->exit;
}

# POD is not written intentionally as these methods are added just to
# avoid failures to existing script which would have worked currently as
# instance method call on STask object.

sub initialize_ls_set {
    my ( $pkg_or_obj, %orig_args ) = @_;
    $Log->enter();
    $pkg_or_obj->_ls_set( _method => 'initialize_ls_set', %orig_args );
    $Log->exit;
}

sub update_ls_set {
    my ( $pkg_or_obj, %orig_args ) = @_;
    $Log->enter();
    $pkg_or_obj->_ls_set( _method => 'update_ls_set', %orig_args );
    $Log->exit;
}

# Overridden version of _common_validate_with which does the following:
# 1. Adds source-path and destination-path to the additional_spec, both of
#    which are allowed to be either an object or a scalar
# 2. Converts options passed in into source-path and destination-path strings.
#    If a combination of volume+vserver+cluster are provided, then
#    NACL::C::Snapmirror::_common_validate_with does the conversion. If
#    objects are passed for either/both of source-path/destination-path then
#    the method _construct_path is invoked to perform the conversion. This
#    method also takes care of validating that the appropriate objects
#    are passed in (C/STask Volume/Qtree objects)
# 2. Defaults allow_extra to 1
# 3. Defaults method-timeout to be SNAPMIRROR_TIMEOUT
sub _common_validate_with {
    my ( $pkg_or_obj, %orig_opts ) = @_;

    # If the destination-path is of type "NACL::C::Volume"
    # then fetch the command_interface from the destination-path object.
    my %params = @{ $orig_opts{params} };

    if (   !$params{command_interface}
        && ( defined $params{'destination-path'} )
        && ( !ref $pkg_or_obj ) )
    {
        my $dest_obj = $params{'destination-path'};
        if ( ( blessed $dest_obj) && $dest_obj->isa('NACL::C::Volume') ) {
            push( @{ $orig_opts{params} }, "command_interface", $dest_obj->command_interface() );
        }
    }

    # Overloaded component options
    foreach my $src_or_dest (qw(source-path destination-path)) {
        $orig_opts{additional_spec}{$src_or_dest} ||= { type => SCALAR | OBJECT, optional => 1 };

    }

    $orig_opts{allow_extra} = 1 if ( !exists $orig_opts{allow_extra} );

    my %opts = $pkg_or_obj->SUPER::_common_validate_with(%orig_opts);
    if ( ref $opts{"source-path"} && !$opts{source_ci} ) {
        $opts{source_ci} = $opts{"source-path"}->command_interface();
    }
    $pkg_or_obj->_construct_path( opts => \%opts );
    $opts{'method-timeout'} //= SNAPMIRROR_TIMEOUT;

    $Log->exit();
    return %opts;
} ## end sub _common_validate_with

# Construct a string path for the source-path and destination-path
# arguments when provided objects (validation of the type of objects
# is also performed)
sub _construct_path {
    $Log->enter();
    my ( $pkg_or_obj, %args ) = @_;
    my $opts = $args{opts};
    $opts->{command_interface} = $pkg_or_obj->command_interface() if ( !defined $opts->{command_interface} );
    if ( $opts->{command_interface}->mode() eq 'CMode' ) {
        $pkg_or_obj->_construct_path_CM( opts => $opts );
    } else {
        $pkg_or_obj->_construct_path_7M( opts => $opts );
    }
    $Log->exit();
} ## end sub _construct_path

sub _construct_path_CM {
    $Log->enter();
    my ( $pkg_or_obj, %args ) = @_;
    my $opts = $args{opts};
    foreach my $src_or_dest (qw(source destination)) {
        my $flag = 0;
        if ( blessed( $opts->{"$src_or_dest-path"} ) ) {
            $flag = 1 if ( $src_or_dest eq "source" );
            if ( $opts->{"$src_or_dest-path"}->isa('NACL::C::Vserver') ) {
                $opts->{"$src_or_dest-path"} = $pkg_or_obj->_construct_vserver_path(
                    'vserver_obj' => $opts->{"$src_or_dest-path"},
                    %$opts,
                );
            } elsif ( $opts->{"$src_or_dest-path"}->isa('NACL::C::Volume') ) {
                $opts->{"$src_or_dest-path"} = $pkg_or_obj->construct_volpath(
                    vol_obj_or_volpath => $opts->{"$src_or_dest-path"},
                    cluster_name       => delete( $opts->{"$src_or_dest-cluster"} ),
                    is_source          => $flag,
                    %$opts
                );
            } else {
                NATE::Exceptions::Argument->throw( "Invalid object passed as source/destination-path.  "
                        . "Expected a NACL::C|STask::Volume or NACL::C|STask::Vserver, "
                        . "but was passed a "
                        . ref( $opts->{"$src_or_dest-path"} )
                        . "." );
            }
        }
    }

    $Log->exit();
} ## end sub _construct_path_CM

sub _construct_path_7M {
    $Log->enter();
    my ( $pkg_or_obj, %args ) = @_;
    my $opts            = $args{opts};
    my $throw_exception = sub {
        my %opts = @_;
        my $ref  = ref $opts{obj};
        NATE::Exceptions::Argument->throw( 'The argument ' . "'"
                . $opts{argument} . "'"
                . "was provided as an object of type '$ref' but should "
                . 'have been provided as a Volume or Qtree Component or '
                . 'STask object' );
    };

    foreach my $src_or_dest (qw(source-path destination-path)) {
        if ( ref $opts->{$src_or_dest} ) {
            my $obj = $opts->{$src_or_dest};
            if (   !$obj->isa('NACL::C::Volume')
                && !$obj->isa('NACL::C::VolumeQtree') )
            {
                $Log->exit();
                $throw_exception->(
                    argument => $src_or_dest,
                    obj      => $obj
                );
            } elsif ( $obj->isa('NACL::C::Volume') ) {
                $opts->{$src_or_dest} = $pkg_or_obj->_construct_path_from_vol_obj_7m( vol_obj => $obj );
            } else {
                $opts->{$src_or_dest} = $pkg_or_obj->_construct_path_from_qtree_obj_7m( qtree_obj => $obj );
            }
        }
    }

    $Log->exit();
} ## end sub _construct_path_7M

sub _construct_volpath {
    $Log->enter();
    my $self = shift;
    $Log->warn("_construct_volpath is deprecated. Kindly use - construct_volpath");
    $Log->exit();
    return $self->construct_volpath(@_);
} ## end sub _construct_volpath

# 7Mode only
# Given a volume object determine the volume path
# <hostname>:<vol-name>
sub _construct_path_from_vol_obj_7m {
    $Log->enter();
    my $pkg_or_obj  = shift;
    my %opts        = validate( @_, { vol_obj => { type => OBJECT, isa => 'NACL::C::Volume' } } );
    my $vol_obj     = $opts{vol_obj};
    my $vol_ci      = $vol_obj->command_interface();
    my $source_path = $vol_ci->hostrec()->hostname() . ':' . $vol_obj->volume();
    $Log->exit();
    return $source_path;
} ## end sub _construct_path_from_vol_obj_7m

# 7Mode only
# Given a Qtree object, determine the qtree path
# <hostname>:<qtree-path>
sub _construct_path_from_qtree_obj_7m {
    $Log->enter();
    my $pkg_or_obj  = shift;
    my %opts        = validate( @_, { qtree_obj => { type => OBJECT, isa => 'NACL::C::VolumeQtree' } } );
    my $qtree_obj   = $opts{qtree_obj};
    my $qtree_ci    = $qtree_obj->command_interface();
    my $source_path = $qtree_ci->hostrec()->hostname() . ':' . '/vol/' . $qtree_obj->volume() . '/' . $qtree_obj->qtree();
    $Log->exit();
    return $source_path;
} ## end sub _construct_path_from_qtree_obj_7m

=head2 check

  NACL::STask::Snapmirror->check(
      command_interface     =>  $command_interface,
      'source-path'         =>  $src_volume_specification,
      'destination-path'    =>  $dst_volume_specification,
      'nacltask_wait'       =>  $number,
      %other_options
  );


  $sm->check(nacltask_if_mismatch   => $action, #default "die");


  (Class or instance method) Performs a "snapmirror check" to verify the
integrity of transferred data. If the checker detects an inconsistent mirror,
anomalies are recorded in a consolidated log file that is specified as a node
and absolute pathname.
  If checker ran successfully and found inconsistent mirrors and if
nacltask_if_mismatch is pass, check will log a message saying that
inconsistencies are present. If nacltask_if_mismatch is die,  exception is thrown.
  If checker ran successfully and no inconsistencies found, nothing will be returned.
  If nacltask_wait is 0, user is expected to specify job_component => \$job_component
during method call if he wishes to monitor the job completion. This
$job_component can be later used with wait_on_job() to wait for the check to complete.
  Uses a CMode-CLI or CMode-ZAPI.


=over

=item Options

=over


=item C<< "source-path =>  $src_volume_specification" >>

(Optional for class method, Not Applicable for instance method)
$src_volume_specification may be either a string specifying the full path
to the source volume, or it may be a NACL::(C|STask)::Volume with the
caveat that source-cluster must also be passed as an argument if the
Volume object's command_interface is a NACL::C::Vserver.


=item C<< "destination-path =>  $dst_volume_specification" >>

(Required for class method, Not Applicable for instance method)
destination-path takes the same form as the source-path argument, but
for the destination volume (and possibly requiring destination-cluster.)


=item C<< "source-cluster =>  $string" >>

(Optional) Name of cluster where the source volume resides.
This input applies only for CMode.


=item C<< "destination-cluster =>  $string" >>

(Optional) Name of cluster where the destination volume resides.
This input applies only for CMode

=item C<< "nacltask_wait =>  $number" >>

(Optional) A number 0 or 1 that determines if the check should take place in the
background. 1 means to return after the check has completed and 0 means to
return immediately. If nacltask_wait is 0, inconsistencies are not caught.
The default is 1.

=item C<< "nacltask_if_mismatch => $action" >>

(Optional) If 'pass', exits quietly, log a message saying that inconsistencies
are present. If "die",(default) NACL::Exceptions::DataInconsistent exception is
thrown when inconsistencies found. Node name and file path can be obtained from
the exception object. If check fails due to some other reason, then
NATE::BaseException is thrown with error string.

=item C<< nacltask_snaptags_retry_time => $sleep_time >>

(Optional, defaults to 60 seconds)
The total time to retry snaptags_test command for the snapshot to come in place.
snaptags_test will be retried at an interval of 15 seconds for a total time
of nacltask_snaptags_retry_time seconds.

=item C<< nacltask_verify_health >>

(Optional, defaults to '0')
If set to '1', verifies the health of the snapmirror relation after the checker is done.

=item command_interface, apiset_must, apiset_should, method-timeout, etc.

All other options supported by
L<< NACL::C::Snapmirror->check|lib-NACL-C-Snapmirror-pm/check >>
are supported by NACL::STask::Snapmirror->check as well.

=back

=item Exceptions

=over

=item C<NATE::BaseException>

This type of exception is thrown when filer mode is not supportedi or when Snaptags_test failed before running checker .

=item C<NACL::Exceptions::InvalidChoice>

This type of exception is thrown when an invalid option is passed.

=back

=back

=cut

sub check {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {

            # component options which are overloaded
            'source-path'      => { type => OBJECT | SCALAR, optional => 1 },
            'destination-path' => { type => OBJECT | SCALAR, optional => 1 },

            # options which would be consumed by stask method itself apart
            # from passing the information to components.
            'source-cluster'               => { type => SCALAR, optional => 1 },
            'destination-cluster'          => { type => SCALAR, optional => 1 },
            'nacltask_if_mismatch'         => $pkg_or_obj->_if_action_completed_already_validate_spec(),
            'nacltask_wait'                => { type => SCALAR, default  => 1 },
            'nacltask_snaptags_retry_time' => { type => SCALAR, default  => '60' },
            nacltask_verify_health         => { type => SCALAR, default  => '0' },
        },
        allow_extra => 1,
    );

    # After the call to _common_validate_with(), %opts will have common params
    # like method-timeout, apiset_must, apiset_should, command_interface,
    # primary keys, polling_interval, stask specific params, other options
    # taken by NACL::C::Snapmirror->check()
    my $ci = $opts{command_interface};
    $opts{'method-timeout'} ||= SNAPMIRROR_TIMEOUT;

    # The dependencies which need to be handled are different between 7M and CM
    # Hence we will have two backends for 7M and CM separately.
    if ( $ci->mode() eq '7Mode' ) {

        # 7mode support
        my $msg = '7Mode not supported currently';
        $Log->exit();
        NACL::Exceptions::InvalidChoice->throw($msg);

    } elsif ( $ci->mode() eq 'CMode' ) {

        # Invoke CM backend
        $pkg_or_obj->_check_CM(%opts);

    } else {

        # Just to avoid open if-elsif
        NATE::BaseException->throw( "Filer mode '" . $ci->mode() . "' not supported." );
        $Log->exit();
    }
    $Log->exit();
} ## end sub check

sub _check_CM {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;

    # Get the source-path and destination-path. Rest of the code
    # understands only these two inputs.
    # Also store the volume objects if passed.
    my ( $src_vol_obj, $dst_vol_obj );
    my $nacltask_wait                = delete( $opts{'nacltask_wait'} );
    my $nacltask_if_mismatch         = delete( $opts{'nacltask_if_mismatch'} );
    my $nacltask_snaptags_retry_time = delete( $opts{'nacltask_snaptags_retry_time'} );
    my $verify_health                = delete $opts{'nacltask_verify_health'};

    # source-path could be either object or path. if user had specified
    # volume, vserver, cluster names separately, then _common_validate_with()
    # which is overloaded in NACL::C::Snapmirror would have constructed a
    # path out of it.
    if ( ref $opts{'source-path'} ) {
        $src_vol_obj = $opts{'source-path'};
        $opts{'source-path'} = $pkg_or_obj->construct_volpath(
            vol_obj_or_volpath => $src_vol_obj,
            cluster_name       => delete( $opts{'source-cluster'} ),
            %opts
        );
    }

    if ( ref $opts{'destination-path'} ) {
        $dst_vol_obj = $opts{'destination-path'};
        $opts{'destination-path'} = $pkg_or_obj->construct_volpath(
            vol_obj_or_volpath => $dst_vol_obj,
            cluster_name       => delete( $opts{'destination-cluster'} ),
            %opts
        );
    }

    # Handle Job ID
    my $job_component = delete $opts{job_component};
    {

        # We want to know the job ID of the job performed by
        # SUPER::check, whether or not our caller does.
        my $dummy;
        $job_component ||= \$dummy;
    }

    my $sm_fetch = NACL::CS::Snapmirror->fetch(
        command_interface => $opts{command_interface},
        filter            => { 'destination-path' => $opts{'destination-path'} },
    );

    # If the relationship type is XDP, then we have to do some some special
    # operations on the snapshot that is being checked against.
    my $ci = $opts{command_interface};
    if ( $ci->get_version_manager()->is_promoted() ) {
        $Log->warn( " The snaptags_test cannot be done for promoted_builds as the" . " test privilege is not present in those builds" );
    } else {
	# below if condition is restricted to only a 'XDP' type snapmirror relationship (ONTAP-ONTAP)
	# and is not applicable for ONTAP-AVA scenario
        if ( $sm_fetch->type() eq 'XDP' && ( !$opts{destination_ci} ) ) {

            # If the 'target-snapshot' has been passed, then it has to be locked.
            # Else we have to get the newest snapshot from the CS object that we
            # fetched.
            # There are no cdefs for 'snaptags_test' command.
            my ( $vserver, $volume, $path, $snapshot );

            # Use a static call to is_granular to determine if we need to
            # derive the actual volume path:
            if ( NACL::STask::Snapmirror->is_granular(%opts) ) {
                $path = $pkg_or_obj->volume_paths_from_cg(%opts);
            } else {
                $path = $opts{'destination-path'};
            }

            # Get vserver and volume from the path:
            if (   $path =~ m{^(?:[^:/]+)://(.*)/(.*)$}
                || $path =~ m{^(.*):(.*)$} )
            {
                $vserver = $1;
                $volume  = $2;
            } else {
                NATE::Exceptions::Argument->throw( "Invalid path for snaptags_test: $path" );
            }

            # We only need to actually tag a snapshot if this is *not* a
            # MetaWAFL volume. For now, we determine this by looking for
            # the existence of a state attribute "default-extent-size".
            my $is_metawafl;
            try {
                my $volobj = NACL::C::Volume->new(
                    command_interface => $opts{command_interface},
                    volume            => $volume,
                    vserver           => $vserver
                );
                my $value = $volobj->get_one_state_attribute('default-extent-size');

                $is_metawafl = ( $value && ( $value ne '-' ) ) ? 1 : 0;
            }
            otherwise {

                # If we saw an error, for now just assume that it is because
                # the system doesn't yet know about MetaWAFL and the
                # "default-extent-size" attribute.
                $is_metawafl = 0;
            };

            # Only create the tag if this is not a MetaWAFL volume:
            if ( !$is_metawafl ) {
                if ( $opts{'target-snapshot'} ) {
                    $snapshot = $opts{'target-snapshot'};
                } else {
                    $snapshot = $sm_fetch->newest_snapshot();
                }

                # We need to know on which node the volume is residing.
                my $vol_node = NACL::STask::Volume->get_local_node_name(
                    command_interface => $opts{command_interface},
                    volume            => $volume,
                    vserver           => $vserver,
                );

                my $apiset = $ci->get_7m_or_nodescope_apiset();
                my $cmd    = "snaptags_test -s -v $volume -S $snapshot " . "VOPL_owner_snapmirror.checkInProgress:" . ",snapmirror,5";
                my ( $raw_output, $snap_found );
                my $count       = 0;
                my $retry_count = ceil( $nacltask_snaptags_retry_time / 15 );
                while ( $count <= $retry_count ) {
                    $raw_output = $apiset->execute_command(
                        'nodescope-node-name' => $vol_node,
                        command               => $cmd,
                    );
                    if ( ( grep( /SUCCESS/, $raw_output ) ) ) {
                        $snap_found = 1;
                        last;
                    }
                    $count++;
                    last if ( ( $count > $retry_count ) || ( grep( /snaptags_test not found/, $raw_output ) ) );
                    Tharn::sleep(15);
                }
                if ( !$snap_found ) {
                    NATE::BaseException->throw( 'Snaptags_test failed before running checker.' );
                }
            }
        }
    }

    # Perform check
    # Always make class method calls so that we don't have to bother
    # about skipping primary keys for instance method call

    my $pkg = ref $pkg_or_obj || $pkg_or_obj;
    $pkg->SUPER::check( %opts, job_component => $job_component );

    $pkg_or_obj->job_component_check( ${$job_component} );

    if ($nacltask_wait) {
        $pkg->wait_for_check(
            %opts,
            'destination-path'     => $opts{'destination-path'},
            job_component          => $$job_component,
            nacltask_if_mismatch   => $nacltask_if_mismatch,
            nacltask_verify_health => $verify_health,
        );
    }
    $Log->exit();
} ## end sub _check_CM

=head2 wait_for_check

    $snapmirror->wait_for_check(
        nacltask_if_mismatch   => $action,
    );


    NACL::STask::Snapmirror->wait_for_check(
      command_interface  =>  $command_interface,
      'destination-path' =>  $destination_path,
      job_component      =>  $job_component,
    );

(Class or instance method) Wait for the Snapmirror check to complete. If a job
for checking this Snapmirror is known then this waits for that job.
At the end of check operation, the health of the relation is checked and exception
is thrown if the health is false.

=over

=item Options

=over

=item C<< "command_interface=>$command_interface" >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< "destination-path=>$destination_path" >>

(Required for class method, Not Applicable for instance method)
The destination path of the snapmirror relationship.

=item C<< "job_component=>$job_component" >>

(Required for class method, Not Applicable for instance method) A
NACL::C::Job object representing the C-mode Job used to check Snapmirror. This
is not required when called as in instance method, provided, check is also
called as an instance method.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=item Exceptions

=over

=item C<NATE::BaseException>

This type of exception is thrown when filer mode is not supported or snapmirror check failed due to some other reason or
when no snapshots are found tagged for checker.

=item C<NATE::Exceptions::Argument>

This type of exception is thrown when job_component is missing.

=item C<NACL::Exceptions::DataInconsistent>

This type of exception is thrown when snapmirror check found inconsistent mirror.

=item C<NACL::Exceptions::InvalidChoice>

This type of exception is thrown when an invalid option is specified as mode.

=back

=back

=cut

sub wait_for_check {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component          => { type => OBJECT, optional => 1, isa => 'NACL::C::Job' },
            nacltask_verify_health => { type => SCALAR, default  => 0 },
        },
        allow_extra => 1,
    );

    # After the call to _common_validate_with(), %opts will have common params
    # like method-timeout, apiset_must, apiset_should, command_interface,
    # primary keys, job and polling_interval
    my $ci = $opts{command_interface};
    $opts{'method-timeout'} ||= SNAPMIRROR_TIMEOUT;

    if ( $ci->mode() eq '7Mode' ) {

        # Invoke 7M backend
        $Log->trace("7Mode not supported");
        $Log->exit();
        NACL::Exceptions::InvalidChoice->throw("7Mode not supported currently.");
    } elsif ( $ci->mode() eq 'CMode' ) {

        # Invoke CM backend
        $pkg_or_obj->_wait_for_check_CM(%opts);
    } else {

        # Just to avoid open if-elsif
        NATE::BaseException->throw( "Filer mode '" . $ci->mode() . "' not supported." );
    }
    $Log->exit();
} ## end sub wait_for_check

# CMode backend for wait_for_check
sub _wait_for_check_CM {

    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;
    $opts{job_component} ||= $pkg_or_obj->job_component_check();
    my $verify_health = delete $opts{'nacltask_verify_health'};

    unless ( defined $opts{job_component} ) {
        $Log->exit();
        NATE::Exceptions::Argument->throw( "Missing job_component.  job_component " . "must be passed with the method call" );
    }
    my %common_options = ();
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_options
    );

    my $job_component        = $opts{job_component};
    my $nacltask_if_mismatch = delete( $opts{'nacltask_if_mismatch'} );
    $job_component->wait_on_job( %common_options );
    my $cp = NACL::CS::JobHistory->fetch(
        command_interface => $opts{command_interface},
        filter            => { id => $job_component->id() },
        %common_options,
    );

    my $completion = $cp->completion();

    # We are looking only for success messages and some known failure
    # messages. Message not matching would be considered as
    # inconsistencies between Snapshot copy
    # and would throw NACL::Exceptions::DataInconsistent exception.
    if ( $completion =~ /Snapmirror check done with Snapshot copy/i || $completion =~ /SnapMirror\:\s+done\"?$/i ) {
    } elsif ( $completion =~ /found possible issues/i ) {
        $Log->warn($completion);
    } else {
        if ( $completion =~ /Found inconsisten.* Find details on node "(\S+)" in file "(\S+)"/i ) {
            my ( $node, $path ) = ( $1, $2 );
            if ( $nacltask_if_mismatch =~ /die/i ) {
                my %exception_args = ( 'node' => $node, 'path' => $path );
                $Log->exit();
                NACL::Exceptions::DataInconsistent->throw( "Snapmirror check found inconsistencies. Full message:\n" . $completion,
                    %exception_args );
            } else {
                $Log->comment( "Detailed log about inconsistencies can be found in node $node in file $path" );
                $Log->exit();
            }
        } elsif ( ( $completion =~ /SnapMirror: (done|error:)(\s+)(.+)/i )
            && ( defined $3 ) )
        {
            $Log->exit();
            NATE::BaseException->throw( "Snapmirror check failed due to some other reason " . $cp->completion() );
        } else {
            NACL::Exceptions::DataInconsistent->throw( "Snapmirror check found inconsistencies. Full message:\n" . $completion, );
        }
    }

    $pkg_or_obj->verify_health(%opts) if ($verify_health);

    # The 'destination-path' should be filled in by common_validate_with if
    # this is a method call. Otherwise, the caller should have passed the
    # 'destination-path' parameter.
    my $sm_fetch = NACL::CS::Snapmirror->fetch(
        command_interface => $opts{command_interface},
        filter            => { 'destination-path' => $opts{'destination-path'} },
    );
    my $ci = $opts{command_interface};
    if ( $ci->get_version_manager()->is_promoted() ) {
        $Log->warn( " The snaptags_test cannot be done for promoted_builds as the" . " test privilege is not present in those builds" );
    } else {
        if ( $sm_fetch->type() eq 'XDP' ) {
            my ( $vserver, $volume, $path );
            if ( ref $opts{'destination-path'} ) {
                $opts{'destination-path'} = $opts{'destination-path'}->fullpath;
            }

            # Use a static call to is_granular to determine if we need to
            # derive the actual volume path:
            if ( NACL::STask::Snapmirror->is_granular(%opts) ) {
                $path = $pkg_or_obj->volume_paths_from_cg(%opts);
            } else {
                $path = $opts{'destination-path'};
            }

            # Get vserver and volume from the path:
            if (   $path =~ m{^(?:[^:/]+)://(.*)/(.*)$}
                || $path =~ m{^(.*):(.*)$} )
            {
                $vserver = $1;
                $volume  = $2;
            } else {
                NATE::Exceptions::Argument->throw( "Invalid path for snaptags_test: $path" );
            }

            # We only tagged a snapshot if this is *not* a MetaWAFL volume. For
            # now, we determine this by looking for the existence of a state
            # attribute "default-extent-size". This is a duplication of the
            # same evaluation in _check_CM.
            my $is_metawafl;
            try {
                my $volobj = NACL::C::Volume->new(
                    command_interface => $opts{command_interface},
                    volume            => $volume,
                    vserver           => $vserver
                );
                my $value = $volobj->get_one_state_attribute('default-extent-size');

                $is_metawafl = ( $value && ( $value ne '-' ) ) ? 1 : 0;
            }
            otherwise {

                # If we saw an error, for now just assume that it is because
                # the system doesn't yet know about MetaWAFL and the
                # "default-extent-size" attribute.
                $is_metawafl = 0;
            };

            # We only have to remove the tag if this is *not* a MetaWAFL
            # volume:
            if ( !$is_metawafl ) {

                # We need to know on which node the volume is residing.
                my $vol_node = NACL::STask::Volume->get_local_node_name(
                    command_interface => $opts{command_interface},
                    volume            => $volume,
                    vserver           => $vserver,
                );

                # Armed with vserver and volume, find the snapshot that has the
                # checker tag that we added before starting the checker.
                my $snap_info = NACL::STask::VolumeSnapshot->find(
                    command_interface => $opts{command_interface},
                    filter            => {
                        vserver => $vserver,
                        volume  => $volume,
                        tags    => ['*VOPL_owner_snapmirror.checkInProgress:=snapmirror*'],
                    }
                );

                if ($snap_info) {
                    my $snapshot = $snap_info->snapshot();
                    my $apiset   = $opts{command_interface}->get_7m_or_nodescope_apiset();
                    my $cmd      = "snaptags_test -d -v $volume -S $snapshot " . "VOPL_owner_snapmirror.checkInProgress:" . ",snapmirror,5";
                    my $raw_output = $apiset->execute_command(
                        'nodescope-node-name' => $vol_node,
                        command               => $cmd,
                    );

                    if ( !( grep( /SUCCESS/, $raw_output ) ) ) {
                        NATE::BaseException->throw( 'Snaptags_test failed before running checker.' );
                    }
                } else {    # If snapshot is not found, throw an error.
                    NATE::BaseException->throw('No snapshot was found tagged for checker.');
                }
            }
        }
    }
} ## end sub _wait_for_check_CM

# Constructs and returns vserver path using vserver object
sub _construct_vserver_path {
    $Log->enter();
    my ( $self, %opts ) = @_;
    my $vserver_obj = $opts{vserver_obj};

    $Log->exit();
    return $vserver_obj->vserver() . ':';
}

=head2 construct_volpath

    $vol_path = $pkg->construct_volpath(
        vol_obj_or_volpath => $src_vol_obj,
        cluster_name       => $source_cluster,
    );

(Class or Instance method) Constructs and returns the volume path based on
the given parameters. This method is applicable for CMode only.

=over

=item Options

=over

=item C<< "vol_obj_or_volpath=>$vol_obj_or_volpath" >>

(Required)
This parameter expects a NACL::C::Volume object or the volume path.
When it is not an object, this method just returns same path as vol_path.

=item C<< "cluster_name" >>

(Optional)
Name of the cluster.

=item C<< "is_source" => 0|1 >>

(Optional)
If is_source is set to 1, then source_ci would be used to find the source
cluster name ,else 'command_interface' would be used to find the destination
cluster name Default value is '0'.

=item C<< source_ci => $source_ci >>

(Optional)
See documentation at the beginning of this page.

=item C<< "command_interface=>$command_interface" >>

(Optional)
Represents command interface of the destination cluster or node. This option is
required if you are having intercluster snapmirror relationship and if the
command_interface passed in the destination volume object is of type NACL::C::Vserver.
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=item Exceptions

=over

=item C<NATE::Exceptions::Argument>

This type of exception is thrown when either of vserver name or volume name are missing for the construction of a full volpath or
Either of cluster_name or command_interface/source command_interface is missing or
when invalid path is passed as source/destination-path.

=back

=back

=cut

sub construct_volpath {
    $Log->enter();

    my $self = shift;
    my %opts = NACL::C::Component->_common_validate_with(
        params          => \@_,
        additional_spec => {
            'vol_obj_or_volpath' => { type => SCALAR | OBJECT },
            'cluster_name'       => { type => SCALAR | UNDEF, optional => 1 },
            'command_interface' => {
                isa      => 'NACL::C::CommandInterface::ONTAP',
                type     => OBJECT,
                optional => 1,
            },
            'source_ci' => {
                isa      => 'NACL::C::CommandInterface::ONTAP',
                type     => OBJECT,
                optional => 1,
            },
            'is_source' => { type => BOOLEAN, default => 0 }
        },
        ignore_primary_keys => 1,
        allow_extra         => 1,
    );

    my $vol_obj_or_volpath = delete( $opts{vol_obj_or_volpath} );
    my $cluster_name       = delete( $opts{cluster_name} );
    my $is_source          = delete( $opts{is_source} );

    my %common_options = ();
    $self->_move_common_component_params(
        source => \%opts,
        target => \%common_options
    );

    my $ci;

    if ($is_source) {
        $ci = $opts{'source_ci'};
    } else {
        $ci = $opts{'command_interface'} || $self->command_interface();
    }
    my $volpath = undef;
    if ( ref($vol_obj_or_volpath) ) {
        if ( $vol_obj_or_volpath->isa('NACL::C::Volume') ) {
            my $vol_obj = $vol_obj_or_volpath;

            my $vserver_name = $vol_obj->vserver();
            my $volume_name  = $vol_obj->volume();

            if ( $vserver_name && $volume_name ) {
                $volpath = "$vserver_name:$volume_name";
            } else {
                $Log->exit();
                NATE::Exceptions::Argument->throw( "Missing values for the construction of a full volpath: "
                        . "vserver_name = \"$vserver_name\","
                        . " volume_name = \"$volume_name\"." );
            }
        } else {
            $Log->exit();
            NATE::Exceptions::Argument->throw( "Invalid object passed as source/destination-path.  "
                    . "Expected a NACL::C::Volume or NACL::STask::Volume, "
                    . "but was passed a "
                    . ref($vol_obj_or_volpath)
                    . "." );
        }
    } else {
        $volpath = $vol_obj_or_volpath;
    }

    $Log->exit();
    return $volpath;
} ## end sub construct_volpath

=head2 restore

  my $sm = NACL::STask::Snapmirror->restore(
      command_interface     =>  $command_interface,
      'source-path'         =>  $src_volume_specification,
      'destination-path'    =>  $dst_volume_specification,
      'nacltask_wait'       =>  $boolean,
      'file-list'           => ['/file1', '/file2', ...],
      %other_options
  );

 (or)

  my $sm = NACL::STask::Snapmirror->restore(
      command_interface     =>  $command_interface,
      'source-path'         =>  $src_volume_specification,
      'destination-path'    =>  $dst_volume_specification,
      'nacltask_wait'       =>  $boolean,
      'file-list'           => [$file_obj1, $file_obj2, ...],
      %other_options
  );


  (or)

  my $sm = $snapmirror_object->restore(nacltask_wait  =>  0 );  # return after
                                                        issuing restore command


(Class or Instance method) This method restores Snapshot copies from the
source volume to a destination volume. This provides additional
functionality like wait till the restore is complete.
If restore is called as an instance method, then source and destination paths
will be swapped so that original snapmirror destination will  become restore
source and source volume will become restore destination.

Returns a new snapmirror object representing the transient relationship created.
Supports CMode CLI.

=over

=item Options

=over

=item C<< "command_interface=>$command_interface" >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>
This represents the source end point (source cluster /source node / source
vserver) of the snapmirror relationship.

=item C<< source_ci => $source_ci >>

(Optional)
See documentation at the beginning of this page.
This command interface would be used in the returned snapmirror object.

=item C<< "source-path =>  $src_volume_specification" >>

(Optional for class method, Not Applicable for instance method)
This is the sending side of the restore. Original snapmirror destination becomes
source-path during restore. $src_volume_specification may be either
a string specifying the full path to the source volume, or it may be a
NACL::(C|STask)::Volume with the caveat that source-cluster must also be passed
as an argument if the Volume object's command_interface is a NACL::C::Vserver.

=item C<< "destination-path =>  $dst_volume_specification" >>

(Required for class method, Not Applicable for instance method)
Original snapmirror source volume becomes destination-path during restore.
This is the receiving side of the restore. destination-path takes the same form
as the source-path argument.

=item C<< "source-cluster =>  $string" >>

(Optional) Name of cluster where the source volume resides.
This input applies only for CMode.

=item C<< "destination-cluster =>  $string" >>

(Optional) Name of cluster where the destination volume resides.
This input applies only for CMode.

=item C<< "nacltask_wait =>  $boolean" >>

(Optional) A number 0 or 1 that determines if the transfer has to happen in the
background.  1 means to return after the transfer has completed and 0 means to
return immediately. The default is 1.

=item C<< nacltask_verify =E<gt> $boolean >>

(Optional) A true/false (1 or 0) value that determines if B<restore> should
verify the restored snapshot on the source volume after the operation has
completed. Verification is only done if C<nacltask_wait> is also true (1).
See L</verify_restore>, below. The default is 0 (do not verify).

=item apiset_must, apiset_should, method-timeout, etc.

All other options supported by
L<< NACL::C::Snapmirror->restore|lib-NACL-C-Snapmirror-pm/restore >>
are supported by NACL::STask::Snapmirror->restore as well.

=item file-list

(Optional) The 'file-list' argument takes either a list of paths or can accept
a list of VolumeFile objects.

=back

=item Exceptions

Any exceptions thrown by B<NACL::C::Snapmirror::restore()> may be thrown
here, as well. Additionally, if C<nacltask_verify> is set to a true value,
a failure to verify the restored snapshot will throw an exception of the
class B<NACL::Exceptions::VerifyFailure>.

=over

=item C<NATE::BaseException>

This type of exception is thrown when filer mode is not supported.

=item C<NACL::Exceptions::InvalidChoice>

This type of exception is thrown when an invalid option is passed.

=item C<NACL::Exceptions::VerifyFailure>

This type of exception is thrown when verification for restore operation has failed.

=back

=back

=cut

sub restore {

    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;
    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %args = @args;

    #Flag is to check if the user passed the source_ci
    #or if it was returned by _common_validate_with
    my $flag = 0;
    $flag = 1 if ( !$args{source_ci} );

    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {

            # component options which are overloaded
            'source-path'      => { type => OBJECT | SCALAR, optional => 1 },
            'destination-path' => { type => OBJECT | SCALAR, optional => 1 },

            # options which would be consumed by stask method itself apart
            # from passing the information to components.
            'source-cluster'      => { type => SCALAR,   optional => 1 },
            'destination-cluster' => { type => SCALAR,   optional => 1 },
            'file-list'           => { type => ARRAYREF, optional => 1 },
            'nacltask_wait'       => { type => SCALAR,   default  => 1 },
            'nacltask_verify'     => { type => BOOLEAN,  default  => 0 },
        },
        allow_extra => 1,
    );

    #Delete the source_ci if it was returned for _common_validate_with
    delete $opts{source_ci} if ( $flag == 1 );

    # After the call to _common_validate_with(), %opts will have common params
    # like method-timeout, apiset_must, apiset_should, command_interface,
    # primary keys, polling_interval, stask specific params, other options
    # taken by NACL::C::Snapmirror->restore()
    my $ci = $opts{command_interface};

    $opts{'method-timeout'} ||= SNAPMIRROR_TIMEOUT;
    my $rst_obj;

    my @files;
    if ( defined $opts{'file-list'} ) {
        foreach my $file ( @{ $opts{'file-list'} } ) {
            if ( blessed($file) && $file->isa('NACL::C::VolumeFile') ) {
                my $volume_file = $file->_extract_volume_file_from_path( path => $file->path() )->file();
                push @files, "/$volume_file";
            } else {
                push @files, $file;
            }
        }
        $opts{'file-list'} = \@files;
    }

    # The dependencies which need to be handled are different between 7M and CM
    # Hence we will have two backends for 7M and CM separately.
    if ( $ci->mode() eq '7Mode' ) {

        # Invoke 7M backend
        $Log->trace("7Mode restore not supported");
        $Log->exit();
        NACL::Exceptions::InvalidChoice->throw("7Mode not supported currently.");
    } elsif ( $ci->mode() eq 'CMode' ) {

        # Invoke CM backend
        $rst_obj = $pkg_or_obj->_restore_CM(%opts);
    } else {

        # Just to avoid open if-elsif
        $Log->exit();
        NATE::BaseException->throw( "Filer mode '" . $ci->mode() . "' not supported." );
    }
    $Log->exit();
    return $rst_obj;
} ## end sub restore

sub _restore_CM {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;

    # save these for use after component restore()
    my $wait_for_restore = delete $opts{nacltask_wait};
    my $polling_interval = delete $opts{polling_interval};
    my $do_verify        = delete $opts{nacltask_verify};

    my %common_options = ();
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_options
    );

    # If restore is called as an instance method, then source and destination
    # paths has to be swapped so that original snapmirror destination will
    # become restore source and source volume will become restore destination.
    if ( ref $pkg_or_obj ) {
        my $temp = $opts{'source-path'};
        $opts{'source-path'}      = $opts{'destination-path'};
        $opts{'destination-path'} = $temp;
    }
    $opts{'source_ci'} = $pkg_or_obj->_get_source_ci(%opts);

    # Handle Job ID
    my $job_component = delete $opts{job_component};
    {

        # We want to know the job ID of the job performed by
        # SUPER::restore, whether or not our caller does.
        my $dummy;
        $job_component ||= \$dummy;
    }
    my $sm_rst;

    # Perform restore
    # Always make class method calls so that we don't have to bother
    # about skipping primary keys for instance method call
    my $apiset;
    my $pkg = ref $pkg_or_obj || $pkg_or_obj;
    $sm_rst = $pkg->SUPER::restore(
        %opts,
        job_component => $job_component,

    );

    $pkg_or_obj->last_job( ${$job_component} ) if ( ref $pkg_or_obj );
    $sm_rst->last_job( ${$job_component} );

    if ( $wait_for_restore == 1 ) {

        $opts{polling_interval} = $polling_interval
            if ( defined $polling_interval );

        # Swap source and destination paths since during restore, destination acts
        # as source and vice versa. This has to be changed for snapmirror show.
        ( $opts{'source-path'}, $opts{'destination-path'} ) = ( $opts{'destination-path'}, $opts{'source-path'} );

        # Wait till the restore job is complete.
        $pkg->wait_for_restore(
            %opts,
            'destination-path' => $opts{'destination-path'},
            job_component      => $$job_component,
        );

        if ($do_verify) {
            if ( !$pkg->verify_restore(%opts) ) {
                $Log->exit();
                NACL::Exceptions::VerifyFailure->throw( 'No "restore" snapshot found on source for snapshot ' . $opts{'source-snapshot'} );
            }
        }
    }

    # add destination_ci to created stask object
    $sm_rst->{destination_ci} = $opts{destination_ci} if ( defined $opts{destination_ci} );

    $Log->exit();
    return $sm_rst;
} ## end sub _restore_CM

=head2 wait_for_restore

    $snapmirror->wait_for_restore();


    NACL::STask::Snapmirror->wait_for_restore(
      command_interface  =>  $command_interface,
      'destination-path' =>  $destination_path,
      job_component      =>  $job_component,
    );

(Class or instance method) Wait for the Snapmirror restore to complete. If a job
for restoring this Snapmirror is known then this waits for that job.
Defaule timeout is 1200 seconds and Polling interval 10 seconds.

=over

=item Options

=over

=item C<< "command_interface=>$command_interface" >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< "destination-path=>$destination_path" >>

(Required for class method, Not Applicable for instance method)
The actual destination path (original destination path before restore)of the
snapmirror relationship.

=item C<< "job_component=>$job_component" >>

(Required for class method, Not Applicable for instance method) A
NACL::C::Job object representing the C-mode Job used to restore Snapmirror. This
is not required when called as in instance method, provided, restore is also
called as an instance method.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=back

=back

=cut

sub wait_for_restore {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = @_;

    $opts{job_component} ||= $pkg_or_obj->last_job();
    my @opts = %opts;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@opts );

    %opts = $pkg_or_obj->_common_validate_with(
        params          => \@opts,
        additional_spec => { job_component => { type => OBJECT, optional => 0, isa => 'NACL::C::Job' }, },
        allow_extra     => 1,
    );

    # After the call to _common_validate_with(), %opts will have common params
    # like method-timeout, apiset_must, apiset_should, command_interface,
    # primary keys, job and polling_interval
    my $ci = $opts{command_interface};
    $opts{'method-timeout'}   ||= SNAPMIRROR_TIMEOUT;
    $opts{'polling_interval'} ||= POLL_DELTA;

    my %common_options = ();
    $pkg_or_obj->_copy_common_component_params(
        source => \%opts,
        target => \%common_options
    );

    $opts{job_component}->wait_on_job( %common_options, 'polling_interval' => $opts{'polling_interval'} );

    $Log->exit();
} ## end sub wait_for_restore

=head2 verify_restore

    if ($snapmirror->verify_restore(snapshot => 's1')) {
        # The restore-version of s1 was found on the src
        ...
    }

(Class or instance method)

This method verifies that the specified snapshot was correctly restored from
the destination to the source, by looking for the restore-snapshot on the
source volume. It returns a true/false (1/0) value based on whether the target
snapshot was found, or not. If an exception occurs the exception is thrown
rather than a return value sent back.

=over

=item Options

B<verify_restore> is usually called from within B<restore>, when the
C<nacltask_verify> option is passed to that method. As such, all the
options usually passed to B<restore> can be passed here. These are the
basic core options for specifying a snapmirror relationship (C<source-path>,
C<destination-path>, C<command_interface>, C<source_ci>, etc.).

In addition, the following options are recognized. At least one of these two
must be present:

=over

=item C<< snapshot =E<gt> $snapshot_name >>

=item C<< source-snapshot =E<gt> $snapshot_name >>

Specify the snapshot to be searched for. The second option (C<source-snapshot>)
is allowed for option-compatibility with B<restore>. If both options are
passed, C<snapshot> will take precedence.

=back

=item Exceptions

=over

=item B<NATE::Exceptions::Argument>

If the C<source-path> cannot be parsed, or if neither of C<snapshot> or
C<source-snapshot> are passed in, then an exception of this type will be
thrown.

=back

=back

=cut

###############################################################################
#
#   Sub Name:       verify_restore
#
#   Description:    Verify that the "source-snapshot" did in fact make it to
#                   the src volume as a restore snapshot.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $pkg_or_obj in    scalar    Package name or object of this
#                                                 class
#                   %opts     in      list      Rest of arguments converted by
#                                                 call to _common_validate_with
#
#   Returns:        Success:    1 or 0, restore snapshot is/not present
#                   Failure:    exception thrown
#
###############################################################################
sub verify_restore {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {

            # component options which are overloaded
            'source-path'      => { type => OBJECT | SCALAR, optional => 1 },
            'destination-path' => { type => OBJECT | SCALAR, optional => 1 },
            snapshot           => { type => SCALAR,          optional => 1 },
            'source-snapshot'  => { type => SCALAR,          optional => 1 },
        },
        allow_extra => 1,
    );

    my ( $snapshot, $volume, $vserver, $ci );

    # for ONTAP-AVA scenario we needs to process 'destination-path' instead of 'source-path'
    # since 'destination-path' does contain entity details where restore actually happens
    $volume = $opts{destination_ci} ? $opts{'destination-path'} : $opts{'source-path'};
    my $path = ( ref $volume ) ? $volume->fullpath : $volume;
    if ( $path =~ m{^(?:[^:/]+)://(.*)/(.*)$} || $path =~ m{^(.*):(.*)$} ) {
        $vserver = $1;
        $volume  = $2;
    } else {
        $Log->exit();
        NATE::Exceptions::Argument->throw( "Value '$path' for source-path is not parsable" );
    }

    # We get the command_interface to use by looking first for a 'source_ci'
    # option, then falling back on 'command_interface'. But if this is an
    # instance call, 'command_interface' will probably not work as expected.
    if ( ref($pkg_or_obj) && $pkg_or_obj->source_ci() ) {
        $opts{source_ci} ||= $pkg_or_obj->source_ci();
    }
    $ci = $opts{source_ci} || $opts{command_interface};

    # Lastly, they needed to specify what snapshot was used for the restore.
    # At present, restore() actually lets this be optional, but it isn't
    # optional here.
    $snapshot = $opts{snapshot} || $opts{'source-snapshot'};
    if ( !$snapshot ) {
        $Log->exit();
        NATE::Exceptions::Argument->throw( 'Missing either "snapshot" or "source-snapshot"' );
    }

    # while working on burt#1026004 it has been identified that
    # below NACL::STask::VolumeSnapshot->find() call
    # fetches snapshot details by prefexing snapshot
    # name with"restored" string. But this call doesn't
    # make any sense because no such snapshot is available
    # on particular volume.

    my @sslist = NACL::STask::VolumeSnapshot->find(
        command_interface => $ci,
        filter            => {
            vserver  => $vserver,
            volume   => $volume,
            snapshot => "restored.*.$snapshot.*",
        },
        allow_empty => 1
    );

    $Log->exit();
    return (@sslist) ? 1 : 0;
}

=head2 snapshot_checkpoint

    my $checkpoint = $snapmirror->snapshot_checkpoint;

(Instance method only, works in all modes)

Get the checkpoint value from the currently-running update/transfer
operation. This is generally used in conjunction (or soon followed by) the
next method, B<checkpoint_verify>. This method reads the snapmirror's state
and returns the value that corresponds to the snapshot checkpoint value.

=over

=item Options

None

=item Exceptions

Only exceptions thrown by the call to B<state> or thrown by trying to call
this as a static method will appear.

=back

=cut

###############################################################################
#
#   Sub Name:       snapshot_checkpoint
#
#   Description:    Fetch the value of snapshot_checkpoint from the current
#                   state of the calling object and return it.
#
#                   Yes, this is a bit of syntactic sugar around a simple
#                   call to get_one_state_attribute(). But it is easier to
#                   understand in the context of get-checkpoint/verify-cp.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#
#   Returns:        Success:    value
#                   Failure:    throws exception
#
###############################################################################
sub snapshot_checkpoint {
    $Log->enter();
    my ($self) = @_;
    $self->_verify_invocation( style => 'instance_only' );

    $Log->exit();
    return $self->get_one_state_attribute('snapshot-checkpoint');
} ## end sub snapshot_checkpoint

=head2 checkpoint_verify

    if (! $snapmirror->checkpoint_verify(checkpoint => '500M')) {
        # Log a message that the operation was not working right
    }

(Instance method only, works in all modes)

Check that an update operation is at or past the given checkpoint. Most often,
this will be used after an B<abort> has been done and the B<update> resumed.
It will be used to confirm that the B<update> resumed at the correct point.

Returns a boolean result, 1 (true) if the snapshot checkpoint field is at or
past the checkpoint value, 0 (false) otherwise. May throw an exception if
the call to B<state> (to retrieve a B<NACL::CS::Snapmirror> instance) fails,
or if there is no transfer currently taking place (the C<Snapshot Progress>
field does not appear to contain a size value).

=over

=item Options

This method takes just one parameter, which is required:

=over

=item C<< checkpoint =E<gt> $value >>

The value that the C<Snapshot Progress> field should be at or past. This value
should be a number, with an optional suffix indicating units (C<B>, C<K>/C<KB>,
C<M>/C<MB>, C<G>/C<GB> or C<T>/C<TB>). The suffix is not case-sensitive.

=back

=item Exceptions

If there is no transfer taking place, there is no snapshot checkpoint value to
compare the checkpoint to. In this case, a B<NACL::Exceptions::VerifyFailure>
exception is thrown. The only other exception(s) thrown by this method would
come from the attempt to read the snapmirror's state. See
L<NACL::C::Snapmirror|lib-NACL-C-Snapmirror-pm/state> for any exception
information.

=back

=cut

###############################################################################
#
#   Sub Name:       checkpoint_verify
#
#   Description:    Verify that a currently-running update operation is at
#                   or past the given checkpoint.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Remaining arguments, passed
#                                                 to validate()
#
#   Returns:        Success:    0/1 (false/true)
#                   Failure:    throws exception
#
###############################################################################
sub checkpoint_verify {
    $Log->enter;
    my ( $self, @opts ) = @_;
    my %opts = validate(
        @opts,
        {   checkpoint => {
                type  => SCALAR,
                regex => qr/^\d+(?:[.]\d+)?(?:[kmgt]?b?)?$/i,
            },
        }
    );

    my $checkpoint = NACL::C::UnitNormalization::to_unit( $opts{checkpoint}, 'b' );
    chop($checkpoint);    # Lose the trailing 'b', we don't want it
    $checkpoint = int($checkpoint);    # Round it down to a whole int

    # Get the current value of "Snapshot Progress" to compare to $checkpoint
    my $snap_checkpoint = $self->get_one_state_attribute('snapshot-checkpoint');
    if ( $snap_checkpoint !~ /^\d+(?:[.]\d+)?(?:[kmgt]?b?)?$/i ) {

        # No operation is actually underway, or it's already finished?
        $Log->exit();
        NACL::Exceptions::VerifyFailure->throw(
            '"Snapshot Checkpoint" should have been a size measurement, but ' . "was '$snap_checkpoint'",
            unexpected_values => { snapshot_checkpoint => $snap_checkpoint } );
    }
    $snap_checkpoint = NACL::C::UnitNormalization::to_unit( $snap_checkpoint, 'b' );
    chop($snap_checkpoint);    # Lose the trailing 'b', we don't want it

    $Log->exit();
    return ( $snap_checkpoint >= $checkpoint );
} ## end sub checkpoint_verify

=head2 flip_resync

    # Do a flip/resync on the Snapmirror relationship in $sm
    $sm->flip_resync;

    # Do the flip/resync, but return immediately and do the
    # resync portion in the background:
    $sm->flip_resync(nacltask_wait => 0);

(Instance method) This method is called for CMode only.

It will do the following steps:

=over

=item 1. Break the snapmirror relationship

=item 2. Delete the snapmirror relationship

=item 3. Create a new relationship from the destination to the source (of the same
relationship-type as the orginal)

=item 4. Resync the (new) relationship

=item 5. (If required by Sierra Nevada)
Do a C<snapmirror release> on the old relationship from the source cluster

=back

This method will perform the steps above and return the newly created
snapmirror relationship. The method also allows the user to control whether to
wait for the C<resync> to complete by specifying the C<nacltask_wait> option.

Note that the calling object is not altered, but the underlying snapmirror
relationship it represents will have been deleted. It is up to the user to
dispose of the object when it is no longer needed.

=over

=item Options

=over

=item C<< command_interface =E<gt> $object >>

(Optional) If it is necessary to use a different command-interface to create
the new snapmirror relationship (such as in cross-cluster situations), then
this parameter lets you provide the command-interface to be used. If this is
not passed, then the command-interface associated with the existing snapmirror
is used.

=item C<< nacltask_wait =E<gt> 0|1 >>

(Optional) This option allows the user to choose whether to wait for the resync
to complete before returning control (if the value is 1) or to return
immediately after the resync is started (if the value is 0). The default is 1.

Note that this does not affect the third step (creating the new snapmirror
relationship). That part of the operation will always wait for completion
before moving on to the fourth (resync) step.

=item C<< job_component =E<gt> $scalar_ref >>

(Optional) If C<nacltask_wait> is set to 0, then the user may optionally pass
a scalar reference here to get back the B<NACL::C::Job> object that represents
the running C<resync> operation. If C<nacltask_wait> is 1, this option is
ignored.

=back

Any additional options are passed to the call to C<resync>, with the exception
of those parameters that specify the source and destination volumes (which are
already known). Parameters that are passed through are: C<force>, C<throttle>,
C<transfer-priority>, C<source-snapshot> and C<preserve>.

=item Exceptions

Each of the steps are capable of throwing exceptions if an error occurs. This
method makes no attempt to catch and re-throw these, instead letting the
user deal with them as symptoms of failure in this method.

If this method is called on a snapmirror object that is on a filer in 7-mode,
an exception of the type B<NATE::BaseException> will be thrown. This method
can only be performed on C-Mode relationships.

=back

=cut

###############################################################################
#
#   Sub Name:       flip_resync
#
#   Description:    Perform the "flip/resync" operation on a snapmirror
#                   relationship, returning a new snapmirror with the
#                   end-points swapped.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $self     in      ref       Object of this class
#                   @opts     in      list      Remaining options, see POD
#
#   Returns:        Success:    new NACL::STask::Snapmirror object
#                   Failure:    throws exception
#
###############################################################################
sub flip_resync {
    $Log->enter();
    my ( $self, @opts ) = @_;
    $self->_verify_invocation( style => 'instance_only' );
    my %opts = validate_with(
        params => \@opts,
        spec   => {
            command_interface => {
                isa      => 'NACL::C::CommandInterface::ONTAP',
                optional => 1,
            },
            nacltask_wait => {
                type    => BOOLEAN,
                default => 1,
            },
            job_component => {
                type     => SCALARREF,
                optional => 1,
            },
        },
        allow_extra => 1
    );

    if ( $self->command_interface->mode ne 'CMode' ) {
        NATE::BaseException->throw( q(Filer mode ') . $self->command_interface->mode . q(' not supported for flip_resync) );
    }

    my $command_interface = $opts{command_interface}
        || $self->command_interface;

    # We'll need these current values from the existing snapmirror when we
    # create the new snapmirror:
    my $current_sm_state = $self->state( requested_fields => [qw(type schedule policy tries throttle)] );

    # Any of the following steps may throw an exception. We aren't catching
    # them, we're letting them go directly back to the caller as the error-case
    # for this method.

    # 1. Break
    $self->break;

    # 2. Delete
    $self->purge( nacltask_release => 0 );

    # 3. Create
    my $class = ref($self);    # In case we're sub-classed
    my %create_opts;
    my $job_ref;
    for my $key (qw(type schedule tries throttle)) {
        my $value = $current_sm_state->$key;
        if ( $value ne q{-} ) {
            $create_opts{$key} = $value;
        }
    }
    my $value = $current_sm_state->policy;
    if ( $value ne q{-} ) {
        $create_opts{policy} = $value;
    }

    my $new_sm = $class->create(
        command_interface  => $command_interface,
        'source-path'      => $self->destination_path,
        'destination-path' => $self->source_path,
        do_initialize      => 0,
        nacltask_wait      => 0,
        job_component      => \$job_ref,
        %create_opts
    );

    # Pre-Sierra Nevada should have given us back a job component:
    if ($job_ref) {
        $job_ref->wait_on_job();
    }

    # 4. Resync
    $new_sm->resync(%opts);

    # 5. Release on the source volume
    $self->release( source_ci => $command_interface );
    $new_sm->source_ci( $self->command_interface );

    $Log->exit();
    return $new_sm;
} ## end sub flip_resync

=head2 verify_health

    $snapmirror->verify_health();


    NACL::STask::Snapmirror->verify_health(
      command_interface  =>  $command_interface,
      'destination-path' =>  $destination_path,
    );

(Class or instance method) Verify if the last Snapmirror operation is successful.
Checks the health of the relationship and throws exception if there is last
transfer error. In case of RR, job history completion string will be thrown as
error message.

Supports CMode CLI/ZAPI.

=over

=item Options

=over

=item C<< "command_interface=>$command_interface" >>

(Required for class method, Not Applicable for instance method)
See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>

=item C<< "destination-path=>$destination_path" >>

(Required for class method, Not Applicable for instance method)
The destination path of the snapmirror relationship.

=item C<< "job_component=>$job_component" >>

(Optional for class method, Not Applicable for instance method) A
NACL::C::Job object representing the C-mode Job used to perform the last
Snapmirror operation.

=item C<< "last-transfer-type=>last transfer option" >>

(Optional) If we have to verify the health only after having some particular operation.
Ex: if we have to perform verify_health after initialize operation, pass this option as follows:
'last-transfer-type' => initialize.
verify_health() should do nothing if the last-transfer-type from the CS call is NOT initialize.

=item C<< apiset_must=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_must>

=item C<< apiset_should=>$ruleset >>

(Optional)See L<NACL::C::Component|lib-NACL-C-Component-pm/apiset_should>

=item Exceptions

If the health of the relationship is false, a B<NACL::Exceptions::VerifyFailure>
exception is thrown. See L<NACL::C::Snapmirror|lib-NACL-C-Snapmirror-pm/state>
for any exception information.

=back

=back

=cut

sub verify_health {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;

    $pkg_or_obj->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@args,
        additional_spec => {
            job_component => {
                type      => OBJECT | UNDEF,
                optional  => 1,
                callbacks => {
                    "object should be of type NACL::C::Job" => sub {
                        if   ( defined $_[0] ) { return $_[0]->isa('NACL::C::Job') }
                        else                   { return 1 }
                        }
                }
            },
            'last-transfer-type' => { type => SCALAR | UNDEF, optional => 1 },
        },
        allow_extra => 1,
    );

    # Check that the last snapmirror operation was successful by looking at the
    # "healthy" and "last_transfer_error" fields of the current state:
    my %common_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );
    my $state = NACL::CS::Snapmirror->fetch(
        filter => { 'destination-path' => $opts{'destination-path'} },
        %common_opts,
    );

    if ( exists $opts{'last-transfer-type'} ) {
        if ( $state->last_transfer_type ne $opts{'last-transfer-type'} ) {
            $Log->exit();
            return;
        }
    }
    if ( lc $state->healthy eq 'false' ) {

        # For some reason, the error message is coming back from ONTAP
        # braced within "{" and "}" characters. Do not want.
        ( my $errtext = $state->last_transfer_error ) =~ s/^{(.*)}$/$1/;
        $Log->exit();
        NACL::Exceptions::VerifyFailure->throw( "Operation failed with reason $errtext",
            unexpected_values => { last_transfer_error => $errtext } );
    }
} ## end sub verify_health

=head2 compare_luns

     NACL::STask::Snapmirror->compare_luns(
       command_interface  => $command_interface,
       'source-path'      => $source_path,
       'destination-path' =>  $destination_path,
       skip               => @skip_fields,
       must_mismatch      => @mismatch_fields
    );
                    (or)

    NACL::STask::Snapmirror->compare_luns(
       command_interface  => $command_interface,
       'source-path'      => $source_volume_object,
       'destination-path' =>  $destination_volume_object,
        skip              => @skip_fields,
       must_mismatch      => @mismatch_fields
    );

            (or)

    $snapmirror_obj->compare_luns(source_ci => $ci);

(Class and Instance method) Compare luns between source and destination.

1. This method will compare all the fields of all source luns with  its
corresponding destination luns, other than the fields specified in "skip"
parameter. And the "must_mismatch" parameter should contain the fields which
will not be same for two luns.

2. After comparison, if there are any other extra or less mismatches than
what is specified in 'must_mismatch' , it will throw an exception of type
NACL::Exceptions::LunSetFailure.

Supports CMode CLI

=over

=item Options

=over

=item C<< skip => @skip_fields >>

(Optional)

Array of fields to ignore during compare. Defaults are: ('volume', 'lun-path', 'vserver').

=item C<< must_mismatch => @mismatch_fields >>

(Optional)

Array of required disimilar fields. Method will return entry unless they are disimilar.
Default values are: ('serial', 'uuid', 'vdiskId')

=item C<< 'source-path' => $path_specification >>

(Required for Class method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< "destination-path =>  $volume_specification" >>

(Required for Class method)

See documentation at the beginning of this page for the various ways
this argument can be provided.

=item C<< "source-cluster =>  $string" >>

(Optional) Name of cluster where the source volume resides.


=item C<< "destination-cluster =>  $string" >>

(Optional) Name of cluster where the destination volume resides.
This input applies only for CMode.

=item C<< source_ci => $source_ci >>

(Optional)
See documentation at the beginning of this page.
If this parameter is not defined, then the method will try to construct the
source_ci by itself from the cluster name and fail gracefully by throwing
NATE::BaseException if tharnhost file for source cluster doesn't exist.

See L<NACL::C::Component|lib-NACL-C-Component-pm/command_interface>.

=item command_interface, apiset_must, apiset_should, method-timeout.

=back

=item Exceptions

=over

=item C<NATE::Exceptions::Argument>

This type of exception is thrown when improper path format is provided.

=item C<NATE::BaseException>

This type of exception is thrown when a tharnhost file in the name of source cluster doesn't exist.

=item C<NACL::Exceptions::LunSetFailure>

This type of exception is thrown when there is a mismatch while comparing the luns.

=back

=back

=cut

sub compare_luns {
    $Log->enter();
    my $pkg_or_obj = shift;
    my %opts       = $pkg_or_obj->_common_validate_with(
        params          => \@_,
        additional_spec => {
            'skip' => {
                type    => ARRAYREF,
                default => [ 'volume', 'path', 'vserver', 'space_resv', 'lun' ],
            },
            'must_mismatch' => {
                type    => ARRAYREF,
                default => [ 'serial', 'uuid', 'vdiskId', 'read_only', 'alignment', 'serial_hex' ]
            },
        },
    );
    my ($src_volume,   $dest_volume, @src_lun,                @dest_lun,        $exception_object,
        $cluster_type, %common_opts, %common_opts_without_ci, @sorted_src_objs, @sorted_dest_objs,
        @src_luns,     @dest_luns,   $cluster_name
    );

    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );

    my %opts_without_ci = %common_opts;
    delete $opts_without_ci{'command_interface'};
    my $CS_obj = $pkg_or_obj->get_CS_package_name();

    # fetching the source-path through the destination-path
    # if source-path is not defined
    if ( !defined( $opts{'source-path'} ) ) {
        my $sm = $CS_obj->fetch(
            %common_opts,
            filter           => { 'destination-path' => $opts{'destination-path'} },
            requested_fields => ['source-path']
        );
        $opts{'source-path'} = $sm->source_path();
    }

    # Defining the source_ci parameter, if it is an instance call
    if ( ref($pkg_or_obj) && ( !defined( $opts{'source_ci'} ) ) ) {
        $opts{'source_ci'} = $pkg_or_obj->source_ci()
            if ( defined( $pkg_or_obj->source_ci() ) );
    }

    $src_volume  = $pkg_or_obj->source_volume( $opts{'source-path'} );
    $dest_volume = $pkg_or_obj->destination_volume( $opts{'destination-path'} );

    unless ( defined $opts{'source_ci'} ) {

        # Checking the type of snapmirror relation ship
        $cluster_type = $pkg_or_obj->check_if_intercluster( %common_opts, %opts );

        # Fetching all the luns under source volume
        if ($cluster_type) {
            if ( $opts{'source-path'} =~ /(.+):\/\/(.+)\/(.+)/ ) {
                $cluster_name = $1;
            } elsif ( $opts{'source-path'} =~ /(.+):(.+)/ ) {
                if ( defined( $opts{'source-cluster'} ) ) {
                    $cluster_name = $opts{'source-cluster'};
                } else {
                    $Log->exit();
                    NATE::Exceptions::Argument->throw(
                        "if the path format is of the form vserver:volume " . "then source_ci parameter must be passed" );
                }
            }
            $opts{'source_ci'} = NACL::C::Cluster->new( name => $cluster_name );
            unless ( $opts{'source_ci'}->can_use_as_command_interface ) {
                $Log->exit();
                NATE::BaseException->throw( "A tharnhost file in the name of source"
                        . " cluster '$cluster_name' doesn't exist. So source_ci parameter is"
                        . " expected to be passed for reading the information about the luns"
                        . " present in the source volume" );
            }

        } else {
            $opts{'source_ci'} = $opts{command_interface};
        }
    }
    @src_lun = NACL::CS::Lun->fetch(
        %opts_without_ci,
        command_interface => $opts{'source_ci'},
        filter            => { volume => $src_volume },
    );

    # Fetching all the luns under destination volume
    @dest_lun = NACL::CS::Lun->fetch( %common_opts, filter => { volume => $dest_volume } );
    foreach (@src_lun) {
        push @src_luns, $_->path();
    }
    @src_luns = sort(@src_luns);
    for ( my $j = 0; $j < @src_luns; $j++ ) {
        foreach (@src_lun) {
            if ( $src_luns[$j] eq $_->path() ) {
                push @sorted_src_objs, $_;
            }
        }
    }
    foreach (@dest_lun) {
        push @dest_luns, $_->path();
    }
    @dest_luns = sort(@dest_luns);
    for ( my $k = 0; $k < @dest_luns; $k++ ) {
        foreach (@dest_lun) {
            if ( $dest_luns[$k] eq $_->path() ) {
                push @sorted_dest_objs, $_;
            }
        }
    }

    # Comparing the source lun to that particular destination lun
    my $msgs = '';
    my ( @errors, $unexpected_values );
    my $mismatched = 0;
    for ( my $i = 0; $i < @sorted_dest_objs; $i++ ) {
        my $compare_opts = $pkg_or_obj->_hash_copy(
            source => \%opts,
            copy   => [qw(must_mismatch skip)],
        );
        $compare_opts->{'lun1'} = $sorted_src_objs[$i];
        $compare_opts->{'lun2'} = $sorted_dest_objs[$i];
        try {
            NACL::STask::Lun->compare_luns( %common_opts, %{$compare_opts} );
        }
        catch NACL::Exceptions::CompareLunFailure with {
            my $e = shift;
            $msgs .= $e->text();

            $unexpected_values->{'exception_object'}     = $e;
            $unexpected_values->{'source_lun_path'}      = $compare_opts->{'lun1'}->path();
            $unexpected_values->{'destination_lun_path'} = $compare_opts->{'lun2'}->path();

            push( @errors, $unexpected_values );
            $mismatched = 1;
        };
    }

    if ($mismatched) {
        NACL::Exceptions::LunSetFailure->throw( $msgs, exception_objects => \@errors );
    }
    $Log->exit();
} ## end sub compare_luns

=head2 common_snapshots

    # Get the common snapshots (by name) between the source
    # and destination volumes:
    @common_snapshots = $mirror->common_snapshots();

    # Same, but as a class method instead of an instance method:
    @common_snapshots = NACL::STask::Snapmirror->common_snapshots(
        command_interface  => $ci,
        source_ci          => $source_ci,
        'source-path'      => $source_path,
        'destination-path' => $destination_path
    );

    # Get the common snapshots, but return objects instead of names:
    @common_ss_objs = $mirror->common_snapshots(objects => 1);

    # Get the snapshot objects, but get them for the snapshots on
    # the destination, rather than the source:
    @common_snapshots = $mirror->common_snapshots(
        objects          => 1,
        from_destination => 1
    );

(Instance or class method)

Works on CMode CLI.

Interrogate both end-point volumes of the snapmirror relationship for lists of
snapshots, then return a list of the snapshots that are common between the two
lists. Here, "common" is defined as snapshots having the same UUID (value of
the C<instance-uuid> field).

The returned list is the names of the snapshots, unless objects are requested.
The list may be empty if there are no common snapshots found between the two
volumes. If objects are requested, the elements of the returned list will be
objects of the B<NACL::STask::VolumeSnapshot> class.

If a common snapshot has a different name on the source and destination
volumes, the name returned in the list will be the name it is known by on the
source. Likewise, if objects are requested the objects will represent the
snapshots on the source volume, not the destination. This can be controlled
by the C<from_destination> option, explained below.

Because a snapmirror object saves the command interface associated with the
destination, if the source volume is on a different node than the destination
it may be necessary to include the C<source_ci> option to specify the
command interface from which to get the snapshots on the source volume.

B<Note for intercluster scenarios:> If the snapmirror source and destination
volumes are on different clusters, you I<must> give the C<source_ci> option to
specify the command-interface for the source volume. Otherwise, the volume will
be searched for on the same cluster as the destination volume. This could lead
to either an error if there is no such volume on the destination cluster or
worse, inconsistent results if there is. If you are calling this as an instance
method and you provided C<source_ci> when creating the
B<NACL::STask::Snapmirror> object, this should not be an issue.

=over

=item Options

The following options are recognized by this method:

=over

=item C<< objects =E<gt> 0|1 >>

If passed and true, the returned list will be made up of objects (of the
class B<NACL::STask::VolumeSnapshot>, rather than just the snapshot names.
The default is 0 (false), which means names are returned.

=item C<< from_destination =E<gt> 0|1 >>

If passed and true, the returned list will have names (or objects) from the
perspective of the destination volume rather than the source volume. This
means that if a snapshot has a different name on the destination than it
has on the source, it is the name from the destination that will be in the
list. Or, if C<objects> is also set, the returned objects will represent the
snapshots on the destination rather than on the source. The default is 0
(false), which means the return values are from the source.

=back

In addition to these options, all the common options (C<command_interface>,
C<source_ci>, C<source-path>, etc.) are recognized. This allows for calling
this as a class method.

=item Exceptions

This method calls in to the B<common_snapshots> method of the
B<NACL::STask::Volume> class. All exceptions that can be thrown by that method
may appear here. See
L<NACL::STask::Volume|lib-NACL-STask-Volume-pm/common_snapshots>.
Additionally, if an incorrect option is passed in here, an exception of the
class B<NATE::Exceptions::Argument> will be thrown.

=over

=item C<NATE::Exceptions::Argument>

This type of exception is thrown when invalid path passed as source-path or  destination-path or

When source-path and destination-path are on different clusters, but either source_ci was not passed or was the same as command_interface.

=back

=back

=cut

###############################################################################
#
#   Sub Name:       common_snapshots
#
#   Description:    Find a list of the common snapshots (by name) between the
#                   source and destination volumes for the given snapmirror.
#                   Return the names as a list.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $pkg_or_obj in    ref       Object of this class, or the
#                                                 class name if static
#                   @opts     in      array     Remaining opts, handled by
#                                                 _common_validate_with
#
#   Returns:        Success:    list of common snapshots, possibly empty
#                   Failure:    throws exception
#
###############################################################################
sub common_snapshots {
    $Log->enter();
    my ( $pkg_or_obj, @opts ) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@opts,
        additional_spec => {
            objects          => { type => BOOLEAN, default => 0 },
            from_destination => { type => BOOLEAN, default => 0 },
        }
    );
    my ( $src_ci, $src_volume, $src_vserver, $dst_ci, $dst_volume, $dst_vserver, $src_cluster, $dst_cluster, %args );

    # Command interfaces
    $dst_ci = $opts{command_interface};
    $src_ci = $opts{source_ci} ? $opts{source_ci} : $dst_ci;

    # Get cluster/vserver/volume data from paths
    if ( $opts{'source-path'} =~ m{^(.+?)://(.+?)/(.+?)$} ) {
        $src_cluster = $1;
        $src_vserver = $2;
        $src_volume  = $3;
    } elsif ( $opts{'source-path'} =~ m{^(.+?):(.+?)$} ) {
        $src_vserver = $1;
        $src_volume  = $2;
    } else {
        $Log->exit();
        NATE::Exceptions::Argument->throw( "Invalid path ($opts{'source-path'}) passed as source-path" );
    }
    if ( $opts{'destination-path'} =~ m{^(.+?)://(.+?)/(.+?)$} ) {
        $dst_cluster = $1;
        $dst_vserver = $2;
        $dst_volume  = $3;
    } elsif ( $opts{'destination-path'} =~ m{^(.+?):(.+?)$} ) {
        $dst_vserver = $1;
        $dst_volume  = $2;
    } else {
        $Log->exit();
        NATE::Exceptions::Argument->throw( "Invalid path ($opts{'destination-path'}) passed as " . 'destination-path' );
    }

    # Do a very basic sanity-check if $src_ci and $dst_ci are the same
    if ( $src_ci == $dst_ci ) {
        if ( $src_cluster && $dst_cluster && ( $src_cluster ne $dst_cluster ) ) {
            $Log->exit();
            NATE::Exceptions::Argument->throw( '"source-path" and "destination-path" are on different '
                    . 'clusters, but either source_ci was not passed or was '
                    . 'the same as command_interface' );
        }
    }

    # If they want the return data to be from the perspective of the dest
    # volume, then swap the ci/vserver/volume triples
    if ( $opts{from_destination} ) {
        %args = (
            command_interface   => $dst_ci,
            vserver             => $dst_vserver,
            volume              => $dst_volume,
            destination_ci      => $src_ci,
            destination_vserver => $src_vserver,
            destination_volume  => $src_volume,
        );
    } else {
        %args = (
            command_interface   => $src_ci,
            vserver             => $src_vserver,
            volume              => $src_volume,
            destination_ci      => $dst_ci,
            destination_vserver => $dst_vserver,
            destination_volume  => $dst_volume,
        );
    }

    $Log->exit();
    return NACL::STask::Volume->common_snapshots( %args, objects => $opts{objects} );
}

# Should not be used by external callers, hence is not documented. This is a
# wrapper around the component's delete that also waits on the job if one was
# started.
# Being a private routine, it does away with the validation, assuming that
# everything is passed in correctly.
sub delete {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;

    # Job is started for infinite volumes, so get the job-id.
    my $dummy;
    $opts{job_component} ||= \$dummy;

    $pkg_or_obj->SUPER::delete(%opts);

    my $job = ${ $opts{job_component} };
    my %job_args;
    if ( defined $job ) {
        $pkg_or_obj->last_job($job) if ( ref $pkg_or_obj );
        $job_args{job_component} = $job;
    }
    $pkg_or_obj->_construct_path( opts => \%opts );
    my %common_opts;
    $pkg_or_obj->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_opts
    );
    my %primary_keys;
    $pkg_or_obj->_copy_primary_keys(
        source => \%opts,
        target => \%primary_keys
    );

    $pkg_or_obj->wait_for_deletion( 'method-timeout' => 300, %primary_keys, %common_opts );

    $Log->exit();
}

=head2 parents

    @parents = $snapmirror_obj->parents();

( Class or Instance method ) This method returns the list containing the names of
the parent objects of this resource. Used for finding dependencies between resources.

=cut

sub parents {
    $Log->enter();
    $Log->exit();
    return qw(NACL::STask::Volume);
}

=head2 compare_snapshots

    # Validate snapshot lists between source and destination volumes
    $sm->compare_snapshots();

    # Same, but as a class method
    NACL::STask::Snapmirror->compare_snapshots(
        command_interface  => $ci,
        source_ci          => $source_ci,
        'source-path'      => $source_path,
        'destination-path' => $destination_path
    );

    # Extract the snapshot list differences from the exception
    try {
        $sm->compare_snapshots;
    } catch NACL::STask::Exceptions::SnapshotMismatch with {
        my $exception = shift;
        my $text = $exception->text();
        @missing_from_source = $exception->missing_from_source;
        @missing_from_destination = $exception->missing_from_destination;
    };


(Instance or class method)

Works on CMode CLI.

Validate the snapshot lists between the source and destination volumes
excluding SnapMirror snapshots are identical. Identical means that
the snapshots have the same instance-uuid. Two snapshots with the
same name but differing instance-uuids are not the same snapshot.

Snapshots that are found on the destination but not the source
are available via the missing_from_source accessor on the
NACL::STask::Exceptions::SnapshotMismatch exception. This
returns an array of NACL::CS::VolumeSnapshot objects.

Snapshots that are found on the source but not the destination
are available via the missing_from_destination accessor on the
NACL::STask::Exceptions::SnapshotMismatch exception. This
returns an array of NACL::CS::VolumeSnapshot objects.

Under load, C<volume snapshot show> can periodically fail with
NACL::APISet::Exceptions::CommandFailedException. The command
will be retried either until success or the retry limit is reached.
The default retry limit is 10.

=over

=item Options

The following options are recognized by this method:

=over

=item C<command_interface>

Required. Either already in the NACL::STask::Snapmirror object
or passed in. Component object representing node holding the
destination volume.

=item C<source_ci>

Required. Either already in the NACL::STask::Snapmirror object
or passed in. Component object representing node holding the
source volume.

=item C<nacltask_retries>

Optional. Maximum number of attempts to get volume snapshot
information from a filer per volume.

=item C<source-path>

Required. Either already in the NACL::STask::Snapmirror object
or passed in. Source path of the SnapMirror relationship

=item C<destination-path>

Required. Either already in the NACL::STask::Snapmirror object
or passed in. Destination path of the SnapMirror relationship

=back

=item Exceptions

=over

=item C<NACL::Exceptions::RetriesExceeded>

This exception is thrown once the maximum retry limit
has been reached in querying snapshot information from the filer.

=item C<NACL::STask::Exceptions::SnapshotMismatch>

This exception is thrown when the snapshot lists differ
between source and destination volumes excluding SnapMirror
snapshots.

=back

=back

=cut

###############################################################################
#
#   Sub Name:       compare_snapshots
#
#   Description:    Validate that source and destination volumes have same
#                   snapshot lists excluding SnapMirror snapshots.
#                   Returns if both lists match.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $pkg_or_obj in    ref       Object of this class, or the
#                                                 class name if static
#                   @opts     in      array     Remaining opts, handled by
#                                                 _common_validate_with
#
#   Returns:        Success:    returns nothing
#                   Failure:    throws exception
#
###############################################################################
sub compare_snapshots {
    $Log->enter();
    my ( $pkg_or_obj, @opts ) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params          => \@opts,
        additional_spec => { nacltask_retries => { type => SCALAR, default => $VOLSNAPRETRYCOUNT } },
    );
    my ( $src_ci, $dst_ci, $retries, @src_snapshots, @dst_snapshots, $mismatch, %snaphash, $count, $dst_path,
        @missing_from_source, @missing_from_destination, $src_path, );

    # Command interfaces
    $dst_ci = $opts{command_interface};

    # acquire source_ci via either an instance or method call
    if ( ref($pkg_or_obj) && $pkg_or_obj->source_ci() ) {
        $opts{source_ci} ||= $pkg_or_obj->source_ci();
    }
    $src_ci = $opts{source_ci};

    if ( !defined $src_ci ) {
        $Log->exit();
        NATE::Exceptions::Argument->throw('source_ci missing');
    }

    # Retry count
    $retries = $opts{nacltask_retries};

    # Get cluster/vserver/volume data from paths
    $src_path = $pkg_or_obj->_split_path( path => $opts{'source-path'} );
    $dst_path = $pkg_or_obj->_split_path( path => $opts{'destination-path'} );

    @src_snapshots = $pkg_or_obj->_fetch_snapshots(
        command_interface => $src_ci,
        vserver           => $src_path->{vserver},
        volume            => $src_path->{volume},
        retries           => $retries,
    );
    @dst_snapshots = $pkg_or_obj->_fetch_snapshots(
        command_interface => $dst_ci,
        vserver           => $dst_path->{vserver},
        volume            => $dst_path->{volume},
        retries           => $retries,
    );

    $Log->debug( __PACKAGE__ . ' Adding destination snapshots to hash' );
    foreach my $element (@dst_snapshots) {
        if ( $element->snapshot !~ /^snapmirror[.]/ ) {
            $snaphash{ $element->version_uuid } = $element;
        }
    }

    $Log->debug( __PACKAGE__ . ' Iterating over source snapshots in hash' );
    foreach my $element (@src_snapshots) {
        if ( $element->snapshot !~ /^snapmirror[.]/ ) {
            if ( exists $snaphash{ $element->version_uuid } ) {
                delete $snaphash{ $element->version_uuid };
            } else {
                push @missing_from_destination, $element;
                $Log->debug( __PACKAGE__ . ' Missing snapshot ' . $element->snapshot . ' version-uuid ' . $element->version_uuid );
                $mismatch = 1;
            }
        }
    }

    $Log->debug( __PACKAGE__ . ' Checking for extraneous snapshots in hash' );
    $count = scalar keys %snaphash;
    if ( $count > 0 ) {
        $mismatch = 1;
        foreach my $element ( keys %snaphash ) {
            push @missing_from_source, $snaphash{$element};
            $Log->debug( __PACKAGE__ . ' Extra snapshot ' . $snaphash{$element}->snapshot . ' version-uuid ' . $element );
        }
    }

    if ($mismatch) {
        $Log->exit();
        my $error
            = 'mismatched snapshots between '
            . "$src_path->{vserver}:$src_path->{volume} and "
            . "$dst_path->{vserver}:$dst_path->{volume}";
        NACL::STask::Exceptions::SnapshotMismatch->throw(
            text                     => $error,
            missing_from_source      => \@missing_from_source,
            missing_from_destination => \@missing_from_destination,
        );
    }

    $Log->debug( __PACKAGE__ . ' no snapshot differences found' );

    $Log->exit();

    return;
} ## end compare_snapshots

###############################################################################
#
#   Sub Name:       _fetch_snapshots
#
#   Description:    Private routine for compare_snapshots that fetchs
#                   snapshot list from a volume with limited retries.
#
#   Arguments:      NAME      IN/OUT  TYPE      DESCRIPTION
#                   $pkg_or_obj in    ref       Object of this class, or the
#                                                 class name if static
#                   %opts     in      hash      Remaining opts
#                                               expects command_interface,
#                                               volume, vserver, and retries
#
#   Returns:        Success:    returns NACL::CS::VolumeSnapshot object array
#                   Failure:    throws exception
#
###############################################################################
sub _fetch_snapshots {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;
    my ( $lastflag, $counter );
    my @snapshots;

    $Log->debug( __PACKAGE__ . ' retrieving volume snapshot list' );
    $lastflag = 0;
    $counter  = 0;
    foreach ( 1 .. $opts{retries} ) {
        try {
            @snapshots = NACL::CS::VolumeSnapshot->fetch(
                command_interface => $opts{command_interface},
                allow_empty       => 1,
                filter            => {
                    volume  => $opts{volume},
                    vserver => $opts{vserver},
                },
            );
            $lastflag = 1;
        }
        catch NACL::APISet::Exceptions::CommandFailedException with {

            # command fails with a generic Error command failed.
            # This is expected behavior on a loaded filer
            $Log->debug( __PACKAGE__ . ' source volume snapshot show failed' );
        };
        if ($lastflag) {
            last;
        }
        $counter++;
    }

    if ( $counter == $opts{retries} ) {
        NACL::Exceptions::RetriesExceeded->throw( ' unable to volume snapshot within ' . "$opts{retries} tries" );
    }
    $Log->exit();
    return @snapshots;
} ## end sub _fetch_snapshots

# The helper subroutine method to create the destination volume
# applicable only for CMode
sub _create_dest_volume {
    $Log->enter();

    my ( $pkg, %opts ) = @_;

    my $src_path = $opts{'source-path'};

    my $src_volume_obj;

    # If the command_interface is provided, then it's for the destination
    my $ci = $opts{command_interface};

    if ($src_path) {
        if ( blessed($src_path) ) {
            if ( $src_path->isa('NACL::C::Volume') ) {
                $src_volume_obj = $src_path;

                # Extract command_interface from source volume to use as
                # command_interface for destination if command_interface was
                # not provided.
                $ci ||= $src_path->command_interface();
            } else {
                $Log->exit();
                NATE::Exceptions::Argument->throw( "'source-path' was "
                        . "provided as an object of type '"
                        . ref($src_path)
                        . "' but it should be a Volume object when destination-path '"
                        . 'is not provided.' );
            }
        } else {

            # source-path is provided as string. Extract volume object from it
            # Here destination-path we are doing hack
            $src_volume_obj = NACL::C::Snapmirror->get_source_volume(
                %opts,
                command_interface  => $ci,
                source_ci          => $ci,
                'destination-path' => 'vserver:volume'
            );

        }
    } else {

        # source-path not provided. Attempt to create path
        $pkg->_construct_single_path( params => \%opts, type => 'source' );

        if ( !$opts{'source-path'} ) {
            $Log->exit();
            NATE::Exceptions::Argument->throw( 'Neither source-path nor ' . 'destination-path were provided in the call to create.' );
        }

        # source-path is provided as string. Extract volume object from it
        # Here destination-path we are doing hack
        $src_volume_obj = $pkg->get_source_volume(
            %opts,
            command_interface  => $ci,
            source_ci          => $ci,
            'destination-path' => 'vserver:volume'
        );
    }

    my %common_opts;
    $pkg->_copy_common_component_params(
        source => \%opts,
        target => \%common_opts
    );

    my $src_volume_size = $src_volume_obj->get_one_state_attribute( 'size', %common_opts );

    my $vserver = $src_volume_obj->vserver();

    my @available_vserver;

    # Get the existing vserver from source
    push(
        @available_vserver,
        NACL::STask::Vserver->new(
            command_interface => $ci,
            vserver           => $vserver
        )
    );

    # Get the peered vserver if given vserver is peered to any
    my @peer_obj = NACL::STask::VserverPeer->find(
        command_interface => $ci,
        filter            => { vserver => $vserver },
        'allow_empty'     => 1,
        %common_opts
    );

    foreach (@peer_obj) {
        push( @available_vserver, $_->get_vserver_instances()->peer_vserver_obj() );
    }

    my $dest_volume;

    foreach my $vserver_obj (@available_vserver) {
        try {
            $dest_volume = NACL::STask::Volume->create(
                command_interface => $vserver_obj->command_interface(),
                vserver           => $vserver_obj->vserver(),
                size              => $src_volume_size,
                type              => 'DP',
                %common_opts
            );
        }
        catch NATE::BaseException with {

            # do nothing
        };

        last if ($dest_volume);
    }

    unless ($dest_volume) {
        $Log->exit();
        NATE::Exceptions::Argument->throw("unable to find the destination volume");
    }

    $Log->exit();
    return $dest_volume;
}

sub _verify_vserver_subtype {
    $Log->enter();
    my ( $pkg, %opts ) = @_;

    # Verification is only for VDR relationship.
    unless ( $pkg->is_operated_on_vserver(%opts) ) {
        $Log->exit();
        return;
    }

    if ( $opts{command_interface}->is_cmode() ) {
        my %common_comp_opts;
        $pkg->_copy_common_component_params(
            source => \%opts,
            target => \%common_comp_opts
        );

        my $dest_vserver   = $pkg->destination_vserver(%opts);
        my $vserver_cs_obj = NACL::CS::Vserver->fetch(
            command_interface => $opts{'command_interface'},
            filter            => { vserver => $dest_vserver },
            requested_fields  => ['subtype'],
            %common_comp_opts,
        );
        unless ( $vserver_cs_obj->subtype() eq 'default' ) {
            $Log->exit();
            NACL::Exceptions::VerifyFailure->throw( 'Destination Vserver' . " $dest_vserver, subtype is not in default state as expected" );
        }
    }
    $Log->exit();
} ## end sub _verify_vserver_subtype

sub _verify_states {
    $Log->enter();
    my ( $pkg, %opts ) = @_;

    # Verification is only for VDR relationship.
    unless ( $pkg->is_operated_on_vserver(%opts) ) {
        $Log->exit();
        return;
    }

    my $command_interface = $opts{command_interface};
    my %common_comp_opts;
    $pkg->_copy_common_component_params_with_ci(
        source => \%opts,
        target => \%common_comp_opts
    );

    $pkg->verify_state(
        %common_comp_opts,
        'state'            => 'Snapmirrored',
        'status'           => 'Idle',
        'destination-path' => $opts{'destination-path'},
    );

    $Log->exit();
} ## end sub _verify_states(%opts)

=head2 config_diff

    $snapmirror_vdr->config_diff();

(Applicable only for VDR relationships)
Performs an MCC config diff for the destination vserver in the VDR
relationship. This provides the same features and accepts the same options as
L<< NACL::STask::Vserver->config_diff|lib-NACL-STask-Vserver-pm/config_diff >>.

=cut

sub config_diff {
    $Log->enter();
    my ( $pkg_or_obj, @args ) = @_;
    my %opts = $pkg_or_obj->_common_validate_with(
        params      => \@args,
        allow_extra => 1,
    );

    if ( !$pkg_or_obj->is_operated_on_vserver(%opts) ) {
        $Log->exit();
        NATE::Exceptions::Argument->throw( "'config_diff' works only for "
                . 'VDR relationship, but the snapmirror object provided '
                . "(path=$opts{'destination-path'}) is not a VDR relationship" );
    }

    # Get vserver name from destination-path
    my $path_hash = $pkg_or_obj->_split_path( path => $opts{'destination-path'} );
    my $vserver = $path_hash->{vserver};

    # Remove the primary keys
    $pkg_or_obj->_move_primary_keys( source => \%opts, target => {} );

    require NACL::STask::Vserver;
    NACL::STask::Vserver->config_diff( %opts, vserver => $vserver, is_dest => 1 );

    $Log->exit();
}

# Validate that the set of parameters includes what we need in cases where the
# destination is an AVA appliance.
sub _validate_ava_args {
    $Log->enter();
    my ( $self, %opts ) = @_;
    my %args = %{$opts{'arglist'}};

    # We must have a source_ci:
    if ( !$args{source_ci} ) {
        $Log->exit();
        NATE::Exceptions::Argument->throw(
            'When the destination is an AltaVault appliance, "source_ci" ' . 'must be present in the call.' );
    }

    # If destination-path is present, it can't be a NACL::C::Volume or
    # subclass thereof:
    if ( blessed $args{'destination-path'}
        && $args{'destination-path'}->isa('NACL::C::Volume') && ( $opts{'caller_method'} !~ /^(restore|verify_restore)$/) )
    {
	$Log->exit();
	NATE::Exceptions::Argument->throw( 'When the destination is an AltaVault appliance, the '
	    . '"destination-path" parameter may not be an object that '
	    . 'derives from NACL::C::Volume.' );
    }

    # We must have at least one of destination-path or destination-volume:
    if ( !( $args{'destination-path'} || $args{'destination-volume'} ) ) {
        $Log->exit();
        NATE::Exceptions::Argument->throw( 'When the destination is an AltaVault appliance, at least one '
                . 'of "destination-path" or "destination-volume" must be present '
                . 'in the call.' );
    }

    $Log->exit();
    return;
}

# update $args->{command_interface} with ONTAP ci value as all snapmirror
# related commands needs to be executed on ONTAP ci only. This is so that later NACL calls that
# default to using command_interface get the CI that is an ONTAP node. Note
# that we do NOT do this to the object properties, only to $args.
sub _swap_command_interface {
    $Log->enter();
    my ( $self, $args ) = @_;

    #Since $args->{command_interface} is having AVA command interface value
    #hence we need to store back the same into a new key so that it's further available
    #throughout the code. this new key "destination_ci" will be having only NON-ONTAP
    #command interface value
    $args->{destination_ci} = $args->{command_interface};

    # (since an earlier check should have verified that source_ci is present).
    # simply update $args->{command_interface} with ONTAP ci value as all snapmirror
    # related commands needs to be executed on ONTAP ci only.
    $args->{command_interface} = $args->{source_ci};
    $Log->exit();
    return;
}

# Preemptively convert the value of destination-path to an AVA-style path.
sub _build_ava_destination_path {
    $Log->enter();
    my ( $self, $args ) = @_;
    my $command_interface = $args->{destination_ci};

    # Prefer destination-path over destination-volume:
    my $volume = $args->{'destination-path'} || $args->{'destination-volume'};

    # The Full path, assigned back to the hashref that was passed in:
    $args->{'destination-path'} = $command_interface->get_destination_path( volume => $volume ) if ( $volume !~ /.+\:\/share\/.+/ );
    $Log->exit();
    return;
}

# Encapsulate the three previous methods into a single call that validates
# and massages the parameters.
sub _check_and_setup_for_ava {
    $Log->enter();
    my ( $pkg, %opts ) = @_;

    #Perform few basic validations for ONTAP-AVA snapmirroring and related functonalities
    $pkg->_validate_ava_args( %opts );

    #Update command_interface field to contain ONTAP command_interface, since all
    #commands will be executed on ONTAP rather than AVA
    $pkg->_swap_command_interface( $opts{'arglist'} );

    #Formulate destination-path as per AVA format
    #No need to formulate destination-path in case scenario is for restore
    $pkg->_build_ava_destination_path( $opts{'arglist'} ) if ( ($opts{'caller_method'} !~ /^(restore|verify_restore)$/) );

    $Log->exit();
    return;
}

sub _check_ava_perform_actions {
    $Log->enter();
    my ( $pkg_or_obj, %opts ) = @_;

    # below call is to get the parent method name
    my $caller_method = _get_parent_method_name();

    my %tmp_opts = @{$opts{'arglist'}};
    my $command_interface = $pkg_or_obj->command_interface() || $tmp_opts{command_interface};

    # Below condition is added in response to burt#1026701 to accomodate the
    # scenario where command_interface is neither present in object nor
    # within argument-list. The logic is similar to what is implemented in
    # overridden version of _common_validate_with().

    # Note: For ONTAP-AVA restore() scenario command_interface (i.e. AVA command_interface)
    # must be passed since restore volume is always present at ONTAP side but command_interface
    # should point to AVA command_interface only.

    # If the destination-path is of type "NACL::C::Volume"
    # then fetch the command_interface from the destination-path object.
    if ( ( !$command_interface ) && ( !ref $pkg_or_obj ) ) {
	my $destination_path = $pkg_or_obj->destination_path() || $tmp_opts{'destination-path'};
	if ( ( defined $destination_path ) && ( blessed $destination_path ) && ( $destination_path->isa('NACL::C::Volume') ) ) {
            $command_interface = $destination_path->command_interface();
	}
    }

    if ( ( defined $command_interface ) && ( $command_interface->is_ava()) ) {
	# only for ONTAP-AVA scenario update command_interface into argument list
	# before calling validations/command interface swapping utilities
	$tmp_opts{command_interface} = $command_interface if ( !$tmp_opts{command_interface} );

        $pkg_or_obj->_check_and_setup_for_ava( 'arglist' => \%tmp_opts, 'caller_method' => $caller_method );
	@{$opts{'arglist'}} = %tmp_opts;
    }

    $Log->exit();
    return;
}

sub _get_parent_method_name {
    $Log->enter();

    my $caller_method = ( caller(2) )[3];
    $caller_method =~ s/NACL::STask::Snapmirror:://g;

    $Log->exit();
    return $caller_method;
}

=head2 new

This is an override of the B<new> method from the super-class. The purpose
of this is just to accomodate relationships where the destination is not an
ONTAP filer. For full documentation of this method and its parameters, see
the documentation in B<NACL::C::Component>.

=cut

sub new {
    $Log->enter();
    my ( $pkg, @args ) = @_;

    $pkg->_check_ava_perform_actions( 'arglist' => \@args );

    my %opts = @args;
    my $destination_ci = delete $opts{destination_ci};
    my $snap_obj = $pkg->SUPER::new(%opts);

    # add destination_ci to created stask object
    $snap_obj->{destination_ci} = $destination_ci if ( defined $destination_ci );

    $Log->exit();
    return $snap_obj;
}

1;
